/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************
 
    Module Name:
    connect.c
 
    Abstract:
    Routines to deal Link UP/DOWN and build/update BEACON frame contents
 
    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
    John Chang  08-04-2003    created for 11g soft-AP
 */

#include "rt_config.h"

/*
    ==========================================================================
    Description:
        Pre-build a BEACON frame in the shared memory
    ==========================================================================
*/
VOID MakeBssBeacon(
    IN PRTMP_ADAPTER pAd) 
{
    UCHAR         SsidIe = IE_SSID, DsIe = IE_DS_PARM, SuppIe = IE_SUPP_RATES,
                  DsLen = 1, SsidLen, RSNIe = IE_WPA, RSNIe2 = IE_WPA2, SupportedRatesLen;
    MACHDR        BcnHdr;
    LONGLONG      FakeTimestamp;
    ULONG         FrameLen;
    PTXD_STRUC	  pTxD = (PTXD_STRUC)pAd->BeaconRing.va_addr;
    CHAR         *pBeaconFrame = (CHAR *)pAd->BeaconRing.va_data_addr;

    // Bridge mode doesn't send beacon.
    if (pAd->WdsTab.Mode == WDS_BRIDGE_MODE)
        return;

    if ((pAd->PortCfg.PhyMode == PHY_11BG_MIXED) || (pAd->PortCfg.PhyMode == PHY_11G))
        SupportedRatesLen = 4;
    else
        SupportedRatesLen = pAd->PortCfg.SupportedRatesLen;

    if (pAd->PortCfg.HideSsid)
        SsidLen = 0;
    else
        SsidLen = pAd->PortCfg.MSSID[MAIN_MSSID].SsidLen;
    
    MgtMacHeaderInit(pAd, &BcnHdr, SUBTYPE_BEACON, 0, &pAd->PortCfg.Broadcast, &pAd->PortCfg.Bssid);
    
    if ((pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPA) || (pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPAPSK))
    	RSNIe = IE_WPA;
    else if ((pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPA2) || (pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPA2PSK))
    	RSNIe = IE_WPA2;
    
    // Append RSN_IE when  WPA OR WPAPSK, 
    if ((pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPA1WPA2) || (pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
    {
        MakeOutgoingFrame(pBeaconFrame,                             &FrameLen,
                        MAC_HDR_LEN,                                &BcnHdr,
                        TIMESTAMP_LEN,                              &FakeTimestamp,
                        2,                                          &pAd->PortCfg.BeaconPeriod,
                        2,                                          &pAd->PortCfg.MSSID[MAIN_MSSID].CapabilityInfo,
                        1,                                          &SsidIe,
                        1,                                          &SsidLen,
                        SsidLen,                                    pAd->PortCfg.MSSID[MAIN_MSSID].Ssid,
                        1,                                          &SuppIe,
                        1,                                          &SupportedRatesLen,
                        SupportedRatesLen,                          pAd->PortCfg.SupportedRates,
                        1,                                          &DsIe,
                        1,                                          &DsLen,
                        1,                                          &pAd->PortCfg.Channel,
                        1,                                          &RSNIe,
                        1,                                          &pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[0],
                        pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[0],   pAd->PortCfg.MSSID[MAIN_MSSID].RSN_IE[0],
                        1,                                          &RSNIe2,
                        1,                                          &pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[1],
                        pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[1],   pAd->PortCfg.MSSID[MAIN_MSSID].RSN_IE[1],
                        END_OF_ARGS);
    }
	else if (pAd->PortCfg.MSSID[MAIN_MSSID].AuthMode >= Ndis802_11AuthModeWPA)
    {
        MakeOutgoingFrame(pBeaconFrame,                             &FrameLen,
                        MAC_HDR_LEN,                                &BcnHdr,
                        TIMESTAMP_LEN,                              &FakeTimestamp,
                        2,                                          &pAd->PortCfg.BeaconPeriod,
                        2,                                          &pAd->PortCfg.MSSID[MAIN_MSSID].CapabilityInfo,
                        1,                                          &SsidIe,
                        1,                                          &SsidLen,
                        SsidLen,                                    pAd->PortCfg.MSSID[MAIN_MSSID].Ssid,
                        1,                                          &SuppIe,
                        1,                                          &SupportedRatesLen,
                        SupportedRatesLen,                          pAd->PortCfg.SupportedRates,
                        1,                                          &DsIe,
                        1,                                          &DsLen,
                        1,                                          &pAd->PortCfg.Channel,
                        1,                                          &RSNIe,
                        1,                                          &pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[0],
                        pAd->PortCfg.MSSID[MAIN_MSSID].RSNIE_Len[0],   pAd->PortCfg.MSSID[MAIN_MSSID].RSN_IE[0],
                        END_OF_ARGS);
    }
    else
    {
        MakeOutgoingFrame(pBeaconFrame,                  &FrameLen,
                        MAC_HDR_LEN,                     &BcnHdr,
                        TIMESTAMP_LEN,                   &FakeTimestamp,
                        2,                               &pAd->PortCfg.BeaconPeriod,
                        2,                               &pAd->PortCfg.MSSID[MAIN_MSSID].CapabilityInfo,
                        1,                               &SsidIe,
                        1,                               &SsidLen,
                        SsidLen,                         pAd->PortCfg.MSSID[MAIN_MSSID].Ssid,
                        1,                               &SuppIe,
                        1,                               &SupportedRatesLen,
                        SupportedRatesLen,               pAd->PortCfg.SupportedRates,
                        1,                               &DsIe,
                        1,                               &DsLen,
                        1,                               &pAd->PortCfg.Channel,
                        END_OF_ARGS);
    }
    
#ifdef BIG_ENDIAN
    RTMPFrameEndianChange(pAd, pBeaconFrame, DIR_WRITE, FALSE);
#endif

    RTMPWriteTxDescriptor(pTxD, FALSE, CIPHER_NONE, FALSE, FALSE, TRUE, SHORT_RETRY, IFS_BACKOFF, 
        pAd->PortCfg.MlmeRate, 4, FrameLen, pAd->PortCfg.TxPreambleInUsed, 0);
    pAd->Mlme.TimIELocationInBeacon = (UCHAR)FrameLen; 
    pAd->Mlme.CapabilityInfoLocationInBeacon = MAC_HDR_LEN + TIMESTAMP_LEN + 2;
}

/*
    ==========================================================================
    Description:
        Update the BEACON frame in the shared memory. Because TIM IE is variable
        length. other IEs after TIM has to shift and total frame length may change
        for each BEACON period.
    Output:
        pAd->PortCfg.CapabilityInfo
        pAd->Mlme.ErpIeContent
    ==========================================================================
*/
VOID UpdateBeaconFrame(
    IN PRTMP_ADAPTER pAd) 
{
    PTXD_STRUC pTxD = (PTXD_STRUC)pAd->BeaconRing.va_addr;
    MACHDR *pBcnHdr = (MACHDR *)pAd->BeaconRing.va_data_addr;
    UCHAR  *ptr, apidx;
    ULONG FrameLen = pAd->Mlme.TimIELocationInBeacon;

    // Bridge mode doesn't send beacon.
    if (pAd->WdsTab.Mode == WDS_BRIDGE_MODE)
        return;
    
    //
    // step 1 - update BEACON frame's sequence number
    // 
    pAd->Sequence = ((pAd->Sequence) + 1) & (MAX_SEQ_NUMBER);
    pBcnHdr->Seq = pAd->Sequence;
#ifdef BIG_ENDIAN
    // swab 16 bit field - Sequence Control field
    *(USHORT*)((UCHAR*)pBcnHdr + 22) = SWAP16(*(USHORT*)((UCHAR*)pBcnHdr + 22) & 0xFFF0);
#endif

    // step 1.1 - update BEACON's Capability
    ptr = (UCHAR *)pBcnHdr + pAd->Mlme.CapabilityInfoLocationInBeacon;
    *ptr = (UCHAR)(pAd->PortCfg.MSSID[MAIN_MSSID].CapabilityInfo & 0x00ff);
    *(ptr+1) = (UCHAR)((pAd->PortCfg.MSSID[MAIN_MSSID].CapabilityInfo & 0xff00) >> 8);
    
    //
    // step 2 - update TIM IE
    // 
    {
        short   N2;
        UCHAR   i, N1, timlen, multicastbit=0;

        if (pAd->Mlme.DtimCount == 0)
            pAd->Mlme.DtimCount = pAd->Mlme.DtimPeriod - 1;
        else
            pAd->Mlme.DtimCount -= 1;
        ptr = (UCHAR *)pBcnHdr + pAd->Mlme.TimIELocationInBeacon;
        *ptr = IE_TIM;
        *(ptr + 2) = pAd->Mlme.DtimCount;
        *(ptr + 3) = pAd->Mlme.DtimPeriod;

        // handle multicast bit(Aid=0)
        multicastbit = pAd->Mlme.TimBitmap[0] & 0x01;
        pAd->Mlme.TimBitmap[0] &= 0xFE;
        
        // find N1
        for (N1 = 0; N1 < MAX_TIMBITMAP_SIZE; N1++)
        {
            if(pAd->Mlme.TimBitmap[N1] > 0)
            {
                // N1 is even number
                if((N1 % 2) == 1)
                    N1--;
                break;
            }
        }
        if(N1 == MAX_TIMBITMAP_SIZE)
            N1 = 0;
        
        // find N2
        for (N2 = MAX_TIMBITMAP_SIZE - 1; N2 >= 0; N2--)
        {
            if(pAd->Mlme.TimBitmap[N2] > 0)
            {
                break;
            }
        }

        DBGPRINT(RT_DEBUG_INFO, "N1=%d,N2=%d,multicastbit=%d\n", N1,N2, multicastbit);

        // copy partial virtual bitmap
        if ((N1 == 0) && (N2 == -1))
        {
            *(ptr + 5) = 0;
            N2 = 0;
        }
        else
        {
            for (i = N1; i < (N2 - N1 + 1); i++)
            {
                *(ptr + 5 + i) = pAd->Mlme.TimBitmap[N1 + i];
                DBGPRINT(RT_DEBUG_TRACE, "pAd->Mlme.TimBitmap[N1 + %d] = %02x\n", i, pAd->Mlme.TimBitmap[N1 + i]);
            }
        }
        
        // Tim length
        timlen = N2 - N1 + 4;
        *(ptr + 1) = timlen;// IE length

        // bit0 means backlogged mcast/bcast, and others bit means the bitmap offset
        *(ptr + 4) = ((N1/2) | multicastbit);

        // adjust BEACON length according to the new TIM
        FrameLen += (2 + *(ptr+1));

    }

    if ((pAd->PortCfg.PhyMode == PHY_11BG_MIXED) || (pAd->PortCfg.PhyMode == PHY_11G))
    {
        UCHAR ExtendedRatesLen = pAd->PortCfg.SupportedRatesLen - 4;
        
        //
        // step 3 - fill ERP IE
        // 
        ptr = (UCHAR *)pBcnHdr + FrameLen; // pTxD->DataByteCnt;
        *ptr = IE_ERP;
        *(ptr + 1) = 1;
        *(ptr + 2) = pAd->Mlme.ErpIeContent;

        //
        // step 4 - fill up EXTENDED RATE IE. The first 4 rates (1/2/5.5/11) are already in
        //          SUPPORTED RATES IE.
        // 
        *(ptr + 3) = IE_EXT_SUPP_RATES;
        *(ptr + 4) = ExtendedRatesLen;
        NdisMoveMemory(ptr + 5, &pAd->PortCfg.SupportedRates[4], *(ptr + 4));
        ptr      += (3 + 2 + ExtendedRatesLen);
        FrameLen += (3 + 2 + ExtendedRatesLen);
    }

    //
    // step 5 - fill up SSN/WPA IE
    //

    // Since FrameLen may change, update TXD
    RTMPWriteTxDescriptor(pTxD, FALSE, CIPHER_NONE, FALSE, FALSE, TRUE, SHORT_RETRY, IFS_BACKOFF, 
        pAd->PortCfg.MlmeRate, 4, FrameLen, pAd->PortCfg.TxPreambleInUsed, 0);

    //
    // step 6 - if DTIM, then move backlogged bcast/mcast frames from PSQ to TXQ whenever DtimCount==0
    // NOTE: This updated BEACON frame will be sent at "next" TBTT instead of at cureent TBTT. The reason is
    //       because ASIC already fecth the BEACON content down to TX FIFO before driver can make any
    //       modification. To compenstate this effect, the actual time to deilver PSQ frames will be
    //       at the time that we wrapping around DtimCount from 0 to DtimPeriod-1
    if ((pAd->Mlme.DtimCount + 1) == pAd->Mlme.DtimPeriod)
    {
        PQUEUE_ENTRY	pEntry;

        for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
        {
            NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
            NdisAcquireSpinLock(&pAd->TxSwQueueLock);
            while (pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue.Head)
            {
                pEntry = RemoveHeadQueue(&pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue);
                InsertTailQueue(&pAd->TxSwQueue0, pEntry);
                DBGPRINT(RT_DEBUG_TRACE, "DTIM=%d/%d, tx mcast/bcast out...\n", pAd->Mlme.DtimCount, pAd->Mlme.DtimPeriod);
            }
            NdisReleaseSpinLock(&pAd->TxSwQueueLock);
            NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
            pAd->PortCfg.MSSID[apidx].MacTab.PsQIdleCount = 0;
        }
        pAd->Mlme.TimBitmap[0] &= ~(0x01);
        
        // Dequeue one frame from TxSwQueue0..3 queue and process it
        	RTMPDeQueuePacket(pAd);        	
    }
}

/*  ==========================================================================
    Description:
    ==========================================================================
*/
VOID MlmeCntlConfirm(
    IN PRTMP_ADAPTER pAd, 
    IN ULONG MsgType, 
    IN USHORT Msg) 
{
    MlmeEnqueue(&pAd->Mlme.Queue, MLME_CNTL_STATE_MACHINE, MsgType, sizeof(USHORT), &Msg);
}

/*
    ==========================================================================
    Description:
    ==========================================================================
*/
VOID ScanParmFill(
    IN PRTMP_ADAPTER pAd, 
    IN OUT MLME_SCAN_REQ_STRUCT *ScanReq, 
    IN CHAR Ssid[], 
    IN UCHAR SsidLen, 
    IN UCHAR BssType, 
    IN UCHAR ScanType) 
{
    ScanReq->SsidLen = SsidLen;
    NdisMoveMemory(ScanReq->Ssid, Ssid, SsidLen);
    ScanReq->BssType = BssType;
    ScanReq->ScanType = ScanType;
}
