/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rtmp_data.c

    Abstract:
    Data path subroutines

    Revision History:
    Who         When            What
    --------    ----------      ----------------------------------------------
    Paul Lin    08-01-2002      created
    Paul Lin    07-01-2003      add encryption/decryption data flow
    John Chang  08-05-2003      modify 802.11 header for AP purpose
*/
#include "rt_config.h"

static const UCHAR Bit[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};
        
// static  UCHAR   SNAP_802_1H[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};
UCHAR   SNAP_802_1H[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00};

static  UCHAR   SNAP_BRIDGE_TUNNEL[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8};

// static	UCHAR   EAPOL[] = {0x88, 0x8e};
UCHAR   EAPOL[] = {0x88, 0x8e};


static  UCHAR   TPID[] = {0x81, 0x00};
static  UCHAR   IPX[] = {0x81, 0x37};
static  UCHAR   APPLE_TALK[] = {0x80, 0xf3};
static  UCHAR   PlcpSignal[12] = { 
     0, /* RATE_1 */    1, /* RATE_2 */     2, /* RATE_5_5 */   3, /* RATE_11 */    // see BBP spec
    11, /* RATE_6 */   15, /* RATE_9 */    10, /* RATE_12 */   14, /* RATE_18 */    // see IEEE802.11a-1999 p.14
     9, /* RATE_24 */  13, /* RATE_36 */    8, /* RATE_48 */   12  /* RATE_54 */ }; // see IEEE802.11a-1999 p.14

/*
    ========================================================================
    Routine Description:
        Process RxDone interrupt, running in DPC level

    Arguments:
        pAdapter    Pointer to our adapter

    Return Value:
        None

    Note:
        This routine has to maintain Rx ring read pointer.
    ========================================================================
*/
VOID    RTMPHandleRxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PRXD_STRUC      pRxD;
#ifdef BIG_ENDIAN
    PRXD_STRUC      pDestRxD;
    RXD_STRUC       RxD;
#endif
    PHEADER_802_11  pHeader;
    PUCHAR          pData;
    PVOID           pManage;
#ifndef VINCENT_WDSTKIP
    PUCHAR          pDestMac, pSrcMac;
#endif
    UCHAR           Count, KeyIdx, apidx = MAIN_MSSID;
    PWPA_KEY        pWpaKey = (PWPA_KEY) NULL;
    NDIS_STATUS     Status;
    BOOLEAN         bDropFrame;
    //ULONG           RegValue, Address;
    //ULONG           HwDecryptIndex;
    MAC_TABLE_ENTRY *pEntry;
    NDIS_802_11_WEP_STATUS	EncryptType = Ndis802_11EncryptionDisabled;

    // Make sure Rx ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->RxRingLock);
#if 0
    // Verify Hardware Decryption pointer with Software Decryption pointer
    RTMP_IO_READ32(pAdapter, SECCSR0, &RegValue);
    HwDecryptIndex = (RegValue - pAdapter->RxRing[0].pa_addr) / RING_DESCRIPTOR_SIZE;

	Address = pAdapter->RxRing[pAdapter->CurDecryptIndex].pa_addr;
    if (Address != RegValue)
    {
        DBGPRINT(RT_DEBUG_ERROR,"pAdapter->CurDecryptIndex =  %d, \n", pAdapter->CurDecryptIndex);
        DBGPRINT(RT_DEBUG_ERROR,"Decrypt pointer not matched SW = 0x%x, HW = 0x%x\n", Address, RegValue);
        DBGPRINT(RT_DEBUG_ERROR,"Sw Decr Ptr = %d, Rx ptr = %d Hw ptr = %d\n",
            pAdapter->CurDecryptIndex, pAdapter->CurRxIndex, HwDecryptIndex);
    }
#endif

    Count = 0;
    do
    {
        // Point to Rx indexed rx ring descriptor
#ifndef BIG_ENDIAN
        pRxD = (PRXD_STRUC) pAdapter->RxRing[pAdapter->CurRxIndex].va_addr;
#else
        pDestRxD = (PRXD_STRUC) pAdapter->RxRing[pAdapter->CurRxIndex].va_addr;
        RxD = *pDestRxD;
        pRxD = &RxD;
        RTMPDescriptorEndianChange((PUCHAR)pRxD, TYPE_RXD);
#endif
        // Initialize drop frame flag
        bDropFrame = FALSE;

        // In case of false alarm or processed at last instance
        if (pRxD->Owner != DESC_OWN_HOST)
        {
            break;
        }

        // Decrypt engine stuck
        if (pRxD->CipherOwner != DESC_OWN_HOST)
        {
            pAdapter->RalinkCounters.RxRingErrCount++;
            break;
        }
        
#ifdef RALINK_ATE
        if(pAdapter->ate.Mode == ATE_RXFRAME)
        {
            bDropFrame = TRUE;
        }
#endif	//#ifdef RALINK_ATE
        
        // Point to Rx ring buffer where stores the real data frame
        pData   = (PUCHAR) (pAdapter->RxRing[pAdapter->CurRxIndex].va_data_addr);
        // Cast to 802.11 header for flags checking
        pHeader = (PHEADER_802_11) pData;

#ifdef BIG_ENDIAN
        RTMPFrameEndianChange(pAdapter, (PUCHAR)pHeader, DIR_READ, TRUE);
#endif

        // Increase Total receive byte counter after real data received no mater any error or not
        pAdapter->RalinkCounters.ReceivedByteCount +=  pRxD->DataByteCnt;
        
        // Check for all RxD errors
        Status = RTMPCheckRxDescriptor(pRxD);
        
        // Paul 04-03 for OFDM Rx length issue
        if (pRxD->DataByteCnt > 1600)
            Status = NDIS_STATUS_FAILURE;
            
        if (Status == NDIS_STATUS_SUCCESS)
        {
            INC_COUNTER(pAdapter->WlanCounters.ReceivedFragmentCount);
            pAdapter->Counters.GoodReceives++;

			// collect current antenna's average RSSI
			if (pRxD->U2M)
			{
    			if (pAdapter->PortCfg.RxAnt.PrimaryInUsed)
    			{
    			    short AvgRssi = pAdapter->PortCfg.RxAnt.AvgRssi[pAdapter->PortCfg.RxAnt.PrimaryRxAnt];
    			    if (AvgRssi > 0) 
    			        AvgRssi = AvgRssi - (AvgRssi>>3) + pRxD->BBR1;
    			    else
    			        AvgRssi = pRxD->BBR1<<3;
    			    pAdapter->PortCfg.RxAnt.AvgRssi[pAdapter->PortCfg.RxAnt.PrimaryRxAnt] = AvgRssi;
    			}
    			else
    			{
    			    short AvgRssi = pAdapter->PortCfg.RxAnt.AvgRssi[pAdapter->PortCfg.RxAnt.SecondaryRxAnt];
    			    if (AvgRssi > 0) 
    			        AvgRssi = AvgRssi - (AvgRssi>>3) + pRxD->BBR1;
    			    else
    			        AvgRssi = pRxD->BBR1<<3;
    			    pAdapter->PortCfg.RxAnt.AvgRssi[pAdapter->PortCfg.RxAnt.SecondaryRxAnt] = AvgRssi;
    			}
			}
        }
        else
        {
            pAdapter->Counters.RxErrors++;
        }

        // All frames to AP are directed except probe_req. IEEE 802.11/1999 - p.463
        // Do this before checking "duplicate frame".
        // 2003-08-20 accept BEACON to decide if OLBC (Overlapping Legacy BSS Condition) happens
        if (pRxD->U2M)
		{
            if (pHeader->Controlhead.Frame.Type == BTYPE_DATA)
            {
                pAdapter->PortCfg.AvgRssi  = (pAdapter->PortCfg.AvgRssi * 7 + pRxD->BBR1) >> 3;
                pAdapter->PortCfg.NumOfAvgRssiSample ++;
            }
            // ignore all CNTL frames except PS-POLL
            if ((pHeader->Controlhead.Frame.Type == BTYPE_CNTL) && (pHeader->Controlhead.Frame.Subtype != SUBTYPE_PS_POLL))
                Status = NDIS_STATUS_FAILURE;
        }
        else if ((pHeader->Controlhead.Frame.Type == BTYPE_MGMT) && 
            ((pHeader->Controlhead.Frame.Subtype == SUBTYPE_BEACON) || 
             (pHeader->Controlhead.Frame.Subtype == SUBTYPE_PROBE_REQ)))
            ;
        else
            Status = NDIS_STATUS_FAILURE;

        // Gather PowerSave information from all DATA frames. IEEE 802.11/1999 p.461
        if ((Status == NDIS_STATUS_SUCCESS) && (pRxD->U2M) && (pHeader->Controlhead.Frame.Type == BTYPE_DATA))
        {
            if (pHeader->Controlhead.Frame.PwrMgt)
                PsIndicate(pAdapter, &pHeader->Controlhead.Addr2, pRxD->BBR1, PWR_SAVE);
            else
                PsIndicate(pAdapter, &pHeader->Controlhead.Addr2, pRxD->BBR1, PWR_ACTIVE);
        }
        
        // Duplicate frame check. This check only apply to unicast data & management frames
        if ((Status == NDIS_STATUS_SUCCESS) && (pRxD->U2M) && 
            ((pHeader->Controlhead.Frame.Type == BTYPE_DATA) || (pHeader->Controlhead.Frame.Type == BTYPE_MGMT)))
            Status = RTMPCheckDuplicateFrame(pAdapter, pHeader);

        //
        // Do RxD release operation for all failure frames
        //
        if (Status == NDIS_STATUS_SUCCESS)
        {
            // Saved data pointer for management frame which will pass to MLME block
            pManage = (PVOID) pData;

            // pData : Pointer skip the first 24 bytes, 802.11 HEADER
            if((pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
            {
                pData += LENGTH_802_11_WITH_ADDR4;
            }
            else
            {
                pData += LENGTH_802_11;
            }

            //
            // Start of main loop to parse receiving frames.
            // The sequence will be Type first, then subtype...
            //
            switch (pHeader->Controlhead.Frame.Type)
            {
                // Frame with data type
                case BTYPE_DATA:
                    pEntry = MacTableLookup(pAdapter, &pHeader->Controlhead.Addr2);
                    if (pEntry)
                    {
                    	EncryptType = pEntry->WepStatus;
                        apidx = pEntry->ApIdx;
                        INC_COUNTER(pAdapter->PortCfg.MSSID[apidx].RxOkCount);
                        pAdapter->PortCfg.MSSID[apidx].ReceivedByteCount +=  pRxD->DataByteCnt;
                        if (pAdapter->PortCfg.bEnableHSCounter)
                        {
                            pEntry->HSCounter.TotalRxByteCount += pRxD->DataByteCnt;
                            pEntry->HSCounter.LastDataPacketTime = jiffies;
                        }
                    }
                    else
                    	EncryptType = pAdapter->PortCfg.MSSID[MAIN_MSSID].GroupKeyWepStatus;
                    
                    if (pHeader->Controlhead.Frame.ToDs == 0)
                    {
                        bDropFrame = TRUE;
                        break;
                    }

                    // check if Class2 or 3 error
                    if(!(pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
                    {
						if (RTMPCheckClass2Class3Error(pAdapter, pHeader))
						{
							bDropFrame = TRUE;
							break;
						}
                    }

                    // Drop NULL, CF-ACK(no data), CF-POLL(no data), and CF-ACK+CF-POLL(no data) data frame
                    if (pHeader->Controlhead.Frame.Subtype & 0x04) // bit 2 : no DATA
                    {
                        bDropFrame = TRUE;
                        break;
                    }
                    
                    // Find the WPA key, either Group or Pairwise Key
                    if (EncryptType >= Ndis802_11Encryption2Enabled)
                    {
                        pWpaKey = (PWPA_KEY) NULL;
                        
						if ((pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
						{
							pWpaKey = (PWPA_KEY) &pAdapter->WdsTab.Wpa_key;
						}
						else
						{
							if (pHeader->Controlhead.Addr1.Octet[0] &0x01)
							{
								if (pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0)
								{
									pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];
								}
							}                       
							else
							{
								if (pEntry && (pEntry->PairwiseKey[0].KeyLen != 0))
								{
									pWpaKey = (PWPA_KEY) &pEntry->PairwiseKey[0];
								}
								// Use default Group Key if there is no Pairwise key present
								if ((pWpaKey == NULL) && (pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0))
								{
									pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];               
								}
							}
						}
                    }
                    // Begin process unicast to me frame
                    pRxD->CipherAlg = CIPHER_NONE;
                    if  (pRxD->U2M)
                    {
                        //
                        // Begin frame processing
                        //

#ifndef VINCENT_WDSTKIP
						// WDS packet's DA & SA is not the same
						if ((pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
						{
							pDestMac = (PUCHAR)	&(pHeader->Addr3);
							pSrcMac	= (PUCHAR)pHeader + sizeof(HEADER_802_11);
						}
						else
						{
							pDestMac = (PUCHAR) &(pHeader->Addr3);
							pSrcMac = (PUCHAR) &(pHeader->Controlhead.Addr2);
						}
#endif

                        // WEP encrypted frame
                        if (EncryptType == Ndis802_11Encryption1Enabled)    // WEP
                        {
                            if (pHeader->Controlhead.Frame.Wep)
                            {
                                NdisMoveMemory((PUCHAR) &pRxD->Iv, pData, 4);   //Get WEP IV
                                KeyIdx = (*(pData + 3) & 0xc0) >> 6;
                                NdisMoveMemory(pRxD->Key, pAdapter->PortCfg.MSSID[apidx].SharedKey[KeyIdx].Key, pAdapter->PortCfg.MSSID[apidx].SharedKey[KeyIdx].KeyLen);                                 
                                if (pAdapter->PortCfg.MSSID[apidx].SharedKey[KeyIdx].KeyLen == 5)
                                    pRxD->CipherAlg = CIPHER_WEP64;
                                else
                                    pRxD->CipherAlg = CIPHER_WEP128;                                    
                            }
                            else if (pAdapter->PortCfg.MSSID[apidx].PrivacyFilter == Ndis802_11PrivFilter8021xWEP)
                            {
                                // In Ndis802_11PrivFilter8021xWEP, this frame is not encrypted,
                                //Check 802.1x frame, if not drop it.
                                if (!RTMPEqualMemory(EAPOL, pData + 6, 2))
                                {
                                    // Not 802.1X frames
                                    // Add error counter
                                    Status = NDIS_STATUS_FAILURE;
                                    bDropFrame = TRUE;
                                    break;
                                }
                            }
                        }
                        else if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL))    // TKIP
                        {
                            if (pHeader->Controlhead.Frame.Wep)
                            {
								UCHAR	Eiv_Tmp[4];
								
                                NdisMoveMemory((PUCHAR) &pRxD->Iv, pData, 4);           //Get  IV
								// Swap EIV byte order, due to ASIC's bug.
								Eiv_Tmp[0] = *(pData + 7);
								Eiv_Tmp[1] = *(pData + 6);
								Eiv_Tmp[2] = *(pData + 5);
								Eiv_Tmp[3] = *(pData + 4);								
								NdisMoveMemory((PUCHAR) &pRxD->Eiv, Eiv_Tmp, 4);	//Get WEP EIV
                                KeyIdx = (*(pData + 3) & 0xc0) >> 6;
                                // Copy TA into RxD
                                NdisMoveMemory(pRxD->TA, &pHeader->Controlhead.Addr2, 6);
                                NdisMoveMemory(pRxD->Key, pWpaKey->Key, 16);                                    
                                pRxD->CipherAlg = CIPHER_TKIP;
                            }
                            else if (pAdapter->PortCfg.MSSID[apidx].PrivacyFilter == Ndis802_11PrivFilter8021xWEP)
                            {
                                // In Ndis802_11PrivFilter8021xWEP, this frame is not encrypted,
                                // Check 802.1x frame, if not drop it.
                                if (!RTMPEqualMemory(EAPOL, pData + 6, 2))
                                {
                                    // Not 802.1X frames
                                    // Add error counter
                                    Status = NDIS_STATUS_FAILURE;
                                    bDropFrame = TRUE;
                                    break;
                                }
                            }
                        }
                        else if ((EncryptType == Ndis802_11Encryption3Enabled) && (pWpaKey != NULL))    // AES
                        {
                            if (pHeader->Controlhead.Frame.Wep)
                            {
                                NdisMoveMemory((PUCHAR) &pRxD->Iv, pData, 4);           //Get IV
                                NdisMoveMemory((PUCHAR) &pRxD->Eiv, (pData + 4), 4);    //Get  EIV
                                // Copy TA into RxD
                                NdisMoveMemory(pRxD->TA, &pHeader->Controlhead.Addr2, 6);                               
                                KeyIdx = (*(pData + 3) & 0xc0) >> 6;
                                NdisMoveMemory(pRxD->Key, pWpaKey->Key, 16);                                    
                                pRxD->CipherAlg = CIPHER_AES;
                            }
                            else if (pAdapter->PortCfg.MSSID[apidx].PrivacyFilter == Ndis802_11PrivFilter8021xWEP)
                            {
                                // In Ndis802_11PrivFilter8021xWEP, this EAPOL frame is not encrypted,
                                // Check 802.1x frame, if not drop it.
                                if (!RTMPEqualMemory(EAPOL, pData + 6, 2))
                                {
                                    // Not 802.1X frames
                                    // Add error counter
                                    Status = NDIS_STATUS_FAILURE;
                                    bDropFrame = TRUE;
                                    break;
                                }
                            }
                        }
                        else if (pHeader->Controlhead.Frame.Wep)
                        {
                            // Drop WEP frame when WepStatus is FALSE
                            Status = NDIS_STATUS_FAILURE;
                            bDropFrame = TRUE;
                            break;                            
                        }                       
                        else    // Not encryptrd frames
                        {
                            pRxD->CipherAlg = CIPHER_NONE;
                        }
                    }
                    break;

                case BTYPE_MGMT:
                    if (pHeader->Controlhead.Frame.Wep)
                    {
                        UCHAR *pPayload = (UCHAR *)pManage + LENGTH_802_11;

                        // must be SHARED_KEY AUTH seq#3, decryption starts after MAC header
                        if (RTMPDecryptData(pAdapter, apidx, pPayload, pRxD->DataByteCnt - LENGTH_802_11) == FALSE)
                            bDropFrame = 1;
                        else 
                            pRxD->DataByteCnt -= 8;  // Minus WEP size
                    }
                        
                    // Always None encrypted
                    pRxD->CipherAlg = CIPHER_NONE;
                    break;

                case BTYPE_CNTL:
                    // Always None encrypted; must be PS-POLL
                    pRxD->CipherAlg = CIPHER_NONE;
                    break;

                default :
                    bDropFrame = TRUE;
                    break;
            }
        }
        else
            bDropFrame = TRUE;

        // Packet will still do NULL cipher operation and drop afterward
        if (bDropFrame == TRUE)
        {
            pRxD->Drop      = 1;
            pRxD->CipherAlg = CIPHER_NONE;
        }
        else
        {
            pRxD->Drop      = 0;
            if ((pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
            {
                pRxD->IvOffset = LENGTH_802_11_WITH_ADDR4;
            }
            else
            {
                pRxD->IvOffset = LENGTH_802_11;
            }
        }
        
        pRxD->CipherOwner = DESC_OWN_NIC;

#ifdef BIG_ENDIAN
        RTMPFrameEndianChange(pAdapter, (PUCHAR)pHeader, DIR_WRITE, TRUE);
        RTMPDescriptorEndianChange((PUCHAR)pRxD, TYPE_RXD);
        //*pDestRxD = RxD;
        WriteBackToDescriptor((PUCHAR)pDestRxD, (PUCHAR)pRxD, TRUE, TYPE_RXD);
#endif

        pAdapter->CurRxIndex++;
        if (pAdapter->CurRxIndex >= RX_RING_SIZE)
        {
            pAdapter->CurRxIndex = 0;
        }
        Count++;
        
        pAdapter->RalinkCounters.RxCount ++;
        
    }   while (Count < MAX_RX_PROCESS);
        
    // Kick Decrypt Control Register, based on ASIC's implementation
    // We have to kick decrypt & encrypt every frame.
    RTMP_IO_WRITE32(pAdapter, SECCSR0, 0x1);

    // Make sure to release Rx ring resource
    NdisReleaseSpinLock(&pAdapter->RxRingLock);
}

/*
    ========================================================================

    Routine Description:
        Process TxRing TxDone interrupt, running in DPC level

    Arguments:
        Adapter     Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPHandleTxRingTxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PTXD_STRUC      pTxD;
#ifdef	BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif
    PHEADER_802_11  pHeader;
    UCHAR           Count;
    
    // Make sure Tx ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->TxRingLock);
    
    Count = 0;
    do
    {
#ifndef BIG_ENDIAN
        pTxD = (PTXD_STRUC) (pAdapter->TxRing[pAdapter->NextTxDoneIndex].va_addr);
#else
        pDestTxD = (PTXD_STRUC) (pAdapter->TxRing[pAdapter->NextTxDoneIndex].va_addr);
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
        pHeader = (PHEADER_802_11) (pAdapter->TxRing[pAdapter->NextTxDoneIndex].va_data_addr);

        // Check for the descriptor ownership
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->CipherOwn == DESC_OWN_NIC) || (pTxD->Valid == FALSE))
        {
            break;
        }
        
        RTMPHardTransmitDone(
            pAdapter, 
            pTxD, 
            &pHeader->Controlhead.Addr1,
            pAdapter->TxRing[pAdapter->NextTxDoneIndex].FrameType);
        
        // It might happend with no Ndis packet to indicate back to upper layer
        // Clear for NdisSendComplete request
        pTxD->Valid = FALSE;
        
        // Increase Total transmit byte counter after real data sent out
        pAdapter->RalinkCounters.TransmittedByteCount +=  pTxD->DataByteCnt;
        pAdapter->RalinkCounters.TxDoneCount++;

#ifdef BIG_ENDIAN
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
        *pDestTxD = TxD;
#endif
        
        pAdapter->NextTxDoneIndex++;
        if (pAdapter->NextTxDoneIndex >= TX_RING_SIZE)          
        {
            pAdapter->NextTxDoneIndex = 0;
            
            // Group Rekey counter based on packets number tramsmitted
            if ((pAdapter->PortCfg.WPAREKEY.ReKeyMethod == PKT_REKEY)
                && (pAdapter->PortCfg.REKEYTimerRunning == TRUE))
            {   
                pAdapter->PortCfg.REKEYCOUNTER++;
            }
        }
    }   while (++Count < MAX_TX_PROCESS);

#ifdef RALINK_ATE
    if((pAdapter->ate.Mode == ATE_TXCONT) || (pAdapter->ate.Mode == ATE_TXCARR) || ((pAdapter->ate.Mode == ATE_TXFRAME)))
    {
    	if (pAdapter->ate.TxDoneCount < pAdapter->ate.TxCount)
    	{
        	pAdapter->ate.TxDoneCount++;
        	pTxD = (PTXD_STRUC)pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;

        	RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
            	SHORT_RETRY, IFS_BACKOFF, pAdapter->ate.TxRate, 4,
            	pAdapter->ate.TxLength, pAdapter->PortCfg.TxPreambleInUsed, 0);

        	pAdapter->CurEncryptIndex++;
        	if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
        	{
            	pAdapter->CurEncryptIndex = 0;
        	}

        	RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
		}
    	else if (pAdapter->ate.Mode == ATE_TXFRAME)
    	{
    		DBGPRINT(RT_DEBUG_TRACE, "ATE TXFRAME completed!\n");
    	}
	}
#endif	//#ifdef RALINK_ATE

    // Make sure to release Tx ring resource
    NdisReleaseSpinLock(&pAdapter->TxRingLock);

#if 1

    if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)) &&
      (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)))
    {
      // Some Tx ring resource freed, check for pending send frame for hard transmit
      RTMPDeQueuePacket(pAdapter);
    } 
    
#else 
     
    // Some Tx ring resource freed, check for pending send frame for hard transmit
    if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && 
        (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RADIO_OFF)) &&
        (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
    {
        if(pAdapter->bNetDeviceStopQueue)
        {
            if(pAdapter->TxSwQueue0.Number < (MAX_PACKETS_IN_QUEUE >> 2))
            {
                DBGPRINT(RT_DEBUG_TRACE, "NetDevice start queue!!!\n\n");
                pAdapter->bNetDeviceStopQueue = FALSE;
                //netif_start_queue(pAdapter->net_dev);
				netif_wake_queue(pAdapter->net_dev);
            }
        }
        RTMPDeQueuePacket(pAdapter);
    }
#endif    
}

/*
    ========================================================================

    Routine Description:
        Process Priority ring TxDone interrupt, running in DPC level

    Arguments:
        Adapter     Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPHandlePrioRingTxDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif
    PHEADER_802_11  pHeader;
    UCHAR           Count;
    PMGMT_STRUC     pMgmt;
    
    // Make sure Prio ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->PrioRingLock);   
    
    Count = 0;
    do
    {
#ifndef BIG_ENDIAN
        pTxD = (PTXD_STRUC) (pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].va_addr);
#else
        pDestTxD = (PTXD_STRUC) (pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].va_addr);
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
        pHeader = (PHEADER_802_11) (pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].va_data_addr);

        // Check for the descriptor ownership
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->Valid == FALSE))
        {
#ifdef BIG_ENDIAN
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
#endif
            break;
        }
        
        // No need to put in reply for MLME
        RTMPHardTransmitDone(
            pAdapter, 
            pTxD, 
            &pHeader->Controlhead.Addr1,
            pAdapter->PrioRing[pAdapter->NextPrioDoneIndex].FrameType);
        
        // It might happend with no Ndis packet to indicate back to upper layer
        pTxD->Valid = FALSE;        
        
        // Increase Total transmit byte counter after real data sent out
        pAdapter->RalinkCounters.TransmittedByteCount +=  pTxD->DataByteCnt;

#ifdef BIG_ENDIAN
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
        *pDestTxD = TxD;
#endif

        pAdapter->NextPrioDoneIndex++;
        if (pAdapter->NextPrioDoneIndex >= PRIO_RING_SIZE)
        {
            pAdapter->NextPrioDoneIndex = 0;
        }
    }   while (++Count < MAX_TX_PROCESS);

    // Make sure to release Prio ring resource
    NdisReleaseSpinLock(&pAdapter->PrioRingLock);   
    
    if (RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RADIO_OFF))
        return;
    
    if (pAdapter->PushMgmtIndex != pAdapter->PopMgmtIndex)
    {
        if (RTMPFreeDescriptorRequest(pAdapter, PRIO_RING, 1) == NDIS_STATUS_SUCCESS)
        {
            pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PopMgmtIndex];
            if (pMgmt->Valid == TRUE)
            {
                MlmeHardTransmit(pAdapter, pMgmt->pBuffer, pMgmt->Length);
                MlmeConfirm((PVOID)pAdapter, pMgmt->pBuffer, 0);
                pMgmt->Valid = FALSE;
                pAdapter->PopMgmtIndex++;
                pAdapter->MgmtQueueSize--;
                if (pAdapter->PopMgmtIndex >= MGMT_RING_SIZE)
                {
                    pAdapter->PopMgmtIndex = 0;
                }
            }
        }
    }   
}


/*
    ========================================================================

    Routine Description:
        Process Rx ring DecryptionDone interrupt, running in DPC level

    Arguments:
        Adapter     Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPHandleDecryptionDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PRXD_STRUC      pRxD;
#ifdef BIG_ENDIAN
    PRXD_STRUC      pDestRxD;
    RXD_STRUC       RxD;
#endif
    PHEADER_802_11  pHeader;
    UINT            PacketSize;
    PVOID           pManage;
    UCHAR           Header802_3[14];
    UCHAR           LLC_Len[2];
    UCHAR           Count, apidx = MAIN_MSSID;
    ULONG           RegValue;
    ULONG           HwDecryptIndex;
    PUCHAR          pDestMac, pSrcMac;
    PUCHAR          pData;
    PUCHAR          pEncap;
    NDIS_STATUS     Status;
    PWPA_KEY        pWpaKey = (PWPA_KEY) NULL;
    MAC_TABLE_ENTRY *pEntry;
    NDIS_802_11_PRIVACY_FILTER   Privacy;
    WPA_STATE		WpaState;
    struct sk_buff	*skb;
    struct net_device	*net_dev = pAdapter->net_dev;
    ULONG           i;
    BOOLEAN         bWdsPacket;
    USHORT          TCI;
    NDIS_802_11_WEP_STATUS	EncryptType = Ndis802_11EncryptionDisabled;

    // Make sure Rx ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->RxRingLock);
    
    RTMP_IO_READ32(pAdapter, SECCSR0, &RegValue);
    HwDecryptIndex = (RegValue - pAdapter->RxRing[0].pa_addr) / RING_DESCRIPTOR_SIZE;

    Count = 0;
    //do
    while (pAdapter->CurDecryptIndex != HwDecryptIndex)
    {
        // Point to Rx indexed rx ring descriptor
#ifndef BIG_ENDIAN
        pRxD = (PRXD_STRUC) pAdapter->RxRing[pAdapter->CurDecryptIndex].va_addr;
#else
        pDestRxD = (PRXD_STRUC) pAdapter->RxRing[pAdapter->CurDecryptIndex].va_addr;
        RxD = *pDestRxD;
        pRxD = &RxD;
        RTMPDescriptorEndianChange((PUCHAR)pRxD, TYPE_RXD);
#endif
    
        // In case of false alarm or processed at last instance
        if (pRxD->Owner != DESC_OWN_HOST)
        {
            break;
        }
            
        if (pRxD->CipherOwner != DESC_OWN_HOST)
        {
            break;
        }
    
        // Point to Rx ring buffer where stores the real data frame
        pData   = (PUCHAR) (pAdapter->RxRing[pAdapter->CurDecryptIndex].va_data_addr);
        // Cast to 802.11 header for flags checking
        pHeader = (PHEADER_802_11) pData;

#ifdef BIG_ENDIAN
        RTMPFrameEndianChange(pAdapter, (PUCHAR)pHeader, DIR_READ, FALSE);
#endif
        // Patch for ASIC: There are some occasions ASIC will false turned on this bit
        // Driver will check the decrypt algorithm and decide whether this ICV is true or not
        if ((pRxD->IcvError == 1) && (pRxD->CipherAlg == CIPHER_NONE))
            pRxD->IcvError = 0;
		
        // Since we already process header at RxDone interrupt, there is no need to proces
        // header sanity again, the only thing we have to check is icv_err bit
/*        if (pRxD->IcvError == 1)
        {
            DBGPRINT(RT_DEBUG_TRACE,"DecryptDone - ICV error\n");
            pRxD->Drop =1;          // Drop frame with icv error
        }*/
            
        // Saved data pointer for management frame which will pass to MLME block
        pManage = (PVOID) pData;

        // pData : Pointer skip the first 24 bytes, 802.11 HEADER
        pData += LENGTH_802_11;

        bWdsPacket = FALSE;
		if ((pRxD->Drop == 0) && (pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
		{
        	// Lazy WDS mode, auto learning, match with FrDs=1 and ToDs=1
	        if (pAdapter->WdsTab.Mode == WDS_LAZY_MODE)
	        {
	            // Lookup entry, if exist, then update the Rcv beacon time and return
	            for (i = 0; i < MAX_WDS_ENTRY; i++)
	            {
	                if ((pAdapter->WdsTab.WdsEntry[i].Valid == TRUE) && (MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, &pHeader->Controlhead.Addr2)))
	                {
	                    // Update wds received beacon time for age out usage
	                    pAdapter->WdsTab.WdsEntry[i].WdsRcvBeaconTime = jiffies;
	                    break;
	                }
	            }

	            // Look for empty entry, and add to our WdsTable
	            if (i == MAX_WDS_ENTRY)
	            {
		            for (i = 0; i < MAX_WDS_ENTRY; i++)
		            {
		                if (pAdapter->WdsTab.WdsEntry[i].Valid == FALSE)
		                {
		                    COPY_MAC_ADDR(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, &pHeader->Controlhead.Addr2);
		                    pAdapter->WdsTab.WdsEntry[i].WdsRcvBeaconTime = jiffies;

		                    pAdapter->WdsTab.WdsEntry[i].CurrTxRate = RATE_11;
		                    if(pAdapter->PortCfg.PhyMode == PHY_11B)
		                        pAdapter->WdsTab.WdsEntry[i].MaxSupportedRate = RATE_11;
		                    else
		                        pAdapter->WdsTab.WdsEntry[i].MaxSupportedRate = RATE_54;
		                    pAdapter->WdsTab.WdsEntry[i].Valid = TRUE;
		                    
		                    DBGPRINT(RT_DEBUG_TRACE, "Lazy WDS mode, AP(%02x:%02x:%02x:%02x:%02x:%02x) has 4-addr packet and add to entry(%d) in our WDSTable \n", 
		                        pHeader->Controlhead.Addr2.Octet[0], pHeader->Controlhead.Addr2.Octet[1], pHeader->Controlhead.Addr2.Octet[2],
		                        pHeader->Controlhead.Addr2.Octet[3], pHeader->Controlhead.Addr2.Octet[4], pHeader->Controlhead.Addr2.Octet[5], i);
		                    break;
		                }
		            }
		            if (i == MAX_WDS_ENTRY)
		            {
		                DBGPRINT(RT_DEBUG_WARN, "WDS Table is full!\n");
		            }
	            }
	        }
	        
			if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
			{
				for (i = 0; i < MAX_WDS_ENTRY; i++)
				{
					if ((pAdapter->WdsTab.WdsEntry[i].Valid == TRUE) && (MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, &pHeader->Controlhead.Addr2)))
					{
						net_dev = pAdapter->WdsTab.WdsEntry[i].dev;
						bWdsPacket = TRUE;
						pData += ETH_LENGTH_OF_ADDRESS;	// offset the addr4
						DBGPRINT(RT_DEBUG_INFO, "WDS packet from (%d)!!!\n", i);
						break;
					}
				}
				if (i == MAX_WDS_ENTRY)
				{
					DBGPRINT(RT_DEBUG_ERROR, "WDS packet, but not our WDS group!!!\n");
					pRxD->Drop = 1;
				}
			}
		}

#if defined(VINCENT_WDSFAST)
		// Insert DA to DA table, find out DA is to MSSID or WDS
		if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
		{
			if ((pRxD->Drop == 0) && (pHeader->Controlhead.Frame.FrDs == 1) && (pHeader->Controlhead.Frame.ToDs == 1))
			{
				// use SA/TA to record DA table
				unsigned char *TA, *SA;
			
				TA = (unsigned char *) pHeader->Controlhead.Addr2.Octet;
				SA = (PUCHAR)pHeader + sizeof(HEADER_802_11);

				Insert_DA(pAdapter, TA, SA);
			}
		}
#endif

        //
        // Start of main loop to parse receiving frames.
        // The sequence will be Type first, then subtype...
        //
        if (pRxD->Drop == 0)
        {
			// The total available payload should exclude 24-byte 802.11 Header
			// If Security is enabled, IV, EIV, ICV size is excluded by ASIC
            if (bWdsPacket)
            {
                PacketSize = pRxD->DataByteCnt - LENGTH_802_11_WITH_ADDR4;
            }
            else
            {
                PacketSize = pRxD->DataByteCnt - LENGTH_802_11;
            }

			// Find the WPA key, either Group or Pairwise Key
			// Although the data has been decrypted by ASIC,
			// driver has to calculate the RxMIC which required the key.
			// The failed case should not happen. If it did, drop it.

			// port access control
			pEntry = PACInquiry(pAdapter, &pHeader->Controlhead.Addr2, &Privacy,&WpaState);
			if(pEntry)
			{
			    apidx = pEntry->ApIdx;
			    EncryptType = pEntry->WepStatus;
			}
			else
				EncryptType = pAdapter->PortCfg.MSSID[MAIN_MSSID].GroupKeyWepStatus;

			// Indicate to virtual I/F(not WDS)
			if (pEntry && (apidx != 0))
			{
			    net_dev = pAdapter->PortCfg.MSSID[apidx].MSSIDDev;
			    DBGPRINT(RT_DEBUG_INFO, "Receive packet to virutal I/F(ra%d)\n", apidx);
			}

			if (EncryptType >= Ndis802_11Encryption2Enabled)
			{
				pWpaKey = (PWPA_KEY) NULL;
            
				if (bWdsPacket)
				{
					Privacy = Ndis802_11PrivFilterAcceptAll;
					pWpaKey = (PWPA_KEY) &pAdapter->WdsTab.Wpa_key;
				}
				else
				{
					if (pHeader->Controlhead.Addr1.Octet[0] &0x01)
					{
						if (pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0)
						{
							pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];
						}
					}
					// Try to find the Pairwise Key
					else
					{
						if (pEntry && (pEntry->PairwiseKey[0].KeyLen != 0))
						{
							pWpaKey = (PWPA_KEY) &pEntry->PairwiseKey[0];
						}
						// Use default Group Key if there is no Pairwise key present
						if ((pWpaKey == NULL) && (pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0))
						{
							pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];               
							DBGPRINT(RT_DEBUG_INFO, "Rx Use Group Key\n");
						}
					}
				}
			}

            switch (pHeader->Controlhead.Frame.Type)
            {
                // Frame with data type
                case BTYPE_DATA:    
                    if(pAdapter->PortCfg.MSSID[apidx].BANClass3Data==TRUE)
                    	break;

					/* unknown source - illegal packet  */
					if (!bWdsPacket && pEntry==NULL)
						break;

					// WDS packet's DA & SA is not the same
					if (bWdsPacket)
					{
						pDestMac = (PUCHAR)	&(pHeader->Addr3);
						pSrcMac	= (PUCHAR)pHeader + sizeof(HEADER_802_11);
					}
					else
					{
						pDestMac = (PUCHAR) &(pHeader->Addr3);
						pSrcMac = (PUCHAR) &(pHeader->Controlhead.Addr2);
					}

                    switch(Privacy)
                    {
                        case Ndis802_11PrivFilter8021xWEP:
                            if (pEntry == NULL)   // this should not happen
                                break;
                            if (!RTMPCheckWPAframe(pAdapter, (UCHAR *)pHeader, pRxD->DataByteCnt))
                                break;
                            
                            if ((pEntry->WpaState < AS_INITPMK)
                            	&& ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2) || pAdapter->PortCfg.MSSID[apidx].IEEE8021X == TRUE))
                            {
                                NdisMoveMemory(LLC_Len, EAPOL, 2);
                                MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                DBGPRINT(RT_DEBUG_TRACE, "Apidx=%d,pData %x %x %x %x %x \n",apidx,*pData,*(pData+1),*(pData+2),*(pData+3),*(pData+4));
#ifdef RTMP_EMBEDDED
                                if ((skb = __dev_alloc_skb(PacketSize + LENGTH_802_3 + 2, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
                                if ((skb = dev_alloc_skb(PacketSize + LENGTH_802_3 + 2)) != NULL)
#endif
                                {
                                    	skb->dev = net_dev;
                                    	skb_reserve(skb, 2);	// 16 byte align the IP header
                                    	memcpy(skb_put(skb, LENGTH_802_3), Header802_3, LENGTH_802_3);
                                    	memcpy(skb_put(skb, PacketSize), pData, PacketSize);
                                    	skb->protocol = eth_type_trans(skb, net_dev);
                                    	netif_rx(skb);
                                    	net_dev->last_rx = jiffies;
                                    	pAdapter->stats.rx_packets++;
                                }
                            }
                            else
                            {
                                MlmeEnqueueForRecv(
                                    pAdapter,
                                    &pAdapter->Mlme.Queue,  
                                    (UCHAR)pRxD->BBR1, 
                                    pRxD->DataByteCnt, 
                                    pManage);
                            }
                            break;
                            
                        case Ndis802_11PrivFilterAcceptAll:
                
                            if (pHeader->Frag == 0)     // First or Only fragment
                            {
                                // For TKIP frame, calculate the MIC value
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {
                                    if (pRxD->CipherAlg == CIPHER_TKIP)
                                    {
                                        // Minus MIC length
                                        PacketSize -= 8;
                                        if (RTMPTkipCompareMICValue(pAdapter, pData, pDestMac, pSrcMac, pWpaKey->RxMic, PacketSize) == FALSE)
                                        {
                                            HandleCounterMeasure(pAdapter, pEntry);
                                            
                                            DBGPRINT(RT_DEBUG_ERROR,"Rx MIC Value error\n");                          
                                            Status = NDIS_STATUS_FAILURE;
                                            break;
                                        } 
                                            //  TODO:  Getting RxTSC from Rx descriptor
                                    }               
                                }                                
                                // Save encapaturation starting pointer
                                pEncap = pData;                             
                                pAdapter->FragFrame.Flags &= 0xFFFFFFFE;                                
                                // Check for encapsulataion other than RFC1042 & Bridge tunnel
                                if ((!RTMPEqualMemory(SNAP_802_1H, pEncap, 6)) && (!RTMPEqualMemory(SNAP_BRIDGE_TUNNEL, pEncap, 6)))
                                {
                                    LLC_Len[0] = PacketSize / 256;
                                    LLC_Len[1] = PacketSize % 256;
                                    MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                }
                                else
                                {
                                    // Remove 802.11 H header & reconstruct 802.3 header
                                    pData += (LENGTH_802_1_H - LENGTH_802_3_TYPE);
                                    if (RTMPEqualMemory(EAPOL, pData, 2)) 
                                    {
                                        PacketSize+=LENGTH_802_11;
                                        if(pEntry )
                                            MlmeEnqueueForRecv(pAdapter, &pAdapter->Mlme.Queue, (CHAR)pRxD->BBR1, PacketSize,  pManage);
                                        break;
                                    }
                                    // Patch for WHQl only, which did not turn on Netbios but use IPX within its payload
                                    if ((RTMPEqualMemory(IPX, pData, 2) || RTMPEqualMemory(APPLE_TALK, pData, 2)) && RTMPEqualMemory(SNAP_802_1H, pEncap, 6))
                                    {
                                        LLC_Len[0] = PacketSize / 256;
                                        LLC_Len[1] = PacketSize % 256;
                                        pData = pData - LENGTH_802_1_H;
                                        MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, ((PUCHAR) LLC_Len));
                                    }
                                    else
                                    {
                                        MAKE_802_3_HEADER(Header802_3, pDestMac, pSrcMac, pData);
                                        // The total available payload should exclude 24-byte 802.11 Header
                                        // and 8-byte 802.2 LLC
                                        PacketSize -= LENGTH_802_1_H;
                                        NdisMoveMemory(pAdapter->FragFrame.Header_LLC, pEncap, 8);
                                        pAdapter->FragFrame.Flags |= 0x01;
                                    }                                
                                    // Point to read 802.3 payload
                                    pData += LENGTH_802_3_TYPE;
                                }                                    
                                // One & The only fragment
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {
                                    pAdapter->PortCfg.LedCntl.fRxActivity = TRUE; // for RX ACTIVITY LED                        
                                   	// check if DA is another associted WSTA reachable via wireless bridging,
                                   	// if it is, then no need to indicate to LLC
									if (RTMPBridgeToWirelessSta(pAdapter, Header802_3, LENGTH_802_3, pData, PacketSize, apidx))
                                   		break;

#if defined(VINCENT_WDSFAST)
									if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
									{
                                   		if (RTMPBridgeToWirelessWDS(pAdapter, Header802_3, LENGTH_802_3, pData, PacketSize, apidx))
                                   			break;
									}
#endif
                    
                                    // Acknolwdge upper layer the received frame.
                                    // Extra reserve 60 byte in packet tail for VLAN tag or other purpose
#ifdef RTMP_EMBEDDED
                                    if ((skb = __dev_alloc_skb(PacketSize + LENGTH_802_3 + 62, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
                                    if ((skb = dev_alloc_skb(PacketSize + LENGTH_802_3 + 62)) != NULL)
#endif
                                    {
                                    	skb->dev = net_dev;
                                    	skb_reserve(skb, 2);	// 16 byte align the IP header
                                    	if (pAdapter->PortCfg.MSSID[apidx].VLAN_VID != 0)
                                        {
#ifdef BIG_ENDIAN
                                            TCI = (pAdapter->PortCfg.MSSID[apidx].VLAN_Priority << 13) + pAdapter->PortCfg.MSSID[apidx].VLAN_VID;
#else
                                            TCI = SWAP_16((pAdapter->PortCfg.MSSID[apidx].VLAN_Priority << 13) + pAdapter->PortCfg.MSSID[apidx].VLAN_VID);
#endif
                                            memcpy(skb_put(skb, 12), Header802_3, 12);
                                            memcpy(skb_put(skb, 2), TPID, 2);
                                            memcpy(skb_put(skb, 2), &TCI, 2);
                                            memcpy(skb_put(skb, 2), &Header802_3[12], 2);
                                        }
                                    	else
                                    	{
                                    	    memcpy(skb_put(skb, LENGTH_802_3), Header802_3, LENGTH_802_3);
                                    	}
                                    	memcpy(skb_put(skb, PacketSize), pData, PacketSize);
                                    	skb->protocol = eth_type_trans(skb, net_dev);
                                    	netif_rx(skb);
                                    	net_dev->last_rx = jiffies;
                                    	pAdapter->stats.rx_packets++;
                                    }
                                    NdisZeroMemory(Header802_3, LENGTH_802_3);                                    
                                }
                                // First fragment of fragmented frames
                                else
                                {
                                    NdisMoveMemory(pAdapter->FragFrame.Buffer,  pData, PacketSize);
                                    NdisMoveMemory(pAdapter->FragFrame.Header802_3, Header802_3, LENGTH_802_3);
                                    NdisZeroMemory(Header802_3, LENGTH_802_3);
                                    pAdapter->FragFrame.RxSize   = PacketSize;
                                    pAdapter->FragFrame.Sequence = pHeader->Sequence;
                                    pAdapter->FragFrame.LastFrag = pHeader->Frag;       // Should be 0
                                }
                            }
                            // Middle & End of fragment burst fragments
                            else
                            {
                                // No LLC-SNAP header in except the first fragment frame                                
                                if ((pHeader->Sequence != pAdapter->FragFrame.Sequence) ||
                                    (pHeader->Frag != (pAdapter->FragFrame.LastFrag + 1)))
                                {
                                    // Fragment is not the same sequence or out of fragment number order
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    Status = NDIS_STATUS_FAILURE;
                                    break;
                                }   
                                else if ((pAdapter->FragFrame.RxSize + PacketSize) > MAX_FRAME_SIZE)
                                {
                                    // Fragment frame is too large, it exeeds the maximum frame size.
                                    // We have to drop it.
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    Status = NDIS_STATUS_FAILURE;
                                    break;
                                }
                                NdisMoveMemory(&pAdapter->FragFrame.Buffer[pAdapter->FragFrame.RxSize], pData, PacketSize);
                                pAdapter->FragFrame.RxSize  += PacketSize;
                                pAdapter->FragFrame.LastFrag = pHeader->Frag;       // Update fragment number
                                    
                                // Last fragment
                                if (pHeader->Controlhead.Frame.MoreFrag == FALSE)
                                {
                                    pAdapter->PortCfg.LedCntl.fRxActivity = TRUE; // for RX ACTIVITY LED
                        
                                    // For TKIP frame, calculate the MIC value
                                    if (pRxD->CipherAlg == CIPHER_TKIP)
                                    {
                                        // Minus MIC length
                                        pAdapter->FragFrame.RxSize -= 8;
                                        
                                        if (pAdapter->FragFrame.Flags & 0x00000001)
                                        {
                                            if (RTMPTkipCompareMICValueWithLLC(
                                                pAdapter,
                                                pAdapter->FragFrame.Header_LLC,
                                                pAdapter->FragFrame.Buffer,
                                                pDestMac,
                                                pSrcMac,
                                                pWpaKey->RxMic,
                                                pAdapter->FragFrame.RxSize) == FALSE)
                                            {
                                                HandleCounterMeasure(pAdapter, pEntry);
                                                
                                                DBGPRINT(RT_DEBUG_ERROR,"Rx MIC Value error 2\n");                            
                                                Status = NDIS_STATUS_FAILURE;
                                                break;
                                            }
                                        }
                                        else
                                        {
                                            if (RTMPTkipCompareMICValue(
                                                pAdapter,
                                                pAdapter->FragFrame.Buffer,
                                                pDestMac,
                                                pSrcMac,
                                                pWpaKey->RxMic,
                                                pAdapter->FragFrame.RxSize) == FALSE)
                                            {
                                                HandleCounterMeasure(pAdapter, pEntry);
                                                
                                                DBGPRINT(RT_DEBUG_ERROR,"Rx MIC Value error 2\n");                            
                                                Status = NDIS_STATUS_FAILURE;
                                                break;
                                            }
                                        }                            
                                            //  TODO:
                                            // Getting RxTSC from Rx descriptor
                                    }

                                   	// check if DA is another associted WSTA reachable via wireless bridging,
                                   	// if it is, then no need to indicate to LLC
									if (RTMPBridgeToWirelessSta(pAdapter, Header802_3, LENGTH_802_3, pData, PacketSize, apidx))
                                   		break;

#if defined(VINCENT_WDSFAST)
									if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
									{
                                   		if (RTMPBridgeToWirelessWDS(pAdapter, Header802_3, LENGTH_802_3, pData, PacketSize, apidx))
                                   			break;
									}
#endif
                    
                                    // Acknolwdge upper layer the received frame
#ifdef RTMP_EMBEDDED
                                    if ((skb = __dev_alloc_skb(pAdapter->FragFrame.RxSize + LENGTH_802_3 + 2, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
                                    if ((skb = dev_alloc_skb(pAdapter->FragFrame.RxSize + LENGTH_802_3 + 2)) != NULL)
#endif
                                    {
                                    	skb->dev = net_dev;
                                    	skb_reserve(skb, 2);	// 16 byte align the IP header
                                    	memcpy(skb_put(skb, LENGTH_802_3), pAdapter->FragFrame.Header802_3, LENGTH_802_3);
                                    	memcpy(skb_put(skb, pAdapter->FragFrame.RxSize), (PVOID)pAdapter->FragFrame.Buffer, pAdapter->FragFrame.RxSize);
                                    	skb->protocol = eth_type_trans(skb, net_dev);
                                    	netif_rx(skb);
                                    	net_dev->last_rx = jiffies;
                                    	pAdapter->stats.rx_packets++;
                                    }
                                    
                                    // Clear Fragment frame contents
                                    NdisZeroMemory(&pAdapter->FragFrame, sizeof(FRAGMENT_FRAME));
                                    DBGPRINT(RT_DEBUG_INFO, "!!! Frame with Fragment Indicated !!!\n");
                                }
                            }
                            break;
                    }
                    break;
                case BTYPE_MGMT:
                    // Read required regsiter for MLME engine
                    // RTMP_IO_READ32(pAdapter, CSR17, &High32TSF);     // TSF value
                    // RTMP_IO_READ32(pAdapter, CSR16, &Low32TSF);      // TSF vlaue
                    if(pAdapter->PortCfg.MSSID[apidx].BANClass3Data==TRUE)
                    {
                        // disallow new association
                        MACFRAME    *Fr = (MACFRAME *)pManage;
                        if ((Fr->Hdr.SubType == SUBTYPE_ASSOC_REQ) ||(Fr->Hdr.SubType == SUBTYPE_AUTH))
                            break;
                        DBGPRINT(RT_DEBUG_INFO, "   Disallow new Assciation \n ");
					}
                
                    // Enqueue this frame to MLME engine
                    MlmeEnqueueForRecv(
                        pAdapter,
                        &pAdapter->Mlme.Queue,  
                        // High32TSF, 
                        // Low32TSF,
                        (UCHAR)pRxD->BBR1, 
                        pRxD->DataByteCnt, 
                        pManage);                   
                    break;
        
                case BTYPE_CNTL:
                    // handle PS-POLL here
                    if ((pRxD->U2M) && (pHeader->Controlhead.Frame.Subtype == SUBTYPE_PS_POLL))
                    {
                        USHORT Aid = pHeader->Controlhead.Duration & 0x3fff;
                        PMACADDR pAddr = &pHeader->Controlhead.Addr2;
                        
                        DBGPRINT(RT_DEBUG_TRACE,"rcv PS-POLL (AID=%d) from %02x:%02x:%02x:%02x:%02x:%02x\n", 
                            Aid, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5]);
                        RTMPHandleRxPsPoll(pAdapter, pAddr, Aid);
                    }
					else
                    {
                        DBGPRINT(RT_DEBUG_TRACE,"ignore CNTL (subtype=%d)\n", pHeader->Controlhead.Frame.Subtype);
                    }
                    break;
    
                default :
                    break;
            }
        }
            
        pAdapter->CurDecryptIndex++;
        if (pAdapter->CurDecryptIndex >= RX_RING_SIZE)
        {
            pAdapter->CurDecryptIndex = 0;
        }
        Count++;
            
        pAdapter->RalinkCounters.DecryptCount ++;
            
        // Clear Cipherowner bit & Rx Owner bit for all drop & non-drop frames
        pRxD->CipherOwner = DESC_OWN_HOST;
        pRxD->Owner       = DESC_OWN_NIC;
#ifdef BIG_ENDIAN
        RTMPDescriptorEndianChange((PUCHAR)pRxD, TYPE_RXD);
        //*pDestRxD = RxD;
        WriteBackToDescriptor((PUCHAR)pDestRxD, (PUCHAR)pRxD, FALSE, TYPE_RXD);
#endif
    }
    //} while (Count < RX_RING_SIZE);
    //} while (pAdapter->CurDecryptIndex != HwDecryptIndex);
        
    // Make sure to release Rx ring resource
    NdisReleaseSpinLock(&pAdapter->RxRingLock);
}

/*
    ========================================================================

    Routine Description:
        Process Tx ring EncryptionDone interrupt, running in DPC level

    Arguments:
        Adapter     Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPHandleEncryptionDoneInterrupt(
    IN  PRTMP_ADAPTER   pAdapter)
{
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif
    UCHAR           Count;
    ULONG           RegValue;
    ULONG           HwEncryptIndex;
    
    // Make sure Prio ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->TxRingLock); 
    
    RTMP_IO_READ32(pAdapter, SECCSR1, &RegValue);
    HwEncryptIndex = (RegValue - pAdapter->TxRing[0].pa_addr) / RING_DESCRIPTOR_SIZE;

    Count = 0;
    //do
    while (pAdapter->NextEncryptDoneIndex != HwEncryptIndex)
    {
#ifndef BIG_ENDIAN
        pTxD = (PTXD_STRUC) (pAdapter->TxRing[pAdapter->NextEncryptDoneIndex].va_addr);
#else
        pDestTxD = (PTXD_STRUC) (pAdapter->TxRing[pAdapter->NextEncryptDoneIndex].va_addr);
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif

        // Check for the descriptor cipher ownership
        if ((pTxD->CipherOwn == DESC_OWN_NIC) || (pTxD->Owner == DESC_OWN_NIC))
        {
            pAdapter->RalinkCounters.TxRingErrCount++;
#ifdef BIG_ENDIAN
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
#endif
            break;
        }

		if (pTxD->IvOffset == LENGTH_802_11_WITH_ADDR4)
		{
			UCHAR	Iv_Tmp[4];
			PUCHAR	pTmp;

			NdisMoveMemory(Iv_Tmp, &pTxD->Iv, 4);
			pTmp        = (PUCHAR) &pTxD->Iv;
			*pTmp       = Iv_Tmp[3];
			*(pTmp + 1) = Iv_Tmp[2];
			*(pTmp + 2) = Iv_Tmp[1];
			*(pTmp + 3) = Iv_Tmp[0];			
		}

		if (pTxD->CipherAlg == CIPHER_TKIP)
		{
			UCHAR	Eiv_Tmp[4];
			PUCHAR	pTmp;

			NdisMoveMemory(Eiv_Tmp, &pTxD->Eiv, 4);
			pTmp        = (PUCHAR) &pTxD->Eiv;
			*pTmp       = Eiv_Tmp[3];
			*(pTmp + 1) = Eiv_Tmp[2];
			*(pTmp + 2) = Eiv_Tmp[1];
			*(pTmp + 3) = Eiv_Tmp[0];			
		}
        // Sanity Check, CurTxIndex should equal to NextEncryptDoneIndex
        // ASSERT(pAdapter->CurTxIndex == pAdapter->NextEncryptDoneIndex);
        
        pTxD->Valid = TRUE;
        pTxD->Owner = DESC_OWN_NIC;

#ifdef BIG_ENDIAN
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
        //*pDestTxD = TxD;
        WriteBackToDescriptor((PUCHAR)pDestTxD, (PUCHAR)pTxD, FALSE, TYPE_TXD);
#endif

        pAdapter->NextEncryptDoneIndex++;
        if (pAdapter->NextEncryptDoneIndex >= TX_RING_SIZE)
        {
            pAdapter->NextEncryptDoneIndex = 0;
        }
        pAdapter->CurTxIndex = pAdapter->NextEncryptDoneIndex;
        pAdapter->RalinkCounters.KickTxCount++;

        if (pAdapter->CurTxIndex == pAdapter->CurEncryptIndex)
            break;
    }
    //} while (++Count < MAX_TX_PROCESS);
    //} while (pAdapter->NextEncryptDoneIndex != HwEncryptIndex);

    // Kick Tx Control Register at the end of all ring buffer preparation
    RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x1);
    
    // Make sure to release Tx ring resource
    NdisReleaseSpinLock(&pAdapter->TxRingLock); 
}

/*
    ========================================================================
    Routine Description:
    Arguments:
        Adapter     Pointer to our adapter
    ========================================================================
*/
VOID    RTMPHandleTbcnInterrupt(IN PRTMP_ADAPTER pAdapter)
{
    UpdateBeaconFrame(pAdapter);
}

/*
    ========================================================================
    Routine Description:
        Process all transmit ring Tx Done interrupt, running    in DPC level

    Arguments:
        Adapter     Pointer to our adapter
    ========================================================================
*/
VOID    RTMPHardTransmitDone(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PTXD_STRUC      pTxD,
    IN  PMACADDR        pAddr,
    IN  UCHAR           FrameType)
{
    MAC_TABLE_ENTRY     *pMacEntry;
    
    switch (pTxD->TxResult)
    {
        case SUCCESS_WITHOUT_RETRY:         // Success without any retry
            // Return send complete status
//          DBGPRINT(RT_DEBUG_INFO, ("TX %02x:%02x:%02x:%02x:%02x:%02x Success without retry<<<\n",
//              pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]));
            if (pTxD->RTS)
            {
                // Increase 802.11 counters
                INC_COUNTER(pAdapter->WlanCounters.RTSSuccessCount);
                pTxD->RTS = 0;
            }
            
            // Increase general counters
            pAdapter->Counters.GoodTransmits++;
            INC_COUNTER(pAdapter->WlanCounters.TransmittedFragmentCount);

            // update DRS related counters
            if (pTxD->ACK && (FrameType == BTYPE_DATA))
            {
                pMacEntry = MacTableLookup(pAdapter, pAddr);
                if (pMacEntry)
               	{
                    pMacEntry->OneSecTxOkCount ++;
                }
                else
                {
                    INT	i;

                    if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
                    {
                        for(i = 0; i < MAX_WDS_ENTRY; i++)
                        {
                            if ((pAdapter->WdsTab.WdsEntry[i].Valid == TRUE) && (MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, pAddr)))
                            {
                                pAdapter->WdsTab.WdsEntry[i].OneSecTxOkCount ++;
                                break;
                            }
                        }
                    }
                }
            }
                            
            break;
              
        case SUCCESS_WITH_RETRY:            // Success with some retry
            // DBGPRINT(RT_DEBUG_INFO, "TX Success with retry(=%d)<<<\n",pTxD->RetryCount);
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.RetryCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            INC_COUNTER(pAdapter->WlanCounters.TransmittedFragmentCount);
            
            // Increase general counters
            pAdapter->Counters.GoodTransmits++;
            
            if (pTxD->RetryCount > 1)
            {
                // Increase 802.11 counters
                INC_COUNTER(pAdapter->WlanCounters.MultipleRetryCount);
                
                // Increase general counters
                pAdapter->Counters.MoreCollisions++;
            }
            else
            {
                // Increase general counters
                pAdapter->Counters.OneCollision++;
            }
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSSuccessCount);
                pTxD->RTS = 0;
            }

            // update DRS related counters
            if (pTxD->ACK && (FrameType == BTYPE_DATA))
            {
                pMacEntry = MacTableLookup(pAdapter, pAddr);
                if (pMacEntry)
                {
                    if (pTxD->TxRate > pMacEntry->CurrTxRate)
                    {
                        // must be NULL frame retried @ UpRate; downgrade TxQuality[UpRate]
                        // so that rate upgrade won't happen
                        pMacEntry->TxQuality[pTxD->TxRate] += 3;
                        if (pMacEntry->TxQuality[pTxD->TxRate] > DRS_TX_QUALITY_WORST_BOUND)
                            pMacEntry->TxQuality[pTxD->TxRate] = DRS_TX_QUALITY_WORST_BOUND;
                    }
                    else if (pTxD->TxRate == pMacEntry->CurrTxRate)
						pMacEntry->OneSecTxRetryOkCount ++;
				}
                else
                {
                    INT	i;

                    if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
                    {
                        for(i = 0; i < MAX_WDS_ENTRY; i++)
                        {
                            if (((pAdapter->WdsTab.WdsEntry[i].Valid == TRUE) && MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, pAddr)))
                            {
                                if (pTxD->TxRate > pAdapter->WdsTab.WdsEntry[i].CurrTxRate)
                                {
                                    // must be NULL frame retried @ UpRate; downgrade TxQuality[UpRate]
                                    // so that rate upgrade won't happen
                                    pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] += 3;
                                    if (pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] > DRS_TX_QUALITY_WORST_BOUND)
                                    pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] = DRS_TX_QUALITY_WORST_BOUND;
                                }
                                else
                                    pAdapter->WdsTab.WdsEntry[i].OneSecTxRetryOkCount ++;

                                break;
                            }
                        }
                    }
                }
            }
                            
            break;

        case FAIL_RETRY_LIMIT:              // Fail on hitting retry count limit
            // DBGPRINT(RT_DEBUG_TRACE, "TX %02x:%02x:%02x:%02x:%02x:%02x @ %d Mbps Failed(RETRY LIMIT)<<<\n",
            //     pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5], RateIdToMbps[pTxD->TxRate]);
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FailedCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }

            // update DRS related counters
            if (pTxD->ACK && (FrameType == BTYPE_DATA))
            {
                pMacEntry = MacTableLookup(pAdapter, pAddr);
                if (pMacEntry)
                {
                    if (pTxD->TxRate > pMacEntry->CurrTxRate)
                    {
                        // must be NULL frame retried @ UpRate; downgrade TxQuality[UpRate]
                        // so that rate upgrade won't happen
                        pMacEntry->TxQuality[pTxD->TxRate] += 4;
                        if (pMacEntry->TxQuality[pTxD->TxRate] > DRS_TX_QUALITY_WORST_BOUND)
                            pMacEntry->TxQuality[pTxD->TxRate] = DRS_TX_QUALITY_WORST_BOUND;
                    }
                    else if (pTxD->TxRate == pMacEntry->CurrTxRate)
						pMacEntry->OneSecTxFailCount ++;
				}
                else
                {
                    INT	i;

                    if (pAdapter->WdsTab.Mode != WDS_DISABLE_MODE)
                    {
                        for(i = 0; i < MAX_WDS_ENTRY; i++)
                        {
                            if ((pAdapter->WdsTab.WdsEntry[i].Valid == TRUE) && (MAC_ADDR_EQUAL(&pAdapter->WdsTab.WdsEntry[i].WdsAddr, pAddr)))
                            {
                                if (pTxD->TxRate > pAdapter->WdsTab.WdsEntry[i].CurrTxRate)
                                {
                                    // must be NULL frame retried @ UpRate; downgrade TxQuality[UpRate]
                                    // so that rate upgrade won't happen
                                    pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] += 4;
                                    if (pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] > DRS_TX_QUALITY_WORST_BOUND)
                                    pAdapter->WdsTab.WdsEntry[i].TxQuality[pTxD->TxRate] = DRS_TX_QUALITY_WORST_BOUND;
                                }
                                else
                                    pAdapter->WdsTab.WdsEntry[i].OneSecTxFailCount ++;

                                break;
                            }
                        }
                    }
                }
            }
                            
            break;
            
        case FAIL_INVALID:
            // DBGPRINT(RT_DEBUG_WARN, ("TX Failed (INVALID)<<<\n"));
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }
            break;          
            
        case FAIL_OTHER:
        default:
            // DBGPRINT(RT_DEBUG_ERROR, ("TX Failed (other=%d)<<<\n",pTxD->TxResult));
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FailedCount);
            INC_COUNTER(pAdapter->WlanCounters.ACKFailureCount);
            
            // Increase general counters
            pAdapter->Counters.TxErrors++;
            
            if (pTxD->RTS)
            {
                INC_COUNTER(pAdapter->WlanCounters.RTSFailureCount);
                pTxD->RTS = 0;
            }
            break;          
    }
}

/*
    ========================================================================
    Routine Description:
        API for MLME to transmit management frame to AP (BSS Mode)
    or station (IBSS Mode)
    
    Arguments:
        pAdapter    Pointer to our adapter
        Buffer      Pointer to  memory of outgoing frame
        Length      Size of outgoing management frame
        
    Return Value:
        NDIS_STATUS_FAILURE
        NDIS_STATUS_PENDING
        NDIS_STATUS_SUCCESS
    ========================================================================
*/
NDIS_STATUS MiniportMMRequest(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PVOID           pBuffer,
    IN  ULONG           Length)
{
    PMGMT_STRUC     pMgmt;
    NDIS_STATUS     Status = NDIS_STATUS_SUCCESS;    

    DBGPRINT(RT_DEBUG_INFO, "---> MiniportMMRequest\n");
    // Check management ring free avaliability
    pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PushMgmtIndex];
    
    // This management cell has been occupied
    if (pMgmt->Valid == TRUE)   
    {
        // No Management ring buffer avaliable
        if (pBuffer)
        {
            MlmeConfirm((PVOID)pAdapter, pBuffer, 0);
        }
        Status = NDIS_STATUS_FAILURE; 
        DBGPRINT(RT_DEBUG_WARN, "<--- MiniportMMRequest (error:: MgmtRing full)\n");
        pAdapter->RalinkCounters.MgmtRingFullCount++;
        return (Status);
    }
    
    // Insert this request into software managemnet ring
    if (pBuffer)
    {
        pMgmt->pBuffer = pBuffer;       
        pMgmt->Length  = Length;
        pMgmt->Valid   = TRUE;
        pAdapter->PushMgmtIndex++;
        pAdapter->MgmtQueueSize++;
        if (pAdapter->PushMgmtIndex >= MGMT_RING_SIZE)
        {
            pAdapter->PushMgmtIndex = 0;
        }
    }   
    else
    {
        // Null pBuffer, no need to free memory buffer.
        // This should not happen
        DBGPRINT(RT_DEBUG_WARN, "<--- MiniportMMRequest (error:: NULL msg)\n");
        Status = NDIS_STATUS_FAILURE; 
        return (Status);
    }
    
    // Check Free priority queue
    if (RTMPFreeDescriptorRequest(pAdapter, PRIO_RING, 1) == NDIS_STATUS_SUCCESS)
    {
        pMgmt = (PMGMT_STRUC) &pAdapter->MgmtRing[pAdapter->PopMgmtIndex];
        if (pMgmt->Valid == TRUE)
        {
            MlmeHardTransmit(pAdapter, pMgmt->pBuffer, pMgmt->Length);
            MlmeConfirm((PVOID)pAdapter, pMgmt->pBuffer, 0);
            pMgmt->Valid = FALSE;
            pAdapter->PopMgmtIndex++;
            pAdapter->MgmtQueueSize--;
            if (pAdapter->PopMgmtIndex >= MGMT_RING_SIZE)
            {
                pAdapter->PopMgmtIndex = 0;
            }
        }
    }
    else
    {
        DBGPRINT(RT_DEBUG_INFO, "not enough space in PrioRing\n");
    }

    DBGPRINT(RT_DEBUG_INFO, "<--- MiniportMMRequest\n");
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        Copy frame from waiting queue into relative ring buffer and set 
    appropriate ASIC register to kick hardware transmit function
    
    Arguments:
        pAdapter    Pointer to our adapter
        pBuffer     Pointer to  memory of outgoing frame
        Length      Size of outgoing management frame
    ========================================================================
*/
VOID    MlmeHardTransmit(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PVOID           pBuffer,
    IN  ULONG           Length)
{
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif
    PUCHAR          pDest;  
    PHEADER_802_11  pHeader_802_11;
    BOOLEAN         AckRequired, InsertTimestamp;
    
    DBGPRINT(RT_DEBUG_INFO, "MlmeHardTransmit\n");
    
    // Make sure Prio ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->PrioRingLock);
        
    pDest = (PUCHAR) pAdapter->PrioRing[pAdapter->CurPrioIndex].va_data_addr;              
#ifndef BIG_ENDIAN
    pTxD  = (PTXD_STRUC) pAdapter->PrioRing[pAdapter->CurPrioIndex].va_addr;
#else
    pDestTxD  = (PTXD_STRUC) pAdapter->PrioRing[pAdapter->CurPrioIndex].va_addr;
    TxD = *pDestTxD;
    pTxD = &TxD;
    RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
        
    if (pTxD->Owner == DESC_OWN_NIC)
    {
        // Descriptor owned by NIC. No descriptor avaliable
        // This should not happen since caller guaranteed.
        // Make sure to release Prio ring resource
        NdisReleaseSpinLock(&pAdapter->PrioRingLock);
        return;
    }
    if (pTxD->Valid == TRUE)
    {
        // Ndis packet of last round did not cleared.
        // This should not happen since caller guaranteed.
        // Make sure to release Prio ring resource
        NdisReleaseSpinLock(&pAdapter->PrioRingLock);
        return;
    }
    
    pHeader_802_11           = (PHEADER_802_11) pBuffer;
    InsertTimestamp = FALSE;
    if (pHeader_802_11->Controlhead.Frame.Type == BTYPE_CNTL) // must be PS-POLL
    {
        AckRequired = FALSE;
        pAdapter->PrioRing[pAdapter->CurPrioIndex].FrameType = BTYPE_CNTL;
    }
    else // BTYPE_MGMT or BMGMT_DATA(must be NULL frame)
    {
        pAdapter->PrioRing[pAdapter->CurPrioIndex].FrameType = BTYPE_MGMT;
        pAdapter->Sequence       = ((pAdapter->Sequence) + 1) & (MAX_SEQ_NUMBER);
        pHeader_802_11->Sequence = pAdapter->Sequence;

        if (pHeader_802_11->Controlhead.Addr1.Octet[0] & 0x01) // MULTICAST, BROADCAST
        {
            AckRequired = FALSE;
            pHeader_802_11->Controlhead.Duration = 0;
        }
        else
        {
            AckRequired = TRUE;
            pHeader_802_11->Controlhead.Duration = RTMPCalcDuration(pAdapter, RATE_2, 14);
            if (pHeader_802_11->Controlhead.Frame.Subtype == SUBTYPE_PROBE_RSP)
            {
                InsertTimestamp = TRUE;
            }
        }
    }
#ifdef BIG_ENDIAN
    RTMPFrameEndianChange(pAdapter, (PUCHAR)pBuffer, DIR_WRITE, FALSE);
#endif
    NdisMoveMemory(pDest, pBuffer, Length);
   
    // Initialize Priority Descriptor
    // For inter-frame gap, the number is for this frame and next frame
    // For MLME rate, we will fix as 2Mb to match other vendor's implement
#ifdef BIG_ENDIAN
    RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
    *pDestTxD = TxD;
    pTxD = pDestTxD;
#endif
    RTMPWriteTxDescriptor(pTxD, FALSE, CIPHER_NONE, AckRequired, FALSE, InsertTimestamp, SHORT_RETRY, IFS_BACKOFF, 
        pAdapter->PortCfg.MlmeRate, 4, Length, pAdapter->PortCfg.TxPreambleInUsed, 0);

    // Increase & maintain Tx Ring Index
    pAdapter->CurPrioIndex++;
    if (pAdapter->CurPrioIndex >= PRIO_RING_SIZE)
    {
        pAdapter->CurPrioIndex = 0;
    }
        
    // Kick priority ring transmit
    RTMP_IO_WRITE32(pAdapter,TXCSR0,0x4);
    
    // Make sure to release Prio ring resource
    NdisReleaseSpinLock(&pAdapter->PrioRingLock);
}   

/*
    ========================================================================
    Routine Description:
        This routine is used to en-queue outgoing packets when
        there is no enough shread memory
        
    Arguments:
        pAdapter    Pointer to our adapter
        pPacket     Pointer to send packet
        
    Return Value:
        None
    ========================================================================
*/
NDIS_STATUS RTMPSendPacket(
    IN  PRTMP_ADAPTER   pAdapter,
    IN	struct sk_buff	*skb)
{
    PVOID           pVirtualAddress;
    UCHAR           NumberOfFrag = 1;
    UCHAR           RTSRequired;
    NDIS_STATUS     Status = NDIS_STATUS_FAILURE;

    SST             Sst;
    UCHAR           PsMode, Rate, apidx = MAIN_MSSID;
    USHORT          Aid;
    MAC_TABLE_ENTRY *pMacEntry = NULL;

    PQUEUE_HEADER   pTxQueue = NULL;
    ULONG           Priority;
#ifdef QOS
    UCHAR           AccessCategory = 0;
#endif
    DBGPRINT(RT_DEBUG_INFO, "<==== RTMPSendPacket\n");

	// Init priority value
	Priority = 0;
#ifdef QOS
    if (skb)
    {
        Priority = skb->priority;
        // 802.11e/d4.4 June, 2003
        if (Priority <= 2)
            AccessCategory = 0;
        else if (Priority == 3)
            AccessCategory = 1;
        else if (Priority <= 5)
            AccessCategory = 2;
        else
            AccessCategory = 3;
        DBGPRINT(RT_DEBUG_INFO, "Priority = %d, AC = %d\n", Priority, AccessCategory);
    }
#endif

    pVirtualAddress = (PVOID)skb->data;

    // If protocol type is equal to 0x81 0x00, then this skb is attached with VLAN tag.
    // We need to remove VLAN tag before transmit it.
    if (RTMPEqualMemory(TPID, (PUCHAR)pVirtualAddress + 12, 2))
    {
        UCHAR   buffer[16];

        memcpy(buffer, skb->data, 16);          // backup DA/SA field
        skb_pull(skb, 16);                      // remove DA/SA/VLAN tag
        memcpy(skb_push(skb, 12), buffer, 12);  // add DA/SA

        pVirtualAddress = (PVOID)skb->data;
    }

    if ((RTMP_GET_PACKET_NET_DEVICE(skb) < 0x20) && (RTMP_GET_PACKET_NET_DEVICE(skb) >= 0x10))
        apidx = RTMP_GET_PACKET_NET_DEVICE(skb) - 0x10;
    
    // check if target STA is in power-saving mode. care only those associated STAs
    if (RTMP_GET_PACKET_NET_DEVICE(skb) >= 0x20)
    {// WDS packet
    	Rate = pAdapter->WdsTab.WdsEntry[RTMP_GET_PACKET_NET_DEVICE(skb) - 0x20].CurrTxRate;
    }
    else
    {// Normal packet or MSSID packet
        pMacEntry = SsPsInquiry(pAdapter, pVirtualAddress, &Sst, &Aid, &PsMode, &Rate);
    }
    
    // For TKIP, MIC value is treated as payload, it might be fragmented through
    // different MPDUs.
    if ((*((PUCHAR) pVirtualAddress) & 0x01) != 0)
    {
	    if (pAdapter->PortCfg.MSSID[apidx].GroupKeyWepStatus == Ndis802_11Encryption2Enabled)
	    {
	        skb->len += 8;
	    }
    }
    else if (pMacEntry)
    {
    	if (pMacEntry->WepStatus == Ndis802_11Encryption2Enabled)
	    {
	        skb->len += 8;
	    }
    }

    // Check for virtual address allocation, it might fail !!!
    if (pVirtualAddress == NULL)
    {
        // Resourece is low, system did not allocation virtual address
        // return NDIS_STATUS_FAILURE directly to upper layer
        return (Status);
    }

    //
    // Check for multicast or broadcast (First byte of DA)
    // For multicast & broadcast, there is no fragment allowed (NumberOfFrag = 1)
    // Only needs to calculate unicast packet.
    //
    if ((*((PUCHAR) pVirtualAddress) & 0x01) == 0)
    {
        INT PacketSizeNeedFrag = skb->len - LENGTH_802_3 + LENGTH_802_1_H;

        // Calculate fragments required
        while(PacketSizeNeedFrag > pAdapter->PortCfg.AllowFragSize * NumberOfFrag)
        {
            NumberOfFrag++;
        }
#if 0
        // Check for payload allowed for each fragment
        AllowFragSize = (pAdapter->PortCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;

        // Calculate fragments required
        NumberOfFrag = ((skb->len - LENGTH_802_3 + LENGTH_802_1_H) / AllowFragSize) + 1;
        // Minus 1 if the size just match to allowable fragment size
        if (((skb->len - LENGTH_802_3 + LENGTH_802_1_H) % AllowFragSize) == 0)
        {
            NumberOfFrag--;
        }
#endif
    }

    // Check for requirement of RTS 
    if (NumberOfFrag > 1)
    {
        // If multiple fragment required, RTS is required only for the first fragment
        // if the fragment size large than RTS threshold
        RTSRequired = (pAdapter->PortCfg.FragmentThreshold > pAdapter->PortCfg.RtsThreshold) ? 1 : 0;
    }
    else
    {
        RTSRequired = (skb->len > pAdapter->PortCfg.RtsThreshold) ? 1 : 0;
    }
    DBGPRINT(RT_DEBUG_INFO, "Number of fragments include RTS :%d\n", NumberOfFrag + RTSRequired);

    // HotSpot counter
    if (pMacEntry && pAdapter->PortCfg.bEnableHSCounter)
    {
        pMacEntry->HSCounter.TotalTxByteCount += skb->len;
        pMacEntry->HSCounter.LastDataPacketTime = jiffies;
    }

    // CTS-to-self is required for OFDM if USE_PROTECTION bit is ON
    // When CTS-to-self is used, RTS/CTS is not required
    if ((Rate >= RATE_FIRST_OFDM_RATE) && ERP_IS_USE_PROTECTION(pAdapter->Mlme.ErpIeContent))
        RTSRequired = 2; // use CTS-to-self

    // use the CurrTxRate that this client supported
    RTMP_SET_PACKET_TXRATE(skb, Rate);
    
    // Save fragment number to Ndis packet reserved field
    RTMP_SET_PACKET_FRAGMENTS(skb, NumberOfFrag);

    // Save RTS requirement to Ndis packet reserved field
    RTMP_SET_PACKET_RTS(skb, RTSRequired);

#if 1
    // 1. multicast/broadcast frames are put to PSQ as long as there's any associated STA in power-save mode
    // 2. if the associted STA in power-save mode, frame also goes to PSQ
    if ((*(CHAR *)pVirtualAddress & 0x01) && pAdapter->PortCfg.MSSID[apidx].MacTab.fAnyStationInPsm)
    {
        // 0. check PsQueue packet number
        if(pAdapter->PortCfg.MSSID[apidx].MacTab.McastPsQueue.Number > MAX_PACKETS_IN_QUEUE)
        {
            DBGPRINT(RT_DEBUG_WARN, "McastPsQueue is full 1!!!\n\n");
            return  NDIS_STATUS_FAILURE;
        }

        // 1. move outgoing frame to power saving queue 
        NdisAcquireSpinLock(&pAdapter->PortCfg.MSSID[apidx].MacTabLock);
        InsertTailQueue(&pAdapter->PortCfg.MSSID[apidx].MacTab.McastPsQueue, skb);
        NdisReleaseSpinLock(&pAdapter->PortCfg.MSSID[apidx].MacTabLock);
        
        // 2. mark corresponding TIM bit in outgoing BEACON frame
        pAdapter->Mlme.TimBitmap[0] |= Bit[0];
        // DBGPRINT(RT_DEBUG_TRACE, "at least 1 STA in psm, move M/BCAST to PSQ, TIM bitmap=%08x\n", pAdapter->Mlme.TimBitmap);
    }
    else if (pMacEntry && (Sst == SST_ASSOC) && (PsMode == PWR_SAVE))
    {
        // return  NDIS_STATUS_FAILURE;
        // 0. check PsQueue packet number
        if(pMacEntry->PsQueue.Number > MAX_PACKETS_IN_QUEUE)
        {
            DBGPRINT(RT_DEBUG_WARN, "PsQueue is full 2!!!\n\n");
            return  NDIS_STATUS_FAILURE;
        }

        // 1. move outgoing frame to power saving queue 
        NdisAcquireSpinLock(&pAdapter->PortCfg.MSSID[apidx].MacTabLock);
        InsertTailQueue(&pMacEntry->PsQueue, skb);
        NdisReleaseSpinLock(&pAdapter->PortCfg.MSSID[apidx].MacTabLock);
            
        // 2. mark corresponding TIM bit in outgoing BEACON frame
        pAdapter->Mlme.TimBitmap[Aid/8] |= Bit[Aid%8];
        DBGPRINT(RT_DEBUG_TRACE, "target (AID=%d) in psm, move to PSQ, TIM bitmap=%02x-%02x-%02x-%02x\n", Aid, pAdapter->Mlme.TimBitmap[0], pAdapter->Mlme.TimBitmap[1], pAdapter->Mlme.TimBitmap[2], pAdapter->Mlme.TimBitmap[3]);
    }
    // 3. otherwise, transmit the frame
    else // (PsMode == PWR_ACTIVE) || (PsMode == PWR_UNKNOWN)
#endif  
    {
        // Enqueue Ndis packet to end of Tx wait queue
        NdisAcquireSpinLock(&pAdapter->TxSwQueueLock);

#ifdef QOS
    	// Select the right priority queue
	    // There should be no else statement since it should always fall within 0-3
    	if (AccessCategory== 0)
	    	pTxQueue = &pAdapter->TxSwQueue0;
    	else if (AccessCategory== 1)
	    	pTxQueue = &pAdapter->TxSwQueue1;
    	else if (AccessCategory== 2)
	    	pTxQueue = &pAdapter->TxSwQueue2;
    	else if (AccessCategory== 3)
	    	pTxQueue = &pAdapter->TxSwQueue3;
#else
	    pTxQueue = &pAdapter->TxSwQueue0;
#endif
        // check packet number in queue
        if(pTxQueue->Number > MAX_PACKETS_IN_QUEUE)
        {
            DBGPRINT(RT_DEBUG_WARN, "TxQueue is full 3!!!\n\n");
            NdisReleaseSpinLock(&pAdapter->TxSwQueueLock);
            return  NDIS_STATUS_FAILURE;
        }
	
        InsertTailQueue(pTxQueue, skb);
        NdisReleaseSpinLock(&pAdapter->TxSwQueueLock);
    }

    Status = NDIS_STATUS_SUCCESS;
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        To do the enqueue operation and extract the first item of waiting 
        list. If a number of available shared memory segments could meet 
        the request of extracted item, the extracted item will be fragmented
        into shared memory segments.
        
    Arguments:
        pAdapter    Pointer to our adapter
        pQueue      Pointer to Waiting Queue
    ========================================================================
*/
VOID    RTMPDeQueuePacket(
    IN  PRTMP_ADAPTER   pAdapter)
{
    struct sk_buff  *skb;
    UCHAR           FragmentRequired;
    NDIS_STATUS     Status;
    UCHAR           Count = 0;
    PQUEUE_HEADER   pQueue;
    ULONG           Number;
    UCHAR           AccessCategory;
    
    // Make sure TxSwQueue0..3 queue resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->TxSwQueueLock);

    while (((pQueue = RTMPCheckTxSwQueue(pAdapter, &Number, &AccessCategory)) != NULL) && (Count < MAX_TX_PROCESS))
    // Check queue before dequeue
    // while ((pQueue->Head != NULL) && (Count < MAX_TX_PROCESS)) 
    {
        // Dequeue the first entry from head of queue list
        skb = (struct sk_buff*)RemoveHeadQueue(pQueue);

        // Total fragment required = number of fragment + RST if required
        FragmentRequired = RTMP_GET_PACKET_FRAGMENTS(skb) + (RTMP_GET_PACKET_RTS(skb)? 1:0);
        
        if (RTMPFreeDescriptorRequest(pAdapter, TX_RING, FragmentRequired) == NDIS_STATUS_SUCCESS)
        {
            // Avaliable ring descriptors are enough for this frame
            // Call hard transmit
            // Nitro mode / Normal mode selection
            if ((pAdapter->PortCfg.EnableTxBurst == 1) && (Number > 1))
                Status = RTMPHardEncrypt(pAdapter, skb, FragmentRequired, TRUE, AccessCategory, (Number > 1));
            else
                Status = RTMPHardEncrypt(pAdapter, skb, FragmentRequired, FALSE, AccessCategory, (Number > 1));
            if (Status == NDIS_STATUS_FAILURE)
            {
                // Packet failed due to various Ndis Packet error
                RTMPFreeSkbBuffer(skb);
                break;
            }
            else if (Status == NDIS_STATUS_RESOURCES)
            {
                // Not enough free tx ring, it might happen due to free descriptor inquery might be not correct
                // It also might change to NDIS_STATUS_FAILURE to simply drop the frame
                // Put the frame back into head of queue
                InsertHeadQueue(pQueue, skb);
                break;
            }           
            Count++;
        }   
        else
        {
            InsertHeadQueue(pQueue, skb);
            pAdapter->PrivateInfo.TxRingFullCnt++;
            DBGPRINT(RT_DEBUG_INFO,"RTMPDequeuePacket --> Not enough free Tx Ring descriptor !!!\n");
            break;
        }
    }

    // Release TxSwQueue0..3 resources
    NdisReleaseSpinLock(&pAdapter->TxSwQueueLock);
}    

/*
    ========================================================================
    Routine Description:
        This subroutine will scan through releative ring descriptor to find
        out avaliable free ring descriptor and compare with request size.
        
    Arguments:
        pAdapter    Pointer to our adapter
        RingType    Selected Ring
        
    Return Value:
        NDIS_STATUS_FAILURE     Not enough free descriptor
        NDIS_STATUS_SUCCESS     Enough free descriptor
    ========================================================================
*/
NDIS_STATUS RTMPFreeDescriptorRequest(
    IN      PRTMP_ADAPTER   pAdapter,
    IN      UCHAR           RingType,
    IN      UCHAR           NumberRequired)
{
    UCHAR           FreeNumber = 0;
    UINT            Index;
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif
    NDIS_STATUS     Status = NDIS_STATUS_FAILURE;

    switch (RingType)
    {
        case TX_RING:
            NdisAcquireSpinLock(&pAdapter->TxRingLock);
            Index = pAdapter->CurEncryptIndex;
            do
            {
#ifndef BIG_ENDIAN
                pTxD  = (PTXD_STRUC) pAdapter->TxRing[Index].va_addr;
#else
                pDestTxD  = (PTXD_STRUC) pAdapter->TxRing[Index].va_addr;
                TxD = *pDestTxD;
                pTxD = &TxD;
                RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
                
                // While Owner bit is NIC, obviously ASIC still need it.
                // If valid bit is TRUE, indicate that TxDone has not process yet
                // We should not use it until TxDone finish cleanup job
                if ((pTxD->Owner == DESC_OWN_HOST) && (pTxD->CipherOwn == DESC_OWN_HOST) && (pTxD->Valid == FALSE))
                {
                    // This one is free
                    FreeNumber++;
                }
                else
                {
                    break;
                }
                    
                Index++;
                if (Index >= TX_RING_SIZE)              // Wrap around issue
                {
                    Index = 0;
                }
                
            }   while (FreeNumber < NumberRequired);    // Quit here ! Free number is enough !
            
            if (FreeNumber >= NumberRequired)
            {
                Status = NDIS_STATUS_SUCCESS;
            }
            
            // Make sure to release Tx ring resource
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            break;
            
        case PRIO_RING:
            NdisAcquireSpinLock(&pAdapter->PrioRingLock);
            Index = pAdapter->CurPrioIndex;
            do
            {
#ifndef BIG_ENDIAN
                pTxD  = (PTXD_STRUC) pAdapter->PrioRing[Index].va_addr;
#else
                pDestTxD  = (PTXD_STRUC) pAdapter->PrioRing[Index].va_addr;
                TxD = *pDestTxD;
                pTxD = &TxD;
                RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
                
                // While Owner bit is NIC, obviously ASIC still need it.
                // If valid bit is TRUE, indicate that TxDone has not process yet
                // We should not use it until TxDone finish cleanup job
                if ((pTxD->Owner == DESC_OWN_HOST) && (pTxD->Valid == FALSE))
                {
                    // This one is free
                    FreeNumber++;
                }
                else
                {
                    break;
                }
                    
                Index++;
                if (Index >= PRIO_RING_SIZE)            // Wrap around issue
                {
                    Index = 0;
                }
                
            }   while (FreeNumber < NumberRequired);    // Quit here ! Free number is enough !
            
            if (FreeNumber >= NumberRequired)
            {
                Status = NDIS_STATUS_SUCCESS;
            }
            
            // Make sure to release Prio ring resource
            NdisReleaseSpinLock(&pAdapter->PrioRingLock);
            break;

        default:
            break;
    }
    
    return (Status);
}

/*
    ========================================================================
    Routine Description:
        Copy frame from waiting queue into relative ring buffer and set 
    appropriate ASIC register to kick hardware encryption before really
    sent out to air.
        
    Arguments:
        pAdapter        Pointer to our adapter
        PNDIS_PACKET    Pointer to outgoing Ndis frame
        NumberOfFrag    Number of fragment required
        
    Return Value:
        None
    ========================================================================
*/
NDIS_STATUS RTMPHardEncrypt(
    IN  PRTMP_ADAPTER   pAdapter,
    IN	struct sk_buff	*skb,
    IN  UCHAR           NumberRequired,
    IN  ULONG           EnableTxBurst,
    IN  UCHAR           AccessCategory,
    IN  BOOLEAN		MoreData)
{
    UINT            Length;
    UINT            BytesCopied;
    UINT            TxSize;
    UINT            FreeFragSize;
    UINT            RemainSize;
    UCHAR           ThisFrameKeyID = 0;
    ULONG           Iv16;
    ULONG           Iv32;
    UCHAR           FrameGap;
    UCHAR           TxRate;
    UCHAR           RetryMode = SHORT_RETRY;
    UCHAR           AckRate;
    UCHAR           CipherAlg, apidx;
    PVOID           pVirtualAddress;
    PUCHAR          pDest;
    PUCHAR          pSrc;
    PUCHAR          pEncap;
    USHORT          Protocol;
    USHORT          AckDuration;
    BOOLEAN         StartOfFrame;
    BOOLEAN         EAPOLFrame;
    BOOLEAN         Encapped;
    BOOLEAN         MICFrag;
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
    PUCHAR          pOriginDest;
#endif
    HEADER_802_11   Header_802_11;
    PWPA_KEY        pWpaKey = (PWPA_KEY) NULL;
    MAC_TABLE_ENTRY *pEntry = NULL;
    BOOLEAN         bWdsPacket = FALSE;
    NDIS_802_11_WEP_STATUS	EncryptType = Ndis802_11EncryptionDisabled;

    // Make sure Tx ring resource won't be used by other threads
    NdisAcquireSpinLock(&pAdapter->TxRingLock);

    if ((skb->data == NULL) || (skb->len < 14))
    {
        DBGPRINT(RT_DEBUG_ERROR, "RTMPHardTransmit --> Ndis Packet buffer error!!!\n");
        // Make sure to release Tx ring resource
        NdisReleaseSpinLock(&pAdapter->TxRingLock);
        return (NDIS_STATUS_FAILURE);
    }

    // Find out the packet comes from which interface
    if (RTMP_GET_PACKET_NET_DEVICE(skb) >= 0x20)
    {// WDS packet
        bWdsPacket = TRUE;
        apidx = MAIN_MSSID;
    }
    else if (RTMP_GET_PACKET_NET_DEVICE(skb) >= 0x10)
    {// Multiple SSID packet
        bWdsPacket = FALSE;
        apidx = RTMP_GET_PACKET_NET_DEVICE(skb) - 0x10;
    }
    else
    {// otherwise
        bWdsPacket = FALSE;
        apidx = MAIN_MSSID;
    }

    if (EnableTxBurst == 1)
        FrameGap = IFS_SIFS;
    else
        FrameGap = IFS_BACKOFF;		// Default frame gap mode

    pVirtualAddress = (PVOID)skb->data;
    Length = skb->len;
    
    // Sequence Number is identical for all fragments belonged to the same frame
    // Sequence is 0 - 4095
    pAdapter->Sequence = ((pAdapter->Sequence) + 1) & (MAX_SEQ_NUMBER);
    // pAdapter->TxRing[pAdapter->CurEncryptIndex].FrameType = BTYPE_DATA;

    TxRate = RTMP_GET_PACKET_TXRATE(skb);
    AckRate = pAdapter->PortCfg.ExpectedACKRate[TxRate];
    AckDuration = RTMPCalcDuration(pAdapter, AckRate, 14);

    // Initialize 802.11 header for each frame
    NdisZeroMemory(&Header_802_11, sizeof(HEADER_802_11));

    if ((*((PUCHAR) pVirtualAddress) & 0x01) != 0)  // Multicast or Broadcast
    {
        INC_COUNTER(pAdapter->WlanCounters.MulticastTransmittedFrameCount);
    }

    //
    // Start making 802.11 frame header
    //
    if (bWdsPacket)
    {
//		char *pSrc = ((char *)pVirtualAddress) + 6; // SA
        // Addr1: RA, Addr2: TA, Addr3: DA, Addr4: SA
        Header_802_11.Controlhead.Frame.FrDs = 1;
        Header_802_11.Controlhead.Frame.ToDs = 1;
        if (MoreData == TRUE)
            Header_802_11.Controlhead.Frame.MoreData = 1;
        NdisMoveMemory(&Header_802_11.Controlhead.Addr1, &pAdapter->WdsTab.WdsEntry[(RTMP_GET_PACKET_NET_DEVICE(skb) - 0x20)].WdsAddr, ETH_LENGTH_OF_ADDRESS);// RA
        NdisMoveMemory(&Header_802_11.Controlhead.Addr2, &pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);// TA
        NdisMoveMemory(&Header_802_11.Addr3, pVirtualAddress, ETH_LENGTH_OF_ADDRESS);// DA
        DBGPRINT(RT_DEBUG_INFO,"WDS-DA=%02x:%02x:%02x:%02x:%02x:%02x, SA=%02x:%02x:%02x:%02x:%02x:%02x\n",
            *((PUCHAR)pVirtualAddress+0),*((PUCHAR)pVirtualAddress+1),*((PUCHAR)pVirtualAddress+2),*((PUCHAR)pVirtualAddress+3),*((PUCHAR)pVirtualAddress+4),*((PUCHAR)pVirtualAddress+5),
            *((PUCHAR)pVirtualAddress+6),*((PUCHAR)pVirtualAddress+7),*((PUCHAR)pVirtualAddress+8),*((PUCHAR)pVirtualAddress+9),*((PUCHAR)pVirtualAddress+10),*((PUCHAR)pVirtualAddress+11));
    }
    else
    {
        char *pSrc = ((char *)pVirtualAddress) + 6; // SA
        // Addr1: DA, Addr2: BSSID, Addr3: SA
        Header_802_11.Controlhead.Frame.FrDs = 1;
        Header_802_11.Controlhead.Frame.ToDs = 0;
        if (MoreData == TRUE)
            Header_802_11.Controlhead.Frame.MoreData = 1;
        NdisMoveMemory(&Header_802_11.Controlhead.Addr1, pVirtualAddress, ETH_LENGTH_OF_ADDRESS);
        NdisMoveMemory(&Header_802_11.Controlhead.Addr2, &pAdapter->PortCfg.Bssid, ETH_LENGTH_OF_ADDRESS);
        NdisMoveMemory(&Header_802_11.Addr3, pSrc, ETH_LENGTH_OF_ADDRESS);
    }

    pEntry = MacTableLookup(pAdapter, &Header_802_11.Controlhead.Addr1);
    if (((*((PUCHAR) pVirtualAddress) & 0x01) == 0) && pEntry)
    {
    	EncryptType = pEntry->WepStatus;
    }
    else
    {
    	EncryptType = pAdapter->PortCfg.MSSID[apidx].GroupKeyWepStatus;
    }

    Header_802_11.Sequence = pAdapter->Sequence;        // Sequence number
    Header_802_11.Controlhead.Frame.Type = BTYPE_DATA;  // Frame type
    Header_802_11.Controlhead.Frame.PwrMgt = 0;

    if (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01)
    {
        // No ACK expected for multicast frame      
        Header_802_11.Controlhead.Duration = 0;
    }
    else
    {
        // ACK size is 14 include CRC, and its rate is 2Mb
        Header_802_11.Controlhead.Duration = pAdapter->PortCfg.Dsifs + AckDuration;

        // record the last TX rate (unicast DATA frame only) that this AP used. for UI to display only.
        // only unicast frames counts, because mcast/bcast frames always use basic rate
    
        // pAdapter->PortCfg.TxRate = TxRate;
    }
    
    // For the purpose to calculate duration for the second last fragment
    RemainSize = skb->len - LENGTH_802_3 + LENGTH_CRC;

    // Add 802.11x protocol check.
    // For non-WPA network, 802.11x message should not encrypt even
    // the privacy is on.
    if (RTMPEqualMemory(EAPOL, ((PUCHAR) pVirtualAddress) + 12, 2))
    {
        EAPOLFrame = TRUE;

    	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) RTMPHardEncrypt ==>> EAPOLFrame=TRUE, skb->len=%d , Length=%d\n", apidx, skb->len, Length);
    }
    else
    {
        EAPOLFrame = FALSE;
    }

#ifdef VINCENT_WDSTKIP
	if(!bWdsPacket)
	{
    	// patch for Lenght += 8; in SendPacket 
    	if (EncryptType == Ndis802_11Encryption2Enabled)
    	{
        	skb->len -= 8;
        	Length -= 8;            
    	}
	}
#else
    // patch for Lenght += 8; in SendPacket 
    if (EncryptType == Ndis802_11Encryption2Enabled)
    {
        skb->len -= 8;
        Length -= 8;            
    }
#endif

    // WPA 802.1x secured port control
    if(!bWdsPacket)
    {
        // in WPA mode, AP does not send packets before port secured.
        if (((pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA) || (pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA2)
        	|| (pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPAPSK) || (pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA2PSK)
        	|| (pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK) || (pAdapter->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA1WPA2))
        	&& (pAdapter->PortCfg.MSSID[apidx].PortSecured == WPA_802_1X_PORT_NOT_SECURED) && (EAPOLFrame == FALSE))
        {
            DBGPRINT(RT_DEBUG_INFO,"I/F(ra%d) RTMPHardEncrypt --> Drop packet before AP secured !!!\n", apidx);
            // Make sure to release Tx ring resource
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            return (NDIS_STATUS_FAILURE);
        }
        // if 802.1x, AP does not send packets other than EAPOL before this session control port set to "accept all".
		if ((pAdapter->PortCfg.MSSID[apidx].IEEE8021X == TRUE) && (pEntry) &&
			(pEntry->PrivacyFilter == Ndis802_11PrivFilter8021xWEP) && (EAPOLFrame == FALSE))
		{
			DBGPRINT(RT_DEBUG_TRACE,"RTMPHardEncrypt --> Drop packet before this control port accept all !!!\n");
			// Make sure to release Tx ring resource
			NdisReleaseSpinLock(&pAdapter->TxRingLock);
			return (NDIS_STATUS_FAILURE);
		}
    }

    MICFrag = FALSE;    // Flag to indicate MIC shall spread into two MPDUs
    Encapped = FALSE;
    pEncap = NULL;
    
    pSrc = (PUCHAR) pVirtualAddress;
    Protocol = *(pSrc + 12) * 256 + *(pSrc + 13);
    if (Protocol > 1500)    // CHeck for LLC encaped
    {
        pEncap = SNAP_802_1H;
        Encapped = TRUE;
        if (RTMPEqualMemory(IPX, pSrc + 12, 2))
        {
            pEncap = SNAP_BRIDGE_TUNNEL;
        }
        
        if (RTMPEqualMemory(APPLE_TALK, pSrc + 12, 2))
        {
            pEncap = SNAP_BRIDGE_TUNNEL;
        }
    }

    //
    // Make RTS or CTS-to-self frame if required
    //
    if (RTMP_GET_PACKET_RTS(skb))
    {
        PCONTROL_HEADER     pControlHeader;
        UINT                DATAFrameSize;
        
        pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;              
#ifndef BIG_ENDIAN
        pTxD = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
#else
        pDestTxD = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
                
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->CipherOwn == DESC_OWN_NIC))
        {
            // Descriptor owned by NIC. No descriptor avaliable
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        if (pTxD->Valid == TRUE)
        {
            // This might happen when HardTransmit process faster than TxDone interrupt.
            // However, Since we did call free descriptor number check before Tx HardTransmit.
            // This case should not happened. We should return to higher caller and requeue this
            // acket until next Tx hardtransmit. Hopefully this situation is corrected then.
            // Ndis packet of last round did not cleared.
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            pTxD->Valid = FALSE;
#ifdef BIG_ENDIAN
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
#endif
                
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        
        pAdapter->TxRing[pAdapter->CurEncryptIndex].FrameType = BTYPE_CNTL;
        pControlHeader = (PCONTROL_HEADER) pDest;
        NdisZeroMemory(pControlHeader, sizeof(CONTROL_HEADER));

        // Calculate duration
        if (RTMP_GET_PACKET_FRAGMENTS(skb) > 1)
        {
            // If fragment required, size is maximum fragment size
            DATAFrameSize = pAdapter->PortCfg.FragmentThreshold;
        }
        else
        {
            // Size should be frame with 802.11 header & CRC
            if (bWdsPacket)
            {
                DATAFrameSize = skb->len + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_1_H + LENGTH_CRC - LENGTH_802_3;// the length of addr4
            }
            else
            {
                DATAFrameSize = skb->len + LENGTH_802_11 + LENGTH_802_1_H + LENGTH_CRC - LENGTH_802_3;
            }

            // Decrease the LLC header size if it's already encapped from protocol layer
            if (Encapped == FALSE)
            {
                DATAFrameSize -= LENGTH_802_11;
            }

            // Add for WPA support
            if ((EncryptType == Ndis802_11Encryption1Enabled) && 
                (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0))
            {
                // WEP encryption
                DATAFrameSize += 8;
            }
            // The following statement might need change, because it might use peer key
            // However, it will complicate the control sequence.
            // Therefore we will simply add the 12 bytes into CTS duration no matter
            // the key exists or not.
            else if (EncryptType == Ndis802_11Encryption2Enabled)
            {
                // TKIP encryption
                DATAFrameSize += 12;        // IV + EIV + ICV, MIC already added to TotalPacketLength
            }
            else if (EncryptType == Ndis802_11Encryption3Enabled)
            {
                // AES encryption
                DATAFrameSize += 16;        // IV + EIV + Hardware MIC
            }
        }

        if (RTMP_GET_PACKET_RTS(skb) == 1)
        {
            DBGPRINT(RT_DEBUG_TRACE,"Making RTS Frame\n");
            
            // FrameControl
            pControlHeader->Frame.Type    = BTYPE_CNTL;
            pControlHeader->Frame.Subtype = SUBTYPE_RTS;        

            // RTS Duration = SIFS + CTS + SIFS + DATA + SIFS + ACK
            pControlHeader->Duration = pAdapter->PortCfg.Dsifs
                + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.RtsRate, 14)
                + pAdapter->PortCfg.Dsifs
                + RTMPCalcDuration(pAdapter, TxRate, DATAFrameSize)
                + pAdapter->PortCfg.Dsifs
                + AckDuration; 

            // RA & TA
            if (bWdsPacket)
            {
                NdisMoveMemory(&pControlHeader->Addr1, &pAdapter->WdsTab.WdsEntry[(RTMP_GET_PACKET_NET_DEVICE(skb) - 0x20)].WdsAddr, ETH_LENGTH_OF_ADDRESS);
            }
            else
            {
                NdisMoveMemory(&pControlHeader->Addr1, (PUCHAR) pVirtualAddress, ETH_LENGTH_OF_ADDRESS);
            }
            NdisMoveMemory(&pControlHeader->Addr2, pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);

            // Write Tx descriptor
            // Don't kick tx start until all frames are prepared
            // RTS has to set more fragment bit for fragment burst
            // RTS did not encrypt
#ifdef BIG_ENDIAN
            RTMPFrameEndianChange(pAdapter, (PUCHAR)pControlHeader, DIR_WRITE, FALSE);
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
            pTxD = pDestTxD;
#endif
            RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, TRUE, TRUE, FALSE, SHORT_RETRY, FrameGap, 
                pAdapter->PortCfg.RtsRate, 4, sizeof(CONTROL_HEADER), pAdapter->PortCfg.TxPreambleInUsed, AccessCategory);

            pTxD->RTS = 1;
        
            // RTS-protected frames should use LONG_RETRY (=4), other frames use SHORT_RETRY(=7)
            RetryMode = LONG_RETRY;
        
        }
        else // CTS-to-self
        {
            DBGPRINT(RT_DEBUG_INFO,"Making CTS-to-self Frame\n");
            
            // FrameControl
            pControlHeader->Frame.Type    = BTYPE_CNTL;
            pControlHeader->Frame.Subtype = SUBTYPE_CTS;        

            // CTS-to-self duration = SIFS + DATA + SIFS + ACK
            pControlHeader->Duration = pAdapter->PortCfg.Dsifs
                + RTMPCalcDuration(pAdapter, TxRate, DATAFrameSize)
                + pAdapter->PortCfg.Dsifs
                + RTMPCalcDuration(pAdapter, pAdapter->PortCfg.ExpectedACKRate[TxRate], 14); 

            NdisMoveMemory(&pControlHeader->Addr1, pAdapter->CurrentAddress, ETH_LENGTH_OF_ADDRESS);

            // Write Tx descriptor
            // Don't kick tx start until all frames are prepared
            // CTS has to set more fragment bit for fragment burst
            // CTS did not encrypt      
            // CTS-to-self will never receive ACK
#ifdef BIG_ENDIAN
            RTMPFrameEndianChange(pAdapter, (PUCHAR)pControlHeader, DIR_WRITE, FALSE);
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
            pTxD = pDestTxD;
#endif
            RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, TRUE, FALSE, SHORT_RETRY, FrameGap, 
                pAdapter->PortCfg.RtsRate, 4, 10, pAdapter->PortCfg.TxPreambleInUsed, AccessCategory);
        }
 
        FrameGap = IFS_SIFS;        // Init frame gap for coming data after RTS
        NumberRequired--;
        
        // Increase & maintain Tx Ring Index
        pAdapter->CurEncryptIndex++;
        if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
        {
            pAdapter->CurEncryptIndex = 0;
        }
        pAdapter->RalinkCounters.EncryptCount++;        
    }

    // Find the WPA key, either Group or Pairwise Key
    if (EncryptType >= Ndis802_11Encryption2Enabled)
    {
        pWpaKey = (PWPA_KEY) NULL;
        // First lokup the DA, if it's a group address, use GROUP key
        if (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01)
        {
            if (pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0)
            {
                pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];
                DBGPRINT(RT_DEBUG_INFO, "Tx Use Group Key\n");
            }
            ThisFrameKeyID = pAdapter->PortCfg.MSSID[apidx].WPAGKeyID;
        }
        else
        {
            ThisFrameKeyID = 0;
            if ((pEntry) && (pEntry->PairwiseKey[0].KeyLen != 0))
            {
                pWpaKey = (PWPA_KEY) &pEntry->PairwiseKey[0];
            }
            // Use default Group Key if there is no Pairwise key present
            if ((pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].KeyLen != 0) && (pWpaKey == NULL))
            {
                pWpaKey = (PWPA_KEY) &pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID];
                DBGPRINT(RT_DEBUG_INFO, "Tx Use Group Key\n");
                ThisFrameKeyID = pAdapter->PortCfg.MSSID[apidx].WPAGKeyID;
            }
	}

	if (bWdsPacket)
	{
	    pWpaKey = (PWPA_KEY) &pAdapter->WdsTab.Wpa_key;
	}
#if 0
        if (pWpaKey != NULL)
        {
            INT i;
                
            DBGPRINT_RAW(RT_DEBUG_TRACE, ("RTMPHardEncrypt TKIP Key = "));
            for (i = 0; i < 16; i++)
            {
                DBGPRINT_RAW(RT_DEBUG_TRACE, ("%02x:", pWpaKey->Key[i]));
            }
            DBGPRINT_RAW(RT_DEBUG_TRACE, ("\n"));                       
        }
#endif      
    }

    StartOfFrame = TRUE;
    // Start Copy Ndis Packet into Ring buffer.
    // For frame required more than one ring buffer (fragment), all ring buffers
    // have to be filled before kicking start tx bit.
    do
    {
        // Get the Tx Ring descriptor & Dma Buffer address
#ifndef BIG_ENDIAN
        pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;              
        pTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
#else
        pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;              
        pOriginDest = pDest;
        pDestTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif

        // Maximum allowable payload with one ring buffer, bound by fragment size
        FreeFragSize = pAdapter->PortCfg.FragmentThreshold - LENGTH_CRC;
        
        if ((pTxD->Owner == DESC_OWN_NIC) || (pTxD->CipherOwn == DESC_OWN_NIC))
        {
            // Descriptor owned by NIC. No descriptor avaliable
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            pAdapter->RalinkCounters.TxRingErrCount++;
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        if (pTxD->Valid == TRUE)
        {
            // Ndis packet of last round did not cleared.
            // This should not happen since caller guaranteed.
            // Make sure to release Tx ring resource
            pTxD->Valid = FALSE;

#ifdef BIG_ENDIAN
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
#endif
                
            pAdapter->RalinkCounters.TxRingErrCount++;
            NdisReleaseSpinLock(&pAdapter->TxRingLock);
            return (NDIS_STATUS_RESOURCES);
        }
        pAdapter->TxRing[pAdapter->CurEncryptIndex].FrameType = BTYPE_DATA;

        // Make fragment number & more fragment bit of 802.11 header
        if (StartOfFrame == TRUE)
        {
            Header_802_11.Frag = 0;         // Start of fragment burst / Single Frame
        }
        else
        {
            Header_802_11.Frag++;           // Rest of fragmented frames.
        }
        
        // Turn on with no frames after this one
        if (NumberRequired > 1)
        {
            ULONG FragSize;
            
            Header_802_11.Controlhead.Frame.MoreFrag = 1;
            // Duration should include next fragment
            // Add stuff for WPA support
            if ((EncryptType == Ndis802_11Encryption1Enabled) &&
                (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0) &&
                (EAPOLFrame == FALSE))
                FragSize = pAdapter->PortCfg.FragmentThreshold + 8;
            else if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey->KeyLen != 0) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                FragSize = pAdapter->PortCfg.FragmentThreshold + 12;
            else if ((EncryptType == Ndis802_11Encryption3Enabled) && (pWpaKey->KeyLen != 0) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                FragSize = pAdapter->PortCfg.FragmentThreshold + 16;
            else
                FragSize = pAdapter->PortCfg.FragmentThreshold;

            Header_802_11.Controlhead.Duration = 3 * pAdapter->PortCfg.Dsifs
                + 2 * AckDuration       
                + RTMPCalcDuration(pAdapter, TxRate, FragSize);
        }
        else
        {
            Header_802_11.Controlhead.Frame.MoreFrag = 0;
            // ACK size is 14 include CRC, and its rate is MaxBasicRate
            Header_802_11.Controlhead.Duration = pAdapter->PortCfg.Dsifs + AckDuration;
        }

        if (NumberRequired == 2)
        {
            ULONG FragSize;

            if (bWdsPacket)
            {
                RemainSize = RemainSize - pAdapter->PortCfg.FragmentThreshold + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_11 + LENGTH_CRC;// the length of addr4
            }
            else
            {
                RemainSize = RemainSize - pAdapter->PortCfg.FragmentThreshold + LENGTH_802_11 + LENGTH_802_11 + LENGTH_CRC;
            }

            // Add stuff for WPA support
            if ((EncryptType == Ndis802_11Encryption1Enabled) &&
                (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
                FragSize = RemainSize + 8;
            else if ((EncryptType == Ndis802_11Encryption2Enabled) &&
                (pWpaKey->KeyLen != 0) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))                
                FragSize = RemainSize + 12;
            else if ((EncryptType == Ndis802_11Encryption3Enabled) &&
                (pWpaKey->KeyLen != 0) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))                
                FragSize = RemainSize + 16;
            else
                FragSize = RemainSize;
            
            Header_802_11.Controlhead.Duration = 3 * pAdapter->PortCfg.Dsifs
                + 2 * AckDuration + RTMPCalcDuration(pAdapter, TxRate, FragSize);
        }
#if 0
		// Support Nitro mode, add additional 100 us for all data frame duration
		if (EnableTxBurst == 1)
		{
			Header_802_11.Controlhead.Duration += 100;
		}
#endif
        // Check for WEP enable bit and prepare for software WEP
        if ((EncryptType == Ndis802_11Encryption1Enabled) && (EAPOLFrame == FALSE) &&
            (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0))                            
        {
            Header_802_11.Controlhead.Frame.Wep = 1;
        }
        else if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))        
        {
            Header_802_11.Controlhead.Frame.Wep = 1;
        }
        else if ((EncryptType == Ndis802_11Encryption3Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))        
        {
            Header_802_11.Controlhead.Frame.Wep = 1;
        }
        
        // Copy 802.11 header to Tx ring buffer
        if (bWdsPacket)
        {
            NdisMoveMemory(pDest, &Header_802_11, sizeof(Header_802_11));
            NdisMoveMemory(pDest + sizeof(Header_802_11), (PUCHAR)pVirtualAddress + ETH_LENGTH_OF_ADDRESS, ETH_LENGTH_OF_ADDRESS);// SA
            pDest        += LENGTH_802_11_WITH_ADDR4;
            FreeFragSize -= (sizeof(Header_802_11) + ETH_LENGTH_OF_ADDRESS);
        }
        else
        {
            NdisMoveMemory(pDest, &Header_802_11, sizeof(Header_802_11));
            pDest        += sizeof(Header_802_11);
            FreeFragSize -= sizeof(Header_802_11);
        }

        if ((EncryptType == Ndis802_11Encryption1Enabled) && (EAPOLFrame == FALSE) &&
            (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0))
        {
            // Prepare IV, IV offset, Key for Hardware encryption
            RTMPInitWepEngine(
                pAdapter,
                pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].Key,
                pAdapter->PortCfg.MSSID[apidx].DefaultKeyId,
                pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen,
                (PUCHAR) &pTxD->Iv);
            if (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen == 5)
            {
                CipherAlg = CIPHER_WEP64;
            }
            else
            {
                CipherAlg = CIPHER_WEP128;
            }

            // Set Iv offset in TxD
			if (bWdsPacket)
			{
	            pTxD->IvOffset = LENGTH_802_11_WITH_ADDR4;
			}
			else
			{
				pTxD->IvOffset = LENGTH_802_11;
			}

            // Copy Encrypt Key to TxD
            NdisMoveMemory(
                pTxD->Key,
                pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].Key,
                pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen);            
        }
        else if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) && 
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))
        {
            INT     i = 0;

            // Prepare 8 bytes TKIP encapsulation for MPDU
            {
                TKIP_IV	tkipIv;

                tkipIv.IV16.field.rc0 = *(pWpaKey->TxTsc + 1);
                tkipIv.IV16.field.rc1 = (tkipIv.IV16.field.rc0 | 0x20) & 0x7f;
                tkipIv.IV16.field.rc2 = *pWpaKey->TxTsc;
                tkipIv.IV16.field.CONTROL.field.ExtIV = 1;  // 0: non-extended IV, 1: an extended IV
                tkipIv.IV16.field.CONTROL.field.KeyID = ThisFrameKeyID;
                tkipIv.IV32 = *(PULONG)(pWpaKey->TxTsc + 2);

#if BIG_ENDIAN == TRUE
                pTxD->Iv = (tkipIv.IV16.field.rc0 << 24) | (tkipIv.IV16.field.rc1 << 16) | (tkipIv.IV16.field.rc2 << 8) | (tkipIv.IV16.field.CONTROL.Byte);
#elif RTMP_EMBEDDED == TRUE
                pTxD->Iv = (tkipIv.IV16.field.CONTROL.Byte << 24) | (tkipIv.IV16.field.rc2 << 16) | (tkipIv.IV16.field.rc1 << 8) | (tkipIv.IV16.field.rc0);
#else
                pTxD->Iv = tkipIv.IV16.word;
#endif

                *((PUCHAR) &pTxD->Eiv) = *((PUCHAR) &tkipIv.IV32 + 3);
                *((PUCHAR) &pTxD->Eiv + 1) = *((PUCHAR) &tkipIv.IV32 + 2);
                *((PUCHAR) &pTxD->Eiv + 2) = *((PUCHAR) &tkipIv.IV32 + 1);
                *((PUCHAR) &pTxD->Eiv + 3) = *((PUCHAR) &tkipIv.IV32);
            }
            
#ifdef VINCENT_WDSTKIP
			if(!bWdsPacket)
			{
            	// Increase TxTsc value for next transmission
            	while (++pWpaKey->TxTsc[i] == 0x0)
            	{
                	i++;
                	if (i == 6)
                    	break;
            	}
			}
#else
            // Increase TxTsc value for next transmission
            while (++pWpaKey->TxTsc[i] == 0x0)
            {
                i++;
                if (i == 6)
                    break;
            }
#endif
            
            // Set IV offset
            if (bWdsPacket)
            {
                pTxD->IvOffset = LENGTH_802_11_WITH_ADDR4;
            }
            else
            {
                pTxD->IvOffset = LENGTH_802_11;
            }

            // Copy TKey
            NdisMoveMemory(pTxD->Key, pWpaKey->Key, 16);
            
            // Set Cipher suite
            CipherAlg = CIPHER_TKIP;
        }
        else if ((EncryptType == Ndis802_11Encryption3Enabled) && (pWpaKey != NULL) &&
            ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))                 
        {
            INT     i;
            PUCHAR  pTmp;

            i = 0;
            pTmp = (PUCHAR) &Iv16;
            *pTmp       = pWpaKey->TxTsc[0];
            *(pTmp + 1) = pWpaKey->TxTsc[1];
            *(pTmp + 2) = 0;
            *(pTmp + 3) = (ThisFrameKeyID << 6) | 0x20;
            Iv32 = *(PULONG)(&pWpaKey->TxTsc[2]);
            
            // Increase TxTsc value for next transmission
#ifdef VINCENT_WDSTKIP
			if(!bWdsPacket)
			{
            	while (++pWpaKey->TxTsc[i] == 0x0)
            	{
                	i++;
                	if (i == 6)
                    	break;
            	}
			}
#else
            while (++pWpaKey->TxTsc[i] == 0x0)
            {
                i++;
                if (i == 6)
                    break;
            }
#endif
            if (i == 6)
            {
                // TODO: TSC has done one full cycle, do re-keying stuff follow specs
                // Should send a special event microsoft defined to request re-key
            }
            
            // Copy IV
            NdisMoveMemory(&pTxD->Iv, &Iv16, 4);
            
            // Copy EIV
            NdisMoveMemory(&pTxD->Eiv, &Iv32, 4);
            
            // Set IV offset
            if (bWdsPacket)
            {
                pTxD->IvOffset = LENGTH_802_11_WITH_ADDR4;
            }
            else
            {
                pTxD->IvOffset = LENGTH_802_11;
            }

            // Copy TKey
            NdisMoveMemory(pTxD->Key, pWpaKey->Key, 16);

            // Set Cipher suite
            CipherAlg = CIPHER_AES;
        }
        else
            CipherAlg = CIPHER_NONE;
        
        //
        // Only the first fragment required LLC-SNAP header !!!
        //
        if ((StartOfFrame == TRUE) && (Encapped == TRUE))
        {
            // For WEP & no encryption required frame, just copy LLC header into buffer,
            // Hardware will do the encryption job.
            // For TKIP, we have to calculate MIC and store it first
            if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED)))) 
            {
				//printk("----%s----%d----\n", __FILE__, __LINE__);
                // Calculate MSDU MIC Value
#ifdef VINCENT_WDSTKIP
                RTMPCalculateMICValue(pAdapter,skb, (PUCHAR)(&Header_802_11), 
						bWdsPacket,pEncap, 6, pWpaKey);
#else
                RTMPCalculateMICValue(pAdapter, skb, pEncap, 6, pWpaKey);
#endif
            }
            
            // Copy LLC header
            NdisMoveMemory(pDest, pEncap, 6);
            pDest += 6;

            // Copy protocol type
            pSrc = (PUCHAR) pVirtualAddress;
            NdisMoveMemory(pDest, pSrc + 12, 2);
            pDest += 2;
            
            // Exclude 802.3 header size, we will recalculate the size at
            // the end of fragment preparation.
            Length -= 14;
            pSrc += 14;
            FreeFragSize -= LENGTH_802_1_H;
        }
        else if ((StartOfFrame == TRUE) && (Encapped == FALSE))
        {
            if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) && 
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED)))) 
            {
				//printk("----%s----%d----\n", __FILE__, __LINE__);
                // Calculate MSDU MIC Value
#ifdef VINCENT_WDSTKIP
                RTMPCalculateMICValue(pAdapter, skb,(PUCHAR)(&Header_802_11), 
						bWdsPacket,pEncap, 0, pWpaKey);
#else
                RTMPCalculateMICValue(pAdapter, skb, pEncap, 0, pWpaKey);
#endif
            }
            pSrc = (PUCHAR) pVirtualAddress + 14;
            Length -= 14;
        }
        
        // Start copying payload
        BytesCopied = 0;
        do
        {
            if (Length >= FreeFragSize)
            {
                // Copy only the free fragment size, and save the pointer
                // of current buffer descriptor for next fragment buffer.
                NdisMoveMemory(pDest, pSrc, FreeFragSize);
                BytesCopied += FreeFragSize;
                pSrc        += FreeFragSize;
                pDest       += FreeFragSize;
                Length      -= FreeFragSize;
                break;
            }
            else
            {
                // Copy the rest of this buffer descriptor pointed data
                // into ring buffer.
                NdisMoveMemory(pDest, pSrc, Length);
                BytesCopied  += Length;
                pDest        += Length;
                FreeFragSize -= Length;
            }

            // No more buffer descriptor
            // Add MIC value if needed
            if ((EncryptType == Ndis802_11Encryption2Enabled) && (MICFrag == FALSE) && (pWpaKey != NULL) &&
                ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))         
            {
                Length = 8;     // Set length to MIC length
#if 0
				{
					int tmpi;
                	printk("Calculated TX MIC value =\n");
                	for (tmpi = 0; tmpi < 8; tmpi++)
                	{
                    	printk("%02x:", pAdapter->PrivateInfo.Tx.MIC[tmpi]);  
                	}
                	printk("\n"); 
				}
#endif
                if (FreeFragSize >= Length)
                {
                    NdisMoveMemory(pDest, pAdapter->PrivateInfo.Tx.MIC, Length);
                    BytesCopied  += Length;
                    pDest        += Length;
                    FreeFragSize -= Length;
                    Length = 0;
                    RemainSize   += 8;  // Need to add MIC as payload
                }
                else
                {
                    NdisMoveMemory(pDest, pAdapter->PrivateInfo.Tx.MIC, FreeFragSize);
                    BytesCopied  += FreeFragSize;
                    pSrc          = pAdapter->PrivateInfo.Tx.MIC + FreeFragSize;
                    pDest        += FreeFragSize;
                    Length       -= FreeFragSize;
                    MICFrag       = TRUE;
                    RemainSize   += (8 - FreeFragSize); // Need to add MIC as payload
                }                
            }
        }   while (FALSE);       // End of copying payload
                
        // Real packet size, No 802.1H header for fragments except the first one.
        if ((StartOfFrame == TRUE) && (Encapped == TRUE))
        {
			if (bWdsPacket)
			{
				TxSize = BytesCopied + LENGTH_802_11_WITH_ADDR4 + LENGTH_802_1_H; // the length of addr4
			}
			else
			{
				TxSize = BytesCopied + LENGTH_802_11 + LENGTH_802_1_H;
			}
        }
        else
        {
			if (bWdsPacket)
			{
				TxSize = BytesCopied + LENGTH_802_11_WITH_ADDR4; // contain the length of addr4
			}
			else
			{
				TxSize = BytesCopied + LENGTH_802_11;
			}
        }

        RemainSize = RemainSize - BytesCopied;
            
        if ((EncryptType == Ndis802_11Encryption1Enabled) && (EAPOLFrame == FALSE) &&
            (pAdapter->PortCfg.MSSID[apidx].SharedKey[pAdapter->PortCfg.MSSID[apidx].DefaultKeyId].KeyLen != 0))
        {
            // IV + ICV which ASIC added after encryption done
            TxSize += 8;
        }
        else if ((EncryptType == Ndis802_11Encryption2Enabled) && (pWpaKey != NULL) &&
        ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))                 
        {
            // IV + EIV + ICV which ASIC added after encryption done
            TxSize += 12;
        }
        else if ((EncryptType == Ndis802_11Encryption3Enabled) && (pWpaKey != NULL) &&
        ((EAPOLFrame == FALSE) || (pEntry && (pEntry->PortSecured == WPA_802_1X_PORT_SECURED))))         
        {
            // IV + EIV + HW MIC
            TxSize += 16;
        }
                
        // Prepare Tx descriptors before kicking tx.
        // The BBP register index in Tx descriptor has to be configured too.
#ifdef BIG_ENDIAN
        RTMPFrameEndianChange(pAdapter, pOriginDest, DIR_WRITE, FALSE);
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
        *pDestTxD = TxD;
        pTxD = pDestTxD;
#endif
        if (Header_802_11.Controlhead.Addr1.Octet[0] & 0x01)
        {
            // Multicast, retry bit is off
            RTMPWriteTxDescriptor(pTxD, TRUE, CipherAlg, FALSE, FALSE, FALSE, RetryMode, FrameGap, TxRate, 4, TxSize, pAdapter->PortCfg.TxPreambleInUsed, AccessCategory);
        }
        else
        {
            RTMPWriteTxDescriptor(pTxD, TRUE, CipherAlg, TRUE, FALSE, FALSE, RetryMode, FrameGap, TxRate, 4, TxSize, pAdapter->PortCfg.TxPreambleInUsed, AccessCategory);
        }

#if 0
        // debug print
        if (CipherAlg == CIPHER_WEP64)
        {
            PUCHAR ptr = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;
            int i;
            DBGPRINT_RAW(RT_DEBUG_TRACE, ("RTMPHardEncrypt - Algo=%d, len=%d\n", CipherAlg, TxSize));
            for (i=0;i<3;i++)
            {
                DBGPRINT_RAW(RT_DEBUG_TRACE, ("%02x%02x%02x%02x-%02x%02x%02x%02x  %02x%02x%02x%02x-%02x%02x%02x%02x\n", 
                    *ptr, *(ptr+1), *(ptr+2), *(ptr+3), *(ptr+4), *(ptr+5), *(ptr+6), *(ptr+7), 
                    *(ptr+8), *(ptr+9), *(ptr+10), *(ptr+11), *(ptr+12), *(ptr+13), *(ptr+14), *(ptr+15) ));
                ptr += 16;
            }
        }
#endif

        // Set frame gap for the rest of fragment burst.
        // It won't matter if there is only one fragment (single fragment frame).
        StartOfFrame = FALSE;
        FrameGap     = IFS_SIFS;
        NumberRequired--;
        
        // Increase & maintain Tx Ring Index
        pAdapter->CurEncryptIndex++;
        if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
        {
            pAdapter->CurEncryptIndex = 0;
        }
        
        pAdapter->RalinkCounters.EncryptCount++;
		INC_COUNTER(pAdapter->PortCfg.MSSID[apidx].TxOkCount);
		pAdapter->PortCfg.MSSID[apidx].TransmittedByteCount +=  pTxD->DataByteCnt;
    }   while (NumberRequired > 0);

    // Kick Encrypt Control Register at the end of all ring buffer preparation
    RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
    
    // Acknowledge protocol send complete of pending packet.
    RTMPFreeSkbBuffer(skb);
    
    // Make sure to release Tx ring resource
    NdisReleaseSpinLock(&pAdapter->TxRingLock);

    return (NDIS_STATUS_SUCCESS);
}

/*
    ========================================================================
    Routine Description:
        Calculates the duration which is required to transmit out frames 
    with given size and specified rate.
        
    Arguments:
        pAdapter        Pointer to our adapter
        Rate            Transmit rate
        Size            Frame size in units of byte
        
    Return Value:
        Duration number in units of usec
    ========================================================================
*/
inline USHORT  RTMPCalcDuration(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  UCHAR           Rate,
    IN  ULONG           Size)
{
    ULONG   Duration = 0;

    if (Rate < RATE_FIRST_OFDM_RATE) // CCK
    {
        if ((Rate > RATE_1) && (pAdapter->PortCfg.TxPreambleInUsed == Rt802_11PreambleShort))
            Duration = 96;  // 72+24 preamble+plcp
        else
            Duration = 192; // 144+48 preamble+plcp
            
        Duration += (USHORT)((Size << 4) / RateIdTo500Kbps[Rate]);
        if ((Size << 4) % RateIdTo500Kbps[Rate])
            Duration ++;
    }
    else // OFDM rates
    {
        Duration = 20 + 6;      // 16+4 preamble+plcp + Signal Extension
        Duration += 4 * (USHORT)((11 + Size * 4) / RateIdTo500Kbps[Rate]);
        if ((11 + Size * 4) % RateIdTo500Kbps[Rate])
            Duration += 4;
    }
    
    return (USHORT)Duration;
    
}

/*
    ========================================================================
    Routine Description:
        Calculates the duration which is required to transmit out frames 
    with given size and specified rate.
        
    Arguments:
        pTxD        Pointer to transmit descriptor
        Ack         Setting for Ack requirement bit
        Fragment    Setting for Fragment bit
        RetryMode   Setting for retry mode
        Ifs         Setting for IFS gap
        Rate        Setting for transmit rate
        Service     Setting for service
        Length      Frame length
    ========================================================================
*/
inline VOID    RTMPWriteTxDescriptor(
    IN  PTXD_STRUC  pSourceTxD,
    IN  BOOLEAN     DoEncrypt,
    IN  UCHAR       CipherAlg,
    IN  BOOLEAN     Ack,
    IN  BOOLEAN     Fragment,
    IN  BOOLEAN     InsTimestamp,
    IN  UCHAR       RetryMode,
    IN  UCHAR       Ifs,
    IN  UINT        Rate,
    IN  UCHAR       Service,
    IN  UINT        Length,
    IN  USHORT      TxPreamble,
	IN  UCHAR       AccessCategory)
{
    UINT            Residual;
    PTXD_STRUC      pTxD;
    
#ifndef BIG_ENDIAN
    pTxD = pSourceTxD;
#else
    TXD_STRUC       TxD;
    
    TxD = *pSourceTxD;
    pTxD = &TxD;
    RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif

    pTxD->MoreFrag    = Fragment;
    pTxD->ACK         = Ack;
    pTxD->Timestamp   = InsTimestamp;
    pTxD->RetryMd     = RetryMode;
    pTxD->IFS         = Ifs;
    pTxD->DataByteCnt = Length;
    pTxD->TxRate      = Rate;
	switch (AccessCategory) // 802.11e/d4.4 June/2003
	{
	    case 3: // TC3, <AIFSN, CwMin, CwMax> = <1, aCwMin/4, aCwMin/2>
        	pTxD->CWmin       = CW_MIN_IN_BITS - 2;
        	pTxD->CWmax       = CW_MIN_IN_BITS - 1;
        	pTxD->Aifs        = 1;
        	break;
        case 2: // TC2, <AIFSN, CwMin, CwMax> = <1, aCwMin/2, aCwMin>
        	pTxD->CWmin       = CW_MIN_IN_BITS - 1;
        	pTxD->CWmax       = CW_MIN_IN_BITS;
        	pTxD->Aifs        = 1;
        	break;
        case 1: // TC1, <AIFSN, CwMin, CwMax> = <1, aCwMin, aCwMax>
        	pTxD->CWmin       = CW_MIN_IN_BITS;
        	pTxD->CWmax       = CW_MAX_IN_BITS;
        	pTxD->Aifs        = 1;
        	break;
        case 0: // TC0, <AIFSN, CwMin, CwMax> = <2, aCwMin, aCwMax>
        default:
        	pTxD->CWmin       = CW_MIN_IN_BITS;
        	pTxD->CWmax       = CW_MAX_IN_BITS;
        	pTxD->Aifs        = 2;
        	break;
	}
        
    if (Rate < RATE_FIRST_OFDM_RATE)
        pTxD->Ofdm = 0;
    else
        pTxD->Ofdm = 1;

    // fill up PLCP SIGNAL field
    pTxD->PlcpSignal = PlcpSignal[Rate];
    if (((Rate == RATE_2) || (Rate == RATE_5_5) || (Rate == RATE_11)) && (TxPreamble == Rt802_11PreambleShort)) // no short preamble for RATE_1
    {
        pTxD->PlcpSignal |= 0x0008;
    }

    // fill up PLCP SERVICE field, not used for OFDM rates
    pTxD->PlcpService = Service;

    // file up PLCP LENGTH_LOW and LENGTH_HIGH fields
    Length += 4;
    if (Rate < RATE_FIRST_OFDM_RATE)    // 11b - RATE_1, RATE_2, RATE_5_5, RATE_11
    {
        if ((Rate == RATE_1) || ( Rate == RATE_2))
        {
            Length = Length * 8 / (Rate + 1);
        }
        else
        {
            Residual = ((Length * 16) % (11 * (1 + Rate - RATE_5_5)));
            Length = Length * 16 / (11 * (1 + Rate - RATE_5_5));
            if (Residual != 0)
            {
                Length++;
            }
            if ((Residual <= (3 * (1 + Rate - RATE_5_5))) && (Residual != 0))
            {
                if (Rate == RATE_11)			// Only 11Mbps require length extension bit
                    pTxD->PlcpService |= 0x80; // 11b's PLCP Length extension bit
            }
        }

        pTxD->PlcpLengthHigh = Length / 256;
        pTxD->PlcpLengthLow = Length % 256;
    }
    else    // OFDM - RATE_6, RATE_9, RATE_12, RATE_18, RATE_24, RATE_36, RATE_48, RATE_54
    {
        pTxD->PlcpLengthHigh = Length / 64;  // high 6-bit of total byte count
        pTxD->PlcpLengthLow = Length % 64;   // low 6-bit of total byte count
    }

    if (DoEncrypt == TRUE)      // Do encryption only
    {
        pTxD->Owner     = DESC_OWN_HOST;
        pTxD->Valid     = FALSE;
        pTxD->CipherAlg = CipherAlg;
        pTxD->CipherOwn = DESC_OWN_NIC;
    }
    else    // Hard transmit
    {
        pTxD->Valid     = TRUE;
        pTxD->CipherAlg = CIPHER_NONE;
        pTxD->CipherOwn = DESC_OWN_HOST;
        pTxD->Owner     = DESC_OWN_NIC;
    }
#ifdef BIG_ENDIAN
    RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
    *pSourceTxD = *pTxD;
#endif
}

/*
    ========================================================================
    Routine Description:
        Search tuple cache for receive duplicate frame from unicast frames.
        
    Arguments:
        pAdapter        Pointer to our adapter
        pHeader         802.11 header of receiving frame
        
    Return Value:
        TRUE            found matched tuple cache
        FALSE           no matched found
    ========================================================================
*/
BOOLEAN RTMPSearchTupleCache(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    INT Index;

    for (Index = 0; Index < MAX_CLIENT; Index++)
    {
        if (pAdapter->TupleCache[Index].Valid == FALSE)
            continue;
        
        if (RTMPEqualMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6) &&
            (pAdapter->TupleCache[Index].Sequence == pHeader->Sequence) &&
            (pAdapter->TupleCache[Index].Frag == pHeader->Frag))
        {
//			DBGPRINT(RT_DEBUG_TRACE,"DUPCHECK - duplicate frame hit entry %d\n", Index); 
            return (TRUE);
        }
    }
    return (FALSE);
}

/*
    ========================================================================
    Routine Description:
        Update tuple cache for new received unicast frames.
        
    Arguments:
        pAdapter        Pointer to our adapter
        pHeader         802.11 header of receiving frame
    ========================================================================
*/
VOID    RTMPUpdateTupleCache(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    UCHAR Index;

    for (Index = 0; Index < MAX_CLIENT; Index++)
    {
        if (pAdapter->TupleCache[Index].Valid == FALSE)
        {
            // Add new entry
            NdisMoveMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6);
            pAdapter->TupleCache[Index].Sequence = pHeader->Sequence;
            pAdapter->TupleCache[Index].Frag     = pHeader->Frag;
            pAdapter->TupleCache[Index].Valid    = TRUE;
//			DBGPRINT(RT_DEBUG_INFO,"DUPCHECK - Add Entry %d, MAC=%02x:%02x:%02x:%02x:%02x:%02x\n", 
//			    Index, pAdapter->TupleCache[Index].MAC.Octet[0], pAdapter->TupleCache[Index].MAC.Octet[1],
//			    pAdapter->TupleCache[Index].MAC.Octet[2], pAdapter->TupleCache[Index].MAC.Octet[3],
//			    pAdapter->TupleCache[Index].MAC.Octet[4], pAdapter->TupleCache[Index].MAC.Octet[5]);
            return;
        }
        else if (RTMPEqualMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6))
        {
            // Update old entry
            pAdapter->TupleCache[Index].Sequence = pHeader->Sequence;
            pAdapter->TupleCache[Index].Frag     = pHeader->Frag;
            return;
        }
    }

    // tuple cache full, replace the first inserted one (even though it may not be
    // least referenced one)
	if (Index == MAX_CLIENT)
	{
	    pAdapter->TupleCacheLastUpdateIndex ++;
	    if (pAdapter->TupleCacheLastUpdateIndex >= MAX_CLIENT)
	        pAdapter->TupleCacheLastUpdateIndex = 0;
	    Index = pAdapter->TupleCacheLastUpdateIndex;

		// replace with new entry
		NdisMoveMemory(&pAdapter->TupleCache[Index].MAC, &pHeader->Controlhead.Addr2, 6);
		pAdapter->TupleCache[Index].Sequence = pHeader->Sequence;
		pAdapter->TupleCache[Index].Frag     = pHeader->Frag;
		pAdapter->TupleCache[Index].Valid    = TRUE;
//		DBGPRINT(RT_DEBUG_INFO,"DUPCHECK - replace Entry %d, MAC=%02x:%02x:%02x:%02x:%02x:%02x\n", 
//		    Index, pAdapter->TupleCache[Index].MAC.Octet[0], pAdapter->TupleCache[Index].MAC.Octet[1],
//		    pAdapter->TupleCache[Index].MAC.Octet[2], pAdapter->TupleCache[Index].MAC.Octet[3],
//		    pAdapter->TupleCache[Index].MAC.Octet[4], pAdapter->TupleCache[Index].MAC.Octet[5]);
	}
}

/*
    ========================================================================
    Routine Description:
        Check Rx descriptor, return NDIS_STATUS_FAILURE if any error found
    ========================================================================
*/
NDIS_STATUS RTMPCheckRxDescriptor(
    IN  PRXD_STRUC  pRxD)
{
    if (pRxD->PhyErr)
        return NDIS_STATUS_FAILURE;
    else if (pRxD->Crc)
        return NDIS_STATUS_FAILURE;
    else
        return NDIS_STATUS_SUCCESS;
}

/*
  ========================================================================
  Return:
    NDIS_STATUS_SUCCESS - This received unicast frame is NOT a duplicate
    NDIS_STATUS_FAILURE - received frame is duplicated and should be dropped
  ========================================================================
*/
NDIS_STATUS RTMPCheckDuplicateFrame(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    if (pHeader->Controlhead.Frame.Retry)
    {
        if (RTMPSearchTupleCache(pAdapter, pHeader) == TRUE)
        {
            // Found retry frame in tuple cache, Discard this frame / fragment
            // Increase 802.11 counters
            INC_COUNTER(pAdapter->WlanCounters.FrameDuplicateCount);
            return NDIS_STATUS_FAILURE;
        }
    }
    RTMPUpdateTupleCache(pAdapter, pHeader);
    return NDIS_STATUS_SUCCESS;
}

/*
  ========================================================================
  Description:
    This routine checks if a received frame causes class 2 or class 3
    error, and perform error action (DEAUTH or DISASSOC) accordingly
  ========================================================================
*/
BOOLEAN RTMPCheckClass2Class3Error(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PHEADER_802_11  pHeader)
{
    SST     Sst;
    UCHAR   PsMode, Rate;
    USHORT  Aid;
                        
    SsPsInquiry(pAdapter, &pHeader->Controlhead.Addr2, &Sst, &Aid, &PsMode, &Rate);
    if (Sst == SST_ASSOC)
        ; // okay to receive this DATA frame
    else if (Sst == SST_AUTH)
    {
        Cls3errAction(pAdapter, &pHeader->Controlhead.Addr2);
        return TRUE; 
    }
    else
    {
        Cls2errAction(pAdapter, &pHeader->Controlhead.Addr2);
        return TRUE; 
    }
    return FALSE;
}

/*
  ========================================================================
  Description:
    This routine frees all packets in PSQ that's destined to a specific DA.
    BCAST/MCAST in DTIMCount=0 case is also handled here, just like a PS-POLL 
    is received from a WSTA which has MAC address FF:FF:FF:FF:FF:FF
  ========================================================================
*/
VOID RTMPHandleRxPsPoll(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PMACADDR        pAddr,
    IN  USHORT          Aid)
{
    PQUEUE_ENTRY        pEntry;
    PMAC_TABLE_ENTRY    pMacEntry;
    
    //DBGPRINT(RT_DEBUG_TRACE,("rcv PS-POLL (AID=%d) from %02x:%02x:%02x:%02x:%02x:%02x\n", 
    //    Aid, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5]));
    pMacEntry = MacTableLookup(pAdapter, pAddr);
    if (pMacEntry)
    {
        Aid = pMacEntry->Aid;
        NdisAcquireSpinLock(&pAdapter->PortCfg.MSSID[pMacEntry->ApIdx].MacTabLock);
        NdisAcquireSpinLock(&pAdapter->TxSwQueueLock);
        // cleanup all backlogged frames in PSQ
        while (pMacEntry->PsQueue.Head)
        {
            pEntry = RemoveHeadQueue(&pMacEntry->PsQueue);
            InsertTailQueue(&pAdapter->TxSwQueue0, pEntry);
            DBGPRINT(RT_DEBUG_TRACE, "rx PSPOLL, tx frame out...\n");
        }
        NdisReleaseSpinLock(&pAdapter->TxSwQueueLock);
        NdisReleaseSpinLock(&pAdapter->PortCfg.MSSID[pMacEntry->ApIdx].MacTabLock);
        if ((Aid > 0) && (Aid < MAX_LEN_OF_MAC_TABLE * MAX_MSSID_NUM))
            pAdapter->Mlme.TimBitmap[Aid/8] &= (~Bit[Aid%8]);
        pMacEntry->PsQIdleCount = 0;

        // Dequeue one frame from TxSwQueue0..3 queue and process it
        // There are two place calling dequeue for TX ring.
        // 1. Here, right after queueing the frame.
        // 2. At the end of TxRingTxDone service routine.
        // 3. Within RTMPBridgeToWirelessSta() service routine
//        if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
        {
            RTMPDeQueuePacket(pAdapter);
        }   
    }
}
#if defined(VINCENT_WDSFAST)
/*
    ========================================================================
    Routine Description:
        This function checks DA is in the DA MAC table or not. If not,
		insert the DA and it's TA number according to WDS table into the 
		table.(Actually, DA is SA.)
        
    Arguments:
        pAdapter        Pointer to the Rx descriptor
		damac			Pointer to the DA MAC
        
    Return Value:
		-1		  :	DA table is full or find the DA.
	positive value: DA is inserted in the number of DA table.
    ========================================================================
*/
int	Insert_DA(
		IN	PRTMP_ADAPTER	pAd,
		IN	PUCHAR			TA,
		IN	PUCHAR			SA)
{
	int ii, jj, wdsnum=-1, flag=FALSE;

	if(!memcmp(&pAd->CurrentAddress[0], SA, ETH_LENGTH_OF_ADDRESS))
		return -3;
		
	read_lock(&pAd->da_rwlock);
	for (ii=0; ii<MAX_DA_TABLE; ii++)
	{
		if((atomic_read(&pAd->datable[ii].valid) == TRUE) 
				&& !memcmp(&pAd->datable[ii].SA[0], SA, ETH_LENGTH_OF_ADDRESS))
		{
			pAd->datable[ii].leasetime = jiffies;
			flag = TRUE; //find it
			break;
		}
	}
	read_unlock(&pAd->da_rwlock);

	if(flag != TRUE)
	{
		for(jj=0; jj<MAX_WDS_ENTRY; jj++)
		{
			if ((pAd->WdsTab.WdsEntry[jj].Valid == TRUE) 
					&& (!memcmp(&pAd->WdsTab.WdsEntry[jj].WdsAddr, TA, ETH_LENGTH_OF_ADDRESS)))
			{
				wdsnum = jj;
				break;
			}
		}

		if (wdsnum == MAX_WDS_ENTRY)
			return -2; // TA not find

		// insert it
		write_lock(&pAd->da_rwlock);
		for (ii=0; ii<MAX_DA_TABLE; ii++)
		{
			if(atomic_read(&pAd->datable[ii].valid) == FALSE)
			{
			//	printk("KERN ---- %s %d ---- add WDS entry %d\n", __FILE__, __LINE__, ii);
				memcpy(&pAd->datable[ii].SA[0],SA,ETH_LENGTH_OF_ADDRESS);
				atomic_set(&pAd->datable[ii].wds, wdsnum);
				pAd->datable[ii].leasetime = jiffies;
				atomic_set(&pAd->datable[ii].valid, TRUE);
				break;
			}
		}
		write_unlock(&pAd->da_rwlock);
	}		

	return -1;
}

/*
    ========================================================================
    Routine Description:
        This function checks if the received frames should be sent out to the 
        wireless media based on DA look-up in the wirelss DA MAC table.
        
    Arguments:
        pAdapter        Pointer to the Rx descriptor
		damac			Pointer to the DA MAC
        
    Return Value:
		-1		  :	This DA not exist in our DA table.
	positive value: Match which WDS item in DA table. This value is the same
					with WDS entry table in pAdapter->WdsTab.WdsEntry[i].
    ========================================================================
*/
int	Search_DA(
		IN	PRTMP_ADAPTER	pAd,
		IN	PUCHAR			DA)
{
	register int ii, jj;

	// find the DA from DA table
	write_lock(&pAd->da_rwlock);
	for (ii = 0; ii<MAX_DA_TABLE; ii++)
	{
		// find
		if((atomic_read(&pAd->datable[ii].valid)==TRUE)
				&& !memcmp(&pAd->datable[ii].SA[0], DA, ETH_LENGTH_OF_ADDRESS))
		{
			pAd->datable[ii].leasetime = jiffies;
			write_unlock(&pAd->da_rwlock);
			return atomic_read(&pAd->datable[ii].wds);
		}
	}
	write_unlock(&pAd->da_rwlock);

	// find the DA from WDS entry
	for (jj = 0; jj < MAX_WDS_ENTRY; jj++)
	{
		if ((pAd->WdsTab.WdsEntry[jj].Valid == TRUE) && 
				(MAC_ADDR_EQUAL(&pAd->WdsTab.WdsEntry[jj].WdsAddr, DA)))
		{
			return jj;
		}
	}

	return -1;
}

/*
    ========================================================================
    Routine Description:
        This function checks if the received frames should be sent out to the 
        wireless media based on DA look-up in the wirelss MAC table.
        
    Arguments:
        pRxD        Pointer to the Rx descriptor
        
    Return Value:
        TRUE -  the input frame is destined to a known wireless STA 
                and had being sent out to the wireless media. It's not 
                necessary to make another copy up to LLC nor Portal function.
        FALSE - the input frame still need to deliver to the upper LLC
    ========================================================================
*/
BOOLEAN RTMPBridgeToWirelessWDS(
    IN  PRTMP_ADAPTER   pAd,
    IN  PUCHAR          pHeader802_3,
    IN  UINT            HdrLen,
    IN  PUCHAR          pData,
    IN  UINT            DataLen,
    IN  UINT            FromWhichSSID)
{
    BOOLEAN			result = FALSE;
	int				wdsnum=FALSE, Status;
    struct sk_buff *skb;
    
	wdsnum = Search_DA(pAd, pHeader802_3);

	if(wdsnum < 0)
		return FALSE;

	do{
#ifdef RTMP_EMBEDDED
    	if ((skb = __dev_alloc_skb(HdrLen + DataLen + 2, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
        if ((skb = dev_alloc_skb(HdrLen + DataLen + 2)) != NULL)
#endif
        {
    	    skb->len = HdrLen + DataLen;
           	memcpy((skb->data), pHeader802_3, HdrLen);
           	memcpy((skb->data + HdrLen), pData, DataLen);
       	}
       	else
       	{
           	break;
       	}
            	
		// b skb_buff ̰ WDS number
       	RTMP_SET_PACKET_NET_DEVICE_WDS(skb, wdsnum);
           	
       	Status = RTMPSendPacket(pAd, skb);
       	if (Status == NDIS_STATUS_SUCCESS)
       	{
        	RTMPDeQueuePacket(pAd);
        }
        else // free this packet space
        {
           	RTMPFreeSkbBuffer(skb);
        }

        result = TRUE;
	} while(FALSE);

    return result;
}
#endif
/*
    ========================================================================
    Routine Description:
        This function checks if the received frames should be sent out to the 
        wireless media based on DA look-up in the wirelss MAC table.
        
    Arguments:
        pRxD        Pointer to the Rx descriptor
        
    Return Value:
        TRUE -  the input frame is destined to a known wireless STA 
                and had being sent out to the wireless media. It's not 
                necessary to make another copy up to LLC nor Portal function.
        FALSE - the input frame still need to deliver to the upper LLC
    ========================================================================
*/
BOOLEAN RTMPBridgeToWirelessSta(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PUCHAR          pHeader802_3,
    IN  UINT            HdrLen,
    IN  PUCHAR          pData,
    IN  UINT            DataLen,
    IN  UINT            FromWhichSSID)
{
    SST				Sst;
    USHORT			Aid;
    UCHAR			PsMode, Rate;
    MAC_TABLE_ENTRY	*pEntry;
    BOOLEAN			result = FALSE;
    struct sk_buff *skb;
    
    // decide the return value. TRUE if no need to indicate to LLC, FALSE otherwise
    pEntry = SsPsInquiry(pAdapter, (MACADDR *)pHeader802_3, &Sst, &Aid, &PsMode, &Rate);
    if (pHeader802_3[0] & 0x01)
        result = FALSE;
    else if (pEntry && (Sst == SST_ASSOC)) 
        result = TRUE;
    else
        result = FALSE;
    
	/* unicast packet */
    if ((pHeader802_3[0] & 0x01) != 0x01)
    {
	    if(!pEntry)
	    {// STAs not in wireless
	        return FALSE;
	    }
	    else if (FromWhichSSID == pEntry->ApIdx)
	    {// STAs in same SSID
	        if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].IsolateInterStaTraffic == 1))
	        {
	            return result;
	        }
	        else
	        {   
	        }
	    }
	    else
	    {// STAs in different SSID
	        if (pAdapter->PortCfg.IsolateInterStaTrafficBTNSSID == 1)
	        {
	            return TRUE;
	        }
	        else
	        {
	        }
	    }
	}

    if ((pHeader802_3[0] & 0x01) || (pEntry && (Sst == SST_ASSOC)))
    {
        NDIS_STATUS     Status;

        do {
            // 1. build a NDIS packet and call RTMPSendPacket();
            //    be careful about how/when to release this internal allocated NDIS PACKET buffer
#ifdef RTMP_EMBEDDED
            if ((skb = __dev_alloc_skb(HdrLen + DataLen + 2, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
            if ((skb = dev_alloc_skb(HdrLen + DataLen + 2)) != NULL)
#endif
            {
                skb->len = HdrLen + DataLen;
                memcpy((skb->data), pHeader802_3, HdrLen);
                memcpy((skb->data + HdrLen), pData, DataLen);
            }
            else
            {
                break;
            }
            	
            // 1.1 apidx != 0, then we need set packet mssid attribute.
            if(pEntry && (pEntry->ApIdx != 0))
                RTMP_SET_PACKET_NET_DEVICE_MSSID(skb, pEntry->ApIdx);
            	
            // 2. send out the packet
            DBGPRINT(RT_DEBUG_INFO,"WBRG - len=%d, DA=%02x:%02x:%02x:%02x:%02x:%02x, SA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
                HdrLen+DataLen,pHeader802_3[0],pHeader802_3[1],pHeader802_3[2],pHeader802_3[3],pHeader802_3[4],pHeader802_3[5],
                pHeader802_3[6],pHeader802_3[7],pHeader802_3[8],pHeader802_3[9],pHeader802_3[10],pHeader802_3[11]);

            Status = RTMPSendPacket(pAdapter, skb);
            if (Status == NDIS_STATUS_SUCCESS)
            {
                // Dequeue one frame from TxSwQueue0..3 queue and process it
                // There are three place calling dequeue for TX ring.
                // 1. Here, right after queueing the frame.
                // 2. At the end of TxRingTxDone service routine.
                // 3. Upon NDIS call RTMPSendPackets
                // if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
                {
                    RTMPDeQueuePacket(pAdapter);
                }   
            }
            else // free this packet space
            {
                RTMPFreeSkbBuffer(skb);
            }

        } while (FALSE);
    }

    return result;
}

/*
	========================================================================

	Routine	Description:
		Check and fine the packet waiting in SW queue with highest priority
		
	Arguments:
		pAdapter	Pointer	to our adapter
		
	Return Value:
		pQueue		Pointer to Waiting Queue

	Note:
	
	========================================================================
*/
PQUEUE_HEADER   RTMPCheckTxSwQueue(
    IN	PRTMP_ADAPTER   pAdapter,
    OUT	ULONG           *Number,
    OUT UCHAR           *AccessCategory)
{
#ifdef QOS
    // Calculate total number of packets waiting in queues for Nitro mode 
    *Number = pAdapter->TxSwQueue3.Number + pAdapter->TxSwQueue2.Number + pAdapter->TxSwQueue1.Number +
        pAdapter->TxSwQueue0.Number;

    if (pAdapter->TxSwQueue3.Head != NULL)
    {
        *AccessCategory = 3;
        return (&pAdapter->TxSwQueue3);
    }
    else if (pAdapter->TxSwQueue2.Head != NULL)
    {
        *AccessCategory = 2;
        return (&pAdapter->TxSwQueue2);
    }
    else if (pAdapter->TxSwQueue1.Head != NULL)
    {
        *AccessCategory = 1;
        return (&pAdapter->TxSwQueue1);
    }
    else if (pAdapter->TxSwQueue0.Head != NULL)
    {
        *AccessCategory = 0;
        return (&pAdapter->TxSwQueue0);
    }

    // No packet pending in Tx Sw queue
    *AccessCategory = 0;
    return (NULL);
#else
	*Number = pAdapter->TxSwQueue0.Number;
    *AccessCategory = 0;
    if (pAdapter->TxSwQueue0.Head != NULL)
		return (&pAdapter->TxSwQueue0);
    else
        return NULL;
#endif
}

#ifdef BIG_ENDIAN
/*
	========================================================================

	Routine	Description:
		Endian conversion of Tx/Rx descriptor .
		
	Arguments:
		pAdapter		Pointer	to our adapter
		pData			Pointer	to Tx/Rx descriptor
		DescriptorType  Direction of the frame
		
	Return Value:
		None
		
	Note:
        Call this function when read or update descriptor
	========================================================================
*/
VOID	RTMPDescriptorEndianChange(
	IN	PUCHAR			pData,
	IN	ULONG			DescriptorType)
{
    *((ULONG *)(pData + 40)) = SWAP32(*((ULONG *)(pData + 40)));        // Byte 10
    if(DescriptorType == TYPE_TXD)
        *((ULONG *)(pData + 8)) = SWAP32(*((ULONG *)(pData + 8)));      // Byte 2
    *(ULONG *)pData = SWAP32(*(ULONG *)pData);                          // Byte 0; this must be swapped last
}

VOID	WriteBackToDescriptor(
	IN  PUCHAR			Dest,
 	IN	PUCHAR			Src,
    IN  BOOLEAN			DoEncrypt,
	IN  ULONG           DescriptorType)
{
	PULONG p1, p2;
	UCHAR i;
	
	p1 = ((PULONG)Dest) + 1;
	p2 = ((PULONG)Src) + 1;
	for (i = 1; i < RING_DESCRIPTOR_SIZE/4 ; i++)
		*p1++ = *p2++;
	*(PULONG)Dest = *(PULONG)Src;
}

/*
	========================================================================

	Routine	Description:
		Endian conversion of all kinds of 802.11 frames .
		
	Arguments:
		pAdapter		Pointer	to our adapter
		pData			Pointer	to the 802.11 frame structure
		Dir				Direction of the frame
		FromRxDoneInt   Caller is from RxDone interrupt
		
	Return Value:
		None
		
	Note:
	    Call this function when read or update buffer data
	========================================================================
*/
VOID	RTMPFrameEndianChange(
	IN  PRTMP_ADAPTER   pAdapter, 
	IN  PUCHAR          pData, 
	IN  ULONG           Dir,
	IN  BOOLEAN         FromRxDoneInt)
{
    PMACHDR	pFrame;
    PUCHAR  pMacHdr;

    // swab 16 bit fields - Frame Control field
    if(Dir == DIR_READ)
    {
        *(USHORT *)pData = SWAP16(*(USHORT *)pData);
    }

    pFrame = (PMACHDR) pData;
    pMacHdr = (PUCHAR) pFrame;

    // swab 16 bit fields - Duration/ID field
    *(USHORT *)(pMacHdr + 2) = SWAP16(*(USHORT *)(pMacHdr + 2));
    
    // swab 16 bit fields - Sequence Control field
    *(USHORT *)(pMacHdr + 22) = SWAP16(*(USHORT *)(pMacHdr + 22));

    if(pFrame->Type == BTYPE_MGMT)
    {
        switch(pFrame->SubType)
        {
            case SUBTYPE_ASSOC_REQ:
            case SUBTYPE_REASSOC_REQ:
                // swab 16 bit fields - CapabilityInfo field
                pMacHdr += MAC_HDR_LEN;
                *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);

                // swab 16 bit fields - Listen Interval field
                pMacHdr += 2;
                *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
                break;

            case SUBTYPE_ASSOC_RSP:
            case SUBTYPE_REASSOC_RSP:
                // swab 16 bit fields - CapabilityInfo field
                pMacHdr += MAC_HDR_LEN;
                *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);

                // swab 16 bit fields - Status Code field
                pMacHdr += 2;
                *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
                
                // swab 16 bit fields - AID field
                pMacHdr += 2;
                *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
                break;

            case SUBTYPE_AUTH:
                // If from RTMPHandleRxDoneInterrupt routine, it is still a encrypt format.
                // The convertion is delayed to RTMPHandleDecryptionDoneInterrupt.
                if(!FromRxDoneInt)
                {
                    // swab 16 bit fields - Auth Alg No. field
                    pMacHdr += MAC_HDR_LEN;
                    *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);

                    // swab 16 bit fields - Auth Seq No. field
                    pMacHdr += 2;
                    *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);

                    // swab 16 bit fields - Status Code field
                    pMacHdr += 2;
                    *(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
            	}
            	break;

            case SUBTYPE_BEACON:
            case SUBTYPE_PROBE_RSP:
            	// swab 16 bit fields - BeaconInterval field
            	pMacHdr += MAC_HDR_LEN + TIMESTAMP_LEN;
            	*(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);

            	// swab 16 bit fields - CapabilityInfo field
            	pMacHdr += sizeof(USHORT);
            	*(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
            	break;

            case SUBTYPE_DEAUTH:
            case SUBTYPE_DISASSOC:
            	// swab 16 bit fields - Reason code field
            	pMacHdr += MAC_HDR_LEN;
            	*(USHORT *)pMacHdr = SWAP16(*(USHORT *)pMacHdr);
            	break;
        }
    }
    else if( pFrame->Type == BTYPE_DATA )
    {
    }
    else if(pFrame->Type == BTYPE_CNTL)
    {
    }
    else
    {
        DBGPRINT(RT_DEBUG_ERROR,"Invalid Frame Type!!!\n");
    }

    // swab 16 bit fields - Frame Control
    if(Dir == DIR_WRITE)
    {
        *(USHORT *)pData = SWAP16(*(USHORT *)pData);
    }
}
#endif

/*
    ========================================================================

    Routine Description:
        Suspend MSDU transmission

    Arguments:
        pAdapter        Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPSuspendMsduTransmission(
    IN  PRTMP_ADAPTER   pAdapter)
{
    DBGPRINT(RT_DEBUG_TRACE,"SCANNING, suspend MSDU transmission ...\n");
    RTMP_SET_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS);
}

/*
    ========================================================================

    Routine Description:
        Resume MSDU transmission

    Arguments:
        pAdapter        Pointer to our adapter

    Return Value:
        None

    Note:

    ========================================================================
*/
VOID    RTMPResumeMsduTransmission(
    IN  PRTMP_ADAPTER   pAdapter)
{
    DBGPRINT(RT_DEBUG_TRACE,"SCAN done, resume MSDU transmission ...\n");
    RTMP_CLEAR_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS);

    // Dequeue Tx queue if Reset is not in progress
    if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)) &&
        (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RADIO_OFF)))
    {
        //RTMPDeQueuePacket(pAdapter, &pAdapter->TxSwQueue0);
        // Call dequeue without selected queue, let the subroutine select the right priority
        // Tx software queue
        RTMPDeQueuePacket(pAdapter);
    }
}
