/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************
 
    Module Name:
    soft_ap.c
 
    Abstract:
    Access Point specific routines and MAC table maintenance routines
 
    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
    John Chang  08-04-2003    created for 11g soft-AP

 */

#include "rt_config.h"

#define MAX_NUM_OF_FREE_NDIS_PACKET     32

char const *pEventText[EVENT_MAX_EVENT_TYPE] = {
    "restart access point",
    "successfully associated",
    "has disassociated",
    "has been aged-out and disassociated" ,    
    "active countermeasures",
    "has disassociated with invalid PSK password"};


/*
    ==========================================================================
    Description:
        Initialize AP specific data especially the NDIS packet pool that's
        used for wireless client bridging.
    ==========================================================================
 */
NDIS_STATUS ApInitialize(
	IN	PRTMP_ADAPTER	pAd)
{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;
    UCHAR       GTK[TKIP_GTK_LENGTH], apidx;

    DBGPRINT(RT_DEBUG_TRACE, "---> ApInitialize\n");

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        pAd->PortCfg.MSSID[apidx].WPAGKeyID = 1;
        
        // initialize MAC table and allocate spin lock
        NdisZeroMemory(&pAd->PortCfg.MSSID[apidx].MacTab, sizeof(MAC_TABLE));
        NdisAllocateSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);

        // Init TKIP Group-Key-related variables
        GenRandom(pAd,pAd->PortCfg.MSSID[apidx].GMK);
        GenRandom(pAd,pAd->PortCfg.MSSID[apidx].GNonce);
        CountGTK(pAd->PortCfg.MSSID[apidx].GMK, (UCHAR*)pAd->PortCfg.MSSID[apidx].GNonce, pAd->CurrentAddress, pAd->PortCfg.MSSID[apidx].GroupKey[0].Key, TKIP_GTK_LENGTH);
        NdisMoveMemory(pAd->PortCfg.MSSID[apidx].GroupKey[pAd->PortCfg.MSSID[apidx].WPAGKeyID].Key, GTK, 16);
        NdisMoveMemory(pAd->PortCfg.MSSID[apidx].GroupKey[pAd->PortCfg.MSSID[apidx].WPAGKeyID].TxMic, &GTK[16], 8);
        NdisMoveMemory(pAd->PortCfg.MSSID[apidx].GroupKey[pAd->PortCfg.MSSID[apidx].WPAGKeyID].RxMic, &GTK[24], 8);            

        pAd->PortCfg.MSSID[apidx].PortSecured = WPA_802_1X_PORT_NOT_SECURED;
    }

    // Init Group key update timer, and countermeasures timer
    init_timer(&pAd->PortCfg.REKEYTimer);
    pAd->PortCfg.REKEYTimer.data = (unsigned long)pAd;
    pAd->PortCfg.REKEYTimer.function = &GREKEYPeriodicExec;

    init_timer(&pAd->PortCfg.CounterMeasureTimer);
    pAd->PortCfg.CounterMeasureTimer.data = (unsigned long)pAd;
    pAd->PortCfg.CounterMeasureTimer.function = &CMTimerExec;

    DBGPRINT(RT_DEBUG_TRACE, "<--- ApInitialize\n");
	return Status;
}

/*
    ==========================================================================
    Description:
        Shutdown AP and free AP specific resources
    ==========================================================================
 */
VOID ApShutdown(
    IN PRTMP_ADAPTER pAd)
{
    DBGPRINT(RT_DEBUG_TRACE, "---> ApShutdown\n");
    ApStop(pAd);
    DBGPRINT(RT_DEBUG_TRACE, "<--- ApShutdown\n");
}

/*
    ==========================================================================
    Description:
        Start AP service. If any vital AP parameter is changed, a STOP-START
        sequence is required to disassociate all STAs.
    Note:
    ==========================================================================
 */
VOID ApStartUp(
    IN PRTMP_ADAPTER pAd) 
{
    UCHAR       apidx;

    DBGPRINT(RT_DEBUG_TRACE, "!!! ApStartUp !!!\n");

    ApLogEvent(pAd, (PMACADDR)pAd->CurrentAddress, EVENT_RESET_ACCESS_POINT, MAIN_MSSID);

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        if ((pAd->PortCfg.MSSID[apidx].SsidLen <= 0) || (pAd->PortCfg.MSSID[apidx].SsidLen > MAX_LEN_OF_SSID))
            return;
        
        pAd->PortCfg.MSSID[apidx].CapabilityInfo = CAP_GENERATE((pAd->PortCfg.MSSID[apidx].WepStatus != Ndis802_11EncryptionDisabled), (pAd->PortCfg.TxPreamble == Rt802_11PreambleShort), pAd->PortCfg.UseShortSlotTime);
        DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) CapabilityInfo=%x, WepStatus=%d\n", apidx,
            pAd->PortCfg.MSSID[apidx].CapabilityInfo,pAd->PortCfg.MSSID[apidx].WepStatus);
    }

#if defined(VINCENT_REGION)
	// update channel list
	BuildChannelList(pAd);
#endif

    AsicSetBssid(pAd, (MACADDR *)pAd->CurrentAddress); 
    AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
    AsicLockChannel(pAd, pAd->PortCfg.Channel);
    MlmeSetTxPreamble(pAd, pAd->PortCfg.TxPreamble);
    MlmeUpdateTxRates(pAd);

    // start sending BEACON out        
    MakeBssBeacon(pAd);
    UpdateBeaconFrame(pAd);
    ApUpdateCapabilityAndErpIe(pAd);
    AsicEnableBssSync(pAd);
    ASIC_LED_ACT_ON(pAd);

    DBGPRINT(RT_DEBUG_TRACE, "ApStartUp : Group rekey method= %d, interval = 0x%x\n",pAd->PortCfg.WPAREKEY.ReKeyMethod,pAd->PortCfg.WPAREKEY.ReKeyInterval);
    // Group rekey related
    if ((pAd->PortCfg.WPAREKEY.ReKeyInterval != 0) && ((pAd->PortCfg.WPAREKEY.ReKeyMethod == TIME_REKEY) || (pAd->PortCfg.WPAREKEY.ReKeyMethod == PKT_REKEY))) 
    {
        // Regularly check the timer
        if (pAd->PortCfg.REKEYTimerRunning == FALSE)
        {
            pAd->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
            add_timer(&pAd->PortCfg.REKEYTimer);
            pAd->PortCfg.REKEYTimerRunning = TRUE;
            pAd->PortCfg.REKEYCOUNTER = 0;
        }
    }
    else
    	pAd->PortCfg.REKEYTimerRunning = FALSE;

	pAd->Mlme.PeriodicRound = 0;

    // send SIGUSR1 to rt2500apd
    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        if ((pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA)
        	|| (pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA2))
        {
    	    PUCHAR					src;
            struct file				*srcf;
            INT 					retval, orgfsuid, orgfsgid;
            mm_segment_t			orgfs;
            CHAR					buffer[MAX_INI_BUFFER_SIZE];
            CHAR					tmpbuf[600];
       	    src = PROFILE_PATH;

    	    // Save uid and gid used for filesystem access.
    	    // Set user and group to 0 (root)	
    	    orgfsuid = current->fsuid;
    	    orgfsgid = current->fsgid;
    	    current->fsuid=current->fsgid = 0;
            orgfs = get_fs();
            set_fs(KERNEL_DS);

    	    if (src && *src) 
    	    {
                srcf = filp_open(src, O_RDONLY, 0);
    		    if (IS_ERR(srcf)) 
    		    {
    			    DBGPRINT(RT_DEBUG_ERROR, "--> Error %ld opening %s\n", -PTR_ERR(srcf),src);    
    		    }
    		    else 
    		    {
    			    /* The object must have a read method */
    			    if (srcf->f_op && srcf->f_op->read) 
    			    {
    				    memset(buffer, 0x00, MAX_INI_BUFFER_SIZE);
    				    retval=srcf->f_op->read(srcf, buffer, MAX_INI_BUFFER_SIZE, &srcf->f_pos);
    				    if (retval < 0)
    				    {
    					    DBGPRINT(RT_DEBUG_ERROR, "--> Read %s error %d\n", src, -retval);
    				    }
    				    else
    				    {
    					    // set file parameter to portcfg
    					    //CountryRegion
    					    if(RTMPGetKeyParameter("Default", "pid", tmpbuf, 255, buffer))
    					    {
    						    pAd->PortCfg.ApdPid= (ULONG) simple_strtol(tmpbuf, 0, 10);
    						    DBGPRINT(RT_DEBUG_TRACE, "ApdPid=%d\n", pAd->PortCfg.ApdPid);
                                kill_proc(pAd->PortCfg.ApdPid, SIGUSR1, 0);
    					    }
    				    }
                    }
                    else
                    {
                        DBGPRINT(RT_DEBUG_ERROR, "--> %s does not have a write method\n", src);
                    }
    		
    			    retval=filp_close(srcf,NULL);
    		
    			    if (retval)
    			    {
    				    DBGPRINT(RT_DEBUG_ERROR, "--> Error %d closing %s\n", -retval, src);
    			    }
                }
    		}
    	    set_fs(orgfs);
    	    current->fsuid = orgfsuid;
    	    current->fsgid = orgfsgid;

    	    break;  // If one SSID operate in WPA, read configuration file only once.
    	}
    }
}

/*
    ==========================================================================
    Description:
        disassociate all STAs and stop AP service.
    Note:
    ==========================================================================
 */
VOID ApStop(
    IN PRTMP_ADAPTER pAd) 
{
    DBGPRINT(RT_DEBUG_TRACE, "!!! ApStop !!!\n");
    //NICDisableInterrupt(pAd);
    AsicDisableSync(pAd);

    if(pAd->PortCfg.REKEYTimerRunning==TRUE)
    {
        del_timer_sync(&pAd->PortCfg.REKEYTimer);
        pAd->PortCfg.REKEYTimerRunning=FALSE;
    }

    if(pAd->PortCfg.CMTimerRunning==TRUE)
    {
        del_timer_sync(&pAd->PortCfg.CounterMeasureTimer);
    	pAd->PortCfg.CMTimerRunning = FALSE;
    }

    MacTableReset(pAd);
    ASIC_LED_ACT_OFF(pAd);
}

/*
    ==========================================================================
    Description:
        This routine is used to clean up a specified power-saving queue. It's
        used whenever a wireless client is deleted.
    ==========================================================================
 */
VOID CleanupPsQueue(
    IN  PRTMP_ADAPTER   pAd,
    IN  PQUEUE_HEADER   pQueue)
{
    struct sk_buff	*skb;
    
    while (pQueue->Head)
    {
        skb = (struct sk_buff *)RemoveHeadQueue(pQueue);
        if(skb)		RTMPFreeSkbBuffer(skb);
    }
}

/*
    ==========================================================================
    Description:
        This routine reset the entire MAC table. All packets pending in
        the power-saving queues are freed here.
    ==========================================================================
 */
VOID MacTableReset(
    IN  PRTMP_ADAPTER  pAd)
{
    int i, apidx;

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
        CleanupPsQueue(pAd, &pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue);
        for (i=0; i<MAX_LEN_OF_MAC_TABLE; i++)
        {
            if (pAd->PortCfg.MSSID[apidx].MacTab.Content[i].Valid)
           {
                if (pAd->PortCfg.MSSID[apidx].MacTab.Content[i].RetryTimerRunning == TRUE)
                {
                    DBGPRINT(RT_DEBUG_TRACE, "pAd->MacTab.Content[%d]  Cancel Retry Timer !\n",i);
                    del_timer_sync(&pAd->PortCfg.MSSID[apidx].MacTab.Content[i].RetryTimer);
                    pAd->PortCfg.MSSID[apidx].MacTab.Content[i].RetryTimerRunning = FALSE;             
                }
                CleanupPsQueue(pAd, &pAd->PortCfg.MSSID[apidx].MacTab.Content[i].PsQueue);
            }
        }
        NdisZeroMemory(&pAd->PortCfg.MSSID[apidx].MacTab, sizeof(MAC_TABLE));
        InitializeQueueHeader(&pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue);
        NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
    }
}

/*
    ==========================================================================
    Description:
        Add and new entry into MAC table
    ==========================================================================
 */
MAC_TABLE_ENTRY *MacTableInsertEntry(
    IN	PRTMP_ADAPTER	pAd, 
    IN  PMACADDR        pAddr,
    IN  UCHAR           apidx) 
{
    UCHAR HashIdx;
    int i;
    MAC_TABLE_ENTRY *pEntry = NULL, *pCurrEntry;

    // if FULL, return
    if (pAd->PortCfg.MSSID[apidx].MacTab.Size >= MAX_LEN_OF_MAC_TABLE) 
        return NULL;

    // allocate one MAC entry
    NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
    for (i = 0; i< MAX_LEN_OF_MAC_TABLE; i++)
    {
        // pick up the first available vacancy
        if (pAd->PortCfg.MSSID[apidx].MacTab.Content[i].Valid == FALSE)
        {
            pEntry = &pAd->PortCfg.MSSID[apidx].MacTab.Content[i];
            NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
            pEntry->Valid = TRUE;

            // Move init entry retry timer to where it first created
            init_timer(&pEntry->RetryTimer);
            
            pEntry->RetryTimerRunning = FALSE;
            pEntry->CMTimerRunning = FALSE;
            pEntry->RSNIE_Len = 0;
            NdisMoveMemory(pEntry->R_Counter, pAd->PortCfg.R_Counter, sizeof(pEntry->R_Counter));
            pEntry->ReTryCounter = PEER_MSG1_RETRY_TIMER_CTR;
            pEntry->AuthMode = pAd->PortCfg.MSSID[apidx].AuthMode;
            pEntry->WepStatus = pAd->PortCfg.MSSID[apidx].WepStatus;
            pEntry->pAd = (unsigned long)pAd;
            pEntry->PMKID_CacheIdx = -1;

            if (pEntry->AuthMode < Ndis802_11AuthModeWPA)
                pEntry->WpaState = AS_NOTUSE;
            else
                pEntry->WpaState = AS_INITIALIZE;
            pEntry->GTKState = REKRY_NEGOTIATING;
            pEntry->PrivacyFilter = Ndis802_11PrivFilterAcceptAll;        
            pEntry->PairwiseKey[0].KeyLen = 0;
            pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
            pEntry->Addr = *pAddr;
            pEntry->Sst = SST_NOT_AUTH;
            pEntry->AuthState = AS_NOT_AUTH;
            pEntry->Aid = 0;
            pEntry->CapabilityInfo = 0;
            pEntry->PsMode = PWR_ACTIVE;
            pEntry->MaxSupportedRate = RATE_11;
            pEntry->CurrTxRate = RATE_11;
            pEntry->PsQIdleCount = 0;
            pEntry->NoDataIdleCount = 0;
            InitializeQueueHeader(&pEntry->PsQueue);

            pEntry->ApIdx = apidx;
            
            pAd->PortCfg.MSSID[apidx].MacTab.Size ++;

            DBGPRINT(RT_DEBUG_TRACE, "MacTableInsertEntry - allocate entry #%d, Total= %d\n",i, pAd->PortCfg.MSSID[apidx].MacTab.Size);
            break;
        }
    }

    // add this MAC entry into HASH table
    if (pEntry)
    {
        HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);
        if (pAd->PortCfg.MSSID[apidx].MacTab.Hash[HashIdx] == NULL)
        {
            pAd->PortCfg.MSSID[apidx].MacTab.Hash[HashIdx] = pEntry;
        }
        else
        {
            pCurrEntry = pAd->PortCfg.MSSID[apidx].MacTab.Hash[HashIdx];
            while (pCurrEntry->pNext != NULL)
                pCurrEntry = pCurrEntry->pNext;
            pCurrEntry->pNext = pEntry;
        }
    }
    
    NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
    return pEntry;
}
    
/*
    ==========================================================================
    Description:
        Delete a specified client from MAC table
    ==========================================================================
 */
BOOLEAN MacTableDeleteEntry(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR pAddr,
    IN UCHAR apidx) 
{
    USHORT HashIdx;
    MAC_TABLE_ENTRY *pEntry, *pPrevEntry;

    NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
    HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);
    pEntry = pAd->PortCfg.MSSID[apidx].MacTab.Hash[HashIdx];
    if (pEntry)
    {
        if(pEntry->RetryTimerRunning==TRUE)
        {
            DBGPRINT(RT_DEBUG_TRACE, "MacTableDeleteEntry : Cancel Retry Timer !\n");
            del_timer_sync(&pEntry->RetryTimer);
            pEntry->RetryTimerRunning=FALSE;                
        }

        if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr))
        {
            pAd->PortCfg.MSSID[apidx].MacTab.Hash[HashIdx] = pEntry->pNext;
            CleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
            NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
            pAd->PortCfg.MSSID[apidx].MacTab.Size --;
            DBGPRINT(RT_DEBUG_TRACE, "MacTableDeleteEntry - Total= %d\n", pAd->PortCfg.MSSID[apidx].MacTab.Size);
        }
        else
        {
            while (pEntry->pNext)
            {
                pPrevEntry = pEntry;
                pEntry = pEntry->pNext;
                if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr))
                {
                    pPrevEntry->pNext = pEntry->pNext;
                    CleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
                    NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
                    pAd->PortCfg.MSSID[apidx].MacTab.Size --;
                    DBGPRINT(RT_DEBUG_TRACE, "MacTableDeleteEntry - Total= %d\n", pAd->PortCfg.MSSID[apidx].MacTab.Size);
                    break;
                }
            }
        }
    }

    NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
    ApUpdateCapabilityAndErpIe(pAd);
    ProbeReqCacheDelete(pAd, pAddr);
    
    return TRUE;
}

/*
    ==========================================================================
    Description:
        Look up the MAC address in the MAC table. Return NULL if not found.
    Return:
        pEntry - pointer to the MAC entry; NULL is not found
    ==========================================================================
*/
MAC_TABLE_ENTRY *MacTableLookup(
    IN PRTMP_ADAPTER pAd, 
    IN MACADDR *pAddr) 
{
    ULONG HashIdx, i;
    MAC_TABLE_ENTRY *pEntry = NULL;
    BOOLEAN bFound = FALSE;

    HashIdx = MAC_ADDR_HASH_INDEX(*pAddr);

    for(i = 0; i < pAd->PortCfg.SsidNum; i++)
    {
        pEntry = pAd->PortCfg.MSSID[i].MacTab.Hash[HashIdx];

        while (pEntry && pEntry->Valid)
        {
            if (MAC_ADDR_EQUAL(&pEntry->Addr, pAddr)) 
            {
                bFound = TRUE;
                break;
            }
            else
                pEntry = pEntry->pNext;
        }
        if (bFound)
            break;
    }
    return pEntry;
}

/*
    ==========================================================================
    Description:
        This routine is called by MlmePeriodicExec() every second to check if
        1. any associated client in PSM. If yes, then TX MCAST/BCAST should be
           out in DTIM only
        2. any client being idle for too long and should be aged-out from MAC table
        3. garbage collect PSQ
    ==========================================================================
*/
VOID MacTableMaintenance(
    IN PRTMP_ADAPTER pAd)
{
    int     i;
    UCHAR   apidx;

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        pAd->PortCfg.MSSID[apidx].MacTab.fAnyStationInPsm = FALSE;
        for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++) 
        {
            MAC_TABLE_ENTRY *pEntry = &pAd->PortCfg.MSSID[apidx].MacTab.Content[i];

            if (pEntry->Valid == FALSE)
                continue;
            
            pEntry->NoDataIdleCount ++;

#ifdef RT2500_DBG
            DBGPRINT(RT_DEBUG_INFO, "(ra%d)%02x:%02x:%02x:%02x:%02x:%02x,TxByte=%d,RxByte=%d,LastDataTime=%d\n", apidx,
                pEntry->Addr.Octet[0],pEntry->Addr.Octet[1],pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],
                pEntry->Addr.Octet[4],pEntry->Addr.Octet[5],pEntry->HSCounter.TotalTxByteCount,
                pEntry->HSCounter.TotalRxByteCount,pEntry->HSCounter.LastDataPacketTime);
#endif

            // 0. STA failed to complete association should be removed to save MAC table space.
            if ((pEntry->Sst != SST_ASSOC) && (pEntry->NoDataIdleCount >= MAC_TABLE_ASSOC_TIMEOUT))
            {
                DBGPRINT(RT_DEBUG_TRACE, "%02x:%02x:%02x:%02x:%02x:%02x fail to complete ASSOC in %d sec\n",
                        pEntry->Addr.Octet[0],pEntry->Addr.Octet[1],pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],
                        pEntry->Addr.Octet[4],pEntry->Addr.Octet[5],MAC_TABLE_ASSOC_TIMEOUT);
                MacTableDeleteEntry(pAd, &pEntry->Addr, apidx);
                continue;
            }
            
            // 1. check if there's any associated STA in power-save mode. this affects outgoing
            //    MCAST/BCAST frames should be stored in PSQ till DtimCount=0
            if (pEntry->PsMode == PWR_SAVE)
                pAd->PortCfg.MSSID[apidx].MacTab.fAnyStationInPsm = TRUE;

            // 2. delete those MAC entry that has been idle for a long time
            if (pEntry->NoDataIdleCount >= MAC_TABLE_AGEOUT_TIME)
            {
                DBGPRINT(RT_DEBUG_TRACE, "ageout %02x:%02x:%02x:%02x:%02x:%02x after %d-sec silence\n",
                        pEntry->Addr.Octet[0],pEntry->Addr.Octet[1],pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],
                        pEntry->Addr.Octet[4],pEntry->Addr.Octet[5],MAC_TABLE_AGEOUT_TIME);
                ApLogEvent(pAd, &pEntry->Addr, EVENT_AGED_OUT, apidx);
                MacTableDeleteEntry(pAd, &pEntry->Addr, apidx);
                continue;
            }
            
            // 3. garbage collect the PsQueue if the STA has being idle for a while
            if (pEntry->PsQueue.Head)
            {
                pEntry->PsQIdleCount ++;  
                if (pEntry->PsQIdleCount > 2) 
                {
                    NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
                    CleanupPsQueue(pAd, &pEntry->PsQueue);
                    NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
                    pEntry->PsQIdleCount = 0;
                }
            }
            else
                pEntry->PsQIdleCount = 0;
        }
        
        // 4. garbage collect pAd->MacTab.McastPsQueue if backlogged MCAST/BCAST frames
        //    stale in queue. Since MCAST/BCAST frames always been sent out whenever 
        //    DtimCount==0, the only case to let them stale is surprise removal of the NIC,
        //    so that ASIC-based Tbcn interrupt stops and DtimCount dead.
        if (pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue.Head)
        {
            pAd->PortCfg.MSSID[apidx].MacTab.PsQIdleCount ++;
            if (pAd->PortCfg.MSSID[apidx].MacTab.PsQIdleCount > 1)
            {
                NdisAcquireSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
                CleanupPsQueue(pAd, &pAd->PortCfg.MSSID[apidx].MacTab.McastPsQueue);
                NdisReleaseSpinLock(&pAd->PortCfg.MSSID[apidx].MacTabLock);
                pAd->PortCfg.MSSID[apidx].MacTab.PsQIdleCount = 0;
            }
        }
        else
            pAd->PortCfg.MSSID[apidx].MacTab.PsQIdleCount = 0;
    }
}

/*
    ==========================================================================
    Description:
        Look up a STA MAC table. Return its Sst to decide if an incoming
        frame from this STA or an outgoing frame to this STA is permitted.
    Return:
    ==========================================================================
*/

MAC_TABLE_ENTRY *SsPsInquiry(
    IN	PRTMP_ADAPTER	pAd, 
    IN MACADDR *pAddr, 
    OUT SST   *Sst, 
    OUT USHORT *Aid,
    OUT UCHAR *PsMode,
    OUT UCHAR *Rate) 
{
    MAC_TABLE_ENTRY *pEntry = NULL;

    if (MAC_ADDR_IS_GROUP(*pAddr)) // mcast & broadcast address
    {
        *Sst        = SST_ASSOC;
        *Aid        = 0;
        *PsMode     = PWR_ACTIVE;
        *Rate       = pAd->PortCfg.MlmeRate; //pAd->PortCfg.MaxBasicRate;
    } 
    else // unicast address
    {
        pEntry = MacTableLookup(pAd, pAddr);
        if (pEntry) 
        {
            *Sst        = pEntry->Sst;
            *Aid        = pEntry->Aid;
            *PsMode     = pEntry->PsMode;
             if((pAd->PortCfg.MSSID[pEntry->ApIdx].AuthMode >= Ndis802_11AuthModeWPA) && (pEntry->GTKState != REKEY_ESTABLISHED))
                *Rate   = RATE_11;
             else
                *Rate   = pEntry->CurrTxRate;
        } 
        else 
        {
            *Sst        = SST_NOT_AUTH;
            *Aid        = 0;
            *PsMode     = PWR_ACTIVE;
            *Rate       = pAd->PortCfg.MlmeRate; //pAd->PortCfg.MaxBasicRate;
        }
    }
    return pEntry;
}

/*
    ==========================================================================
    Description:
        Update the station current power save mode. Calling this routine also
        prove the specified client is still alive. Otherwise AP will age-out
        this client once IdleCount exceeds a threshold.
    ==========================================================================
 */
BOOLEAN PsIndicate(
    IN PRTMP_ADAPTER pAd, 
    IN PMACADDR pAddr, 
    IN UCHAR Rssi,
    IN UCHAR Psm) 
{
    MAC_TABLE_ENTRY *pEntry;

    pEntry = MacTableLookup(pAd, pAddr);
    if (pEntry) 
    {
        if ((pEntry->PsMode == PWR_SAVE) && (Psm == PWR_ACTIVE))
        {
            DBGPRINT(RT_DEBUG_TRACE, "PsIndicate - %02x:%02x:%02x:%02x:%02x:%02x wakes up, act like rx PS-POLL\n", pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]);
            // sleep station awakes, move all pending frames from PSQ to TXQ if any
            RTMPHandleRxPsPoll(pAd, pAddr, pEntry->Aid);
        }
        else if ((pEntry->PsMode != PWR_SAVE) && (Psm == PWR_SAVE))
        {
            DBGPRINT(RT_DEBUG_TRACE, "PsIndicate - %02x:%02x:%02x:%02x:%02x:%02x sleeps\n", pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]);
        }
        
        pEntry->NoDataIdleCount = 0;
        pEntry->PsMode = Psm;
        pEntry->LastRssi = Rssi;
    } 
    else 
    {
        // not in table, try to learn it ???? why bother?
    }
    return TRUE;
}

#if BIG_ENDIAN == FALSE && RTMP_EMBEDDED == FALSE
extern unsigned long get_cmos_time(void);
#endif
/*
    ==========================================================================
    Description:
        This routine is called to log a specific event into the event table.
        The table is a QUERY-n-CLEAR array that stop at full.
    ==========================================================================
 */
VOID ApLogEvent(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR pAddr,
    IN USHORT   Event,
    IN UCHAR    apidx)
{
    if (pAd->EventTab.Num < MAX_NUM_OF_EVENT)
    {
        RT_802_11_EVENT_LOG *pLog = &pAd->EventTab.Log[pAd->EventTab.Num];
#if BIG_ENDIAN == FALSE && RTMP_EMBEDDED == FALSE
        pLog->SystemTime = get_cmos_time();
#else
        pLog->SystemTime = jiffies;
#endif
        COPY_MAC_ADDR(&pLog->Addr, pAddr);
        pLog->Event = Event;
        pLog->ApIdx = apidx;
        DBGPRINT(RT_DEBUG_TRACE,"LOG#%d %02x:%02x:%02x:%02x:%02x:%02x %s\n",
            pAd->EventTab.Num, pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], 
            pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], pEventText[Event]);
        pAd->EventTab.Num += 1;
    }
}

/*
    ==========================================================================
    Description:
        Update ERP IE and CapabilityInfo based on STA association status.
        The result will be auto updated into the next outgoing BEACON in next
        TBTT interrupt service routine
    ==========================================================================
 */
VOID ApUpdateCapabilityAndErpIe(
    IN PRTMP_ADAPTER pAd)
{
    UCHAR  i, ErpIeContent = 0, apidx;
    BOOLEAN ShortSlotCapable = (BOOLEAN)pAd->PortCfg.UseShortSlotTime;
#if 0
    if (pAd->PortCfg.PhyMode != PHY_11BG_MIXED)
    {
        if (pAd->PortCfg.PhyMode == PHY_11G)
        {
            if((pAd->PortCfg.TxPreamble == Rt802_11PreambleLong) || (ApCheckLongPreambleSTA(pAd) == TRUE))
            {
                pAd->PortCfg.TxPreambleInUsed = Rt802_11PreambleLong;
                pAd->Mlme.ErpIeContent = (ErpIeContent | 0x04);
            }
            else
            {
                pAd->PortCfg.TxPreambleInUsed = Rt802_11PreambleShort;
                pAd->Mlme.ErpIeContent = ErpIeContent;
            }
            AsicSetSlotTime(pAd, ShortSlotCapable);
        }
        return;
    }
#else
    if (pAd->PortCfg.PhyMode == PHY_11B)
        return;
#endif

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        for (i=0; i<MAX_LEN_OF_MAC_TABLE; i++)
        {
            PMAC_TABLE_ENTRY pEntry = &pAd->PortCfg.MSSID[apidx].MacTab.Content[i];
            if ((pEntry->Valid != TRUE) || (pEntry->Sst != SST_ASSOC))
                continue;
            
            // at least one 11b client associated, turn on ERP.NonERPPresent bit
            // almost all 11b client won't support "Short Slot" time, turn off for maximum compatibility
            if (pEntry->MaxSupportedRate < RATE_FIRST_OFDM_RATE)
            {
                ShortSlotCapable = FALSE;
                ErpIeContent |= 0x01;
            }

            // at least one client can't support short slot
            if ((pEntry->CapabilityInfo & 0x0400) == 0)
                ShortSlotCapable = FALSE;
        }
    }

    // decide ErpIR.UseProtection bit, depending on pAd->PortCfg.UseBGProtection
    //    AUTO (0): UseProtection = 1 if any 11b STA associated
    //    ON (1): always USE protection
    //    OFF (2): always NOT USE protection
    if (pAd->PortCfg.UseBGProtection == 0)
    {
        ErpIeContent = (ErpIeContent)? 0x03 : 0x00;
        if ((pAd->Mlme.LastOLBCDetectTime + (10 * HZ)) > pAd->Mlme.Now32) // legacy BSS exist within 10 sec
        {
            DBGPRINT(RT_DEBUG_INFO, "ApUpdateCapabilityAndErpIe - Legacy 802.11b BSS overlaped\n");
            ErpIeContent |= 0x02;                                     // set Use_Protection bit
        }
    }
    else if (pAd->PortCfg.UseBGProtection == 1)   
        ErpIeContent |= 0x02;
    else
        ;

    if((pAd->PortCfg.TxPreamble == Rt802_11PreambleLong) || (ApCheckLongPreambleSTA(pAd) == TRUE))
    {
        pAd->PortCfg.TxPreambleInUsed = Rt802_11PreambleLong;
        pAd->Mlme.ErpIeContent = (ErpIeContent | 0x04);
    }
    else
    {
        pAd->PortCfg.TxPreambleInUsed = Rt802_11PreambleShort;
        pAd->Mlme.ErpIeContent = ErpIeContent;
    }

    //
    // deicide CapabilityInfo.ShortSlotTime bit
    //
    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        if (ShortSlotCapable)
            pAd->PortCfg.MSSID[apidx].CapabilityInfo |= 0x0400;
        else
            pAd->PortCfg.MSSID[apidx].CapabilityInfo &= 0xfbff;

        DBGPRINT(RT_DEBUG_INFO, "ApUpdateCapabilityAndErpIe I/F(ra%d)- Capability= 0x%04x, ERP is 0x%02x\n", 
            apidx, pAd->PortCfg.MSSID[apidx].CapabilityInfo, pAd->Mlme.ErpIeContent);
    }

    AsicSetSlotTime(pAd, ShortSlotCapable);
}

/*
    ==========================================================================
    Description:
        Check to see the exist of long preamble STA in associated list
    ==========================================================================
 */
BOOLEAN ApCheckLongPreambleSTA(
    IN PRTMP_ADAPTER pAd)
{
    UCHAR   i;
    UCHAR   apidx;
    
    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        for (i=0; i<MAX_LEN_OF_MAC_TABLE; i++)
        {
            PMAC_TABLE_ENTRY pEntry = &pAd->PortCfg.MSSID[apidx].MacTab.Content[i];
            if ((pEntry->Valid != TRUE) || (pEntry->Sst != SST_ASSOC))
                continue;
        
            if (!CAP_IS_SHORT_PREAMBLE_ON(pEntry->CapabilityInfo))
            {
                DBGPRINT(RT_DEBUG_TRACE, "Long preamble capable STA exist\n");
                return TRUE;
            }
        }
    }

    return FALSE;
}

/*
    ==========================================================================
    Description:
        Check if the specified STA pass the Access Control List checking.
        If fails to pass the checking, then no authentication nor association 
        is allowed
    Return:
        MLME_SUCCESS - this STA passes ACL checking
        
    ==========================================================================
*/
BOOLEAN ApCheckAccessControlList(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR      pAddr) 
{
    BOOLEAN Result = TRUE;

    if (pAd->PortCfg.AccessControlList.Policy == 0)       // ACL is disabled
        Result = TRUE;
    else
    {
        ULONG i;
        if (pAd->PortCfg.AccessControlList.Policy == 1)   // ACL is a positive list
            Result = FALSE;
        else                                              // ACL is a negative list
            Result = TRUE;
        for (i=0; i<pAd->PortCfg.AccessControlList.Num; i++)
        {
            if (MAC_ADDR_EQUAL(pAddr, pAd->PortCfg.AccessControlList.Entry[i].Addr))
            {
                Result = !Result;
                break;
            }
        }
    }

    if (Result == FALSE)
    {
        DBGPRINT(RT_DEBUG_TRACE, "%02x:%02x:%02x:%02x:%02x:%02x failed ACL checking\n",
        pAddr->Octet[0],pAddr->Octet[1],pAddr->Octet[2],pAddr->Octet[3],pAddr->Octet[4],pAddr->Octet[5]);
    }

    return Result;
}

/*
    ==========================================================================
    Description:
        This routine update the current MAC table based on the current ACL.
        If ACL change causing an associated STA become un-authorized. This STA
        wil be kicked out immediately.
    ==========================================================================
*/
VOID ApUpdateAccessControlList(
    IN PRTMP_ADAPTER pAd)
{
    USHORT      AclIdx, MacIdx;
    BOOLEAN     Matched;
    UCHAR       apidx;

    // ACL is disabled. do nothing about the MAC table
    if (pAd->PortCfg.AccessControlList.Policy == 0)
        return;

    for (apidx = 0; apidx < pAd->PortCfg.SsidNum; apidx++)
    {
        for (MacIdx=0; MacIdx < MAX_LEN_OF_MAC_TABLE; MacIdx++)
        {
            if (! pAd->PortCfg.MSSID[apidx].MacTab.Content[MacIdx].Valid) 
                continue;

            Matched = FALSE;
            for (AclIdx = 0; AclIdx < pAd->PortCfg.AccessControlList.Num; AclIdx++)
            {
                if (MAC_ADDR_EQUAL(&pAd->PortCfg.MSSID[apidx].MacTab.Content[MacIdx].Addr, pAd->PortCfg.AccessControlList.Entry[AclIdx].Addr))
                {
                    Matched = TRUE;
                    break;
                }
            }
                
            if ((Matched == FALSE) && (pAd->PortCfg.AccessControlList.Policy == 1))
            {
                DBGPRINT(RT_DEBUG_TRACE, "STA not on positive ACL. remove it...\n");
                MacTableDeleteEntry(pAd, &pAd->PortCfg.MSSID[apidx].MacTab.Content[MacIdx].Addr, apidx);
            }
            else if ((Matched == TRUE) && (pAd->PortCfg.AccessControlList.Policy == 2))
            {
                DBGPRINT(RT_DEBUG_TRACE, "STA on negative ACL. remove it...\n");
                MacTableDeleteEntry(pAd, &pAd->PortCfg.MSSID[apidx].MacTab.Content[MacIdx].Addr, apidx);
            }
        }
    }
}

/* 
    ==========================================================================
    Description:
        Send out a NULL frame to a specified STA at a higher TX rate. The 
        purpose is to ensure the designated client is okay to received at this
        rate.
    ==========================================================================
 */
VOID ApEnqueueNullFrame(
    IN PRTMP_ADAPTER pAd,
    IN PMACADDR      pAddr,
    IN UCHAR         TxRate,
    IN int			ToWhichWds) 
{
    MACHDR         *NullFr;

    // since TxRate may change, we have to change Duration each time
    NullFr = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
    if (NullFr != NULL) 
    {
    	if(ToWhichWds == -1)
    	{
	        MgtMacHeaderInit(pAd, NullFr, SUBTYPE_NULL_FUNC, 0, pAddr, &pAd->PortCfg.Bssid);
	        NullFr->Type = BTYPE_DATA;
	        NullFr->Frds = 1;
	        NullFr->Duration = RTMPCalcDuration(pAd, TxRate, 14);
	        ApSendNullFrame(pAd, (VOID *)NullFr, sizeof(MACHDR), TxRate);
    	}
    	else
    	{
    		NdisZeroMemory(NullFr, LENGTH_802_11_WITH_ADDR4);
		    NullFr->Type = BTYPE_DATA;
		    NullFr->SubType = SUBTYPE_NULL_FUNC;
		    NullFr->Frds = 1;
	        NullFr->Tods = 1;
		    COPY_MAC_ADDR(&NullFr->Addr1, &pAd->WdsTab.WdsEntry[ToWhichWds].WdsAddr);
		    COPY_MAC_ADDR(&NullFr->Addr2, &pAd->CurrentAddress);
		    COPY_MAC_ADDR(&NullFr->Addr3, &pAd->WdsTab.WdsEntry[ToWhichWds].WdsAddr);
		    COPY_MAC_ADDR(&NullFr->Addr3 + MAC_ADDR_LEN, pAd->CurrentAddress);
	        NullFr->Duration = RTMPCalcDuration(pAd, TxRate, 14);
	        ApSendNullFrame(pAd, (VOID *)NullFr, LENGTH_802_11_WITH_ADDR4, TxRate);
		}
    }
}

VOID	ApSendNullFrame(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuffer,
	IN	ULONG			Length,
	IN  UCHAR           TxRate)
{
	PUCHAR          pDest;
    PTXD_STRUC      pTxD;
#ifdef BIG_ENDIAN
    PTXD_STRUC      pDestTxD;
    TXD_STRUC       TxD;
#endif

	{
		// Make sure Tx ring resource won't be used by other threads
		NdisAcquireSpinLock(&pAdapter->TxRingLock);
	
		// Get the Tx Ring descriptor & Dma Buffer address
		pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;
#ifndef BIG_ENDIAN
        pTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
#else
        pDestTxD  = (PTXD_STRUC) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
        TxD = *pDestTxD;
        pTxD = &TxD;
        RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
#endif
		
		if ((pTxD->Owner == DESC_OWN_HOST) && (pTxD->CipherOwn == DESC_OWN_HOST) && (pTxD->Valid == FALSE))
		{
			DBGPRINT(RT_DEBUG_TRACE, "send NULL Frame @%d Mbps...\n", RateIdToMbps[TxRate]);
#ifdef BIG_ENDIAN
            RTMPFrameEndianChange(pAdapter, (PUCHAR)pBuffer, DIR_WRITE, FALSE);
#endif
			NdisMoveMemory(pDest, pBuffer, Length);
            pAdapter->TxRing[pAdapter->CurEncryptIndex].FrameType = BTYPE_DATA;

#ifdef BIG_ENDIAN
            RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
            *pDestTxD = TxD;
            pTxD = pDestTxD;
#endif

			RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, TRUE, FALSE, FALSE, SHORT_RETRY, IFS_BACKOFF, 
			    TxRate, 4, Length, pAdapter->PortCfg.TxPreambleInUsed, 0);

			// Increase & maintain Tx Ring Index
			pAdapter->CurEncryptIndex++;
			if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
			{
				pAdapter->CurEncryptIndex = 0;
			}
			
			pAdapter->RalinkCounters.EncryptCount++;

			// Kick Encrypt Control Register at the end of all ring buffer preparation
			RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
			
		}		
		NdisReleaseSpinLock(&pAdapter->TxRingLock);
	}
	MlmeConfirm((PVOID)pAdapter, pBuffer, 0);	
}

/* 
    ==========================================================================
    Description:
        This routine is called at initialization. It returns a channel number
        that complies to regulation domain and less interference with current
        enviornment.
    Return:
        ch -  channel number that
    NOTE:
        the retrun channel number is guaranteed to comply to current regulation
        domain that recorded in pAd->PortCfg.CountryRegion
    ==========================================================================
 */
UCHAR ApAutoSelectChannel(
    IN PRTMP_ADAPTER pAd)
{
    UCHAR ch;
    UCHAR dirtyness[MAX_LEN_OF_CHANNELS+1], dirty;
    UCHAR max_rssi[MAX_LEN_OF_CHANNELS+1];
    ULONG FalseCca, FcsError;

    // passive scan channel 1-14. collect statistics
    NdisZeroMemory(dirtyness, MAX_LEN_OF_CHANNELS+1);
    for (ch=14; ch>=1; ch--)
    {
        AsicSwitchChannel(pAd, ch);
        pAd->Counters.GoodReceives = 0;
        pAd->Counters.RxErrors = 0;
        pAd->PortCfg.AutoChannel_MaxRssi = 0;
        pAd->PortCfg.AutoChannel_Channel = ch;
        max_rssi[ch]=0;
        mdelay(500); // 0.5 sec at each channel
        max_rssi[ch] = pAd->PortCfg.AutoChannel_MaxRssi;
        RTMP_IO_READ32(pAd, CNT3, &FalseCca);
        FalseCca &= 0x0000ffff;
        RTMP_IO_READ32(pAd, CNT0, &FcsError);
        
//        if (pAd->Counters.GoodReceives)
        if (max_rssi[ch])
        {
            dirtyness[ch] += 10;
            if (ch > 1) dirtyness[ch-1] += 1;
            if (ch > 2) dirtyness[ch-2] += 1;
            if (ch > 3) dirtyness[ch-3] += 1;
            if (ch > 4) dirtyness[ch-4] += 1;
            if (ch < 14) dirtyness[ch+1] += 1;
            if (ch < 13) dirtyness[ch+2] += 1;
            if (ch < 12) dirtyness[ch+3] += 1;
            if (ch < 11) dirtyness[ch+4] += 1;
        }
        DBGPRINT(RT_DEBUG_TRACE,"Msleep at ch#%d to collect RX=%d, RSSI=%d, CRC error =%d, False CCA =%d\n", 
            ch, pAd->Counters.GoodReceives, max_rssi[ch] - RSSI_TO_DBM_OFFSET, FcsError, FalseCca);
    }
    pAd->PortCfg.AutoChannel_Channel = 0;

    DBGPRINT(RT_DEBUG_TRACE,"Dirtyness = %d.%d.%d.%d - %d.%d.%d.%d - %d.%d.%d.%d - %d.%d\n", 
        dirtyness[1], dirtyness[2], dirtyness[3], dirtyness[4], dirtyness[5], dirtyness[6], dirtyness[7],
        dirtyness[8], dirtyness[9], dirtyness[10], dirtyness[11], dirtyness[12], dirtyness[13], dirtyness[14]);

    // RULE 1. pick up a good channel that no one used
    for (ch = FirstChannel(pAd); ch !=0; ch = NextChannel(pAd, ch))
    {
        if (dirtyness[ch] == 0) break;
    }
    if (ch != 0)
    {
        DBGPRINT(RT_DEBUG_TRACE,"ApAutoSelectChannel pick up ch#%d\n",ch);
        return ch;
    }

    // RULE 2. if not available, then co-use a channel that's no interference (dirtyness=10)
    // RULE 3. if not available, then co-use a channel that has minimum interference (dirtyness=11,12)
    for (dirty = 10; dirty <= 12; dirty++)
    {
        UCHAR candidate[MAX_LEN_OF_CHANNELS+1], candidate_num;
        UCHAR min_rssi = 255, final_channel = 0;
        
        candidate_num = 0;
        NdisZeroMemory(candidate, MAX_LEN_OF_CHANNELS+1);
        for (ch = FirstChannel(pAd); ch != 0; ch = NextChannel(pAd, ch))
        {
            if (dirtyness[ch] == dirty) 
            { 
                candidate[ch]=1; 
                candidate_num++; 
            }
        }
        // if there's more than 1 candidate, pick up the channel with minimum RSSI
        if (candidate_num)
        {
            for (ch = 1; ch <= 14; ch ++)
            {
                if (candidate[ch] && (max_rssi[ch] < min_rssi))
                {
                    final_channel = ch;
                    min_rssi = max_rssi[ch];
                }
            }
            if (final_channel != 0)
            {
                DBGPRINT(RT_DEBUG_TRACE,"ApAutoSelectChannel pick up ch#%d\n",final_channel);
                return final_channel;
            }
        }
    }

    // RULE 4. still not available, pick up the first channel
    ch = FirstChannel(pAd);
    DBGPRINT(RT_DEBUG_TRACE,"ApAutoSelectChannel pick up ch#%d\n",ch);
    return ch;
}

UCHAR FirstChannel(
    IN PRTMP_ADAPTER pAd)
{
    UCHAR ch = 0;
    switch (pAd->PortCfg.CountryRegion)
    {
        case REGION_FCC:    // 1 - 11
            ch = 1;
            break;
        case REGION_IC:     // 1 -11
            ch = 1;
            break;
        case REGION_ISRAEL:  // 3 - 9
            ch = 3;
            break;
        case REGION_ETSI:   // 1 - 13
            ch = 1;
            break;
        case REGION_SPAIN:  // 10 - 11
            ch = 10;
            break;
        case REGION_FRANCE: // 10 -13
            ch = 10;
            break;
        case REGION_MKK:    // 14
            ch = 14;
            break;
        case REGION_MKK1:   // 1 - 14
            ch = 1;
            break;
        default:            // Error. should never happen
            break;
    }   
    return ch;
}

UCHAR NextChannel(
    IN PRTMP_ADAPTER pAd,
    IN UCHAR         ch)
{
    UCHAR next = ch + 1;
    switch (pAd->PortCfg.CountryRegion)
    {
        case REGION_FCC:    // 1 - 11
            if (next > 11) next = 0;
            break;
        case REGION_IC:     // 1 -11
            if (next > 11) next = 0;
            break;
        case REGION_ISRAEL: // 3 - 9
            if (next > 9)  next = 0;
            break;
        case REGION_ETSI:   // 1 - 13
            if (next > 13) next = 0;
            break;
        case REGION_SPAIN:  // 10 - 11
            if (next > 11) next = 0;
            break;
        case REGION_FRANCE: // 10 -13
            if (next > 13) next = 0;
            break;
        case REGION_MKK:    // 14
            if (next > 14) next = 0;
            break;
        case REGION_MKK1:   // 1 - 14
            if (next > 14) next = 0;
            break;
        default:            // Error. should never happen
            next = 0;
            break;
    }   
    return next;
}

UCHAR RTMPFindAPIndex(
    IN PRTMP_ADAPTER    pAd,
    IN CHAR             *Ssid,
    IN UCHAR            SsidLen) 
{
    UINT    i;

    for (i = 0; i < pAd->PortCfg.SsidNum; i++)
    {
        if ((SsidLen == pAd->PortCfg.MSSID[i].SsidLen) && RTMPEqualMemory(Ssid, pAd->PortCfg.MSSID[i].Ssid, (ULONG) SsidLen))
            return i;
    }
    
    return ENTRY_NOT_FOUND;
}

void ProbeReqCacheUpdate(
    IN PRTMP_ADAPTER    pAd,
    IN PMACADDR         pAddr,
    IN UCHAR            apidx) 
{
    UCHAR   i, cacheidx;

    cacheidx = ProbeReqCacheLookup(pAd, pAddr);
    if (cacheidx != ENTRY_NOT_FOUND)
    {
        // update apidx, maybe active scan to another ssid
        pAd->PortCfg.ProbeReqCache[cacheidx].apidx = apidx;
        pAd->PortCfg.ProbeReqCache[cacheidx].timestamp = jiffies;
        DBGPRINT(RT_DEBUG_INFO,"ProbeReqCacheUpdate update %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
            pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], cacheidx, apidx);
        return;
    }
    else
    {
        // append cache item
        for (i = 0; i < MAX_PROBEREQ_CACHE_SIZE; i++)
        {
            if (!pAd->PortCfg.ProbeReqCache[i].Valid)
            {
                pAd->PortCfg.ProbeReqCache[i].Valid = TRUE;
                pAd->PortCfg.ProbeReqCache[i].apidx = apidx;
                COPY_MAC_ADDR(&pAd->PortCfg.ProbeReqCache[i].MAC, pAddr);
                pAd->PortCfg.ProbeReqCache[i].timestamp = jiffies;
                
                DBGPRINT(RT_DEBUG_TRACE,"ProbeReqCacheUpdate append %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
                    pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], i, apidx);
                return;
            }
        }
    }

    DBGPRINT(RT_DEBUG_ERROR,"ProbeReqCacheUpdate can not find available cache!\n");
}

UCHAR ProbeReqCacheLookup(
    IN PRTMP_ADAPTER    pAd,
    IN PMACADDR         pAddr)
{
    INT     i;

    for (i = 0; i < MAX_PROBEREQ_CACHE_SIZE; i++)
    {
        if (pAd->PortCfg.ProbeReqCache[i].Valid && MAC_ADDR_EQUAL(&pAd->PortCfg.ProbeReqCache[i].MAC, pAddr))
        {
            DBGPRINT(RT_DEBUG_INFO,"ProbeReqCacheLookup find %02x:%02x:%02x:%02x:%02x:%02x in cache(%d)from I/F(ra%d)\n", 
                pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], i, pAd->PortCfg.ProbeReqCache[i].apidx);
            
            //pAd->PortCfg.ProbeReqCache[i].timestamp = jiffies;
            return i;
        }
    }

    DBGPRINT(RT_DEBUG_WARN,"ProbeReqCacheLookup %02x:%02x:%02x:%02x:%02x:%02x can not be found\n",
        pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5]);
    return ENTRY_NOT_FOUND;
}

void ProbeReqCacheDelete(
    IN PRTMP_ADAPTER    pAd,
    IN PMACADDR         pAddr)
{
    UCHAR   cacheidx;

    cacheidx = ProbeReqCacheLookup(pAd, pAddr);
    if (cacheidx != ENTRY_NOT_FOUND)
    {
        pAd->PortCfg.ProbeReqCache[cacheidx].Valid = FALSE;
        DBGPRINT(RT_DEBUG_TRACE, "ProbeReqCacheDelete from I/F(ra%d)\n", pAd->PortCfg.ProbeReqCache[cacheidx].apidx);
    }
}

