/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************
     
     Module Name:
     sync.c
     
     Abstract:
     Synchronization state machine related services
     
     Revision History:
     Who         When          What
     --------    ----------    ----------------------------------------------
     John Chang  08-04-2003    created for 11g soft-AP
     
 */

#include "rt_config.h"

NDIS_STATUS MlmeAllocateMemory(
	IN PRTMP_ADAPTER pAd,
	OUT PVOID        *AllocVa);

UCHAR FirstChannel2(
	IN PRTMP_ADAPTER pAd);

UCHAR NextChannel2(
	IN PRTMP_ADAPTER pAd, 
	IN UCHAR channel);


// 2.4 Ghz channel plan
UCHAR Ra24Ghz_FCC[] = {1,2,3,4,5,6,7,8,9,10,11};
UCHAR Ra24Ghz_IC[]  = {1,2,3,4,5,6,7,8,9,10,11};
UCHAR Ra24Ghz_ESTI[]= {1,2,3,4,5,6,7,8,9,10,11,12,13};
UCHAR Ra24Ghz_SPAIN[] = {10,11};
UCHAR Ra24Ghz_FRANCE[] = {10,11,12,13};
UCHAR Ra24Ghz_MKK[] =  {14};
UCHAR Ra24Ghz_MKK1[] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14};
UCHAR Ra24Ghz_ISRAEL[]  = {3,4,5,6,7,8,9};

// 5 Ghz channel plan
UCHAR Ra5Ghz_UNII[] = {36,40,44,48,52,56,60,64,  149,153,157,161};
UCHAR Ra5Ghz_MMAC[] = {34,38,42,46};
UCHAR Ra5Ghz_HyperLAN2[] = {36,40,44,48,52,56,60,64,  100,104,108,112,116,120,124,128,132,136,140};


/*
    ==========================================================================
    Description:
        The sync state machine, 
    Parameters:
        Sm - pointer to the state machine
    Note:
        the state machine looks like the following
        
                        SYNC_IDLE             
    MT2_PEER_PROBE_REQ  peer_probe_req_action    
    ==========================================================================
 */
VOID SyncStateMachineInit(
	IN PRTMP_ADAPTER pAd, 
	IN STATE_MACHINE *Sm, 
	OUT STATE_MACHINE_FUNC Trans[])
{
	StateMachineInit(Sm, (STATE_MACHINE_FUNC*)Trans, MAX_SYNC_STATE, MAX_SYNC_MSG, (STATE_MACHINE_FUNC)Drop, SYNC_IDLE, SYNC_MACHINE_BASE);

	StateMachineSetAction(Sm, SYNC_IDLE, MT2_PEER_PROBE_REQ, (STATE_MACHINE_FUNC)PeerProbeReqAction);
	StateMachineSetAction(Sm, SYNC_IDLE, MT2_PEER_BEACON, (STATE_MACHINE_FUNC)PeerBeaconAction);

	/* for AP-client */

	/* idle state */
	StateMachineSetAction(Sm, SYNC_IDLE, MT2_MLME_SCAN_REQ, (STATE_MACHINE_FUNC)MlmeScanReqAction);

	/* scan_listen state */
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_MLME_SCAN_REQ, (STATE_MACHINE_FUNC)InvalidStateWhenScan);
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_MLME_JOIN_REQ, (STATE_MACHINE_FUNC)InvalidStateWhenScan);
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_MLME_START_REQ, (STATE_MACHINE_FUNC)InvalidStateWhenScan);
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_PEER_BEACON, (STATE_MACHINE_FUNC)PeerBeaconAtScanAction);
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_PEER_PROBE_RSP, (STATE_MACHINE_FUNC)PeerBeaconAtScanAction);
	StateMachineSetAction(Sm, SCAN_LISTEN, MT2_SCAN_TIMEOUT, (STATE_MACHINE_FUNC)ScanTimeoutAction);

	init_timer(&pAd->Mlme.SyncAux.ScanTimer);
	pAd->Mlme.SyncAux.ScanTimer.data = (unsigned long)pAd;
	pAd->Mlme.SyncAux.ScanTimer.function = &ScanTimeout;
}

/* 
    ==========================================================================
    Description:
        Scan timeout handler, executed in timer thread
    ==========================================================================
 */
VOID ScanTimeout(
	IN  unsigned long data)
{
	RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;

	DBGPRINT(RT_DEBUG_INFO,"SYNC - Scan Timeout \n");
	MlmeEnqueue(&pAd->Mlme.Queue, SYNC_STATE_MACHINE, MT2_SCAN_TIMEOUT, 0, NULL);
	MlmeHandler(pAd);
}

/* 
    ==========================================================================
    Description:
    ==========================================================================
 */
VOID InvalidStateWhenScan(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem)
{
	DBGPRINT(RT_DEBUG_TRACE, "AYNC - InvalidStateWhenScan(state=%d). Reset SYNC machine\n", pAd->Mlme.SyncMachine.CurrState);
	pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
	MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_STATE_MACHINE_REJECT);
}


/* 
    ==========================================================================
    Description:
        Scan timeout procedure. basically add channel index by 1 and rescan
    ==========================================================================
 */
VOID ScanTimeoutAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem)
{
	pAd->Mlme.SyncAux.Channel = NextChannel2(pAd, pAd->Mlme.SyncAux.Channel);
	ScanNextChannel(pAd);
}

/* 
    ==========================================================================
    Description:
        peer sends beacon back when scanning
    ==========================================================================
 */
VOID PeerBeaconAtScanAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem)
{
	MACADDR         Bssid, Addr2;
	UCHAR           Ssid[MAX_LEN_OF_SSID], BssType, Channel, Rates[MAX_LEN_OF_SUPPORTED_RATES], RatesLen, 
		SsidLen, DtimCount, DtimPeriod, BcastFlag, MessageToMe, Legacy;
	CF_PARM         CfParm;
	USHORT          BeaconPeriod, AtimWin, CapabilityInfo;
	MACFRAME       *Fr;
//	USHORT          Msg;
	LARGE_INTEGER   TimeStamp;
	BOOLEAN         CfExist = FALSE;
	BOOLEAN         ExtendedRateIeExist;
	UCHAR           Erp;

	// New for WPA security suites
	UCHAR                       VarIE[MAX_VIE_LEN];     // Total VIE length = MAX_VIE_LEN - -5
	NDIS_802_11_VARIABLE_IEs    *pVIE = NULL;

	// NdisFillMemory(Ssid, MAX_LEN_OF_SSID, 0x00);
	Fr = (MACFRAME *) Elem->Msg;
	// Init Variable IE structure
	pVIE = (PNDIS_802_11_VARIABLE_IEs) VarIE;
	pVIE->Length = 0;
	if (PeerBeaconAndProbeRspSanity2(pAd, 
		Elem->Msg, 
		Elem->MsgLen, 
		&Addr2, 
		&Bssid, Ssid, 
		&SsidLen, 
		&BssType, 
		&BeaconPeriod, 
		&Channel, 
		&TimeStamp, 
		&CfExist, 
		&CfParm, 
		&AtimWin, 
		&CapabilityInfo, 
		Rates, 
		&RatesLen,
		&ExtendedRateIeExist,
		&Erp,
		&DtimCount, 
		&DtimPeriod, 
		&BcastFlag, 
		&MessageToMe, 
		&Legacy,
		pVIE)) {
		ULONG Idx;
		UCHAR Rssi = 0;

		// This correct im-proper RSSI indication during SITE SURVEY issue.
		// Always report bigger RSSI during SCANNING when receiving multiple BEACONs from the same AP. 
		// This case happens because BEACONs come from adjacent channels, so RSSI become weaker as we 
		// switch to more far away channels.
		Idx = BssTableSearch(&pAd->PortCfg.BssTab, &Bssid);
		if (Idx != BSS_NOT_FOUND) 
			Rssi = pAd->PortCfg.BssTab.BssEntry[Idx].Rssi;
		if (Elem->Rssi > Rssi)
			Rssi = Elem->Rssi;

		DBGPRINT(RT_DEBUG_INFO, "SYNC - PeerBeaconAtScanAction (Subtype=%d, SsidLen=%d, Ssid=%s)\n", Fr->Hdr.SubType, SsidLen,Ssid);

		// Mask out unnecessary capability information
		CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;

		BssTableSetEntry(pAd, &pAd->PortCfg.BssTab, &Bssid, Ssid, SsidLen, BssType, 
			BeaconPeriod, CfExist, &CfParm, AtimWin, CapabilityInfo, Rates, 
			RatesLen, ExtendedRateIeExist, Channel, Rssi, TimeStamp, pVIE);
	}
	// sanity check fail, ignored
}



/* 
    ==========================================================================
    Description:
        Scan next channel
    ==========================================================================
 */
VOID ScanNextChannel(
	IN PRTMP_ADAPTER pAd)
{
	MACHDR          Hdr;
	UCHAR           SsidIe = IE_SSID, SuppRateIe = IE_SUPP_RATES;
	VOID           *OutBuffer = NULL;
	VOID           *OutBuffer2 = NULL;
//	USHORT          Msg;
	NDIS_STATUS     NStatus;
	ULONG           FrameLen = 0;
	UCHAR           SsidLen = 0;

	if (pAd->Mlme.SyncAux.Channel == 0) {
		DBGPRINT(RT_DEBUG_TRACE, "SYNC - End of SCAN, restore to channel %d\n",pAd->PortCfg.Channel);
		AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
		AsicLockChannel(pAd, pAd->PortCfg.Channel);

		pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
		RTMPResumeMsduTransmission(pAd);
		/* start sending BEACON */
		MakeBssBeacon(pAd);
		UpdateBeaconFrame(pAd);
		AsicEnableBssSync(pAd);

		//        MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_SUCCESS);
	} else {
		AsicSwitchChannel(pAd, pAd->Mlme.SyncAux.Channel);

		// Total SCAN time still limits within 3 sec (DDK constraint). 
		// TODO: We need more intelligent rules here to further improve out-of-service issue.
		// e.g. temporary stop copying NDIS packet to TxRing until SCAN complete
		//      if (INFRA_ON(pAd) || ADHOC_ON(pAd))
		if (pAd->PortCfg.PhyMode == PHY_11ABG_MIXED)
			pAd->Mlme.SyncAux.ScanTimer.expires = jiffies + (MIN_CHANNEL_TIME * HZ)/1000;
		else
			pAd->Mlme.SyncAux.ScanTimer.expires = jiffies + (MAX_CHANNEL_TIME * HZ)/1000;
		add_timer(&pAd->Mlme.SyncAux.ScanTimer);

		if (pAd->Mlme.SyncAux.ScanType == SCAN_ACTIVE) {
			NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
			if (NStatus != NDIS_STATUS_SUCCESS) {
				DBGPRINT(RT_DEBUG_TRACE, "SYNC - ScanNextChannel() allocate memory fail\n");
				pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
				MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_FAIL_NO_RESOURCE);
				return;
			}

			DBGPRINT(RT_DEBUG_INFO, "SYNC - send ProbeReq @ channel=%d...\n", pAd->Mlme.SyncAux.Channel);
			MgtMacHeaderInit(pAd, &Hdr, SUBTYPE_PROBE_REQ, 0, &pAd->PortCfg.Broadcast, &pAd->PortCfg.Broadcast);
			SsidLen = 0;
			MakeOutgoingFrame(OutBuffer,        &FrameLen,
				sizeof(MACHDR),   &Hdr,
				1,                &SsidIe,
				1,                &SsidLen,  
				1,                &SuppRateIe,
				1,                &pAd->PortCfg.SupportedRatesLen,
				pAd->PortCfg.SupportedRatesLen, pAd->PortCfg.SupportedRates, 
				END_OF_ARGS);
			MiniportMMRequest(pAd, OutBuffer, FrameLen);

			// Allocate another for probe scan with SSID
			NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer2);  //Get an unused nonpaged memory
			if (NStatus != NDIS_STATUS_SUCCESS) {
				DBGPRINT(RT_DEBUG_TRACE, "SYNC - ScanNextChannel() allocate memory fail\n");
				pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
				MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_FAIL_NO_RESOURCE);
				return;
			}
			// make another probe scan with SSID from mlme.cntlaux.ssid
			SsidLen = pAd->Mlme.CntlAux.SsidLen;
			MakeOutgoingFrame(OutBuffer2,       &FrameLen,
				sizeof(MACHDR),   &Hdr,
				1,                &SsidIe,
				1,                &SsidLen,
				SsidLen,          pAd->Mlme.CntlAux.Ssid,
				1,                &SuppRateIe,
				1,                &pAd->PortCfg.SupportedRatesLen,
				pAd->PortCfg.SupportedRatesLen, pAd->PortCfg.SupportedRates, 
				END_OF_ARGS);
			MiniportMMRequest(pAd, OutBuffer2, FrameLen);

			DBGPRINT(RT_DEBUG_INFO, "SYNC - send ProbeReq @ channel=%d...\n", pAd->Mlme.SyncAux.Channel);
		} else if (pAd->Mlme.SyncAux.ScanType == SCAN_PASSIVE) {
			NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
			if (NStatus != NDIS_STATUS_SUCCESS) {
				DBGPRINT(RT_DEBUG_TRACE, "SYNC - ScanNextChannel() allocate memory fail\n");
				pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
				MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_FAIL_NO_RESOURCE);
				return;
			}

			DBGPRINT(RT_DEBUG_INFO, "SYNC - send ProbeReq @ channel=%d...\n", pAd->Mlme.SyncAux.Channel);
			MgtMacHeaderInit(pAd, &Hdr, SUBTYPE_PROBE_REQ, 0, &pAd->PortCfg.Broadcast, &pAd->PortCfg.Broadcast);
			SsidLen = 0;
			MakeOutgoingFrame(OutBuffer,        &FrameLen,
				sizeof(MACHDR),   (UCHAR*)&Hdr,
				1,                &SsidIe,
				1,                &SsidLen,  
				1,                &SuppRateIe,
				1,                &pAd->PortCfg.SupportedRatesLen,
				pAd->PortCfg.SupportedRatesLen, pAd->PortCfg.SupportedRates, 
				END_OF_ARGS);
			MiniportMMRequest(pAd, OutBuffer, FrameLen);
		}

		pAd->Mlme.SyncMachine.CurrState = SCAN_LISTEN;
	}
}

/* 
    ==========================================================================
    Description:
        This routine return the first channel number according to the country 
        code selection and RF IC selection (signal band or dual band). It is called
        whenever driver need to start a site survey of all supported channels.
    Return:
        ch - the first channel number of current country code setting
    ==========================================================================
 */
UCHAR FirstChannel2(
	IN PRTMP_ADAPTER pAd)
{
	return pAd->PortCfg.ChannelList[0];
}

/* 
    ==========================================================================
    Description:
        This routine returns the next channel number. This routine is called
        during driver need to start a site survey of all supported channels.
    Return:
        next_channel - the next channel number valid in current country code setting.
    Note:
        return 0 if no more next channel
    ==========================================================================
 */
UCHAR NextChannel2(
	IN PRTMP_ADAPTER pAd, 
	IN UCHAR channel)
{
	int i;
	UCHAR next_channel = 0;

	for (i = 0; i < (pAd->PortCfg.ChannelListNum - 1); i++)
		if (channel == pAd->PortCfg.ChannelList[i]) {
			next_channel = pAd->PortCfg.ChannelList[i+1];
			break;
		}
	return next_channel;
}

/* 
    ==========================================================================
    Description:
        MLME SCAN req state machine procedure
    ==========================================================================
 */
VOID MlmeScanReqAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem)
{
//	MACADDR        Bssid;
	UCHAR          Ssid[MAX_LEN_OF_SSID], SsidLen, ScanType, BssType;
	ULONG          Now;

	// Suspend MSDU transmission here
	RTMPSuspendMsduTransmission(pAd);

	// first check the parameter sanity
	if (MlmeScanReqSanity(pAd, 
		Elem->Msg, 
		Elem->MsgLen, 
		&BssType, 
		Ssid, 
		&SsidLen, 
		&ScanType)) {
		DBGPRINT(RT_DEBUG_TRACE, "SYNC - MlmeScanReqAction\n");
		Now = jiffies;
		pAd->PortCfg.LastScanTime = Now;
		// reset all the timers
		//        del_timer_sync(&pAd->Mlme.SyncAux.BeaconTimer);
		del_timer_sync(&pAd->Mlme.SyncAux.ScanTimer);

		// record desired BSS parameters
		pAd->Mlme.SyncAux.BssType = BssType;
		pAd->Mlme.SyncAux.ScanType = ScanType;
		pAd->Mlme.SyncAux.SsidLen = SsidLen;
		NdisMoveMemory(pAd->Mlme.SyncAux.Ssid, Ssid, SsidLen);

		// start from the first channel
		pAd->Mlme.SyncAux.Channel = FirstChannel2(pAd);
		ScanNextChannel(pAd);
	} else {
		DBGPRINT(RT_DEBUG_ERROR, "SYNC - MlmeScanReqAction() sanity check fail. BUG!!!\n");
		pAd->Mlme.SyncMachine.CurrState = SYNC_IDLE;
		MlmeCntlConfirm(pAd, MT2_SCAN_CONF, MLME_INVALID_FORMAT);
	}
}

/* 
    ==========================================================================
    Description:
        Update PortCfg->ChannelList[] according to 1) Country Region 2) RF IC type,
        and 3) PHY-mode user selected.
        The outcome is used by driver when doing site survey.
    ==========================================================================
 */
VOID BuildChannelList(
	IN PRTMP_ADAPTER pAd)
{
	UCHAR i,  index = 0;
	NdisZeroMemory(pAd->PortCfg.ChannelList, MAX_LEN_OF_CHANNELS);

	// if not 11a-only mode, channel list starts from 2.4Ghz band
	if (pAd->PortCfg.PhyMode != PHY_11A) {
		switch (pAd->PortCfg.CountryRegion) {
		case REGION_FCC:    // 1 - 11
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_FCC, sizeof(Ra24Ghz_FCC));
			index += sizeof(Ra24Ghz_FCC);
			break;
		case REGION_IC:     // 1 -11
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_IC, sizeof(Ra24Ghz_IC));
			index += sizeof(Ra24Ghz_IC);
			break;
		case REGION_ISRAEL:  // 3 - 9
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_ISRAEL, sizeof(Ra24Ghz_ISRAEL));
			index += sizeof(Ra24Ghz_ISRAEL);
			break;
		case REGION_ETSI:   // 1 - 13
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_ESTI, sizeof(Ra24Ghz_ESTI));
			index += sizeof(Ra24Ghz_ESTI);
			break;
		case REGION_SPAIN:  // 10 - 11
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_SPAIN, sizeof(Ra24Ghz_SPAIN));
			index += sizeof(Ra24Ghz_SPAIN);
			break;
		case REGION_FRANCE: // 10 -13
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_FRANCE, sizeof(Ra24Ghz_FRANCE));
			index += sizeof(Ra24Ghz_FRANCE);
			break;
		case REGION_MKK:    // 14
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_MKK, sizeof(Ra24Ghz_MKK));
			index += sizeof(Ra24Ghz_MKK);
			break;
		case REGION_MKK1:   // 1 - 14
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra24Ghz_MKK1, sizeof(Ra24Ghz_MKK1));
			index += sizeof(Ra24Ghz_MKK1);
			break;
		default:            // Error. should never happen
			break;
		}
	}

	if ((pAd->PortCfg.PhyMode == PHY_11A) || (pAd->PortCfg.PhyMode == PHY_11ABG_MIXED)) {
#if 0
		switch (pAd->PortCfg.CountryRegion) {
		case REGION_FCC:     // UNII <36,40,44,48,52,56,60,64> <149,153,157,161>
		case REGION_IC:      // UNII <36,40,44,48,52,56,60,64> <149,153,157,161>
		case REGION_ISRAEL:  // UNII <36,40,44,48,52,56,60,64> <149,153,157,161>
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra5Ghz_UNII, sizeof(Ra5Ghz_UNII));
			index += sizeof(Ra5Ghz_UNII);
			break;
		case REGION_ETSI:    // HiperLAN2 <36,40,44,48,52,56,60,64> <100,104,108,112,116,120,124,128,132,136,140>
		case REGION_SPAIN:   // HiperLAN2 <36,40,44,48,52,56,60,64> <100,104,108,112,116,120,124,128,132,136,140>
		case REGION_FRANCE:  // HiperLAN2 <36,40,44,48,52,56,60,64> <100,104,108,112,116,120,124,128,132,136,140>
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra5Ghz_HyperLAN2, sizeof(Ra5Ghz_HyperLAN2));
			index += sizeof(Ra5Ghz_HyperLAN2);
			break;
		case REGION_MKK:    // Japan MMAC <34,38,42,46>
		case REGION_MKK1:   // Japan MMAC <34,38,42,46>
			NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra5Ghz_MMAC, sizeof(Ra5Ghz_MMAC));
			index += sizeof(Ra5Ghz_MMAC);
			break;
		default:            // Error. should never happen
			break;
		}
#else
		// 2003-10-05 john - use UNII temoparaily for all regulation domains for easy test untill
		// RF guys confirm the supported channel plans
		NdisMoveMemory(&pAd->PortCfg.ChannelList[index], Ra5Ghz_UNII, sizeof(Ra5Ghz_UNII));
		index += sizeof(Ra5Ghz_UNII);
#endif
	}

	pAd->PortCfg.ChannelListNum = index;
	DBGPRINT(RT_DEBUG_TRACE,"country code=%d, RFIC=%d, PHY mode=%d, support %d channels\n", 
		pAd->PortCfg.CountryRegion, pAd->PortCfg.RfType, pAd->PortCfg.PhyMode, pAd->PortCfg.ChannelListNum);
	for (i=0;i<index;i++) {
		DBGPRINT(RT_DEBUG_TRACE,"channel #%d\n", pAd->PortCfg.ChannelList[i]);
	}
#ifdef VINCENT_REGION
	// Check for Channel setting, if wrong, set it to the lowest channel.
	if ( (pAd->PortCfg.Channel < pAd->PortCfg.ChannelList[0]) || 
			(pAd->PortCfg.Channel > pAd->PortCfg.ChannelList[index-1]) )
	{
		DBGPRINT(RT_DEBUG_TRACE,"Change Channel %d to %d\n", pAd->PortCfg.Channel, pAd->PortCfg.ChannelList[0]);
		pAd->PortCfg.Channel = pAd->PortCfg.ChannelList[0];
	}
#endif
}


/*
    ==========================================================================
    Description:
        Process the received ProbeRequest from clients
    Parameters:
        Elem - msg containing the ProbeReq frame
    ==========================================================================
 */
VOID PeerProbeReqAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR       Addr2;
    CHAR          Ssid[MAX_LEN_OF_SSID];
    UCHAR         SsidLen, apidx = 0;
    MACHDR        ProbeRspHdr;
    UCHAR         *OutBuffer = NULL;
    ULONG         FrameLen = 0;
    LONGLONG      FakeTimestamp;
    UCHAR         SsidIe = IE_SSID, DsIe = IE_DS_PARM, SuppIe = IE_SUPP_RATES, 
                  DsLen = 1, RSNIe = IE_WPA, RSNIe2 = IE_WPA2;
    
    // No need to reply probe req. in bridge mode
    if (pAd->WdsTab.Mode == WDS_BRIDGE_MODE)
        return;
    
    if (! PeerProbeReqSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, Ssid, &SsidLen)) //, Rates, &RatesLen))
        return;
    
    if (((SsidLen == 0) && (pAd->PortCfg.HideSsid == 0)) || ((apidx = RTMPFindAPIndex(pAd, &Ssid[0], SsidLen)) != ENTRY_NOT_FOUND))
    {
        // allocate and send out ProbeRsp frame
        OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
        if(OutBuffer == NULL)
	        return;

        // Append or update probe req. cache for MSSID(Multiple SSID) purpose, only when active scan
        // apidx = 0 have no need to append to cache, becaue it is default SSID when auth.
        if ((apidx != ENTRY_NOT_FOUND) && (apidx != 0))
        {
            ProbeReqCacheUpdate(pAd, &Addr2, apidx);
        }
        else
            apidx = MAIN_MSSID; // avoid apidx = 0xFF
            
        DBGPRINT(RT_DEBUG_INFO, "SYNC - I/F(ra%d)Send PROBE_RSP to %02x:%02x:%02x:%02x:%02x:%02x...\n", apidx,
            Addr2.Octet[0],Addr2.Octet[1],Addr2.Octet[2],Addr2.Octet[3],Addr2.Octet[4],Addr2.Octet[5] );

        MgtMacHeaderInit(pAd, &ProbeRspHdr, SUBTYPE_PROBE_RSP, 0, &Addr2, &pAd->PortCfg.Bssid);

        if ((pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA) || (pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPAPSK))
	    	RSNIe = IE_WPA;
	    else if ((pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA2) || (pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA2PSK))
	    	RSNIe = IE_WPA2;

        // Append RSN_IE when  WPA OR WPAPSK, 
        if (((pAd->PortCfg.PhyMode == PHY_11BG_MIXED) || (pAd->PortCfg.PhyMode == PHY_11G)) 
            && (pAd->PortCfg.MSSID[apidx].AuthMode < Ndis802_11AuthModeWPA))
        {
            UCHAR SupportedRatesLen = 4, ErpLen = 1;
            UCHAR ExtendedRatesIe = IE_EXT_SUPP_RATES, ErpIe = IE_ERP;
            UCHAR ExtendedRatesLen = pAd->PortCfg.SupportedRatesLen - SupportedRatesLen;
                
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                          MAC_HDR_LEN,                          &ProbeRspHdr, 
                          TIMESTAMP_LEN,                        &FakeTimestamp,
                          2,                                    &pAd->PortCfg.BeaconPeriod,
                          2,                                    &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                          1,                                    &SsidIe, 
                          1,                                    &pAd->PortCfg.MSSID[apidx].SsidLen, 
                          pAd->PortCfg.MSSID[apidx].SsidLen,    pAd->PortCfg.MSSID[apidx].Ssid,
                          1,                                    &SuppIe, 
                          1,                                    &SupportedRatesLen,
                          SupportedRatesLen,                    pAd->PortCfg.SupportedRates, 
                          1,                                    &DsIe, 
                          1,                                    &DsLen, 
                          1,                                    &pAd->PortCfg.Channel,
                          1,                                    &ErpIe,
                          1,                                    &ErpLen,
                          1,                                    &pAd->Mlme.ErpIeContent,
                          1,                                    &ExtendedRatesIe,
                          1,                                    &ExtendedRatesLen,
                          ExtendedRatesLen,                     &pAd->PortCfg.SupportedRates[SupportedRatesLen],
                          END_OF_ARGS);
        }
        else if(  pAd->PortCfg.MSSID[apidx].AuthMode < Ndis802_11AuthModeWPA)
        {
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                        MAC_HDR_LEN,                            &ProbeRspHdr, 
                        TIMESTAMP_LEN,                          &FakeTimestamp,
                        2,                                      &pAd->PortCfg.BeaconPeriod,
                        2,                                      &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                        1,                                      &SsidIe, 
                        1,                                      &pAd->PortCfg.MSSID[apidx].SsidLen, 
                        pAd->PortCfg.MSSID[apidx].SsidLen,      pAd->PortCfg.MSSID[apidx].Ssid,
                        1,                                      &SuppIe, 
                        1,                                      &pAd->PortCfg.SupportedRatesLen,
                        pAd->PortCfg.SupportedRatesLen,         pAd->PortCfg.SupportedRates, 
                        1,                                      &DsIe, 
                        1,                                      &DsLen, 
                        1,                                      &pAd->PortCfg.Channel,
                        END_OF_ARGS);
        }
        else if (((pAd->PortCfg.PhyMode == PHY_11BG_MIXED) || (pAd->PortCfg.PhyMode == PHY_11G)) 
            && ((pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA1WPA2) ||(pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK)))
        {
            UCHAR SupportedRatesLen = 4, ErpLen = 1;
            UCHAR ExtendedRatesIe = IE_EXT_SUPP_RATES, ErpIe = IE_ERP;
            UCHAR ExtendedRatesLen = pAd->PortCfg.SupportedRatesLen - SupportedRatesLen;
                
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                          MAC_HDR_LEN,                          &ProbeRspHdr, 
                          TIMESTAMP_LEN,                        &FakeTimestamp,
                          2,                                    &pAd->PortCfg.BeaconPeriod,
                          2,                                    &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                          1,                                    &SsidIe, 
                          1,                                    &pAd->PortCfg.MSSID[apidx].SsidLen, 
                          pAd->PortCfg.MSSID[apidx].SsidLen,    pAd->PortCfg.MSSID[apidx].Ssid,
                          1,                                    &SuppIe, 
                          1,                                    &SupportedRatesLen,
                          SupportedRatesLen,                    pAd->PortCfg.SupportedRates, 
                          1,                                    &DsIe, 
                          1,                                    &DsLen, 
                          1,                                    &pAd->PortCfg.Channel,
                          1,                                    &ErpIe,
                          1,                                    &ErpLen,
                          1,                                    &pAd->Mlme.ErpIeContent,
                          1,                                    &ExtendedRatesIe,
                          1,                                    &ExtendedRatesLen,
                          ExtendedRatesLen,                     &pAd->PortCfg.SupportedRates[SupportedRatesLen],
                          1,                                      &RSNIe,
						  1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],
                          pAd->PortCfg.MSSID[apidx].RSNIE_Len[0], pAd->PortCfg.MSSID[apidx].RSN_IE[0],
                          1,                                      &RSNIe2,
                          1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[1],
                          pAd->PortCfg.MSSID[apidx].RSNIE_Len[1], pAd->PortCfg.MSSID[apidx].RSN_IE[1],
                          END_OF_ARGS);
        }
        else if ((pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPA1WPA2) ||(pAd->PortCfg.MSSID[apidx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
        {
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                        MAC_HDR_LEN,                            &ProbeRspHdr, 
                        TIMESTAMP_LEN,                          &FakeTimestamp,
                        2,                                      &pAd->PortCfg.BeaconPeriod,
                        2,                                      &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                        1,                                      &SsidIe, 
                        1,                                      &pAd->PortCfg.MSSID[apidx].SsidLen, 
                        pAd->PortCfg.MSSID[apidx].SsidLen,      pAd->PortCfg.MSSID[apidx].Ssid,
                        1,                                      &SuppIe, 
                        1,                                      &pAd->PortCfg.SupportedRatesLen,
                        pAd->PortCfg.SupportedRatesLen,         pAd->PortCfg.SupportedRates, 
                        1,                                      &DsIe, 
                        1,                                      &DsLen, 
                        1,                                      &pAd->PortCfg.Channel,
                        1,                                      &RSNIe,
                        1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],
                        pAd->PortCfg.MSSID[apidx].RSNIE_Len[0], pAd->PortCfg.MSSID[apidx].RSN_IE[0],
                        1,                                      &RSNIe2,
                        1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[1],
                        pAd->PortCfg.MSSID[apidx].RSNIE_Len[1], pAd->PortCfg.MSSID[apidx].RSN_IE[1],
                        END_OF_ARGS);
        }
        else if ((pAd->PortCfg.PhyMode == PHY_11BG_MIXED) || (pAd->PortCfg.PhyMode == PHY_11G))
        {
            UCHAR SupportedRatesLen = 4, ErpLen = 1;
            UCHAR ExtendedRatesIe = IE_EXT_SUPP_RATES, ErpIe = IE_ERP;
            UCHAR ExtendedRatesLen = pAd->PortCfg.SupportedRatesLen - SupportedRatesLen;
             
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                        MAC_HDR_LEN,                            &ProbeRspHdr, 
                        TIMESTAMP_LEN,                          &FakeTimestamp,
                        2,                                      &pAd->PortCfg.BeaconPeriod,
                        2,                                      &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                        1,                                      &SsidIe, 
                        1,                                      &pAd->PortCfg.MSSID[apidx].SsidLen, 
                        pAd->PortCfg.MSSID[apidx].SsidLen,      pAd->PortCfg.MSSID[apidx].Ssid,
                        1,                                      &SuppIe, 
                        1,                                      &SupportedRatesLen,
                        SupportedRatesLen,                      pAd->PortCfg.SupportedRates, 
                        1,                                      &DsIe, 
                        1,                                      &DsLen, 
                        1,                                      &pAd->PortCfg.Channel,
                        1,                                      &ErpIe,
                        1,                                      &ErpLen,
                        1,                                      &pAd->Mlme.ErpIeContent,
                        1,                                      &ExtendedRatesIe,
                        1,                                      &ExtendedRatesLen,
                        ExtendedRatesLen,                       &pAd->PortCfg.SupportedRates[SupportedRatesLen],
                        1,                                      &RSNIe,
                        1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],
                        pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],    pAd->PortCfg.MSSID[apidx].RSN_IE[0],
                        END_OF_ARGS);
        }
        else
        {
            MakeOutgoingFrame(OutBuffer,                        &FrameLen, 
                        MAC_HDR_LEN,                            &ProbeRspHdr, 
                        TIMESTAMP_LEN,                          &FakeTimestamp,
                        2,                                      &pAd->PortCfg.BeaconPeriod,
                        2,                                      &pAd->PortCfg.MSSID[apidx].CapabilityInfo,
                        1,                                      &SsidIe, 
                        1,                                      &pAd->PortCfg.MSSID[apidx].SsidLen, 
                        pAd->PortCfg.MSSID[apidx].SsidLen,      pAd->PortCfg.MSSID[apidx].Ssid,
                        1,                                      &SuppIe, 
                        1,                                      &pAd->PortCfg.SupportedRatesLen,
                        pAd->PortCfg.SupportedRatesLen,         pAd->PortCfg.SupportedRates, 
                        1,                                      &DsIe, 
                        1,                                      &DsLen, 
                        1,                                      &pAd->PortCfg.Channel,
                        1,                                      &RSNIe,
                        1,                                      &pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],
                        pAd->PortCfg.MSSID[apidx].RSNIE_Len[0],    pAd->PortCfg.MSSID[apidx].RSN_IE[0],
                        END_OF_ARGS);
        }
            
        MiniportMMRequest(pAd, OutBuffer, FrameLen);
    }
}

/* 
    ==========================================================================
    Description:
        parse the received BEACON

    NOTE:
        The only thing AP cares about received BEACON frames is to decide 
        if there's any overlapped legacy BSS condition (OLBC).
        If OLBC happened, this AP should set the ERP->Use_Protection bit in its 
        outgoing BEACON. The result is to tell all its clients to use RTC/CTS
        or CTS-to-self protection to protect B/G mixed traffic
    ==========================================================================
 */
VOID PeerBeaconAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
    MACADDR       Bssid, Addr2;
    CHAR          Ssid[MAX_LEN_OF_SSID];
    UCHAR         SsidLen, BssType, Channel, Rates[MAX_LEN_OF_SUPPORTED_RATES];
    UCHAR         RatesLen;
    USHORT        CapabilityInfo, BeaconPeriod;
    LONGLONG      TimeStamp;
    BOOLEAN       ExtendedRateIeExist;
    BOOLEAN       LegacyBssExist;
    UCHAR         Erp, index;

    if (PeerBeaconAndProbeRspSanity(pAd, 
                                Elem->Msg, 
                                Elem->MsgLen, 
                                &Addr2, 
                                &Bssid, 
                                Ssid, 
                                &SsidLen, 
                                &BssType, 
                                &BeaconPeriod, 
                                &Channel, 
                                &TimeStamp, 
                                &CapabilityInfo, 
                                Rates, 
                                &RatesLen,
                                &ExtendedRateIeExist,
                                &Erp))
    {
        if (Channel == pAd->PortCfg.AutoChannel_Channel)
        {
            // record the max RSSI of any received BEACON frames. APStartUp phase will
            // use this information to help select a less interference channel
            if (Elem->Rssi> pAd->PortCfg.AutoChannel_MaxRssi)
                pAd->PortCfg.AutoChannel_MaxRssi = Elem->Rssi;
        }

        // ignore BEACON not in this channel
        if (Channel != pAd->PortCfg.Channel)
            return;

        if ((Erp & 0x01) || (RatesLen <= 4))
            LegacyBssExist = TRUE;
        else
            LegacyBssExist = FALSE;

        if (LegacyBssExist)
        {
            pAd->Mlme.LastOLBCDetectTime = pAd->Mlme.Now32;
            DBGPRINT(RT_DEBUG_INFO, "%02x:%02x:%02x:%02x:%02x:%02x is a legacy BSS (rate# =%d, ERP=%d), set Use_Protection bit\n", 
                Bssid.Octet[0], Bssid.Octet[1], Bssid.Octet[2], Bssid.Octet[3], Bssid.Octet[4], Bssid.Octet[5], RatesLen, Erp);
        }

		// Lazy WDS mode, auto learning,
        if (pAd->WdsTab.Mode == WDS_LAZY_MODE)
        {
            // Lookup entry, if exist, then update the Rcv beacon time and return
            for (index = 0; index < MAX_WDS_ENTRY; index++)
            {
                if ((pAd->WdsTab.WdsEntry[index].Valid == TRUE) && (MAC_ADDR_EQUAL(&pAd->WdsTab.WdsEntry[index].WdsAddr, &Bssid)))
                {
                    // Update wds received beacon time for age out usage
                    pAd->WdsTab.WdsEntry[index].WdsRcvBeaconTime = jiffies;
                    return;
                }
            }

            // Look for empty entry, and add to our WdsTable
            for (index = 0; index < MAX_WDS_ENTRY; index++)
            {
                if (pAd->WdsTab.WdsEntry[index].Valid == FALSE)
                {
                	COPY_MAC_ADDR(&pAd->WdsTab.WdsEntry[index].WdsAddr, &Bssid);
                    pAd->WdsTab.WdsEntry[index].WdsRcvBeaconTime = jiffies;

                    pAd->WdsTab.WdsEntry[index].CurrTxRate = RATE_11;
                    if(pAd->PortCfg.PhyMode == PHY_11B)
                        pAd->WdsTab.WdsEntry[index].MaxSupportedRate = RATE_11;
                    else
                        pAd->WdsTab.WdsEntry[index].MaxSupportedRate = RATE_54;
                    pAd->WdsTab.WdsEntry[index].Valid = TRUE;
                    
	                DBGPRINT(RT_DEBUG_TRACE, "Lazy WDS mode, add AP(%02x:%02x:%02x:%02x:%02x:%02x) to entry(%d) in WDSTable \n", 
	                    Bssid.Octet[0], Bssid.Octet[1], Bssid.Octet[2], Bssid.Octet[3], Bssid.Octet[4], Bssid.Octet[5], index);
                    break;
                }
            }
            if (index == MAX_WDS_ENTRY)
            {
                DBGPRINT(RT_DEBUG_WARN, "WDS Table is full!\n");
            }
        }
    }
    // sanity check fail, ignore this frame
}

NDIS_STATUS MlmeAllocateMemory(
	IN PRTMP_ADAPTER pAd,
	OUT PVOID        *AllocVa)
{
	NDIS_STATUS                     Status = NDIS_STATUS_SUCCESS;

	*AllocVa = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
	if (*AllocVa == NULL)
	{
		Status = NDIS_STATUS_FAILURE;
	}

	return Status;
}

