/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    wpa.c

    Abstract:

    Revision History:
    Who         When            What
    --------    ----------      ----------------------------------------------
    Jan Lee     03-07-22        Initial
*/
#include "rt_config.h"

extern UCHAR		EAPOL[];
extern UCHAR		SNAP_802_1H[];

static UCHAR       OUI_WPA_WEP40[4]		= {0x00, 0x50, 0xF2, 0x01};
static UCHAR       OUI_WPA_TKIP[4]		= {0x00, 0x50, 0xF2, 0x02};
static UCHAR       OUI_WPA_CCMP[4]		= {0x00, 0x50, 0xF2, 0x04};
static UCHAR       OUI_WPA2_WEP40[4]		= {0x00, 0x0F, 0xAC, 0x01};
static UCHAR       OUI_WPA2_TKIP[4]		= {0x00, 0x0F, 0xAC, 0x02};
static UCHAR       OUI_WPA2_CCMP[4]		= {0x00, 0x0F, 0xAC, 0x04};

BOOLEAN WPAMsgTypeSubst(
    IN UCHAR    EAPType,
    OUT ULONG   *MsgType) 
{
    switch(EAPType)
    {
        case EAPPacket:
            *MsgType = MACHINE_TYPE_EAPPacket;
            break;

        case EAPOLStart:
            *MsgType = MACHINE_TYPE_EAPOLStart;
            break;

        case EAPOLLogoff:
            *MsgType = MACHINE_TYPE_EAPOLLogoff;
            break;

        case EAPOLKey:
            *MsgType = MACHINE_TYPE_EAPOLKey;
            break;

        case EAPOLASFAlert:
            *MsgType = MACHINE_TYPE_EAPOLASFAlert;
            break;

        default:
            DBGPRINT(RT_DEBUG_TRACE, "Unknown WPA MsgType : return FALSE;\n");    
            return FALSE;
    }
    
    return TRUE;
}

/*  
    ==========================================================================
    Description: 
        association state machine init, including state transition and timer init
    Parameters: 
        S - pointer to the association state machine
    ==========================================================================
 */
VOID WpaStateMachineInit(
    IN  PRTMP_ADAPTER       pAd,
    IN  STATE_MACHINE       *S,
    OUT STATE_MACHINE_FUNC  Trans[])
{
    StateMachineInit(S, (STATE_MACHINE_FUNC*)Trans, MAX_WPA_PTK_STATE, MAX_WPA_MSG, (STATE_MACHINE_FUNC)Drop, WPA_PTK, WPA_MACHINE_BASE);

    StateMachineSetAction(S, WPA_PTK, MACHINE_TYPE_EAPPacket, (STATE_MACHINE_FUNC)WpaEAPPacketAction);
    StateMachineSetAction(S, WPA_PTK, MACHINE_TYPE_EAPOLStart, (STATE_MACHINE_FUNC)WpaEAPOLStartAction);
    StateMachineSetAction(S, WPA_PTK, MACHINE_TYPE_EAPOLLogoff, (STATE_MACHINE_FUNC)WpaEAPOLLogoffAction);
    StateMachineSetAction(S, WPA_PTK, MACHINE_TYPE_EAPOLKey, (STATE_MACHINE_FUNC)WpaEAPOLKeyAction);
    StateMachineSetAction(S, WPA_PTK, MACHINE_TYPE_EAPOLASFAlert, (STATE_MACHINE_FUNC)WpaEAPOLASFAlertAction);
}

/*
    ==========================================================================
    Description:
        Port Access Control Inquiry function. Return entry's Privacy and Wpastate.
        Also check if AP needs to initilize 4-way handshake. 
    Return:
        pEntry 
    ==========================================================================
*/
MAC_TABLE_ENTRY *PACInquiry(
    IN PRTMP_ADAPTER                pAd,
    IN MACADDR                      *pAddr,
    OUT NDIS_802_11_PRIVACY_FILTER  *Privacy,
    OUT WPA_STATE                   *WpaState)
{
    MAC_TABLE_ENTRY *pEntry = (MAC_TABLE_ENTRY*)NULL;
    
    *Privacy = Ndis802_11PrivFilterAcceptAll;
    *WpaState = AS_NOTUSE;

    if (MAC_ADDR_IS_GROUP(*pAddr))
    {// mcast & broadcast address
    } 
    else
    {// unicast address
        pEntry = MacTableLookup(pAd, pAddr);
        if (pEntry) 
        {
            *Privacy = pEntry->PrivacyFilter;
            *WpaState = pEntry->WpaState;
        } 
    }
    return pEntry;
}

/*
    ==========================================================================
    Description:
       
    Return:
         TRUE if this is EAP frame
         FALSE otherwise
    ==========================================================================
*/
BOOLEAN RTMPCheckWPAframe(
    IN PRTMP_ADAPTER    pAdapter,
    IN PUCHAR           pHeader,
    IN ULONG            DataByteCount)
{
    PUCHAR  pData;
    
    DBGPRINT(RT_DEBUG_INFO, "RTMPCheckWPAframe ===>\n");
    
    if (DataByteCount < (LENGTH_802_11 + LENGTH_802_1_H + LENGTH_EAPOL_H))
        return FALSE;

    pData = pHeader + LENGTH_802_11;
    DataByteCount -= LENGTH_802_11;
    if (RTMPEqualMemory(SNAP_802_1H, pData, 6)) 
    {
        pData += 6;
    }
    if (RTMPEqualMemory(EAPOL, pData, 2)) 
    {
        pData += 2;
    }
    else
        return FALSE;

    switch (*(pData+1))
    {   
        case EAPPacket:
            DBGPRINT(RT_DEBUG_INFO, "Receive EAP-Packet frame, TYPE = 0 \n");
            break;

        case EAPOLStart:
            DBGPRINT(RT_DEBUG_INFO, "Receive EAPOL-Start frame, TYPE = 1 \n");
            break;

        case EAPOLLogoff:
            DBGPRINT(RT_DEBUG_INFO, "Receive EAPOLLogoff frame, TYPE = 2 \n");
            break;

        case EAPOLKey:
            DBGPRINT(RT_DEBUG_INFO, "Receive EAPOL-Key frame, TYPE = 3, Length =%x\n", *(pData+2));
            break;

        case EAPOLASFAlert:
            DBGPRINT(RT_DEBUG_INFO, "Receive EAPOLASFAlert frame, TYPE = 4 \n");
            break;

        default:
            return FALSE;
    }   
    return TRUE;
}

/*
    ==========================================================================
    Description:
       Check invalidity of multicast cipher selection in RSN IE.
    Return:
         TRUE if match
         FALSE otherwise
    ==========================================================================
*/
BOOLEAN RTMPCheckMcast(
    IN PRTMP_ADAPTER    pAdapter,
    IN PEID_STRUCT      eid_ptr,
    IN MAC_TABLE_ENTRY	*pEntry)
{
	pEntry->AuthMode = pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode;

    if (eid_ptr->Len >= 6)
    {
    	// WPA and WPA2 format not the same in RSN_IE
    	if (eid_ptr->Eid == IE_WPA)
    	{
    		if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2)
    			pEntry->AuthMode = Ndis802_11AuthModeWPA;
    		else if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK)
    			pEntry->AuthMode = Ndis802_11AuthModeWPAPSK;

	        if (RTMPEqualMemory(&eid_ptr->Octet[6], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][6], 4))
	            return TRUE;
    	}
    	else if (eid_ptr->Eid == IE_WPA2)
    	{
    		UCHAR	IE_Idx = 0;

			// When WPA1/WPA2 mix mode, the RSN_IE is stored in different structure
			if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2) || (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
				IE_Idx = 1;
	
    		if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2)
    			pEntry->AuthMode = Ndis802_11AuthModeWPA2;
    		else if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK)
    			pEntry->AuthMode = Ndis802_11AuthModeWPA2PSK;

    		if (RTMPEqualMemory(&eid_ptr->Octet[2], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][2], 4))
	            return TRUE;
    	}
    }

    DBGPRINT(RT_DEBUG_ERROR, "RTMPCheckMcast ==> WPAIE=%d\n", eid_ptr->Eid);

    return FALSE;
}

/*
    ==========================================================================
    Description:
       Check invalidity of unicast cipher selection in RSN IE.
    Return:
         TRUE if match
         FALSE otherwise
    ==========================================================================
*/
BOOLEAN RTMPCheckUcast(
    IN PRTMP_ADAPTER    pAdapter,
    IN PEID_STRUCT      eid_ptr,
    IN MAC_TABLE_ENTRY	*pEntry)
{
	pEntry->WepStatus = pAdapter->PortCfg.MSSID[pEntry->ApIdx].WepStatus;
    
    if (eid_ptr->Len >= 16)
    {
    	if (eid_ptr->Eid == IE_WPA)
    	{
    		if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].WepStatus == Ndis802_11Encryption4Enabled)
			{// multiple cipher (TKIP/CCMP)
				if (RTMPEqualMemory(&eid_ptr->Octet[12], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][12], 4))
				{
					pEntry->WepStatus = Ndis802_11Encryption2Enabled;
					return TRUE;
				}
				else if (RTMPEqualMemory(&eid_ptr->Octet[12], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][16], 4))
				{
					pEntry->WepStatus = Ndis802_11Encryption3Enabled;
					return TRUE;
				}
    		}
    		else
    		{// single cipher
    			if (RTMPEqualMemory(&eid_ptr->Octet[12], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][12], 4))
		            return TRUE;
    		}
    	}
    	else if (eid_ptr->Eid == IE_WPA2)
    	{
    		UCHAR	IE_Idx = 0;

			// When WPA1/WPA2 mix mode, the RSN_IE is stored in different structure
			if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2) || (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
				IE_Idx = 1;
	
			if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].WepStatus == Ndis802_11Encryption4Enabled)
			{// multiple cipher (TKIP/CCMP)
				if (RTMPEqualMemory(&eid_ptr->Octet[8], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][8], 4))
				{
					pEntry->WepStatus = Ndis802_11Encryption2Enabled;
					return TRUE;
				}
				else if (RTMPEqualMemory(&eid_ptr->Octet[8], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][12], 4))
				{
					pEntry->WepStatus = Ndis802_11Encryption3Enabled;
					return TRUE;
				}
			}
			else
			{// single cipher
				if (RTMPEqualMemory(&eid_ptr->Octet[8], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][8], 4))
					return TRUE;
			}
    	}
    }

    DBGPRINT(RT_DEBUG_ERROR, "RTMPCheckUcast ==> WPAIE=%d\n", eid_ptr->Eid);

    return FALSE;
}

/*
    ==========================================================================
    Description:
       Check invalidity of authentication method selection in RSN IE.
    Return:
         TRUE if match
         FALSE otherwise
    ==========================================================================
*/
BOOLEAN RTMPCheckAUTH(
    IN PRTMP_ADAPTER    pAdapter,
    IN PEID_STRUCT      eid_ptr,
    IN MAC_TABLE_ENTRY	*pEntry)
{
    if (eid_ptr->Len >= 16)
    {
    	if (eid_ptr->Eid == IE_WPA)
    	{
    		if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].WepStatus == Ndis802_11Encryption4Enabled)
    		{
    			if (RTMPEqualMemory(&eid_ptr->Octet[18], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][18], 4))
		            return TRUE;
    			else if (RTMPEqualMemory(&eid_ptr->Octet[18], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][22], 4))
		            return TRUE;
    		}
    		else
			{
				if (RTMPEqualMemory(&eid_ptr->Octet[18], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0][18], 4))
		            return TRUE;
			}
    	}
    	else if (eid_ptr->Eid == IE_WPA2)
    	{
    		UCHAR	IE_Idx = 0;

			// When WPA1/WPA2 mix mode, the RSN_IE is stored in different structure
			if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2) || (pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
				IE_Idx = 1;
	
			if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].WepStatus == Ndis802_11Encryption4Enabled)
			{
				if (RTMPEqualMemory(&eid_ptr->Octet[14], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][14], 4))
					return TRUE;
				else if (RTMPEqualMemory(&eid_ptr->Octet[14], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][18], 4))
					return TRUE;
			}
			else
			{
				if (RTMPEqualMemory(&eid_ptr->Octet[14], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[IE_Idx][14], 4))
					return TRUE;
			}
    	}
    }

    DBGPRINT(RT_DEBUG_ERROR, "RTMPCheckAUTH ==> WPAIE=%d\n", eid_ptr->Eid);

    return FALSE;
}

/*
    ==========================================================================
    Description:
        this is state machine function. 
        When receiving EAP packets which is  for 802.1x authentication use. 
        Not use in PSK case
    Return:
    ==========================================================================
*/
VOID WpaEAPPacketAction(
    IN PRTMP_ADAPTER    pAd, 
    IN MLME_QUEUE_ELEM  *Elem) 
{
}

VOID WpaEAPOLLogoffAction(
    IN PRTMP_ADAPTER	pAd, 
    IN MLME_QUEUE_ELEM	*Elem) 
{
}

VOID WpaEAPOLASFAlertAction(
    IN PRTMP_ADAPTER    pAd, 
    IN MLME_QUEUE_ELEM  *Elem) 
{   
}

VOID WpaEAPOLStartAction(
    IN PRTMP_ADAPTER    pAd, 
    IN MLME_QUEUE_ELEM  *Elem) 
{
    MAC_TABLE_ENTRY     *pEntry;
    PHEADER_802_11      pHeader;

    DBGPRINT(RT_DEBUG_TRACE, "WpaEAPOLStartAction ====>> \n");
    
    // Delay for avoiding the misorder of EAPOL-KEY and Assoc Rsp.
    mdelay(500);
    
    pHeader = (PHEADER_802_11)Elem->Msg;
    // TODO: For normaol PSK, we enqueue an EAPOL-Start command to trigger the process.
    if (Elem->MsgLen == 6)
        pEntry = MacTableLookup(pAd, (PMACADDR) Elem->Msg);
    else
	    pEntry = MacTableLookup(pAd, &pHeader->Controlhead.Addr2);
    
    if (pEntry) 
    {
        // TODO: If WPA-PSK mode and port not secured, reset to the init PSK state
        if ((pEntry->PortSecured == WPA_802_1X_PORT_NOT_SECURED)
        	&& ((pEntry->AuthMode == Ndis802_11AuthModeWPAPSK) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK) || ((pEntry->AuthMode == Ndis802_11AuthModeWPA2) && (pEntry->PMKID_CacheIdx != -1))))
        {
            pEntry->PrivacyFilter = Ndis802_11PrivFilter8021xWEP;
            pEntry->WpaState = AS_INITPSK;
            pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
            NdisZeroMemory(pEntry->R_Counter, sizeof(pEntry->R_Counter));
            
            // add timer for retry
            pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
            pEntry->ReTryCounter = PEER_MSG1_RETRY_TIMER_CTR;
            WPAStart4WayHS(pAd, pEntry);
        }
    } 
}

/*
    ==========================================================================
    Description:
        Function to handel countermeasures active attack.  Init 60-sec timer if necessary.
    Return:
    ==========================================================================
*/
VOID HandleCounterMeasure(
    IN PRTMP_ADAPTER    pAd, 
    IN MAC_TABLE_ENTRY  *pEntry) 
{
    INT         i;
    UCHAR       idx;
    
    do{
		if (!pEntry)
			break;

		// if entry not set key done, ignore this RX MIC ERROR
	    if ((pEntry->WpaState < AS_PTKINITDONE) || (pEntry->GTKState != REKEY_ESTABLISHED))
	        break;
			
		pAd->PortCfg.MICFailureCounter++;
		// record which entry causes this MIC error, if this entry sends disauth/disassoc, AP doesn't need to log the CM
		pEntry->CMTimerRunning = TRUE;
		if (pAd->PortCfg.CMTimerRunning == TRUE)
		{
			DBGPRINT(RT_DEBUG_TRACE, "Receive CM Attack Twice Within 60 seconds ====>>> \n");            

			ApLogEvent(pAd, &pEntry->Addr, EVENT_COUNTER_M, pEntry->ApIdx);
			del_timer_sync(&pAd->PortCfg.CounterMeasureTimer);
			pAd->PortCfg.CMTimerRunning = FALSE;
			// happened twice within 60 sec,  AP SENDS disaccociate all associated STAs.  All STA's transition to State 2
			for (idx = 0; idx < pAd->PortCfg.SsidNum; idx++)
			{
				// renew GTK
				GenRandom(pAd, pAd->PortCfg.MSSID[idx].GNonce);
				for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
				{
					if (pAd->PortCfg.MSSID[idx].MacTab.Content[i].Valid == TRUE)
					{
						DisAssocAction(pAd, &pAd->PortCfg.MSSID[idx].MacTab.Content[i], SUBTYPE_DISASSOC, REASON_MIC_FAILURE);
					}
				}
			}

			// Further,  ban all Class 3 DATA transportation for  a period 0f 60 sec
			// disallow new association , too
			pAd->PortCfg.MSSID[pEntry->ApIdx].BANClass3Data = TRUE;        
		}
	}while(FALSE);

    pAd->PortCfg.CounterMeasureTimer.expires = jiffies + 60*MLME_TASK_EXEC_INTV;
    add_timer(&pAd->PortCfg.CounterMeasureTimer);
    pAd->PortCfg.CMTimerRunning = TRUE;
    pAd->PortCfg.PrevaMICFailTime = pAd->PortCfg.aMICFailTime;
    pAd->PortCfg.aMICFailTime = jiffies;
}

/*
    ==========================================================================
    Description:
        This is state machine function. 
        When receiving EAPOL packets which is  for 802.1x key management. 
        Use both in WPA, and WPAPSK case. 
        In this function, further dispatch to different functions according to the received packet.  3 categories are : 
          1.  normal 4-way pairwisekey and 2-way groupkey handshake
          2.  MIC error (Countermeasures attack)  report packet from STA.
          3.  Request for pairwise/group key update from STA
    Return:
    ==========================================================================
*/
VOID WpaEAPOLKeyAction(
    IN PRTMP_ADAPTER    pAdapter, 
    IN MLME_QUEUE_ELEM  *Elem) 
{
    INT                 i;
    MAC_TABLE_ENTRY     *pEntry;
	KEY_DESCRIPTER		*pKeyDesc;
    PHEADER_802_11      pHeader;
    PEAPOL_PACKET       pMsg;

    DBGPRINT(RT_DEBUG_TRACE, "WpaEAPOLKeyAction =====>>> \n");

    pHeader = (PHEADER_802_11)Elem->Msg;
    pMsg = (PEAPOL_PACKET) &Elem->Msg[LENGTH_802_11 + LENGTH_802_1_H];
    pKeyDesc = (PKEY_DESCRIPTER)&Elem->Msg[(LENGTH_802_11+LENGTH_802_1_H+LENGTH_EAPOL_H)];

#ifdef BIG_ENDIAN
    // pMsg->KeyDesc.Keyinfo and pKeyDesc->Keyinfo both point to the same addr.
    // Thus, it only needs swap once.
    *(USHORT *)&pMsg->KeyDesc.Keyinfo = SWAP16(*(USHORT *)&pMsg->KeyDesc.Keyinfo);
#endif

    pEntry = MacTableLookup(pAdapter, &pHeader->Controlhead.Addr2);
	if (!pEntry)
		return;

    if ((pEntry->AuthMode != Ndis802_11AuthModeWPA) && (pEntry->AuthMode != Ndis802_11AuthModeWPAPSK)
    	&& (pEntry->AuthMode != Ndis802_11AuthModeWPA2) && (pEntry->AuthMode != Ndis802_11AuthModeWPA2PSK))
        return;

    if ((pMsg->ProVer!= EAPOL_VER) || ((pMsg->KeyDesc.Type != WPA1_KEY_DESC) && (pMsg->KeyDesc.Type != WPA2_KEY_DESC)))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Key descripter does not match with WPA rule \n");
        return;
    }

    if ((pEntry->WepStatus == Ndis802_11Encryption3Enabled) && (pMsg->KeyDesc.Keyinfo.KeyDescVer != DESC_TYPE_AES))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Key descripter version not match AES \n");
        return;
    }
    else if ((pEntry->WepStatus == Ndis802_11Encryption2Enabled) && (pMsg->KeyDesc.Keyinfo.KeyDescVer != DESC_TYPE_TKIP))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Key descripter version not match TKIP \n");
        return;
    }

    if ((pEntry->Sst == SST_ASSOC) && (pEntry->WpaState >= AS_INITPSK))
    {
        if ((pKeyDesc->Keyinfo.KeyMic) && (pKeyDesc->Keyinfo.Error) && (pKeyDesc->Keyinfo.Request))
        {
            // Category 2. 
            DBGPRINT(RT_DEBUG_TRACE, "MIC, ERROR, REQUEST ALL 1'S, ACTIVE COUNTERMEASURE \n");
            HandleCounterMeasure(pAdapter, pEntry);
        }
        else if ((pKeyDesc->Keyinfo.Secure) && !(pKeyDesc->Keyinfo.Error) && !(pKeyDesc->Keyinfo.Request))
        {
#ifdef BIG_ENDIAN
            // recovery original byte order, before forward Elem to another routine
            *(USHORT *)&pMsg->KeyDesc.Keyinfo = SWAP16(*(USHORT *)&pMsg->KeyDesc.Keyinfo);
#endif
            // SECURE=1, REQUEST=0, ERROR=0
            if ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPAPSK))
            PeerGroupMsg2Action(pAdapter, pEntry, &Elem->Msg[LENGTH_802_11], (Elem->MsgLen-LENGTH_802_11));
            else if ((pEntry->AuthMode == Ndis802_11AuthModeWPA2) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK))
            	Wpa2PeerPairMsg4Action(pAdapter, pEntry, Elem);
        }
        else if (!(pKeyDesc->Keyinfo.Secure) && !(pKeyDesc->Keyinfo.Error) && !(pKeyDesc->Keyinfo.Request))
        {
#ifdef BIG_ENDIAN
            // recovery original byte order, before forward Elem to another routine
            *(USHORT *)&pMsg->KeyDesc.Keyinfo = SWAP16(*(USHORT *)&pMsg->KeyDesc.Keyinfo);
#endif
            // general 4-way handshake step // SECURE=0, REQUEST=0 ,  ERROR= 0
            // Category 1
            if (pEntry->WpaState == AS_PTKSTART)
                PeerPairMsg2Action(pAdapter, pEntry, Elem);
            else if (pEntry->WpaState == AS_PTKINIT_NEGOTIATING)
                Wpa1PeerPairMsg4Action(pAdapter, pEntry, Elem);
        }
        else if (!(pKeyDesc->Keyinfo.Error) && (pKeyDesc->Keyinfo.Request))
        {
            // Category 3
            if (pKeyDesc->Keyinfo.KeyType == GROUPKEY)
            {
                DBGPRINT(RT_DEBUG_TRACE, "REQUEST=1, ERROR= 0, update group key\n");
                GenRandom(pAdapter, pAdapter->PortCfg.MSSID[pEntry->ApIdx].GNonce);
                for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
                {
                    if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].MacTab.Content[i].Valid == TRUE) && (pAdapter->PortCfg.MSSID[pEntry->ApIdx].MacTab.Content[i].WpaState == AS_PTKINITDONE))
                    {
                        pAdapter->PortCfg.MSSID[pEntry->ApIdx].MacTab.Content[i].GTKState = REKRY_NEGOTIATING;
                        WPAHardTransmit(pAdapter, &pAdapter->PortCfg.MSSID[pEntry->ApIdx].MacTab.Content[i]);
                    }
                }
            }
            else
            {
                DBGPRINT(RT_DEBUG_TRACE, "REQUEST=1, ERROR= 0, update pairwise key\n");
                pEntry->PairwiseKey[0].KeyLen = 0;
                pEntry->Sst = SST_ASSOC;
                if ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2))
                    pEntry->WpaState = AS_INITPMK;  
                else if ((pEntry->AuthMode == Ndis802_11AuthModeWPAPSK) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK))
                    pEntry->WpaState = AS_INITPSK;  
                pEntry->GTKState = REKRY_NEGOTIATING;
                
				pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
				pEntry->ReTryCounter = PEER_MSG1_RETRY_TIMER_CTR;
				WPAStart4WayHS(pAdapter, pEntry);
            }
        }
        else
        {
                //
        }
    }
}

/*
    ==========================================================================
    Description:
        This is a function to initilize 4way handshake
    Return:
         
    ==========================================================================
*/
VOID WPAStart4WayHS(
    IN PRTMP_ADAPTER    pAdapter,
    IN MAC_TABLE_ENTRY  *pEntry)
{
    ULONG           FrameLen = 0;
    UCHAR           *OutBuffer = NULL;
    UCHAR           Header802_3[14];
    EAPOL_PACKET    Packet;

    DBGPRINT(RT_DEBUG_TRACE, "==>WPAStart4WayHS\n");

    do
    {
    	if (!pEntry->Valid)
    		break;
    	if ((pEntry->WpaState > AS_PTKSTART) || (pEntry->WpaState < AS_INITPMK))
	    {
	        DBGPRINT(RT_DEBUG_ERROR, "Not expect calling WPAStart4WayHS here ");
	        break;
	    }
    	
        OutBuffer = kmalloc(MAX_LEN_OF_EAP_HS, GFP_KERNEL);
        if(OutBuffer == NULL)
            break;

        WPAMake8023Hdr(pAdapter, pEntry->Addr.Octet, Header802_3);

        // Increment replay counter by 1  
        ADD_ONE_To_64BIT_VAR(pAdapter->PortCfg.R_Counter);
        
        // 0. init Packet and Fill header
        NdisZeroMemory(&Packet, sizeof(Packet));
        Packet.ProVer = EAPOL_VER;
        Packet.ProType = EAPOLKey;
        Packet.Body_Len[1] = 0x5f;
        
        // 1. Fill replay counter
        NdisMoveMemory(pEntry->R_Counter, pAdapter->PortCfg.R_Counter, sizeof(pEntry->R_Counter));
        NdisMoveMemory(Packet.KeyDesc.RCounter, pEntry->R_Counter, sizeof(pEntry->R_Counter));
        
        // 2. Fill key version, keyinfo, key len
        Packet.KeyDesc.Keyinfo.KeyType = PAIRWISEKEY;
        Packet.KeyDesc.Keyinfo.KeyAck = 1;
        if ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPAPSK))
        {
        	Packet.KeyDesc.Type = WPA1_KEY_DESC;
        	Packet.KeyDesc.DataLen[1] = 0;

        	DBGPRINT(RT_DEBUG_TRACE, "pEntry->AuthMode == Ndis802_11AuthModeWPA/WPAPSK\n");
        }
        else if ((pEntry->AuthMode == Ndis802_11AuthModeWPA2) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK))
        {
        	Packet.KeyDesc.Type = WPA2_KEY_DESC;
        	Packet.KeyDesc.DataLen[1] = 0;

			// PMKID match, then start 4-way HS directly
        	if (pEntry->PMKID_CacheIdx != -1)
        	{
        		Packet.KeyDesc.Data[0] = 0xDD;
	        	Packet.KeyDesc.Data[2] = 0x00;
	        	Packet.KeyDesc.Data[3] = 0x0F;
	        	Packet.KeyDesc.Data[4] = 0xAC;
	        	Packet.KeyDesc.Data[5] = 0x04;

        		NdisMoveMemory(&Packet.KeyDesc.Data[6], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMKIDCache.BSSIDInfo[pEntry->PMKID_CacheIdx].PMKID, LEN_PMKID);
        		NdisMoveMemory(&pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK, &pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMKIDCache.BSSIDInfo[pEntry->PMKID_CacheIdx].PMK, PMK_LEN);
        		DBGPRINT(RT_DEBUG_TRACE, "PMK = %x %x %x %x \n",pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK[0],pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK[1],pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK[2],pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK[3]);
        		
        		Packet.KeyDesc.Data[1] = 0x14;// 4+2+LEN_PMKID
        		Packet.KeyDesc.DataLen[1] += 6 + LEN_PMKID;
        		Packet.Body_Len[1] += 6 + LEN_PMKID;
        	}
        	
        	DBGPRINT(RT_DEBUG_TRACE, "pEntry->AuthMode == Ndis802_11AuthModeWPA2/WPA2PSK\n");
        }

        if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
        {
            Packet.KeyDesc.KeyLength[1] = LEN_TKIP_KEY;
            Packet.KeyDesc.Keyinfo.KeyDescVer = DESC_TYPE_TKIP;
        }
        else if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
        {
            Packet.KeyDesc.KeyLength[1] = LEN_AES_KEY;
            Packet.KeyDesc.Keyinfo.KeyDescVer = DESC_TYPE_AES;
        }

#ifdef BIG_ENDIAN
        *(USHORT *)&Packet.KeyDesc.Keyinfo = SWAP16(*(USHORT *)&Packet.KeyDesc.Keyinfo);
#endif
        
        // 3. Fill Anonce
        GenRandom(pAdapter, pEntry->ANonce);
        NdisMoveMemory(Packet.KeyDesc.Nonce, pEntry->ANonce, sizeof(pEntry->ANonce));
        MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            sizeof(Header802_3),    Header802_3,
                            Packet.Body_Len[1] + 4,  &Packet,
                            END_OF_ARGS);

        RTMPToWirelessSta(pAdapter, OutBuffer, FrameLen);

        mod_timer(&pEntry->RetryTimer, pEntry->RetryTimer.expires);
		// Should start from 1, init is 0
		pEntry->ReTryCounter++;
		pEntry->RetryTimerRunning = TRUE;

        pEntry->WpaState = AS_PTKSTART;

        kfree(OutBuffer);
    }while(FALSE);
    
    DBGPRINT(RT_DEBUG_TRACE, "<== WPAStart4WayHS : pEntry->WpaState= %d , FrameLen= %d\n",pEntry->WpaState,FrameLen);
}

/*
    ==========================================================================
    Description:
        When receiving the second packet of 4-way pairwisekey handshake.
    Return:
    ==========================================================================
*/
VOID PeerPairMsg2Action(
    IN PRTMP_ADAPTER    pAdapter, 
    IN MAC_TABLE_ENTRY  *pEntry,
    IN MLME_QUEUE_ELEM  *Elem) 
{
    UCHAR                   PTK[80], mic[LEN_KEY_DESC_MIC], SNonce[LEN_KEY_DESC_NONCE], digest[80];
    ULONG                   FrameLen = 0;
    UCHAR                   *OutBuffer = NULL;
    USHORT                  MICMsgLen, MICOffset;
    PUCHAR                  pSA;
    UCHAR                   Header802_3[LENGTH_802_3];
    EAPOL_PACKET            Packet;
    PEAPOL_PACKET           pMsg2;

    DBGPRINT(RT_DEBUG_TRACE, "PeerPairMsg2Action ===>\n");

    // check Entry in valid State
    if (pEntry->WpaState < AS_PTKSTART)
        return;

    do
    {
        OutBuffer = kmalloc(MAX_LEN_OF_EAP_HS, GFP_KERNEL);
        if(OutBuffer == NULL)
            return;
            
        // Save Data Length to pDesc for receiving packet,   then put in outgoing frame Data Len fields.
        pMsg2 = (PEAPOL_PACKET) &Elem->Msg[LENGTH_802_11 + LENGTH_802_1_H];
            
        // 1. check Replay Counter coresponds to MSG1.,  otherwise discard
        if (!RTMPEqualMemory(pMsg2->KeyDesc.RCounter, pEntry->R_Counter, LEN_KEY_DESC_REPLAY))
        {
            DBGPRINT(RT_DEBUG_ERROR, "Replay Counter Different in Msg 2 of 4-way handshake !! \n");
            DBGPRINT(RT_DEBUG_ERROR, "Receive : %d  %d  %d  %d  %d  %d  %d  %d \n", pMsg2->KeyDesc.RCounter[0],pMsg2->KeyDesc.RCounter[1],\
                pMsg2->KeyDesc.RCounter[2],pMsg2->KeyDesc.RCounter[3],pMsg2->KeyDesc.RCounter[4],pMsg2->KeyDesc.RCounter[5],pMsg2->KeyDesc.RCounter[6],pMsg2->KeyDesc.RCounter[7]);
            DBGPRINT(RT_DEBUG_ERROR, "Current : %d  %d  %d  %d  %d  %d  %d  %d \n",pEntry->R_Counter[0],pEntry->R_Counter[1],pEntry->R_Counter[2],pEntry->R_Counter[3],\
                pEntry->R_Counter[4],pEntry->R_Counter[5],pEntry->R_Counter[6],pEntry->R_Counter[7]);
            break;
        }

        // 2. Derive necessary info for counting PTK and thus count PTK
        MICMsgLen = pMsg2->Body_Len[1] | ((pMsg2->Body_Len[0]<<8) && 0xff00);
        MICMsgLen += LENGTH_EAPOL_H;
        if (MICMsgLen > (Elem->MsgLen -LENGTH_802_11 - LENGTH_802_1_H))
        {
            DBGPRINT(RT_DEBUG_ERROR, "Receive wrong format EAPOL packets \n");
            break;        
        }
        NdisMoveMemory(SNonce, pMsg2->KeyDesc.Nonce, LEN_KEY_DESC_NONCE);
        pSA = pEntry->Addr.Octet;

        CountPTK(pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK,  pEntry->ANonce, pAdapter->CurrentAddress, SNonce, pSA, PTK, LEN_PTK); 
        NdisMoveMemory(pEntry->PTK, PTK, LEN_PTK);
            
	    DBGPRINT(RT_DEBUG_INFO, "ANonce1-%x %x %x %x %x %x %x %x \n",pEntry->ANonce[0],pEntry->ANonce[1],pEntry->ANonce[2],pEntry->ANonce[3],pEntry->ANonce[4],pEntry->ANonce[5],pEntry->ANonce[6],pEntry->ANonce[7]);
	    DBGPRINT(RT_DEBUG_INFO, "ANonce2-%x %x %x %x %x %x %x %x \n",pEntry->ANonce[8],pEntry->ANonce[9],pEntry->ANonce[10],pEntry->ANonce[11],pEntry->ANonce[12],pEntry->ANonce[13],pEntry->ANonce[14],pEntry->ANonce[15]);

        // 3. verify MSG2 MIC, If not valid, discard.
        MICOffset = sizeof(EAPOL_PACKET)-MAX_LEN_OF_RSNIE-2-LEN_KEY_DESC_MIC;
        NdisMoveMemory(OutBuffer, pMsg2, MICMsgLen);  
        NdisZeroMemory((OutBuffer+MICOffset), LEN_KEY_DESC_MIC);
        if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
        {
            hmac_md5(PTK, LEN_EAP_MICK, OutBuffer, MICMsgLen, mic);
        }
        else
        {
            HMAC_SHA1(OutBuffer,  MICMsgLen, PTK, LEN_EAP_MICK, digest);
            NdisMoveMemory(mic, digest, LEN_KEY_DESC_MIC);
        }
        if (!RTMPEqualMemory((PUCHAR)pMsg2+MICOffset, mic, LEN_KEY_DESC_MIC))
        {
            DBGPRINT(RT_DEBUG_ERROR, "MIC Different in Msg 2 of 4-way handshake!! \n");
            break;
        }
        else
            DBGPRINT(RT_DEBUG_TRACE, "MIC VALID in Msg 2 of 4-way handshake!! \n");
            
        // 4. check RSN IE, if not match, send MLME-DEAUTHENTICATE.
        if (!RTMPEqualMemory(&pMsg2->KeyDesc.Data[2], pEntry->RSN_IE, pEntry->RSNIE_Len))
        {       
            DBGPRINT(RT_DEBUG_ERROR, "RSN_IE Different in Msg 2 of 4-way handshake!! \n");
            DBGPRINT(RT_DEBUG_ERROR, "Receive : %x %x %x %x ...\n",pMsg2->KeyDesc.Data[2], pMsg2->KeyDesc.Data[3],pMsg2->KeyDesc.Data[4],pMsg2->KeyDesc.Data[5] );
            DBGPRINT(RT_DEBUG_ERROR, "Current : %x %x %x %x ...\n",pEntry->RSN_IE[0], pEntry->RSN_IE[1],pEntry->RSN_IE[2],pEntry->RSN_IE[3] );
            break;  
        }
        else
            DBGPRINT(RT_DEBUG_TRACE, "RSN_IE VALID in Msg 2 of 4-way handshake!! \n");
            
        pEntry->WpaState = AS_PTKINIT_NEGOTIATING;

        // Increment replay counter by 1       
        ADD_ONE_To_64BIT_VAR(pAdapter->PortCfg.R_Counter);
        NdisMoveMemory(pEntry->R_Counter, pAdapter->PortCfg.R_Counter, sizeof(pEntry->R_Counter));
            
        // 5. CONSTRUCT msg3
        // 5-0. Init Packet and Fill header
        NdisZeroMemory(&Packet, sizeof(Packet));
        Packet.ProVer = EAPOL_VER;
        Packet.ProType = EAPOLKey;
            
        // Fill replay counter
        NdisMoveMemory(Packet.KeyDesc.RCounter, pEntry->R_Counter, sizeof(pEntry->R_Counter));
          
        // Fill key version, keyinfo, key len
        Packet.KeyDesc.Keyinfo.KeyMic = 1;
        Packet.KeyDesc.Keyinfo.KeyType = PAIRWISEKEY;
        Packet.KeyDesc.Keyinfo.Install = 1;
        Packet.KeyDesc.Keyinfo.KeyAck = 1;
        if ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPAPSK))
        {
        	Packet.KeyDesc.Type = WPA1_KEY_DESC;
        	Packet.KeyDesc.DataLen[1] = 0;
        }
        else if ((pEntry->AuthMode == Ndis802_11AuthModeWPA2) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK))
        {
        	Packet.KeyDesc.Type = WPA2_KEY_DESC;
        	Packet.KeyDesc.DataLen[1] = 0;
        	Packet.KeyDesc.Keyinfo.EKD = 1;
        	Packet.KeyDesc.Keyinfo.Secure = 1;
        }

        if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
        {
            Packet.KeyDesc.KeyLength[1] = LEN_TKIP_KEY;
        }
        else if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
        {
            Packet.KeyDesc.KeyLength[1] = LEN_AES_KEY;
        }
        Packet.KeyDesc.Keyinfo.KeyDescVer =
        	(((pEntry->WepStatus == Ndis802_11Encryption3Enabled) || (pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKeyWepStatus == Ndis802_11Encryption3Enabled)) ? (DESC_TYPE_AES) : (DESC_TYPE_TKIP));

#ifdef BIG_ENDIAN
        *(USHORT *)&Packet.KeyDesc.Keyinfo = SWAP16(*(USHORT *)&Packet.KeyDesc.Keyinfo);
#endif

        // 5-3, Fill in Data, Data Len
        NdisMoveMemory(Packet.KeyDesc.Nonce, pEntry->ANonce, sizeof(pEntry->ANonce));
		NdisZeroMemory(Packet.KeyDesc.MIC, sizeof(Packet.KeyDesc.MIC));
        
        if ((pEntry->AuthMode == Ndis802_11AuthModeWPA) || (pEntry->AuthMode == Ndis802_11AuthModeWPAPSK))
        {
        	Packet.KeyDesc.Data[0] = IE_WPA;
        	Packet.KeyDesc.Data[1] = pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[0];
        	Packet.KeyDesc.DataLen[1] = pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[0] + 2;
        	Packet.Body_Len[1] = 93 + 2 + 2 + pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[0];
        	
	        NdisMoveMemory(&Packet.KeyDesc.Data[2], pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0], pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[0]);

        // 5-4, First make a frame for Counting MIC.
        MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            Packet.Body_Len[1] + 4,  &Packet,
                            END_OF_ARGS);
        NdisZeroMemory(mic, sizeof(mic));
            
        // 5-5, count MIC
	        if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
        {
            UCHAR digest[80];
            HMAC_SHA1(OutBuffer,  FrameLen, PTK, LEN_EAP_MICK, digest);
            NdisMoveMemory(mic, digest, LEN_KEY_DESC_MIC);
        }
        else
        {
            hmac_md5(PTK,  LEN_EAP_MICK, OutBuffer, FrameLen, mic);
        }
        NdisMoveMemory(&Packet.KeyDesc.MIC, mic, LEN_KEY_DESC_MIC);        
        }
        else if ((pEntry->AuthMode == Ndis802_11AuthModeWPA2) || (pEntry->AuthMode == Ndis802_11AuthModeWPA2PSK))
        {
        	UCHAR       GTK[TKIP_GTK_LENGTH],pRc4GTK[512], mic[LEN_KEY_DESC_MIC];
			UCHAR       ekey[(LEN_KEY_DESC_IV+LEN_EAP_EK)], Key_Data[512], key_length, RSNIE_Len;

			NdisMoveMemory(Packet.KeyDesc.RSC, pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKey[pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID].TxTsc, 6);

			CountGTK(pAdapter->PortCfg.MSSID[pEntry->ApIdx].GMK, (UCHAR*)pAdapter->PortCfg.MSSID[pEntry->ApIdx].GNonce, pAdapter->CurrentAddress, GTK, TKIP_GTK_LENGTH);

	        NdisMoveMemory(pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKey[pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID].Key, GTK, LEN_TKIP_EK);
	        NdisMoveMemory(pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKey[pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID].TxMic, &GTK[16], LEN_TKIP_TXMICK);
	        NdisMoveMemory(pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKey[pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID].RxMic, &GTK[24], LEN_TKIP_RXMICK);
	        DBGPRINT(RT_DEBUG_INFO, "WPA2 Group Key ID  = %d\n", pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID);
	        DBGPRINT(RT_DEBUG_INFO, "G_Key :%x %x %x %x %x %x %x %x \n",GTK[0],GTK[1],GTK[2],GTK[3],GTK[4],GTK[5],GTK[6],GTK[7]);
	        DBGPRINT(RT_DEBUG_INFO, "       %x %x %x %x %x %x %x %x \n",GTK[8],GTK[9],GTK[10],GTK[11],GTK[12],GTK[13],GTK[14],GTK[15]);
	        DBGPRINT(RT_DEBUG_INFO, "TX Mic:%x %x %x %x %x %x %x %x \n",GTK[16],GTK[17],GTK[18],GTK[19],GTK[20],GTK[21],GTK[22],GTK[23]);
	        DBGPRINT(RT_DEBUG_INFO, "RX Mic:%x %x %x %x %x %x %x %x \n",GTK[24],GTK[25],GTK[26],GTK[27],GTK[28],GTK[29],GTK[30],GTK[31]);

	        NdisZeroMemory(Key_Data, sizeof(Key_Data));

			if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK) ||
				(pAdapter->PortCfg.MSSID[pEntry->ApIdx].AuthMode == Ndis802_11AuthModeWPA1WPA2))
			{
				RSNIE_Len = pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[1];
	        	NdisMoveMemory(&Key_Data[2], pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[1], RSNIE_Len);
			}
			else
			{
				RSNIE_Len = pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSNIE_Len[0];
		        NdisMoveMemory(&Key_Data[2], &pAdapter->PortCfg.MSSID[pEntry->ApIdx].RSN_IE[0], RSNIE_Len);
			}

			Packet.Body_Len[1] = 2+RSNIE_Len;

			Key_Data[0] = IE_WPA2;
        	Key_Data[1] = RSNIE_Len;

			// Key Data Encapsulation format
        	Key_Data[2+RSNIE_Len+0] = 0xDD;
        	Key_Data[2+RSNIE_Len+2] = 0x00;
        	Key_Data[2+RSNIE_Len+3] = 0x0F;
        	Key_Data[2+RSNIE_Len+4] = 0xAC;
        	Key_Data[2+RSNIE_Len+5] = 0x01;

			// GTK Key Data Encapsulation format
        	Key_Data[2+RSNIE_Len+6] = (pAdapter->PortCfg.MSSID[pEntry->ApIdx].WPAGKeyID & 0x03);
			Key_Data[2+RSNIE_Len+7] = 0x00;

			// handle the difference between TKIP and AES-CCMP
			if (pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKeyWepStatus == Ndis802_11Encryption3Enabled)
	        {
	        	Key_Data[2+RSNIE_Len+1] = 0x16;// 4+2+16(OUI+GTK+GTKKEY)
	        	NdisMoveMemory(&Key_Data[2+RSNIE_Len+8], GTK, LEN_EAP_KEY);
	        	Packet.Body_Len[1] += 8+LEN_EAP_KEY;
			}
			else
			{
				Key_Data[2+RSNIE_Len+1] = 0x26;// 4+2+32(OUI+GTK+GTKKEY)
				NdisMoveMemory(&Key_Data[2+RSNIE_Len+8], GTK, TKIP_GTK_LENGTH);
	        	Packet.Body_Len[1] += 8+TKIP_GTK_LENGTH;
			}

			// Still dont know why, but if not append will occur "GTK not include in MSG3"
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH] = 0xDD;
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH+1] = 0;
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH+2] = 0;
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH+3] = 0;
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH+4] = 0;
			Key_Data[2+RSNIE_Len+8+TKIP_GTK_LENGTH+5] = 0;
			Packet.Body_Len[1] += 6;

        	key_length = Packet.Body_Len[1];
        	Packet.KeyDesc.DataLen[1] = key_length;
        	Packet.Body_Len[1] += 93 + 2; 

	        NdisZeroMemory(mic, sizeof(mic));

	        // count MIC
	        //if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
	        if ((pAdapter->PortCfg.MSSID[pEntry->ApIdx].GroupKeyWepStatus == Ndis802_11Encryption3Enabled) || (pEntry->WepStatus == Ndis802_11Encryption3Enabled))
	        {
	            AES_GTK_KEY_WRAP(&pEntry->PTK[16], Key_Data, key_length, pRc4GTK);
	            // AES wrap function will grow 8 bytes in length
	            NdisMoveMemory(Packet.KeyDesc.Data, pRc4GTK, (key_length + 8));
	            Packet.Body_Len[1] += 8;
	            Packet.KeyDesc.DataLen[1] += 8;
	            MakeOutgoingFrame(OutBuffer,            &FrameLen,
	                            Packet.Body_Len[1] + 4, &Packet,
	                            END_OF_ARGS);

	            HMAC_SHA1(OutBuffer, FrameLen, pEntry->PTK, LEN_EAP_MICK, digest);
	            NdisMoveMemory(Packet.KeyDesc.MIC, digest, LEN_KEY_DESC_MIC);
 	        }
	        else
	        {
	            // PREPARE Encrypted  "Key DATA" field.  (Encrypt GTK with RC4, usinf PTK[16]->[31] as Key, IV-field as IV)
	            // put TxTsc in Key RSC field
	            pAdapter->PrivateInfo.FCSCRC32 = PPPINITFCS32;   //Init crc32.

	            // ekey is the contanetion of IV-field, and PTK[16]->PTK[31]
	            NdisMoveMemory(ekey, Packet.KeyDesc.IV, LEN_KEY_DESC_IV);
	            NdisMoveMemory(&ekey[LEN_KEY_DESC_IV], &pEntry->PTK[16], LEN_EAP_EK);
	            ARCFOUR_INIT(&pAdapter->PrivateInfo.WEPCONTEXT, ekey, sizeof(ekey));  //INIT SBOX, KEYLEN+3(IV)
	            pAdapter->PrivateInfo.FCSCRC32 = RTMP_CALC_FCS32(pAdapter->PrivateInfo.FCSCRC32, Key_Data, key_length);
	            WPAARCFOUR_ENCRYPT(&pAdapter->PrivateInfo.WEPCONTEXT, pRc4GTK,Key_Data, key_length);
	            NdisMoveMemory(Packet.KeyDesc.Data, pRc4GTK, key_length);

	            // make a frame for Countint MIC,
	            MakeOutgoingFrame(OutBuffer,            &FrameLen,
	                            Packet.Body_Len[1] + 4, &Packet,
	                            END_OF_ARGS);

	            hmac_md5(pEntry->PTK, LEN_EAP_MICK, OutBuffer, FrameLen, mic);
	            NdisMoveMemory(Packet.KeyDesc.MIC, mic, LEN_KEY_DESC_MIC);
	        }
        }
            
        // 5-6, Make outgoing frame
        WPAMake8023Hdr(pAdapter, pEntry->Addr.Octet, Header802_3);
        MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            LENGTH_802_3,            Header802_3,
                            Packet.Body_Len[1] + 4,  &Packet,
                            END_OF_ARGS);
            
        DBGPRINT(RT_DEBUG_TRACE, "RTMPToWirelessSta : ETHTYPE = %x %x FrameLen = %d! \n",Header802_3[12],Header802_3[13],FrameLen);

        RTMPToWirelessSta(pAdapter, OutBuffer, FrameLen);

        // Should start from 1, init is 0
		pEntry->ReTryCounter = PEER_MSG3_RETRY_TIMER_CTR;
		pEntry->RetryTimer.expires = jiffies + PEER_MSG3_RETRY_EXEC_INTV;
		pEntry->RetryTimerRunning = TRUE;
        mod_timer(&pEntry->RetryTimer, pEntry->RetryTimer.expires);
		DBGPRINT(RT_DEBUG_TRACE, "Send Msg3 and setup timeout timer\n");

        pEntry->WpaState = AS_PTKINIT_NEGOTIATING;
    }while(FALSE);
    kfree(OutBuffer);
}

/*
    ==========================================================================
    Description:
        countermeasures active attack timer execution
    Return:
    ==========================================================================
*/
VOID CMTimerExec(
    IN  unsigned long data) 
{
    UINT            i, j=0, apidx;
    RTMP_ADAPTER    *pAdapter = (RTMP_ADAPTER *)data;
        
    for (apidx = 0; apidx < pAdapter->PortCfg.SsidNum; apidx++)
    {
        pAdapter->PortCfg.MSSID[apidx].BANClass3Data = FALSE;

        for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
        {
            if ((pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Valid == TRUE) && (pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].CMTimerRunning == TRUE))
            {
                pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].CMTimerRunning =FALSE;
                j++;
            }
        }
    }
    if (j > 1)
        DBGPRINT(RT_DEBUG_ERROR, "Find more than one entry which generated MIC Fail ..  \n");

    pAdapter->PortCfg.CMTimerRunning = FALSE;
}
    
VOID WPARetryExec(
    IN  unsigned long data) 
{
    MAC_TABLE_ENTRY     *pEntry = (PMAC_TABLE_ENTRY) data;
    
    if ((pEntry->Valid == TRUE) && (pEntry->RetryTimerRunning == TRUE))
    {
    	RTMP_ADAPTER		*pAdapter = (PRTMP_ADAPTER)pEntry->pAd;
    	
        // Turn off the retry flag first
        pEntry->RetryTimerRunning = FALSE;  
        DBGPRINT(RT_DEBUG_TRACE, "WPARetryExec - ReTryCounter=%d, WpaState=%d \n", pEntry->ReTryCounter, pEntry->WpaState);

        switch (pEntry->AuthMode)
        {
            case Ndis802_11AuthModeWPA:
            case Ndis802_11AuthModeWPAPSK:
            case Ndis802_11AuthModeWPA2:
            case Ndis802_11AuthModeWPA2PSK:
                //  1.   GTK already retried, give up and disconnect client.
                if (pEntry->ReTryCounter > (GRP_HS_RETRY_TIMER_CTR + 1))
                {    
					// Deauthenticate and delete timer and entry
					DisAssocAction(pAdapter, pEntry, SUBTYPE_DISASSOC, REASON_GROUP_KEY_UPDATE_TIMEOUT);
                    DBGPRINT(RT_DEBUG_TRACE, "Delete client after GTK retry failed\n");
                }
				// 2. Retry GTK.
                else if (pEntry->ReTryCounter >= GRP_HS_RETRY_TIMER_CTR)
                {
                    DBGPRINT(RT_DEBUG_TRACE, "ReInitilize 2-way group-key Handshake ===> \n");
                    WPAHardTransmit( pAdapter,  pEntry);
	                pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
    				mod_timer(&pEntry->RetryTimer, pEntry->RetryTimer.expires);
					pEntry->ReTryCounter++;
                    pEntry->RetryTimerRunning = TRUE;  
                }
				// 3. 4-way message 1 retried more than three times. Disconnect client
                else if (pEntry->ReTryCounter > (PEER_MSG1_RETRY_TIMER_CTR + 3))                
                {   
					// Deauthenticate and delete timer and entry
					DisAssocAction(pAdapter, pEntry, SUBTYPE_DEAUTH, REASON_UNSPECIFY);
                    DBGPRINT(RT_DEBUG_TRACE, "Delete client after 3 message 1 retries\n");
                }
				// 4. Retry 4 way message 1, the last try, the timeout is 5 sec for EAPOL-Start
                else if (pEntry->ReTryCounter == (PEER_MSG1_RETRY_TIMER_CTR + 3))                
                {
                    DBGPRINT(RT_DEBUG_TRACE, "Message 1 time out, the last retry\n");
	              pEntry->PrivacyFilter = Ndis802_11PrivFilter8021xWEP;
	                pEntry->WpaState = AS_INITPSK;
	                pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
	                pEntry->RetryTimer.expires = jiffies + (5 * HZ);
                    WPAStart4WayHS(pAdapter, pEntry);
                }
				// 4. Retry 4 way message 1
                else if (pEntry->ReTryCounter < (PEER_MSG1_RETRY_TIMER_CTR + 3))                
                {
                    DBGPRINT(RT_DEBUG_TRACE, "Message 1 time out, retry it\n");
	                pEntry->PrivacyFilter = Ndis802_11PrivFilter8021xWEP;
	                pEntry->WpaState = AS_INITPSK;
	                pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
	                pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
                    WPAStart4WayHS(pAdapter, pEntry);
                }
                break;

            default:
                break;
        }        
    }
}

/*
    ==========================================================================
    Description:
        When receiving the last packet of 4-way pairwisekey handshake.
        Initilize 2-way groupkey handshake following.
    Return:
    ==========================================================================
*/
VOID Wpa1PeerPairMsg4Action(
    IN PRTMP_ADAPTER    pAdapter, 
    IN MAC_TABLE_ENTRY  *pEntry,
    IN MLME_QUEUE_ELEM  *Elem) 
{
    UCHAR               mic[LEN_KEY_DESC_MIC], digest[80];
    ULONG               MICMsgLen;
    EAPOL_PACKET        EAPOLPKT;
    PEAPOL_PACKET       pMsg4;
    NDIS_STATUS         Status;

    DBGPRINT(RT_DEBUG_TRACE, "WpaEAPOL Peer PAIR  Msg4 Action ====>>\n");
        
    if (pEntry->WpaState < AS_PTKINIT_NEGOTIATING)
        return;

    pMsg4 = (PEAPOL_PACKET)&Elem->Msg[LENGTH_802_11 + LENGTH_802_1_H];

    // 1.check Replay Counter
    if (!RTMPEqualMemory(pMsg4->KeyDesc.RCounter, pEntry->R_Counter, LEN_KEY_DESC_REPLAY))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Replay Counter Different in Msg 4 of 4-way handshake!! \n");
        DBGPRINT(RT_DEBUG_ERROR, "Receive : %d %d %d %d  \n", pMsg4->KeyDesc.RCounter[0],pMsg4->KeyDesc.RCounter[1],pMsg4->KeyDesc.RCounter[2],pMsg4->KeyDesc.RCounter[3]);
        DBGPRINT(RT_DEBUG_ERROR, "Current : %d %d %d %d  \n", pEntry->R_Counter[4],pEntry->R_Counter[5],pEntry->R_Counter[6],pEntry->R_Counter[7]);
        return;
    }

    // 2. check MIC, if not valid, discard silently
    MICMsgLen = pMsg4->Body_Len[1] | ((pMsg4->Body_Len[0]<<8) && 0xff00);
    MICMsgLen += LENGTH_EAPOL_H;
    if (MICMsgLen > (Elem->MsgLen -LENGTH_802_11 - LENGTH_802_1_H))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Receive wrong format EAPOL packets \n");
        return;        
    }
    NdisZeroMemory((PUCHAR)&EAPOLPKT, sizeof(EAPOLPKT));
    NdisMoveMemory((PUCHAR)&EAPOLPKT, pMsg4, MICMsgLen);  
    NdisZeroMemory(EAPOLPKT.KeyDesc.MIC, sizeof(EAPOLPKT.KeyDesc.MIC));

    hmac_md5(pEntry->PTK, LEN_EAP_MICK, (PUCHAR)&EAPOLPKT, (MICMsgLen), mic);
    if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
    {
        hmac_md5(pEntry->PTK, LEN_EAP_MICK, (PUCHAR)&EAPOLPKT, (MICMsgLen), mic);
    }
    else
    {
        HMAC_SHA1((PUCHAR)&EAPOLPKT,  MICMsgLen, pEntry->PTK, LEN_EAP_MICK, digest);
        NdisMoveMemory(mic, digest, LEN_KEY_DESC_MIC);
    }
    if (!RTMPEqualMemory(pMsg4->KeyDesc.MIC, mic, LEN_KEY_DESC_MIC))
    {
        DBGPRINT(RT_DEBUG_ERROR, "MIC different in Msg 4 of 4-way handshake!! \n");
        return;
    }
    else
        DBGPRINT(RT_DEBUG_INFO, "MIC valid in Msg 4 of 4-way handshake!! \n");

    // 3. uses the MLME.SETKEYS.request to configure PTK into MAC
    NdisZeroMemory(&pEntry->PairwiseKey[0], sizeof(WPA_KEY));   
    NdisMoveMemory(pEntry->PairwiseKey[0].RxTsc, EAPOLPKT.KeyDesc.RSC, 6);

    pEntry->PairwiseKey[0].KeyLen = LEN_TKIP_EK;
    NdisMoveMemory(pEntry->PairwiseKey[0].Key, &pEntry->PTK[32], LEN_TKIP_EK);
    NdisMoveMemory(pEntry->PairwiseKey[0].RxMic, &pEntry->PTK[TKIP_AP_RXMICK_OFFSET], LEN_TKIP_RXMICK);
    NdisMoveMemory(pEntry->PairwiseKey[0].TxMic, &pEntry->PTK[TKIP_AP_TXMICK_OFFSET], LEN_TKIP_TXMICK);
    
    // 4. upgrade state
    pEntry->PrivacyFilter = Ndis802_11PrivFilterAcceptAll;
    pEntry->WpaState = AS_PTKINITDONE;
    pAdapter->PortCfg.MSSID[pEntry->ApIdx].PortSecured = WPA_802_1X_PORT_SECURED;
    pEntry->PortSecured = WPA_802_1X_PORT_SECURED;

    // 5. init Group 2-way handshake if necessary.
    if ((Status = WPAHardTransmit(pAdapter, pEntry)) != NDIS_STATUS_SUCCESS)
        return;
    
    pEntry->RetryTimerRunning = TRUE;
    pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
    pEntry->ReTryCounter = GRP_HS_RETRY_TIMER_CTR;
	mod_timer(&pEntry->RetryTimer, pEntry->RetryTimer.expires);

    DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) recv WpaEAPOL Peer PAIR Msg4 Action and send GROUP Msg1\n", pEntry->ApIdx);
}

/*
    ==========================================================================
    Description:
        When receiving the last packet of 4-way pairwisekey handshake.
    Return:
    ==========================================================================
*/
VOID Wpa2PeerPairMsg4Action(
    IN PRTMP_ADAPTER	pAdapter, 
    IN MAC_TABLE_ENTRY	*pEntry,
    IN MLME_QUEUE_ELEM	*Elem) 
{
	UINT                Len;
    UCHAR               mic[LEN_KEY_DESC_MIC], digest[80], PMK_key[20];
    PUCHAR              pData;
    PEAPOL_PACKET       pMsg2;
    EAPOL_PACKET        EAPOLPKT;
    VOID				*Msg = &Elem->Msg[LENGTH_802_11];
    UINT				MsgLen = Elem->MsgLen - LENGTH_802_11;

    DBGPRINT(RT_DEBUG_TRACE, "Wpa2PeerPairMsg4Action ====> from MAC : %x %x %x %x %x %x\n",pEntry->Addr.Octet[0],\
        pEntry->Addr.Octet[1], pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],pEntry->Addr.Octet[4],pEntry->Addr.Octet[5]);

    Len = MsgLen - LENGTH_802_1_H;
    pData = (PUCHAR)Msg;
    NdisZeroMemory((PUCHAR)&EAPOLPKT, sizeof(EAPOLPKT));
    NdisZeroMemory(mic, sizeof(mic));

    pMsg2 = (PEAPOL_PACKET) (pData + LENGTH_802_1_H);

    // 1. verify the Reaply counter, if not valid,
    if (!RTMPEqualMemory(pMsg2->KeyDesc.RCounter, pEntry->R_Counter, LEN_KEY_DESC_REPLAY))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Replay Counter Different in Msg 4 of 4-way handshake!!!\n");
        DBGPRINT(RT_DEBUG_ERROR, "Receive : %d %d %d %d \n", pMsg2->KeyDesc.RCounter[4], pMsg2->KeyDesc.RCounter[5], pMsg2->KeyDesc.RCounter[6], pMsg2->KeyDesc.RCounter[7]);
        DBGPRINT(RT_DEBUG_ERROR, "Current : %d %d %d %d \n",pEntry->R_Counter[4],pEntry->R_Counter[5],pEntry->R_Counter[6],pEntry->R_Counter[7]);
        return;
    }
    else
        DBGPRINT(RT_DEBUG_TRACE, "Replay Counter VALID in Msg 4 of 4-way handshake!\n");

    // 2. verify MIC, 
    NdisMoveMemory(&EAPOLPKT, pMsg2, Len);  
    NdisZeroMemory(&EAPOLPKT.KeyDesc.MIC, LEN_KEY_DESC_MIC);
    if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
    {
        hmac_md5(pEntry->PTK, LEN_EAP_MICK, (UCHAR *)&EAPOLPKT, Len, mic);
    }
    else
    {
        HMAC_SHA1((UCHAR *)&EAPOLPKT,   Len, pEntry->PTK, LEN_EAP_MICK, digest);
        NdisMoveMemory(mic, digest, LEN_KEY_DESC_MIC);
    }
    if (!RTMPEqualMemory(pMsg2->KeyDesc.MIC, mic, LEN_KEY_DESC_MIC))
    {
        DBGPRINT(RT_DEBUG_ERROR, "MIC Different in Msg 4 of 4-way handshake!! \n");
        return;
    }
    else
        DBGPRINT(RT_DEBUG_TRACE, "MIC Valid in Msg 4 of 4-way handshake!! \n");

    // 3.  upgrade state
    if (pEntry->RetryTimerRunning == TRUE)
    {
        del_timer_sync(&pEntry->RetryTimer);
        pEntry->RetryTimerRunning = FALSE;              
    }

    NdisZeroMemory(&pEntry->PairwiseKey[0], sizeof(WPA_KEY));   
    NdisMoveMemory(pEntry->PairwiseKey[0].RxTsc, EAPOLPKT.KeyDesc.RSC, 6);

    pEntry->PairwiseKey[0].KeyLen = LEN_TKIP_EK;
    NdisMoveMemory(pEntry->PairwiseKey[0].Key, &pEntry->PTK[32], LEN_TKIP_EK);
    NdisMoveMemory(pEntry->PairwiseKey[0].RxMic, &pEntry->PTK[TKIP_AP_RXMICK_OFFSET], LEN_TKIP_RXMICK);
    NdisMoveMemory(pEntry->PairwiseKey[0].TxMic, &pEntry->PTK[TKIP_AP_TXMICK_OFFSET], LEN_TKIP_TXMICK);
    
    pEntry->PrivacyFilter = Ndis802_11PrivFilterAcceptAll;
    pEntry->WpaState = AS_PTKINITDONE;
    pAdapter->PortCfg.MSSID[pEntry->ApIdx].PortSecured = WPA_802_1X_PORT_SECURED;
    pEntry->PortSecured = WPA_802_1X_PORT_SECURED;
    pEntry->GTKState = REKEY_ESTABLISHED;

	if (pEntry->AuthMode == Ndis802_11AuthModeWPA2)
	{
		// Calculate PMKID
		NdisMoveMemory(&PMK_key[0], "PMK Name", 8);
		NdisMoveMemory(&PMK_key[8], pAdapter->CurrentAddress, MAC_ADDR_LEN);
		NdisMoveMemory(&PMK_key[14], &pEntry->Addr, MAC_ADDR_LEN);
		HMAC_SHA1(PMK_key, 20, pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK, PMK_LEN, digest);
		RTMPAddPMKIDCache(pAdapter, pEntry->ApIdx, &pEntry->Addr, digest, pAdapter->PortCfg.MSSID[pEntry->ApIdx].PMK);
		DBGPRINT(RT_DEBUG_TRACE, "Calc PMKID=%02x:%02x:%02x:%02x:%02x:%02x\n", digest[0],digest[1],digest[2],digest[3],digest[4],digest[5]);
	}

    DBGPRINT(RT_DEBUG_TRACE, "===> AP SETKEYS DONE (ra%d) - WPA2, AuthMode=%d, WepStatus=%d \n\n", pEntry->ApIdx, pEntry->AuthMode, pEntry->WepStatus);
}

/*
    ==========================================================================
    Description:
        When receiving the last packet of 2-way groupkey handshake.
    Return:
    ==========================================================================
*/
VOID PeerGroupMsg2Action(
    IN PRTMP_ADAPTER    pAdapter, 
    IN MAC_TABLE_ENTRY  *pEntry,
    IN VOID             *Msg,
    IN UINT             MsgLen)
{
    UINT                Len;
    UCHAR               mic[LEN_KEY_DESC_MIC], digest[80];
    PUCHAR              pData;
    PEAPOL_PACKET       pMsg2;
    EAPOL_PACKET        EAPOLPKT;

    Len = MsgLen - LENGTH_802_1_H;
    pData = (PUCHAR)Msg;
    NdisZeroMemory((PUCHAR)&EAPOLPKT, sizeof(EAPOLPKT));
    NdisZeroMemory(mic, sizeof(mic));
    DBGPRINT(RT_DEBUG_TRACE, "PeerGroupMsg2Action ====> from MAC : %x %x %x %x %x %x\n",pEntry->Addr.Octet[0],
        pEntry->Addr.Octet[1], pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],pEntry->Addr.Octet[4],pEntry->Addr.Octet[5]);

    if (MsgLen < (LENGTH_802_1_H + LENGTH_EAPOL_H + sizeof(KEY_DESCRIPTER) -MAX_LEN_OF_RSNIE -2 ) )
        return;
            
    if (pEntry->WpaState != AS_PTKINITDONE)
        return;
        
    pMsg2 = (PEAPOL_PACKET) (pData + LENGTH_802_1_H);

    // 1. verify the Reaply counter, if not valid,
    if (!RTMPEqualMemory(pMsg2->KeyDesc.RCounter, pEntry->R_Counter, LEN_KEY_DESC_REPLAY))
    {
        DBGPRINT(RT_DEBUG_ERROR, "Replay Counter Different in Msg 2 of GROUP 2-way handshake!!!\n");
        DBGPRINT(RT_DEBUG_ERROR, "Receive : %d %d %d %d \n", pMsg2->KeyDesc.RCounter[4], pMsg2->KeyDesc.RCounter[5], pMsg2->KeyDesc.RCounter[6], pMsg2->KeyDesc.RCounter[7]);
        DBGPRINT(RT_DEBUG_ERROR, "Current : %d %d %d %d \n",pEntry->R_Counter[4],pEntry->R_Counter[5],pEntry->R_Counter[6],pEntry->R_Counter[7]);
        return;
    }
    else
        DBGPRINT(RT_DEBUG_TRACE, "Replay Counter VALID in Msg 2 of GROUP 2-way handshake!!!\n");

    // 2. verify MIC, 
    NdisMoveMemory(&EAPOLPKT, pMsg2, Len);  
    NdisZeroMemory(&EAPOLPKT.KeyDesc.MIC, LEN_KEY_DESC_MIC);
    if (pEntry->WepStatus == Ndis802_11Encryption2Enabled)
    {
        hmac_md5(pEntry->PTK,   LEN_EAP_MICK, (UCHAR *)&EAPOLPKT, Len, mic);
    }
    else
    {
        HMAC_SHA1((UCHAR *)&EAPOLPKT,   Len, pEntry->PTK, LEN_EAP_MICK, digest);
        NdisMoveMemory(mic, digest, LEN_KEY_DESC_MIC);
    }
    if (!RTMPEqualMemory(pMsg2->KeyDesc.MIC, mic, LEN_KEY_DESC_MIC))
    {
        DBGPRINT(RT_DEBUG_ERROR, "MIC Different in Msg 2 of GROUP 2-way handshake! \n");
        return;
    }
    else
        DBGPRINT(RT_DEBUG_TRACE, "MIC Valid in Msg 2 of GROUP 2-way handshake. \n");

    // 3.  upgrade state
    if (pEntry->RetryTimerRunning == TRUE)
    {
        del_timer_sync(&pEntry->RetryTimer);
        pEntry->RetryTimerRunning = FALSE;              
    }
    pEntry->GTKState = REKEY_ESTABLISHED;

    DBGPRINT(RT_DEBUG_TRACE, "===> AP SETKEYS DONE - (ra%d) WPA1, AuthMode=%d, WepStatus=%d \n\n", pEntry->ApIdx, pEntry->AuthMode, pEntry->WepStatus);
}

/*
    ==========================================================================
    Description:
        Only for sending the first packet of 2-way groupkey handshake
    Return:
    ==========================================================================
*/
NDIS_STATUS WPAHardTransmit(
    IN PRTMP_ADAPTER    pAdapter,
    IN MAC_TABLE_ENTRY  *pEntry)
{
    UCHAR               GTK[TKIP_GTK_LENGTH],pRc4GTK[TKIP_GTK_LENGTH], mic[LEN_KEY_DESC_MIC];
    UCHAR               ekey[LEN_KEY_DESC_IV + LEN_EAP_EK];
    ULONG               FrameLen = 0, apidx;
    UCHAR               *OutBuffer = NULL;
    UCHAR               Header802_3[14];
    EAPOL_PACKET        Packet;

    NdisZeroMemory(ekey, sizeof(ekey));
    NdisZeroMemory(pRc4GTK, sizeof(pRc4GTK));

    apidx = pEntry->ApIdx;

    do
    {
        OutBuffer = kmalloc(MAX_LEN_OF_EAP_HS, GFP_KERNEL);
        if(OutBuffer == NULL)
            return (NDIS_STATUS_FAILURE);

        // Increment replay counter by 1  
        ADD_ONE_To_64BIT_VAR(pAdapter->PortCfg.R_Counter);
        NdisMoveMemory(pEntry->R_Counter, pAdapter->PortCfg.R_Counter, sizeof(pEntry->R_Counter));

        // 0. init Packet and Fill header
        NdisZeroMemory(&Packet, sizeof(Packet));
        Packet.ProVer = EAPOL_VER;
        Packet.ProType = EAPOLKey;
        Packet.Body_Len[1] = LEN_MSG1_2WAY;
        
        // 2, Fill key version, keyinfo, key len
        Packet.KeyDesc.Type = WPA1_KEY_DESC;
        Packet.KeyDesc.Keyinfo.KeyMic = 1;
        Packet.KeyDesc.Keyinfo.Secure = 1;
        Packet.KeyDesc.Keyinfo.KeyDescVer = DESC_TYPE_TKIP;
        Packet.KeyDesc.Keyinfo.KeyType = GROUPKEY;
        Packet.KeyDesc.Keyinfo.KeyIndex = pAdapter->PortCfg.MSSID[apidx].WPAGKeyID;
        Packet.KeyDesc.Keyinfo.KeyAck = 1;
        Packet.KeyDesc.KeyLength[1] = TKIP_GTK_LENGTH;
        Packet.KeyDesc.DataLen[1] = TKIP_GTK_LENGTH;
        if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
        {
            Packet.Body_Len[1] -= 8;
            Packet.KeyDesc.KeyLength[1] = LEN_AES_KEY;
            Packet.KeyDesc.DataLen[1] = LEN_AES_KEY + 8;
            Packet.KeyDesc.Keyinfo.KeyDescVer = DESC_TYPE_AES;
        }

#ifdef BIG_ENDIAN
        *(USHORT *)&Packet.KeyDesc.Keyinfo = SWAP16(*(USHORT *)&Packet.KeyDesc.Keyinfo);
#endif

        CountGTK(pAdapter->PortCfg.MSSID[apidx].GMK, (UCHAR*)pAdapter->PortCfg.MSSID[apidx].GNonce, pAdapter->CurrentAddress, GTK, TKIP_GTK_LENGTH);
        DBGPRINT(RT_DEBUG_INFO, "I/F(ra%d) WPA Group Key ID  = %d\n", apidx, pAdapter->PortCfg.MSSID[apidx].WPAGKeyID);
        DBGPRINT(RT_DEBUG_INFO, "%x %x %x %x %x %x %x %x \n",GTK[0],GTK[1],GTK[2],GTK[3],GTK[4],GTK[5],GTK[6],GTK[7]);
        DBGPRINT(RT_DEBUG_INFO, "%x %x %x %x %x %x %x %x \n",GTK[8],GTK[9],GTK[10],GTK[11],GTK[12],GTK[13],GTK[14],GTK[15]);
        DBGPRINT(RT_DEBUG_INFO, "%x %x %x %x %x %x %x %x \n",GTK[16],GTK[17],GTK[18],GTK[19],GTK[20],GTK[21],GTK[22],GTK[23]);
        DBGPRINT(RT_DEBUG_INFO, "%x %x %x %x %x %x %x %x \n",GTK[24],GTK[25],GTK[26],GTK[27],GTK[28],GTK[29],GTK[30],GTK[31]);
                
        NdisMoveMemory(pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].Key, GTK, LEN_TKIP_EK);
        NdisMoveMemory(pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].TxMic, &GTK[16], LEN_TKIP_TXMICK);
        NdisMoveMemory(pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].RxMic, &GTK[24], LEN_TKIP_RXMICK);            
        
        WPAMake8023Hdr(pAdapter, pEntry->Addr.Octet, Header802_3);
        
        NdisMoveMemory(Packet.KeyDesc.RCounter, pEntry->R_Counter, LEN_KEY_DESC_REPLAY);
        NdisMoveMemory(Packet.KeyDesc.Nonce, pAdapter->PortCfg.MSSID[apidx].GNonce, LEN_KEY_DESC_NONCE);
        NdisMoveMemory(Packet.KeyDesc.IV, &pAdapter->PortCfg.MSSID[apidx].GNonce[16], sizeof(LEN_KEY_DESC_IV));
        
        // Suggest IV be random number plus some number,
        Packet.KeyDesc.IV[15] += 2;
        NdisMoveMemory(Packet.KeyDesc.RSC, pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].TxTsc, 6);
        // Count EAPOL MIC , and encrypt DATA field before Send,   DATA fields includes the encrypted GTK
        if (pEntry->WepStatus == Ndis802_11Encryption3Enabled)
        {
            UCHAR digest[80];  // 80 to prevent overflow
            
            if (pAdapter->PortCfg.MSSID[apidx].GroupKeyWepStatus == Ndis802_11Encryption3Enabled)
            {
            	AES_GTK_KEY_WRAP(&pEntry->PTK[16], GTK, 16, pRc4GTK);
            	NdisMoveMemory(Packet.KeyDesc.Data, pRc4GTK, (LEN_AES_KEY +8));
            }
            else
            {
            	AES_GTK_KEY_WRAP(&pEntry->PTK[16], GTK, TKIP_GTK_LENGTH, pRc4GTK);
            	NdisMoveMemory(Packet.KeyDesc.Data, pRc4GTK, (TKIP_GTK_LENGTH +8));
            	Packet.KeyDesc.KeyLength[1] += 16;
		        Packet.KeyDesc.DataLen[1] += 16;
		        Packet.Body_Len[1] += 16;
            }
            
            // First make a frame  for Countint MIC,
            NdisMoveMemory(Packet.KeyDesc.Data, pRc4GTK, (LEN_AES_KEY +8));
            MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            Packet.Body_Len[1] + 4, &Packet,
                            END_OF_ARGS);

            HMAC_SHA1(OutBuffer,  FrameLen, pEntry->PTK, LEN_EAP_MICK, digest);
            NdisMoveMemory(&Packet.KeyDesc.MIC, digest, LEN_KEY_DESC_MIC);
            
            // make  outgoing frame
            MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            LENGTH_802_3,           &Header802_3,
                            Packet.Body_Len[1] + 4, &Packet,
                            END_OF_ARGS);
        }
        else
        {
            // PREPARE Encrypted  "Key DATA" field.  (Encrypt GTK with RC4, usinf PTK[16]->[31] as Key, IV-field as IV)
            // put TxTsc in Key RSC field
            pAdapter->PrivateInfo.FCSCRC32 = PPPINITFCS32;   //Init crc32.

            // ekey is the contanetion of IV-field, and PTK[16]->PTK[31]
            NdisMoveMemory(ekey, Packet.KeyDesc.IV, LEN_KEY_DESC_IV);
            NdisMoveMemory(&ekey[LEN_KEY_DESC_IV], &pEntry->PTK[16], LEN_EAP_EK);
            ARCFOUR_INIT(&pAdapter->PrivateInfo.WEPCONTEXT, ekey, sizeof(ekey));  //INIT SBOX, KEYLEN+3(IV)
            pAdapter->PrivateInfo.FCSCRC32 = RTMP_CALC_FCS32(pAdapter->PrivateInfo.FCSCRC32, GTK, TKIP_GTK_LENGTH);
            WPAARCFOUR_ENCRYPT(&pAdapter->PrivateInfo.WEPCONTEXT, pRc4GTK,GTK,  TKIP_GTK_LENGTH);
            NdisMoveMemory(Packet.KeyDesc.Data,  pRc4GTK, TKIP_GTK_LENGTH);

            // make a frame for Countint MIC,
            MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            Packet.Body_Len[1] + 4, &Packet,
                            END_OF_ARGS);

            hmac_md5(pEntry->PTK, LEN_EAP_MICK, OutBuffer, FrameLen, mic);
            NdisMoveMemory(Packet.KeyDesc.MIC, mic, LEN_KEY_DESC_MIC);
            
            // make  outgoing frame
            MakeOutgoingFrame(OutBuffer,            &FrameLen,
                            LENGTH_802_3,           &Header802_3,
                            Packet.Body_Len[1] + 4, &Packet,
                            END_OF_ARGS);
        }
         
        DBGPRINT(RT_DEBUG_TRACE,"<== I/F(ra%d) WPAHardTransmit - FrameLen = %d \n", apidx, FrameLen);
        RTMPToWirelessSta(pAdapter, OutBuffer, FrameLen);
    }while (FALSE);
    kfree(OutBuffer);

    return (NDIS_STATUS_SUCCESS);
}

VOID DisAssocAction(
    IN PRTMP_ADAPTER    pAd, 
    IN MAC_TABLE_ENTRY  *pEntry,
    IN USHORT           SubType,
    IN USHORT           Reason)
{
    CHAR            *OutBuffer = NULL;
    ULONG           FrameLen = 0;
    MACHDR          DisassocHdr;
    MACADDR         addr;

    if (pEntry)
    {
	    //  send out a DISASSOC request frame
	    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_KERNEL);
	    if(OutBuffer == NULL)
	        return;
	    
	    NdisMoveMemory(addr.Octet, pEntry->Addr.Octet, MAC_ADDR_LEN);
	    DBGPRINT(RT_DEBUG_ERROR, "WPA - Send DISASSOC  Reason = %d frame  TO %x %x %x %x %x %x \n",Reason,pEntry->Addr.Octet[0],\
	    pEntry->Addr.Octet[1],pEntry->Addr.Octet[2],pEntry->Addr.Octet[3],pEntry->Addr.Octet[4],pEntry->Addr.Octet[5]);
	    MgtMacHeaderInit(pAd, &DisassocHdr, SubType, 0, &addr, &pAd->PortCfg.Bssid);
	    MakeOutgoingFrame(OutBuffer, &FrameLen, sizeof(MACHDR), (CHAR *)&DisassocHdr, 2, (char*)&Reason, END_OF_ARGS);
	    MiniportMMRequest(pAd, OutBuffer, FrameLen);
    
        //ApLogEvent(pAd, &pEntry->Addr, EVENT_DISASSOCIATED, apidx);
        MacTableDeleteEntry(pAd,  &pEntry->Addr, pEntry->ApIdx);
    }
}

/*
    ==========================================================================
    Description:
        Timer execution function for periodically updating group key.
    Return:
    ==========================================================================
*/  
VOID GREKEYPeriodicExec(
    IN  unsigned long data) 
{
    UINT            i;
    ULONG           temp_counter = 0;
    RTMP_ADAPTER    *pAdapter = (RTMP_ADAPTER *)data;
    BOOLEAN         bTimer = TRUE;
    UCHAR           apidx;
        
    DBGPRINT(RT_DEBUG_INFO, "GROUP REKEY PeriodicExec ==>> \n");

    for (apidx = 0; apidx < pAdapter->PortCfg.SsidNum; apidx++)
    {
        if ((pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPA) && (pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPAPSK)
        	&& (pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPA2) && (pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPA2PSK)
        	&& (pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPA1WPA2) && (pAdapter->PortCfg.MSSID[apidx].AuthMode != Ndis802_11AuthModeWPAPSKWPA2PSK))
        {
            bTimer = FALSE;
            pAdapter->PortCfg.REKEYTimerRunning = FALSE;

            continue;
        }    
        else if (pAdapter->PortCfg.WPAREKEY.ReKeyInterval == 0)
        {
            bTimer = FALSE;
            pAdapter->PortCfg.REKEYTimerRunning = FALSE;

            continue;
        }
        
        if ((pAdapter->PortCfg.WPAREKEY.ReKeyMethod == TIME_REKEY) && (pAdapter->PortCfg.REKEYCOUNTER < 0xffffffff))
            temp_counter = (++pAdapter->PortCfg.REKEYCOUNTER);
        // REKEYCOUNTER is incremented every TX_RING_SIZE packets transmitted, 
        // But the unit of Rekeyinterval is 1K packets
        else if (pAdapter->PortCfg.WPAREKEY.ReKeyMethod == PKT_REKEY )
            temp_counter = pAdapter->PortCfg.REKEYCOUNTER/TX_RING_SIZE;
        else
        {
            bTimer = FALSE;
            pAdapter->PortCfg.REKEYTimerRunning = FALSE;

            continue;
        }
        
        if (temp_counter >= (pAdapter->PortCfg.WPAREKEY.ReKeyInterval))
        {
            pAdapter->PortCfg.REKEYCOUNTER = 0;
            DBGPRINT(RT_DEBUG_TRACE, "Rekey Interval Excess, GKeyDoneStations= %d\n", pAdapter->PortCfg.MSSID[apidx].MacTab.Size);
            if (pAdapter->PortCfg.MSSID[apidx].MacTab.Size > 0)
            {
                pAdapter->PortCfg.MSSID[apidx].WPAGKeyID ++;
                if (pAdapter->PortCfg.MSSID[apidx].WPAGKeyID >= 4)
                    pAdapter->PortCfg.MSSID[apidx].WPAGKeyID = 1;
                GenRandom(pAdapter, pAdapter->PortCfg.MSSID[apidx].GNonce);
                for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
                {
                    if ((pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Valid == TRUE) && (pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].WpaState == AS_PTKINITDONE))
                    {
                        pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].GTKState = REKRY_NEGOTIATING;
                        NdisZeroMemory(pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].TxTsc, sizeof(pAdapter->PortCfg.MSSID[apidx].GroupKey[pAdapter->PortCfg.MSSID[apidx].WPAGKeyID].TxTsc));
                        WPAHardTransmit(pAdapter, &pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i]);
                        DBGPRINT(RT_DEBUG_TRACE, "Rekey interval excess, Update Group Key for %x %x %x %x %x %x , WPAGKeyID= %x \n",\
                            pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[0],pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[1],\
                            pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[2],pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[3],\
                            pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[4],pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i].Addr.Octet[5],\
                            pAdapter->PortCfg.MSSID[apidx].WPAGKeyID);
                    }
                }
            }
        }
    }
        
    if(bTimer)
    {
        pAdapter->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
        add_timer(&pAdapter->PortCfg.REKEYTimer);
        pAdapter->PortCfg.REKEYTimerRunning = TRUE;
    }
}

VOID WPAMake8023Hdr(
    IN PRTMP_ADAPTER    pAd, 
    IN PCHAR            pDAddr, 
    IN OUT PCHAR        pHdr)
{    
     // Addr1: DA, Addr2: BSSID, Addr3: SA
    NdisMoveMemory(pHdr, pDAddr, MAC_ADDR_LEN);
    NdisMoveMemory(&pHdr[MAC_ADDR_LEN], pAd->CurrentAddress, MAC_ADDR_LEN);
    NdisMoveMemory(&pHdr[2*MAC_ADDR_LEN], EAPOL, LENGTH_802_3_TYPE);
}

/*
    ==========================================================================
    Description:
        Length of output is in octets rather than bits.  Since length is always a multiple of 8, output array is organized
        so first N octets starting from 0 contains PRF output.
        INPUT : supported input are 16, 32, 48, 64
        OUTPUT : output array should be 80 octets to allow for sha1 overflow
    Return:
    ==========================================================================
*/
VOID PRF(
    IN UCHAR    *Key,
    IN INT      key_len,
    IN UCHAR    *Prefix,
    IN INT      Prefix_len,
    IN UCHAR    *Data,
    IN INT      Data_len,
    OUT UCHAR   *Output,
    IN INT      Output_len)
{
    INT     i;
    UCHAR   input[1024];
    INT     currentindex=0;
    INT     total_len;
    
    NdisMoveMemory(input, Prefix, Prefix_len);
    input[Prefix_len] = 0;
    NdisMoveMemory(&input[Prefix_len+1], Data, Data_len);
    total_len = Prefix_len+1+Data_len;
    input[total_len] = 0;
    total_len++;
    for (i = 0; i < (Output_len+19)/20; i++)
    {
        HMAC_SHA1(input, total_len, Key, key_len, &Output[currentindex]);
        currentindex += 20;
        input[total_len-1]++;
    }
}

VOID CountPTK(
    IN UCHAR    *PMK,
    IN UCHAR    *ANonce,
    IN UCHAR    *AA,
    IN UCHAR    *SNonce,
    IN UCHAR    *SA,
    OUT UCHAR   *output,
    IN UINT     len)
{   
    UCHAR   concatenation[76];
    UINT    CurrPos=0;
    UCHAR   temp[32];
    UCHAR   Prefix[] = {'P', 'a', 'i', 'r', 'w', 'i', 's', 'e', ' ', 'k', 'e', 'y', ' ', 
                        'e', 'x', 'p', 'a', 'n', 's', 'i', 'o', 'n'};

    NdisZeroMemory(temp, sizeof(temp));

    GetSmall(SA, AA, temp, 6);
    NdisMoveMemory(concatenation, temp, 6);
    CurrPos += 6;

    GetLarge(SA, AA, temp, 6);
    NdisMoveMemory(&concatenation[CurrPos], temp, 6);
    CurrPos += 6;

    GetSmall(ANonce, SNonce, temp, 32);
    NdisMoveMemory(&concatenation[CurrPos], temp, 32);
    CurrPos += 32;

    GetLarge(ANonce, SNonce, temp, 32);
    NdisMoveMemory(&concatenation[CurrPos], temp, 32);
    CurrPos += 32;
    
    PRF(PMK, PMK_LEN, Prefix, 22, concatenation, 76 , output, len);
}

VOID CountGTK(
    IN UCHAR    *GMK,
    IN UCHAR    *GNonce,
    IN UCHAR    *AA,
    OUT UCHAR   *output,
    IN UINT     len)
{
    UCHAR   concatenation[76];
    UINT    CurrPos=0;
    UCHAR   temp[80];   
    UCHAR   Prefix[] = {'G', 'r', 'o', 'u', 'p', ' ',  'k', 'e', 'y', ' ', 
                        'e', 'x', 'p', 'a', 'n', 's', 'i', 'o', 'n'};

    NdisMoveMemory(&concatenation[CurrPos], AA, 6);
    CurrPos += 6;

    NdisMoveMemory(&concatenation[CurrPos], GNonce , 32);
    CurrPos += 32;

    PRF(GMK, PMK_LEN, Prefix, 19, concatenation, 38 , temp, len);
    NdisMoveMemory(output, temp, len);
}

VOID GetSmall(
    IN  PVOID   pSrc1,
    IN  PVOID   pSrc2,
    OUT PUCHAR  pOut,
    IN  ULONG   Length)
{
    PUCHAR  pMem1;
    PUCHAR  pMem2;
    ULONG   Index = 0;
    pMem1 = (PUCHAR) pSrc1;
    pMem2 = (PUCHAR) pSrc2;

    for (Index = 0; Index < Length; Index++)
    {
        if (pMem1[Index] != pMem2[Index])
        {
            if (pMem1[Index] > pMem2[Index])        
                NdisMoveMemory(pOut, pSrc2, Length);
            else
                NdisMoveMemory(pOut, pSrc1, Length);             

            break;
        }
    }
}

VOID GetLarge(
    IN  PVOID   pSrc1,
    IN  PVOID   pSrc2,
    OUT PUCHAR  pOut,
    IN  ULONG   Length)
{
    PUCHAR  pMem1;
    PUCHAR  pMem2;
    ULONG   Index = 0;
    pMem1 = (PUCHAR) pSrc1;
    pMem2 = (PUCHAR) pSrc2;

    for (Index = 0; Index < Length; Index++)
    {
        if (pMem1[Index] != pMem2[Index])
        {
            if (pMem1[Index] > pMem2[Index])        
                NdisMoveMemory(pOut, pSrc1, Length);
            else
                NdisMoveMemory(pOut, pSrc2, Length);             

            break;
        }
    }
}

// 802.1i  Annex F.9
VOID GenRandom(
    IN PRTMP_ADAPTER    pAd, 
    OUT UCHAR           *random)
{   
    INT             i, curr;
    UCHAR           local[80];
    UCHAR           result[80];
    LONGLONG        CurrentTime;
    UCHAR           prefix[] = {'I', 'n', 'i', 't', ' ', 'C', 'o', 'u', 'n', 't', 'e', 'r'};

    NdisZeroMemory(result, 80);
    NdisZeroMemory(local, 80);
    NdisMoveMemory(local, pAd->CurrentAddress, ETH_LENGTH_OF_ADDRESS);
    
    for (i = 0; i < 32; i++)
    {       
        curr =  ETH_LENGTH_OF_ADDRESS;
        CurrentTime = jiffies;
        NdisMoveMemory(local,  pAd->CurrentAddress, ETH_LENGTH_OF_ADDRESS);
        curr += ETH_LENGTH_OF_ADDRESS;
        NdisMoveMemory(&local[curr],  &CurrentTime, sizeof(CurrentTime));
        curr += sizeof(CurrentTime);
        NdisMoveMemory(&local[curr],  result, 32);
        curr += 32;
        NdisMoveMemory(&local[curr],  &i,  2);      
        curr += 2;
        PRF(pAd->PortCfg.Key_Counter, 32, prefix,12, local, curr, result, 32); 
    }

    for (i = 32; i > 0; i--)
    {   
        if (pAd->PortCfg.Key_Counter[i-1] == 0xff)
        {
            pAd->PortCfg.Key_Counter[i-1] = 0;
        }
        else
        {
            pAd->PortCfg.Key_Counter[i-1]++;
            break;
        }
    }
    NdisMoveMemory(random, result,  32);
}

/*
    ==========================================================================
    Description:
        ENCRYPT AES GTK before sending in EAPOL frame.
        AES GTK length = 128 bit,  so fix blocks for aes-key-wrap as 2 in this function.
        This function references to RFC 3394 for aes key wrap algorithm.
    Return:
    ==========================================================================
*/  
VOID AES_GTK_KEY_WRAP( 
    IN UCHAR    *key,
    IN UCHAR    *plaintext,
    IN UCHAR	p_len,
    OUT UCHAR   *ciphertext)
{
    UCHAR		A[8], BIN[16], BOUT[16];
    UCHAR		R[512];
    INT			num_blocks = p_len/8;	// unit:64bits
    INT			i, j;
    aes_context	aesctx;
    UCHAR		xor;

    aes_set_key(&aesctx, key, 128);

	// Init IA
	for (i = 0; i < 8; i++)
        A[i] = 0xa6;

	//Input plaintext
	for (i = 0; i < num_blocks; i++)
	{
		for (j = 0 ; j < 8; j++)
			R[8 * (i + 1) + j] = plaintext[8 * i + j];
	}

	// Key Mix
	for (j = 0; j < 6; j++)
	{
		for(i = 1; i <= num_blocks; i++)
		{
			//phase 1
			memcpy(BIN, A, 8);
			memcpy(&BIN[8], &R[8 * i], 8);
			aes_encrypt(&aesctx, BIN, BOUT);

			memcpy(A, &BOUT[0], 8);
			xor = num_blocks * j + i;
			A[7] = BOUT[7] ^ xor;
			memcpy(&R[8 * i], &BOUT[8], 8);
		}
	}

	// Output ciphertext
	memcpy(ciphertext, A, 8);

	for (i = 1; i <= num_blocks; i++)
	{
		for (j = 0 ; j < 8; j++)
			ciphertext[8 * i + j] = R[8 * i + j];
	}
}

/*
    ========================================================================
    Routine Description:
       Send all EAP frames to wireless station.
       These frames don't come from normal SendPackets routine, but are EAPPacket, EAPOL, 
        
    Arguments:
        pRxD        Pointer to the Rx descriptor
        
    Return Value:
None
========================================================================
*/
VOID RTMPToWirelessSta(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PUCHAR          pFrame,
    IN  UINT            FrameLen)
{
    SST             Sst;
    USHORT          Aid;
    UCHAR           PsMode, Rate;
    MAC_TABLE_ENTRY *pEntry;
    struct sk_buff  *skb;

    // decide the return value. TRUE if no need to indicate to LLC, FALSE otherwise
    pEntry = SsPsInquiry(pAdapter, (MACADDR *)pFrame, &Sst, &Aid, &PsMode, &Rate);
    
    if (pEntry && (Sst == SST_ASSOC))
    {
        NDIS_STATUS     Status;

        DBGPRINT(RT_DEBUG_TRACE, "RTMPToWirelessSta ====>> to I/F(ra%d)\n", pEntry->ApIdx);

        do {
            // 1. build a NDIS packet and call RTMPSendPacket();
            //    be careful about how/when to release this internal allocated NDIS PACKET buffer
#ifdef RTMP_EMBEDDED
            if ((skb = __dev_alloc_skb(FrameLen + 2, GFP_DMA|GFP_ATOMIC)) != NULL)
#else
            if ((skb = dev_alloc_skb(FrameLen + 2)) != NULL)
#endif
            {
                skb->len = FrameLen;
                memcpy((skb->data), pFrame, FrameLen);
            }
            else
            {
                break;
            }

            // 1.1 apidx != 0, then we need set packet mssid attribute.
            if(pEntry->ApIdx != 0)
                RTMP_SET_PACKET_NET_DEVICE_MSSID(skb, pEntry->ApIdx);

            // 2. send out the packet
            Status = RTMPSendPacket(pAdapter, skb);
            if (Status == NDIS_STATUS_SUCCESS)
            {
                // Dequeue one frame from TxSwQueue0..3 queue and process it
                // There are three place calling dequeue for TX ring.
                // 1. Here, right after queueing the frame.
                // 2. At the end of TxRingTxDone service routine.
                // 3. Upon NDIS call RTMPSendPackets
                if ((!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && (!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
                {
                    RTMPDeQueuePacket(pAdapter);
                }   
            }
            else // free this packet space
            {
                RTMPFreeSkbBuffer(skb);
            }
        } while (FALSE);
    }
}

/*
    ========================================================================

    Routine Description:
        Sending EAP Req. frame to station in authenticating state.
        These frames come from Authenticator deamon.

    Arguments:
        pAdapter        Pointer to our adapter
        pPacket     Pointer to outgoing EAP frame body + 8023 Header
        Len             length of pPacket
        
    Return Value:
        None
    ========================================================================
*/
VOID WpaSend(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  PUCHAR          pPacket,
    IN  ULONG           Len)
{
    PEAP_HDR        pEapHdr;
    MACADDR         Addr;
    MAC_TABLE_ENTRY *pEntry;
    
    DBGPRINT(RT_DEBUG_TRACE, "WpaSend ==> Len=%d\n",Len);

    NdisMoveMemory(&(Addr.Octet), pPacket, 6);
    pEapHdr = (EAP_HDR*)(pPacket + LENGTH_802_3);
    if ((pEntry = MacTableLookup(pAdapter, &Addr)) == NULL)
        return;

    if (((pEntry->AuthMode >= Ndis802_11AuthModeWPA) && (pEapHdr->ProType != EAPOLKey)) ||
        (pAdapter->PortCfg.MSSID[pEntry->ApIdx].IEEE8021X == TRUE))
        RTMPToWirelessSta(pAdapter, pPacket, Len);

    if (RTMPEqualMemory((pPacket+12), EAPOL, 2) )
    {
        switch (pEapHdr->code)
        {
            case EAP_CODE_SUCCESS:
                if ((pEntry->AuthMode >= Ndis802_11AuthModeWPA) && (pEapHdr->ProType != EAPOLKey))
                {
                    DBGPRINT(RT_DEBUG_TRACE,"Send EAP_CODE_SUCCESS\n\n");
                    if (pEntry && (pEntry->Sst == SST_ASSOC))
                    {
                        pEntry->WpaState = AS_INITPMK;
						// Only set the expire and counters
	                    pEntry->RetryTimer.expires = jiffies + WPA_RETRY_EXEC_INTV;
	                    pEntry->ReTryCounter = PEER_MSG1_RETRY_TIMER_CTR;
	                    WPAStart4WayHS(pAdapter, pEntry);
                    }
                }
                else
                {
                    pEntry->PrivacyFilter = Ndis802_11PrivFilterAcceptAll;
                    pEntry->WpaState = AS_PTKINITDONE;
                    pAdapter->PortCfg.MSSID[pEntry->ApIdx].PortSecured = WPA_802_1X_PORT_SECURED;
                    pEntry->PortSecured = WPA_802_1X_PORT_SECURED;
                    DBGPRINT(RT_DEBUG_TRACE,"IEEE8021X-WEP : Send EAP_CODE_SUCCESS\n\n");
                }
                break;

            case EAP_CODE_FAILURE:
                break;

            default:
                break;    
        }
    }
    else     
    {
        DBGPRINT(RT_DEBUG_TRACE, "Send Deauth, Reason : REASON_NO_LONGER_VALID\n");
        DisAssocAction(pAdapter, pEntry, SUBTYPE_DEAUTH, REASON_NO_LONGER_VALID);
    }
}

VOID RTMPMakeRSNIE(
    IN  PRTMP_ADAPTER   pAdapter,
    IN  UINT            AuthMode,
    IN  UINT            WepStatus,
    IN  UCHAR           apidx)
{
	if (WepStatus == Ndis802_11Encryption4Enabled)
		pAdapter->PortCfg.MSSID[apidx].GroupKeyWepStatus = Ndis802_11Encryption2Enabled;
	else
		pAdapter->PortCfg.MSSID[apidx].GroupKeyWepStatus = WepStatus;

    if ((AuthMode != Ndis802_11AuthModeWPA) && (AuthMode != Ndis802_11AuthModeWPAPSK)
    	&& (AuthMode != Ndis802_11AuthModeWPA2) && (AuthMode != Ndis802_11AuthModeWPA2PSK)
		&& (AuthMode != Ndis802_11AuthModeWPA1WPA2) && (AuthMode != Ndis802_11AuthModeWPAPSKWPA2PSK))
		return;

	pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[0] = 0;
	pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[1] = 0;
    NdisZeroMemory(pAdapter->PortCfg.MSSID[apidx].RSN_IE[0], MAX_LEN_OF_RSNIE);
    NdisZeroMemory(pAdapter->PortCfg.MSSID[apidx].RSN_IE[1], MAX_LEN_OF_RSNIE);

	// For WPA1, RSN_IE=221
	if ((AuthMode == Ndis802_11AuthModeWPA) || (AuthMode == Ndis802_11AuthModeWPAPSK)
		|| (AuthMode == Ndis802_11AuthModeWPA1WPA2) || (AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
	{
		RSNIE				*pRsnie;
		RSNIE_AUTH			*pRsnie_auth;
		RSN_CAPABILITIES	*pRSN_Cap;
		UCHAR				Rsnie_size = 0;

		pRsnie = (RSNIE*)pAdapter->PortCfg.MSSID[apidx].RSN_IE[0];
	    NdisMoveMemory(pRsnie->oui, OUI_WPA_WEP40, 4);
	    pRsnie->version = 1;

	    switch (WepStatus)
	    {
	        case Ndis802_11Encryption2Enabled:
	            NdisMoveMemory(pRsnie->mcast, OUI_WPA_TKIP, 4);
	            pRsnie->ucount = 1;
	            NdisMoveMemory(pRsnie->ucast[0].oui, OUI_WPA_TKIP, 4);
	            Rsnie_size = sizeof(RSNIE);
	            break;

	        case Ndis802_11Encryption3Enabled:
	            NdisMoveMemory(pRsnie->mcast, OUI_WPA_CCMP, 4);
	            pRsnie->ucount = 1;
	            NdisMoveMemory(pRsnie->ucast[0].oui, OUI_WPA_CCMP, 4);
	            Rsnie_size = sizeof(RSNIE);
	            break;
	            
	        case Ndis802_11Encryption4Enabled:
	            NdisMoveMemory(pRsnie->mcast, OUI_WPA_TKIP, 4);
	            pRsnie->ucount = 2;
	            NdisMoveMemory(pRsnie->ucast[0].oui, OUI_WPA_TKIP, 4);
	            NdisMoveMemory(pRsnie->ucast[0].oui + 4, OUI_WPA_CCMP, 4);
	            Rsnie_size = sizeof(RSNIE) + 4;
	            break;
	    }

	    pRsnie_auth = (RSNIE_AUTH*)((PUCHAR)pRsnie + Rsnie_size);

	    switch (AuthMode)
	    {
	        case Ndis802_11AuthModeWPA:
	        case Ndis802_11AuthModeWPA1WPA2:
	            pRsnie_auth->acount = 1;
	            NdisMoveMemory(pRsnie_auth->auth[0].oui, OUI_WPA_WEP40, 4);
	            break;

	        case Ndis802_11AuthModeWPAPSK:
	        case Ndis802_11AuthModeWPAPSKWPA2PSK:
	            pRsnie_auth->acount = 1;
	            NdisMoveMemory(pRsnie_auth->auth[0].oui, OUI_WPA_TKIP, 4);
	            break;
	    }

	    pRSN_Cap = (RSN_CAPABILITIES*)((PUCHAR)pRsnie_auth + sizeof(RSNIE_AUTH));
	    
#ifdef BIG_ENDIAN
	    pRsnie->version = SWAP16(pRsnie->version);
	    pRsnie->ucount = SWAP16(pRsnie->ucount);
	    pRsnie_auth->acount = SWAP16(pRsnie_auth->acount);
#endif
	    
	    pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[0] = Rsnie_size + sizeof(RSNIE_AUTH) + sizeof(RSN_CAPABILITIES);
	}

	// For WPA2, RSN_IE=48, if WPA1WPA2/WPAPSKWPA2PSK mix mode, we store RSN_IE in RSN_IE[1] else RSNIE[0]
	if ((AuthMode == Ndis802_11AuthModeWPA2) || (AuthMode == Ndis802_11AuthModeWPA2PSK)
		|| (AuthMode == Ndis802_11AuthModeWPA1WPA2) || (AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
	{
		RSNIE2				*pRsnie2;
		RSNIE_AUTH			*pRsnie_auth2;
		RSN_CAPABILITIES	*pRSN_Cap;
		UCHAR				Rsnie_size = 0;

		if ((AuthMode == Ndis802_11AuthModeWPA1WPA2) || (AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
			pRsnie2 = (RSNIE2*)pAdapter->PortCfg.MSSID[apidx].RSN_IE[1];
		else
			pRsnie2 = (RSNIE2*)pAdapter->PortCfg.MSSID[apidx].RSN_IE[0];

	    pRsnie2->version = 1;

	    switch (WepStatus)
	    {
	        case Ndis802_11Encryption2Enabled:
	            NdisMoveMemory(pRsnie2->mcast, OUI_WPA2_TKIP, 4);
	            pRsnie2->ucount = 1;
	            NdisMoveMemory(pRsnie2->ucast[0].oui, OUI_WPA2_TKIP, 4);
	            Rsnie_size = sizeof(RSNIE2);
	            break;

	        case Ndis802_11Encryption3Enabled:
	            NdisMoveMemory(pRsnie2->mcast, OUI_WPA2_CCMP, 4);
	            pRsnie2->ucount = 1;
	            NdisMoveMemory(pRsnie2->ucast[0].oui, OUI_WPA2_CCMP, 4);
	            Rsnie_size = sizeof(RSNIE2);
	            break;

	        case Ndis802_11Encryption4Enabled:
	            NdisMoveMemory(pRsnie2->mcast, OUI_WPA2_TKIP, 4);
	            pRsnie2->ucount = 2;
	            NdisMoveMemory(pRsnie2->ucast[0].oui, OUI_WPA2_TKIP, 4);
	            NdisMoveMemory(pRsnie2->ucast[0].oui + 4, OUI_WPA2_CCMP, 4);
	            Rsnie_size = sizeof(RSNIE2) + 4;
	            break;
	    }

	    pRsnie_auth2 = (RSNIE_AUTH*)((PUCHAR)pRsnie2 + Rsnie_size);

	    switch (AuthMode)
	    {
	        case Ndis802_11AuthModeWPA2:
	        case Ndis802_11AuthModeWPA1WPA2:
	            pRsnie_auth2->acount = 1;
	            NdisMoveMemory(pRsnie_auth2->auth[0].oui, OUI_WPA2_WEP40, 4);
	            break;

	        case Ndis802_11AuthModeWPA2PSK:
	        case Ndis802_11AuthModeWPAPSKWPA2PSK:
	            pRsnie_auth2->acount = 1;
	            NdisMoveMemory(pRsnie_auth2->auth[0].oui, OUI_WPA2_TKIP, 4);
	            break;
	    }

	    pRSN_Cap = (RSN_CAPABILITIES*)((PUCHAR)pRsnie_auth2 + sizeof(RSNIE_AUTH));
	    pRSN_Cap->field.PreAuth = (pAdapter->PortCfg.MSSID[apidx].PreAuth == TRUE) ? 1 : 0;

#ifdef BIG_ENDIAN
	    pRsnie2->version = SWAP16(pRsnie2->version);
	    pRsnie2->ucount = SWAP16(pRsnie2->ucount);
	    pRsnie_auth2->acount = SWAP16(pRsnie_auth2->acount);
#endif

		if ((AuthMode == Ndis802_11AuthModeWPA1WPA2) || (AuthMode == Ndis802_11AuthModeWPAPSKWPA2PSK))
			pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[1] = Rsnie_size + sizeof(RSNIE_AUTH) + sizeof(RSN_CAPABILITIES);
		else
			pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[0] = Rsnie_size + sizeof(RSNIE_AUTH) + sizeof(RSN_CAPABILITIES);
	}

    DBGPRINT(RT_DEBUG_TRACE,"RTMPMakeRSNIE : RSNIE_Len[0] = %d, RSNIE_Len[1] = %d\n",
    	pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[0], pAdapter->PortCfg.MSSID[apidx].RSNIE_Len[1]);
}

VOID RTMPAddPMKIDCache(
	IN  PRTMP_ADAPTER   		pAd,
	IN	INT						apidx,
	IN	PMACADDR				pAddr,
	IN	UCHAR					*PMKID,
	IN	UCHAR					*PMK)
{
	INT	i, chcheidx;

	if ((chcheidx = RTMPSearchPMKIDCache(pAd, apidx, pAddr)) != -1)
	{
		pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[chcheidx].RefreshTime = jiffies;
		NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[chcheidx].PMKID, PMKID, LEN_PMKID);
		NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[chcheidx].PMK, PMK, PMK_LEN);
		DBGPRINT(RT_DEBUG_TRACE,"RTMPAddPMKIDCache update %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
           	pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], chcheidx, apidx);
		
		return;
	}
	
	for (i = 0; i < MAX_PMKID_COUNT; i++)
	{
		if (!pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].Valid)
		{
			pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].Valid = TRUE;
			pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].RefreshTime = jiffies;
			COPY_MAC_ADDR(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].MAC, &pAddr->Octet[0]);
			NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].PMKID, PMKID, LEN_PMKID);
			NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].PMK, PMK, PMK_LEN);
			DBGPRINT(RT_DEBUG_TRACE,"RTMPAddPMKIDCache add %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
            	pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], i, apidx);
			break;
		}
	}

	if (i == MAX_PMKID_COUNT)
	{
		ULONG	timestamp = 0, idx = 0;

		DBGPRINT(RT_DEBUG_TRACE, "RTMPAddPMKIDCache(I/F(%d) Cache full\n", apidx);
		for (i = 0; i < MAX_PMKID_COUNT; i++)
		{
			if (pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].Valid)
			{
				if (((timestamp == 0) && (idx == 0)) || ((timestamp != 0) && timestamp < pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].RefreshTime))
				{
					timestamp = pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].RefreshTime;
					idx = i;
				}
			}
		}
		pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].Valid = TRUE;
		pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].RefreshTime = jiffies;
		COPY_MAC_ADDR(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].MAC, &pAddr->Octet[0]);
		NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].PMKID, PMKID, LEN_PMKID);
		NdisMoveMemory(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].PMK, PMK, PMK_LEN);
		DBGPRINT(RT_DEBUG_TRACE,"RTMPAddPMKIDCache add %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
           	pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], idx, apidx);
	}
}

INT RTMPSearchPMKIDCache(
	IN  PRTMP_ADAPTER   pAd,
	IN	INT				apidx,
	IN	PMACADDR		pAddr)
{
	INT	i = 0;
	
	for (i = 0; i < MAX_PMKID_COUNT; i++)
	{
		if ((pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].Valid)
			&& MAC_ADDR_EQUAL(&pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[i].MAC, &pAddr->Octet[0]))
		{
			DBGPRINT(RT_DEBUG_TRACE,"RTMPSearchPMKIDCache %02x:%02x:%02x:%02x:%02x:%02x cache(%d) from I/F(ra%d)\n", 
            	pAddr->Octet[0], pAddr->Octet[1], pAddr->Octet[2], pAddr->Octet[3], pAddr->Octet[4], pAddr->Octet[5], i, apidx);
			break;
		}
	}

	if (i == MAX_PMKID_COUNT)
	{
		DBGPRINT(RT_DEBUG_TRACE, "RTMPSearchPMKIDCache - I/F(%d) not found\n", apidx);
		return -1;
	}

	return i;
}

VOID RTMPDeletePMKIDCache(
	IN  PRTMP_ADAPTER   pAd,
	IN	INT				apidx,
	IN  INT				idx)
{
	if (pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].Valid)
	{
		pAd->PortCfg.MSSID[apidx].PMKIDCache.BSSIDInfo[idx].Valid = FALSE;
		DBGPRINT(RT_DEBUG_TRACE, "RTMPDeletePMKIDCache(I/F(%d), del PMKID CacheIdx=%d\n", apidx, idx);
	}
}

VOID RTMPMaintainPMKIDCache(
	IN  PRTMP_ADAPTER   pAd)
{
	INT	i, j;
	
	for (i = 0; i < MAX_MSSID_NUM; i++)
	{
		for (j = 0; j < MAX_PMKID_COUNT; j++)
		{
			if ((pAd->PortCfg.MSSID[i].PMKIDCache.BSSIDInfo[j].Valid)
				&& ((jiffies - pAd->PortCfg.MSSID[i].PMKIDCache.BSSIDInfo[j].RefreshTime) >= pAd->PortCfg.PMKCachePeriod))
			{
				RTMPDeletePMKIDCache(pAd, i, j);
			}
		}
	}
}

