/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

	Module Name:
	rtmp_info.c

	Abstract:
	IOCTL related subroutines

	Revision History:
	Who         When          What
	--------    ----------    ----------------------------------------------
	Rory Chen   01-03-2003    created

*/

#include	"rt_config.h"


/* The frequency of each channel in MHz */
const long channel_frequency[] = {
        2412, 2417, 2422, 2427, 2432, 2437, 2442,
        2447, 2452, 2457, 2462, 2467, 2472, 2484
};
#define NUM_CHANNELS ( sizeof(channel_frequency) / sizeof(channel_frequency[0]) )


static const USHORT RateIdTo500Kbps2[] = {2, 4, 11, 22, 12, 18, 24, 36, 48, 72, 96, 108, 144, 200};
#define NUM_RateIdTo500Kbps ( sizeof(RateIdTo500Kbps2) / sizeof(RateIdTo500Kbps2[0]) )

struct iw_priv_args privtab[] = {
{ RTPRIV_IOCTL_SET, 
  IW_PRIV_TYPE_CHAR | 1024, 0,
  "set"},  
{ RTPRIV_IOCTL_SITESURVEY,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024,
  "site_survey"},
{ RTPRIV_IOCTL_GSITESURVEY,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024 ,
  "get_site_survey"}, 
{ RTPRIV_IOCTL_GET_MAC_TABLE,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024 ,
  "get_mac_table"}, 
#ifdef RALINK_ATE
{ RTPRIV_IOCTL_BBP,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024,
  "bbp"},
{ RTPRIV_IOCTL_MAC,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024,
  "mac"},
{ RTPRIV_IOCTL_E2P,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024,
  "e2p"},
{ RTPRIV_IOCTL_STATISTICS,
  IW_PRIV_TYPE_CHAR | 1024, IW_PRIV_TYPE_CHAR | 1024,
  "stat"}
#endif
};


static struct {
	char *name;
	int (*set_proc)(PRTMP_ADAPTER pAdapter, PUCHAR arg);
} *PRTMP_PRIVATE_SET_PROC, RTMP_PRIVATE_SUPPORT_PROC[] = {
	{"CountryRegion", Set_CountryRegion_Proc },
	{"SSID", Set_SSID_Proc},
	{"WirelessMode", Set_WirelessMode_Proc},
	{"TxRate", Set_TxRate_Proc},
   	{"BasicRate", Set_BasicRate_Proc},
	{"Channel", Set_Channel_Proc},
	{"BeaconPeriod", Set_BeaconPeriod_Proc},
   	{"DtimPeriod", Set_DtimPeriod_Proc},
   	{"TxPower", Set_TxPower_Proc},
	{"BGProtection", Set_BGProtection_Proc},
	{"TxAntenna", Set_TxAntenna_Proc},
	{"RxAntenna", Set_RxAntenna_Proc},
	{"TxPreamble", Set_TxPreamble_Proc},
	{"RTSThreshold", Set_RTSThreshold_Proc},
	{"FragThreshold", Set_FragThreshold_Proc},
	{"TxBurst", Set_TxBurst_Proc},
	{"TurboRate", Set_TurboRate_Proc},
	{"NoForwardingBTNSSID", Set_NoForwardingBTNSSID_Proc},
	{"NoForwarding", Set_NoForwarding_Proc},
	{"HideSSID", Set_HideSSID_Proc},
	{"ShortSlot", Set_ShortSlot_Proc},
	{"IEEE8021X", Set_IEEE8021X_Proc},
	{"PreAuth", Set_PreAuth_Proc},
	{"VLANID", Set_VLANID_Proc},
	{"VLANPriority", Set_VLANPriority_Proc},
	{"AuthMode", Set_AuthMode_Proc},
	{"EncrypType", Set_EncrypType_Proc},
	{"DefaultKeyID", Set_DefaultKeyID_Proc},
	{"Key1", Set_Key1_Proc},
	{"Key2", Set_Key2_Proc},
	{"Key3", Set_Key3_Proc},
	{"Key4", Set_Key4_Proc},
	{"HSCounter", Set_HSCounter_Proc},
	{"AccessPolicy", Set_AccessPolicy_Proc},
	{"AccessControlList", Set_AccessControlList_Proc},
	{"WPAPSK", Set_WPAPSK_Proc},
	{"ResetCounter", Set_ResetStatCounter_Proc},
	{"RekeyMethod", Set_RekeyMethod_Proc},
	{"RekeyInterval", Set_RekeyInterval_Proc},
	{"PMKCachePeriod", Set_PMKCachePeriod_Proc},
#ifdef VINCENT_K
	{"Rx_Calibration", Set_Rx_Calibration_Proc},
	{"Tx_Cont", Set_Tx_Cont_Proc},
	{"Tx_Start", Set_Tx_Start_Proc},
	{"Rx_Start", Set_Rx_Start_Proc},
	{"InitAsicFromEEPROM",Set_InitAsicFromEEPROM},
#endif
#ifdef RT2500_DBG	
	{"Debug", Set_Debug_Proc},
#endif

#ifdef RALINK_ATE
	{"ATE",       Set_ATE_Proc		},	// set ATE Mode to: STOP, TXCONT, TXCARR, TXFRAME, RXFRAME
	{"ATEDA",     Set_ATE_DA_Proc		},	// set ATE TxFrames ADDR1, DA
	{"ATESA",     Set_ATE_SA_Proc		},	// set ATE TxFrames ADDR2, SA
	{"ATEBSSID",  Set_ATE_BSSID_Proc	},	// set ATE TxFrames ADDR3, BSSID
    {"ATECHANNEL",  Set_ATE_CHANNEL_Proc	},	// set ATE Channel
	{"ATETXPOW",  Set_ATE_TX_POWER_Proc	},	// set ATE TxPower
	{"ATETXLEN",  Set_ATE_TX_LENGTH_Proc},	// set ATE TxLength
	{"ATETXCNT",  Set_ATE_TX_COUNT_Proc	},	// set ATE TxCount
	{"ATETXRATE", Set_ATE_TX_RATE_Proc	},	// set ATE TxRate
#endif	//#ifdef RALINK_ATE

	{NULL,}
};

INT RTMPSetInformation(
	IN	PRTMP_ADAPTER pAdapter,
	IN	OUT	struct ifreq	*rq,
	IN	INT					cmd)
{
	struct iwreq						*wrq = (struct iwreq *) rq;
	INT									Status = NDIS_STATUS_SUCCESS;
	ULONG								phymode;
	BOOLEAN								RestartAPIsRequired = FALSE;
	NDIS_802_11_CONFIGURATION			Configuration;
	NDIS_802_11_ANTENNA					Antenna;
	RT_802_11_PREAMBLE					Preamble;
	NDIS_802_11_RTS_THRESHOLD           RtsThresh;
	NDIS_802_11_FRAGMENTATION_THRESHOLD FragThresh;
	RT_802_11_AP_CONFIG                 ApConfig, *pApConfig=NULL;
	NDIS_802_11_AUTHENTICATION_MODE		AuthMode;
	NDIS_802_11_WEP_STATUS				WepStatus;
	NDIS_802_11_REMOVE_KEY				RemoveKey, *pRemoveKey=NULL;
	NDIS_802_11_KEY						WepKey, *pKey=NULL;
	NDIS_802_11_SSID					Ssid, *pSsid=NULL;	
	RT_802_11_ACL						Acl;
	UCHAR                               CountryRegion;
	ULONG                               KeyIdx;
	UCHAR								TxValue;
	UCHAR								RxValue;
	RT_WPA_REKEY						WPA_REKEY;
	UCHAR                               apidx, i;
	MAC_TABLE_ENTRY                     *pEntry;

	switch(cmd & 0x7FFF) {
		case RT_OID_802_11_COUNTRY_REGION:
			if (wrq->u.data.length != sizeof(CountryRegion))
				Status = -EINVAL;
			else
			{
				copy_from_user(&CountryRegion, wrq->u.data.pointer, wrq->u.data.length);			
				pAdapter->PortCfg.CountryRegion = CountryRegion;
				DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_COUNTRY_REGION (=%d) \n", pAdapter->PortCfg.CountryRegion)
			}
			break;
		case RT_OID_802_11_RESET_COUNTERS:
			NdisZeroMemory(&pAdapter->WlanCounters, sizeof(COUNTER_802_11));
			NdisZeroMemory(&pAdapter->Counters, sizeof(COUNTER_802_3));
			NdisZeroMemory(&pAdapter->RalinkCounters, sizeof(COUNTER_RALINK));
			// Reset HotSpot counter
			for (apidx = 0; apidx < pAdapter->PortCfg.SsidNum; apidx++)
			{
			    for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
			    {
			        pEntry = &pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i];
			        if ((pEntry->Valid == FALSE) || (pEntry->Sst != SST_ASSOC))
			            continue;

			        pEntry->HSCounter.LastDataPacketTime = 0;
			        pEntry->HSCounter.TotalRxByteCount= 0;
			        pEntry->HSCounter.TotalTxByteCount= 0;
			    }
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_RESET_COUNTERS \n")
			break;
		case RT_OID_802_11_ACL:
			if (wrq->u.data.length != sizeof(RT_802_11_ACL))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&Acl, wrq->u.data.pointer, wrq->u.data.length);
				NdisMoveMemory(&pAdapter->PortCfg.AccessControlList, &Acl, sizeof(RT_802_11_ACL));
				if (pAdapter->PortCfg.AccessControlList.Num > MAX_LEN_OF_MAC_TABLE)
					pAdapter->PortCfg.AccessControlList.Num = MAX_LEN_OF_MAC_TABLE;

				// check if ACL change affects any existent associtions
				ApUpdateAccessControlList(pAdapter);
				DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_ACL (Policy=%d, Entry#=%d\n",
							pAdapter->PortCfg.AccessControlList.Policy, pAdapter->PortCfg.AccessControlList.Num);
			}
			break;
		case OID_802_11_SSID:
			if (wrq->u.data.length != sizeof(NDIS_802_11_SSID))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&Ssid, wrq->u.data.pointer, wrq->u.data.length);
				pSsid = &Ssid;

				if(pSsid->SsidLength > MAX_LEN_OF_SSID)
					Status = -EINVAL;
				else
				{
					NdisZeroMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, MAX_LEN_OF_SSID);
					NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, pSsid->Ssid, pSsid->SsidLength);
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen = (UCHAR)pSsid->SsidLength;
				}
				DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_SSID (Len=%d,Ssid=%s)\n", pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid);
			}
			RestartAPIsRequired = TRUE;
			break;

		case RT_OID_802_11_WPA2_ADD_PMKID_CACHE:
			copy_from_user(&WepKey, wrq->u.data.pointer, wrq->u.data.length);
			pKey = &WepKey;

			if (pAdapter->PortCfg.MSSID[MAIN_MSSID].AuthMode >= Ndis802_11AuthModeWPA2)
			{
				if(pKey->KeyLength == 32)
				{
					UCHAR	digest[80], PMK_key[20];
					MACADDR	macaddr;
					
					// Calculate PMKID
					NdisMoveMemory(&PMK_key[0], "PMK Name", 8);
					NdisMoveMemory(&PMK_key[8], pAdapter->CurrentAddress, MAC_ADDR_LEN);
					NdisMoveMemory(&PMK_key[14], pKey->addr, MAC_ADDR_LEN);
					HMAC_SHA1(PMK_key, 20, pKey->KeyMaterial, PMK_LEN, digest);

					NdisMoveMemory(&macaddr.Octet[0], pKey->addr, MAC_ADDR_LEN);
					RTMPAddPMKIDCache(pAdapter, MAIN_MSSID, &macaddr, digest, pKey->KeyMaterial);
					
					DBGPRINT(RT_DEBUG_TRACE, "WPA2(pre-auth):(%02x:%02x:%02x:%02x:%02x:%02x)Calc PMKID=%02x:%02x:%02x:%02x:%02x:%02x\n", 
						pKey->addr[0],pKey->addr[1],pKey->addr[2],pKey->addr[3],pKey->addr[4],pKey->addr[5],digest[0],digest[1],digest[2],digest[3],digest[4],digest[5]);
					DBGPRINT(RT_DEBUG_TRACE, "After ADD_KEY : PMK = %x %x %x %x \n",pKey->KeyMaterial[0],pKey->KeyMaterial[1],pKey->KeyMaterial[2],pKey->KeyMaterial[3]);
				}
				else
                    DBGPRINT(RT_DEBUG_ERROR, "Set::RT_OID_802_11_WPA2_ADD_PMKID_CACHE ERROR or is wep key \n");
			}
			break;
			
		case OID_802_11_ADD_KEY:
			copy_from_user(&WepKey, wrq->u.data.pointer, wrq->u.data.length);
			pKey = &WepKey;

			if (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode >= Ndis802_11AuthModeWPA)
			{
				if(pKey->KeyLength == 32)
				{
					NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK, pKey->KeyMaterial, 32);
                    DBGPRINT(RT_DEBUG_TRACE, "After ADD_KEY : PMK = %x %x %x %x \n",pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK[0],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK[1],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK[2],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK[3]);
				}
				else
                    DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_ADD_KEY ERROR \n");
			}
			else	// Old WEP stuff
			{
				KeyIdx = pKey->KeyIndex & 0x0fffffff;

				// it is a shared key
				if (KeyIdx > 4)
					Status = -EINVAL;
				else
				{
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[KeyIdx].KeyLen = (UCHAR) pKey->KeyLength;
					NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[KeyIdx].Key, &pKey->KeyMaterial, pKey->KeyLength);
					if (pKey->KeyIndex & 0x80000000)
					{
						// Default key for tx (shared key)
						pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId = (UCHAR) KeyIdx;								
					}
				}
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_ADD_KEY (info-len=%d, id=0x%x, Len=%d-byte)\n", wrq->u.data.length, pKey->KeyIndex, pKey->KeyLength);
			break;
		case OID_802_11_REMOVE_KEY:
			if (wrq->u.data.length != sizeof(NDIS_802_11_REMOVE_KEY))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&RemoveKey, wrq->u.data.pointer, wrq->u.data.length);
				pRemoveKey = &RemoveKey;
				if (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode >= Ndis802_11AuthModeWPA)
				{
				}
				else
				{
					KeyIdx = pRemoveKey->KeyIndex;

					if (KeyIdx & 0x80000000)
					{
						// Should never set default bit when remove key
						Status = -EINVAL;
					}
					else
					{
						KeyIdx = KeyIdx & 0x0fffffff;
						if (KeyIdx > 4)
							Status = -EINVAL;
						else
							pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[KeyIdx].KeyLen = 0;
					}
				}
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_REMOVE_KEY (id=0x%x, Len=%d-byte)\n", pRemoveKey->KeyIndex, pRemoveKey->Length);
			break;
		case OID_802_11_WEP_STATUS:
			if (wrq->u.data.length != sizeof(NDIS_802_11_WEP_STATUS))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&WepStatus, wrq->u.data.pointer, wrq->u.data.length);
				// Since TKIP, AES, WEP are all supported. It should not have any invalid setting
				if (WepStatus <= Ndis802_11Encryption4KeyAbsent)
				{
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = WepStatus;
				}
				else
					Status = -EINVAL;
                RTMPMakeRSNIE(pAdapter, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus, pAdapter->PortCfg.IoctlIF);
				DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_WEP_STATUS (=%d)\n",WepStatus);
			}
			break;
		case OID_802_11_AUTHENTICATION_MODE:
			if (wrq->u.data.length != sizeof(NDIS_802_11_AUTHENTICATION_MODE))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&AuthMode, wrq->u.data.pointer, wrq->u.data.length);
				if (AuthMode > Ndis802_11AuthModeMax)
					Status = -EINVAL;
				else
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = AuthMode;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PortSecured = WPA_802_1X_PORT_NOT_SECURED;
            RTMPMakeRSNIE(pAdapter, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus, pAdapter->PortCfg.IoctlIF);
			DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_AUTHENTICATION_MODE (=%d) \n",pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode);
			break;
		case RT_OID_802_11_AP_CONFIG:
			if (wrq->u.data.length != sizeof(RT_802_11_AP_CONFIG))
				Status = -EINVAL;
			else 
			{
				copy_from_user(&ApConfig, wrq->u.data.pointer, wrq->u.data.length);
				pApConfig = &ApConfig;
				pAdapter->PortCfg.EnableTxBurst = pApConfig->EnableTxBurst;
				pAdapter->PortCfg.EnableTurboRate = pApConfig->EnableTurboRate;
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IsolateInterStaTraffic = pApConfig->IsolateInterStaTraffic;
				if (pAdapter->PortCfg.HideSsid != pApConfig->HideSsid)
				{
					pAdapter->PortCfg.HideSsid = pApConfig->HideSsid;
					// re-built BEACON frame format
					AsicDisableSync(pAdapter);
					MakeBssBeacon(pAdapter);
					UpdateBeaconFrame(pAdapter);
					AsicEnableBssSync(pAdapter);
				}
				pAdapter->PortCfg.UseBGProtection = pApConfig->UseBGProtection;
				pAdapter->PortCfg.UseShortSlotTime = pApConfig->UseShortSlotTime;
				ApUpdateCapabilityAndErpIe(pAdapter);
				DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_AP_CONFIG (Burst=%d,72/100=%d,Hotspot=%d,HideSsid=%d,Protection=%d,ShortSlot=%d\n",
									pApConfig->EnableTxBurst,
									pApConfig->EnableTurboRate,
									pApConfig->IsolateInterStaTraffic,
									pApConfig->HideSsid,
									pApConfig->UseBGProtection,
									pApConfig->UseShortSlotTime);
			}
			break;
		case OID_802_11_FRAGMENTATION_THRESHOLD:
			if (wrq->u.data.length != sizeof(NDIS_802_11_FRAGMENTATION_THRESHOLD)) 
				Status = -EINVAL;
			else
			{
				copy_from_user(&FragThresh, wrq->u.data.pointer, wrq->u.data.length);
				pAdapter->PortCfg.bFragmentZeroDisable = FALSE;
				if (FragThresh > MAX_FRAG_THRESHOLD || FragThresh < MIN_FRAG_THRESHOLD)
					if (FragThresh == 0)
					{
						pAdapter->PortCfg.FragmentThreshold = MAX_FRAG_THRESHOLD;
						pAdapter->PortCfg.bFragmentZeroDisable = TRUE;
					}
					else
						Status = -EINVAL;
				else
					pAdapter->PortCfg.FragmentThreshold = (USHORT)FragThresh;
				pAdapter->PortCfg.AllowFragSize = (pAdapter->PortCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;
				DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_FRAGMENTATION_THRESHOLD (=%d) \n",FragThresh);
			}					
			break;
		case OID_802_11_RTS_THRESHOLD:
			if (wrq->u.data.length != sizeof(NDIS_802_11_RTS_THRESHOLD)) 
				Status = -EINVAL;
			else
			{
				copy_from_user(&RtsThresh, wrq->u.data.pointer, wrq->u.data.length);
				if (RtsThresh > MAX_RTS_THRESHOLD)
					Status = -EINVAL;
				else
					pAdapter->PortCfg.RtsThreshold = (USHORT)RtsThresh;
				DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_RTS_THRESHOLD (=%d)\n",RtsThresh);
			}
			break;
		case RT_OID_802_11_PREAMBLE:
			if (wrq->u.data.length != sizeof(RT_802_11_PREAMBLE)) 
				Status = -EINVAL;
			else
			{
			    copy_from_user(&Preamble, wrq->u.data.pointer, wrq->u.data.length);
			    if (Preamble == Rt802_11PreambleShort)
			    {
			        pAdapter->PortCfg.TxPreamble = Rt802_11PreambleShort;
			        // MlmeSetTxPreamble(pAdapter, Rt802_11PreambleShort);
			    }
			    else 
			    {
			        pAdapter->PortCfg.TxPreamble = Rt802_11PreambleLong;
			        // MlmeSetTxPreamble(pAdapter, Rt802_11PreambleLong);
			    }
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_PREAMBLE (=%d)\n", Preamble);
			break;
		case OID_802_11_RX_ANTENNA_SELECTED:
			if (wrq->u.data.length != sizeof(NDIS_802_11_ANTENNA)) 
				Status = -EINVAL;
			else
			{
				copy_from_user(&Antenna, wrq->u.data.pointer, wrq->u.data.length);
				if(Antenna == 0xFFFFFFFF) 
				{// Diversity
					pAdapter->PortCfg.CurrentRxAntenna = (UCHAR)Antenna;
					RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Rx_Configure, &RxValue);
					RxValue = (RxValue & 0xFC) | 0x01;
					RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Rx_Configure, RxValue);
				}
				else if(Antenna >= pAdapter->PortCfg.NumberOfAntenna)
					Status = -EINVAL;
				else
				{
					pAdapter->PortCfg.CurrentRxAntenna = (UCHAR)Antenna;

					RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Rx_Configure, &RxValue);
					if(Antenna == 0)
					{// Antenna A
						RxValue = (RxValue & 0xFC) | 0x00;
					}
					else if(Antenna == 1)
					{// Antenna B
						RxValue = (RxValue & 0xFC) | 0x02;
					}
					RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Rx_Configure, RxValue);
				}
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_RX_ANTENNA_SELECTED (=%d)\n",Antenna);
			break;
		case OID_802_11_TX_ANTENNA_SELECTED:
			if (wrq->u.data.length != sizeof(NDIS_802_11_ANTENNA)) 
				Status = -EINVAL;
			else
			{
				ULONG BbpCsr1;

				RTMP_IO_READ32(pAdapter, BBPCSR1, &BbpCsr1);
				copy_from_user(&Antenna, wrq->u.data.pointer, wrq->u.data.length);
				if (Antenna == 0xFFFFFFFF) // Diversity
				{
					pAdapter->PortCfg.CurrentTxAntenna = (UCHAR)Antenna;
					RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Tx_Configure, &TxValue);
					TxValue = (TxValue & 0xFC) | 0x01;
					RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Tx_Configure, TxValue);
					BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00010001;
					RTMP_IO_WRITE32(pAdapter, BBPCSR1, BbpCsr1);
				}
				else if (Antenna >= pAdapter->PortCfg.NumberOfAntenna)
					Status = -EINVAL;
				else
				{
					pAdapter->PortCfg.CurrentTxAntenna = (UCHAR)Antenna;

					RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Tx_Configure, &TxValue);
					if(Antenna == 0) // Antenna A
					{
						TxValue = (TxValue & 0xFC) | 0x00;
						BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00000000;
					}
					else if(Antenna == 1) // Antenna B
					{
						TxValue = (TxValue & 0xFC) | 0x02;
						BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00020002;
					}
					RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Tx_Configure, TxValue);
					RTMP_IO_WRITE32(pAdapter, BBPCSR1, BbpCsr1);
				}
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_TX_ANTENNA_SELECTED (=0x%08x) \n",Antenna);
			break;
		case OID_802_11_CONFIGURATION:
			if (wrq->u.data.length != sizeof(NDIS_802_11_CONFIGURATION))
				Status = -EINVAL;
			else
			{
				copy_from_user(&Configuration, wrq->u.data.pointer, wrq->u.data.length);
				pAdapter->PortCfg.BeaconPeriod = (USHORT) Configuration.BeaconPeriod;
				//pAdapter->PortCfg.AtimWin = (USHORT) Configuration.ATIMWindow;
				if (Configuration.DSConfig == 2484000)
					pAdapter->PortCfg.Channel = 14;
				else if (Configuration.DSConfig >= 2412000 && Configuration.DSConfig < 2484000) 
					pAdapter->PortCfg.Channel = (UCHAR)((Configuration.DSConfig - 2412000) / 5000) + 1;
				else
					Status = -EINVAL;

				RestartAPIsRequired = TRUE;
			}
		    DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_CONFIGURATION (BeacnPeriod=%d,AtimW=%d,Ch=%d)\n",
		        Configuration.BeaconPeriod, Configuration.ATIMWindow, pAdapter->PortCfg.Channel);
		    break;
		case RT_OID_802_11_PHY_MODE:
			copy_from_user(&phymode, wrq->u.data.pointer, wrq->u.data.length);
			if (phymode != pAdapter->PortCfg.PhyMode)
			{
				RTMPSetPhyMode(pAdapter, phymode);
				RestartAPIsRequired = TRUE;
			}
			DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_PHY_MODE (=%d)\n", phymode);
			break;
		case OID_802_11_DESIRED_RATES:
			if (wrq->u.data.length != sizeof(NDIS_802_11_RATES))
				Status = -EINVAL;
			else 
			{
				NdisZeroMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates, MAX_LEN_OF_SUPPORTED_RATES);
				copy_from_user(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates, wrq->u.data.pointer, wrq->u.data.length);						
				DBGPRINT(RT_DEBUG_TRACE, "Set::OID_802_11_DESIRED_RATES (%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x)\n", 
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1],
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3],
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5],
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] );
				MlmeUpdateTxRates(pAdapter);
			}
			break;
        case RT_OID_802_11_RADIUS_DATA:
            {
                DBGPRINT(RT_DEBUG_TRACE, "SET::RT_OID_802_11_RADIUS_DATA,I/F(ra%d),wrq->u.data.length=%d,wrq->u.data=%02x:%02x\n",
                	pAdapter->PortCfg.IoctlIF,wrq->u.data.length,*(UCHAR*)(wrq->u.data.pointer+12),*(UCHAR*)(wrq->u.data.pointer+13));

                if ((pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode == Ndis802_11AuthModeWPA) 
                	|| (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode == Ndis802_11AuthModeWPA2)
                	|| (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode == Ndis802_11AuthModeWPA1WPA2) 
                	|| (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IEEE8021X == TRUE))
                    WpaSend(pAdapter, wrq->u.data.pointer, wrq->u.data.length);
                break;
            }
		case RT_OID_802_11_WPA_REKEY:
            if (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode < Ndis802_11AuthModeWPA)
            {
                if(pAdapter->PortCfg.REKEYTimerRunning == TRUE)
                {
                    del_timer_sync(&pAdapter->PortCfg.REKEYTimer);
                    pAdapter->PortCfg.REKEYTimerRunning = FALSE;
                }
                break;
            }
            copy_from_user(&WPA_REKEY, wrq->u.data.pointer, wrq->u.data.length);
            // mechanism for rekeying: 0:disable, 1: time-based, 2: packet-based
            if (WPA_REKEY.ReKeyMethod >MAX_REKEY)
                Status = -EINVAL;
            else if (WPA_REKEY.ReKeyInterval > MAX_REKEY_INTER)
                Status = -EINVAL;
            else if ((WPA_REKEY.ReKeyMethod == DISABLE_REKEY) ||(WPA_REKEY.ReKeyInterval == 0))
            {
                DBGPRINT(RT_DEBUG_TRACE, "  DISABLE WPA_REKEY      \n");
                pAdapter->PortCfg.WPAREKEY.ReKeyInterval = WPA_REKEY.ReKeyInterval;                
                pAdapter->PortCfg.WPAREKEY.ReKeyMethod = DISABLE_REKEY;
                if(pAdapter->PortCfg.REKEYTimerRunning == TRUE)
                {
                    del_timer_sync(&pAdapter->PortCfg.REKEYTimer);
                    pAdapter->PortCfg.REKEYTimerRunning = FALSE;
                }
            }
            else if ((WPA_REKEY.ReKeyMethod == PKT_REKEY) || (WPA_REKEY.ReKeyMethod == TIME_REKEY))  //packet-based
            {
                if(pAdapter->PortCfg.REKEYTimerRunning == FALSE)
                {
			        pAdapter->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
			        add_timer(&pAdapter->PortCfg.REKEYTimer);
			        
                    pAdapter->PortCfg.REKEYTimerRunning = TRUE;
                    pAdapter->PortCfg.REKEYCOUNTER = 0;
                }
                pAdapter->PortCfg.WPAREKEY.ReKeyInterval = WPA_REKEY.ReKeyInterval;
                pAdapter->PortCfg.WPAREKEY.ReKeyMethod = WPA_REKEY.ReKeyMethod;                    
            }                         
            DBGPRINT(RT_DEBUG_TRACE, "Set::RT_OID_802_11_SET_WPA_REKEY , ReKeyMethod= %x  , interval= %x    \n",WPA_REKEY.ReKeyMethod,WPA_REKEY.ReKeyInterval);
            break;

		default:
			DBGPRINT(RT_DEBUG_TRACE, "Set::unknown IOCTL's subcmd = 0x%08x\n", cmd);
			Status = -EOPNOTSUPP;
			break;
	}

	mb();//Vincent Tw PCI register
	if (RestartAPIsRequired) 
    {
        ApStop(pAdapter);
        ApStartUp(pAdapter);
    }
	
	return Status;
}

INT RTMPQueryInformation(
	IN	PRTMP_ADAPTER       pAdapter,
	IN	OUT	struct ifreq    *rq,
	IN	INT                 cmd)
{
	struct iwreq						*wrq = (struct iwreq *) rq;
    INT									Status = NDIS_STATUS_SUCCESS;
    ULONG								ulInfo = 0;
    NDIS_802_11_CONFIGURATION			Configuration;
    RT_802_11_PREAMBLE					Preamble;
    RT_802_11_AP_CONFIG                 ApConfig;
    NDIS_802_11_WEP_STATUS				WepStatus;
    NDIS_802_11_SSID					Ssid;
    NDIS_802_11_MAC_ADDRESS             Bssid;
    RT_802_11_MAC_TABLE                 MacTab;
    NDIS_802_11_STATISTICS              Statistics;
    RT_VERSION_INFO						DriverVersionInfo;
    ULONG								FcsValue;
    UINT                                i;
	#ifdef VINCENT_K
	char	msg[1024];
	UCHAR BaseAds, *Resu;
	ULONG PciAds;

	memset(msg, 0, 1024);
	#endif

	switch(cmd) {
		case RT_OID_DEVICE_NAME:
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_DEVICE_NAME\n");
			#ifdef VINCENT_K
		    sprintf(msg, "%s", NIC_DEVICE_NAME);					
			wrq->u.data.length=strlen(msg);
			copy_to_user(wrq->u.data.pointer, &msg, wrq->u.data.length);
			#else
			wrq->u.data.length = sizeof(NIC_DEVICE_NAME);
			copy_to_user(wrq->u.data.pointer, NIC_DEVICE_NAME, wrq->u.data.length);
			#endif
			break;
		case RT_OID_VERSION_INFO:
            DBGPRINT(RT_DEBUG_INFO, "Query::RT_OID_VERSION_INFO \n");
            DriverVersionInfo.DriverVersionW = 2;
            DriverVersionInfo.DriverVersionX = 1;
            DriverVersionInfo.DriverVersionY = 0;
            DriverVersionInfo.DriverVersionZ = 0;
            DriverVersionInfo.DriverBuildYear = 2005;
            DriverVersionInfo.DriverBuildMonth = 1;
            DriverVersionInfo.DriverBuildDay = 1;
            wrq->u.data.length = sizeof(RT_VERSION_INFO);
            copy_to_user(wrq->u.data.pointer, &DriverVersionInfo, wrq->u.data.length);
            break;
		case OID_GEN_RCV_NO_BUFFER:
			DBGPRINT(RT_DEBUG_INFO, "Query::OID_GEN_RCV_NO_BUFFER \n");
			ulInfo = pAdapter->Counters.RxNoBuffer;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			break;
		case OID_GEN_RCV_OK:
			DBGPRINT(RT_DEBUG_INFO, "Query::OID_GEN_RCV_OK \n");
			ulInfo = pAdapter->Counters.GoodReceives;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			break;
		case OID_802_11_STATISTICS:
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_STATISTICS \n");
			// Update FCS counters
		    RTMP_IO_READ32(pAdapter, CNT0, &FcsValue);
		    pAdapter->WlanCounters.FCSErrorCount += FcsValue;
		    
			// Sanity check for calculation of sucessful count
			if (pAdapter->WlanCounters.TransmittedFragmentCount < pAdapter->WlanCounters.RetryCount)
				pAdapter->WlanCounters.TransmittedFragmentCount = pAdapter->WlanCounters.RetryCount;

			Statistics.TransmittedFragmentCount = pAdapter->WlanCounters.TransmittedFragmentCount;
			Statistics.MulticastTransmittedFrameCount = pAdapter->WlanCounters.MulticastTransmittedFrameCount;
			Statistics.FailedCount = pAdapter->WlanCounters.FailedCount;
			Statistics.RetryCount = pAdapter->WlanCounters.RetryCount;
			Statistics.MultipleRetryCount = pAdapter->WlanCounters.MultipleRetryCount;
			Statistics.RTSSuccessCount = pAdapter->WlanCounters.RTSSuccessCount;
			Statistics.RTSFailureCount = pAdapter->WlanCounters.RTSFailureCount;
			Statistics.ACKFailureCount = pAdapter->WlanCounters.ACKFailureCount;
			Statistics.FrameDuplicateCount = pAdapter->WlanCounters.FrameDuplicateCount;
			Statistics.ReceivedFragmentCount = pAdapter->WlanCounters.ReceivedFragmentCount;
			Statistics.MulticastReceivedFrameCount = pAdapter->WlanCounters.MulticastReceivedFrameCount;
			Statistics.FCSErrorCount = pAdapter->WlanCounters.FCSErrorCount;
		    wrq->u.data.length = sizeof(NDIS_802_11_STATISTICS);
			copy_to_user(wrq->u.data.pointer, &Statistics, wrq->u.data.length);
			break;
		case RT_OID_802_11_EVENT_TABLE:
			if (wrq->u.data.length != sizeof(RT_802_11_EVENT_TABLE))
				Status = -EINVAL;
			else 
			{
				wrq->u.data.length = sizeof(RT_802_11_EVENT_TABLE);
				copy_to_user(wrq->u.data.pointer, &pAdapter->EventTab, wrq->u.data.length);

				DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_EVENT_TABLE\n");

				// pAd->EventTab is read-and-clear
				NdisZeroMemory(&pAdapter->EventTab, sizeof(RT_802_11_EVENT_TABLE));
			}
			break;
		case RT_OID_802_11_MAC_TABLE:
			MacTab.Num=0;
			for (i=0; i<MAX_LEN_OF_MAC_TABLE; i++)
			{
				if (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Valid && (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Sst == SST_ASSOC))
				{
					COPY_MAC_ADDR(MacTab.Entry[MacTab.Num].Addr, &pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr);
					MacTab.Entry[MacTab.Num].Aid = (UCHAR)pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Aid;
					MacTab.Entry[MacTab.Num].Psm = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].PsMode;
					MacTab.Entry[MacTab.Num].HSCounter.LastDataPacketTime = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].HSCounter.LastDataPacketTime;
					MacTab.Entry[MacTab.Num].HSCounter.TotalRxByteCount= pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].HSCounter.TotalRxByteCount;
					MacTab.Entry[MacTab.Num].HSCounter.TotalTxByteCount= pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].HSCounter.TotalTxByteCount;
					MacTab.Num += 1;
				}
			}				
			wrq->u.data.length=sizeof(RT_802_11_MAC_TABLE);
			copy_to_user(wrq->u.data.pointer, &MacTab, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_MAC_TABLE \n");
			break;
		case OID_802_11_BSSID:
			COPY_MAC_ADDR(&Bssid, &pAdapter->PortCfg.Bssid);
			#ifdef VINCENT_K
		    sprintf(msg, "%02x.%02x.%02x.%02x.%02x.%02x",
							Bssid[0],Bssid[1],Bssid[2],Bssid[3],Bssid[4],Bssid[5]);					
			wrq->u.data.length=strlen(msg);
			copy_to_user(wrq->u.data.pointer, &msg, wrq->u.data.length);
			#else
			wrq->u.data.length=sizeof(NDIS_802_11_MAC_ADDRESS);
			copy_to_user(wrq->u.data.pointer, &Bssid, wrq->u.data.length);
			#endif
		    DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_BSSID (=%02x:%02x:%02x:%02x:%02x:%02x)\n",
							Bssid[0],Bssid[1],Bssid[2],Bssid[3],Bssid[4],Bssid[5]);					
			break;
		case OID_802_11_SSID:
			#ifdef VINCENT_K
			Ssid.SsidLength = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen;
			NdisZeroMemory(Ssid.Ssid, MAX_LEN_OF_SSID);
			NdisMoveMemory(Ssid.Ssid, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, Ssid.SsidLength);
		    sprintf(msg, "%s", Ssid.Ssid);					
			wrq->u.data.length=strlen(msg);
			copy_to_user(wrq->u.data.pointer, &msg, wrq->u.data.length);
			#else
			Ssid.SsidLength = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen;
			NdisZeroMemory(Ssid.Ssid, MAX_LEN_OF_SSID);
			NdisMoveMemory(Ssid.Ssid, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, Ssid.SsidLength);
			wrq->u.data.length=sizeof(NDIS_802_11_SSID);
			copy_to_user(wrq->u.data.pointer, &Ssid, wrq->u.data.length);
			#endif
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_SSID (Len=%d, ssid=%s)\n", Ssid.SsidLength,Ssid.Ssid);
			break;
		case OID_802_11_RSSI:
			ulInfo = pAdapter->PortCfg.LastRssi - RSSI_TO_DBM_OFFSET;
			wrq->u.data.length=sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_INFO, "Query::OID_802_11_RSSI \n");
			break;
		case OID_GEN_MEDIA_CONNECT_STATUS:
			ulInfo = NdisMediaStateConnected; 
			wrq->u.data.length=sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_INFO, "Query::OID_GEN_MEDIA_CONNECT_STATUS \n");
			break;
		case RT_OID_802_11_LINK_STATUS:
            Status = -EOPNOTSUPP;
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_LINK_STATUS\n");
			break;
		case OID_802_3_CURRENT_ADDRESS:
			wrq->u.data.length = ETH_LENGTH_OF_ADDRESS;
			copy_to_user(wrq->u.data.pointer, &pAdapter->CurrentAddress, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_INFO, "Query::OID_802_3_CURRENT_ADDRESS \n");
			break;
		case OID_802_11_WEP_STATUS:
			WepStatus = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus;
			wrq->u.data.length = sizeof(NDIS_802_11_WEP_STATUS);
			copy_to_user(wrq->u.data.pointer, &WepStatus, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_WEP_STATUS (=%d)\n", WepStatus);
			break;
		case OID_802_11_AUTHENTICATION_MODE:					
			wrq->u.data.length = sizeof(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode);
			copy_to_user(wrq->u.data.pointer, &pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_AUTHENTICATION_MODE (=%d)\n", pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode);
			break;
		case RT_OID_802_11_AP_CONFIG:
			ApConfig.EnableTxBurst = pAdapter->PortCfg.EnableTxBurst;
			ApConfig.EnableTurboRate = pAdapter->PortCfg.EnableTurboRate;
			ApConfig.IsolateInterStaTraffic = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IsolateInterStaTraffic;
			ApConfig.HideSsid = pAdapter->PortCfg.HideSsid;
			ApConfig.UseBGProtection = pAdapter->PortCfg.UseBGProtection;
			ApConfig.UseShortSlotTime = pAdapter->PortCfg.UseShortSlotTime;
			ApConfig.Rsv1 = 0;
			ApConfig.SystemErrorBitmap = pAdapter->PortCfg.SystemErrorBitmap;
			wrq->u.data.length = sizeof(RT_802_11_AP_CONFIG);
			copy_to_user(wrq->u.data.pointer, &ApConfig, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_AP_CONFIG\n");
			break;
		case OID_802_11_FRAGMENTATION_THRESHOLD:
			ulInfo = pAdapter->PortCfg.FragmentThreshold;
			if (pAdapter->PortCfg.bFragmentZeroDisable == TRUE)
				ulInfo = 0;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_FRAGMENTATION_THRESHOLD (=%d)\n", ulInfo);
			break;
		case OID_802_11_RTS_THRESHOLD:
			ulInfo = pAdapter->PortCfg.RtsThreshold;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_RTS_THRESHOLD (=%d)\n", ulInfo);
			break;
		case RT_OID_802_11_PREAMBLE:
			Preamble = pAdapter->PortCfg.TxPreamble;
			wrq->u.data.length = sizeof(RT_802_11_PREAMBLE);
			copy_to_user(wrq->u.data.pointer, &Preamble, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_PREAMBLE (=%d)\n", Preamble);
			break;
		case OID_802_11_RX_ANTENNA_SELECTED:
			if (pAdapter->PortCfg.CurrentRxAntenna == 0xff)
				ulInfo = 0xffffffff;
			else
				ulInfo = pAdapter->PortCfg.CurrentRxAntenna;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_RX_ANTENNA_SELECTED (=%d)\n", ulInfo);
			break;

		case OID_802_11_TX_ANTENNA_SELECTED:
			if (pAdapter->PortCfg.CurrentTxAntenna == 0xff)
				ulInfo = 0xffffffff;
			else
				ulInfo = pAdapter->PortCfg.CurrentTxAntenna;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_TX_ANTENNA_SELECTED (=0x%08x)\n", ulInfo);
			break;
		case OID_802_11_NUMBER_OF_ANTENNAS:
			ulInfo = pAdapter->PortCfg.NumberOfAntenna;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_NUMBER_OF_ANTENNAS (=%d)\n", ulInfo);
			break;
		case OID_802_11_CONFIGURATION:
			Configuration.Length = sizeof(NDIS_802_11_CONFIGURATION);
			Configuration.BeaconPeriod = pAdapter->PortCfg.BeaconPeriod;
			Configuration.ATIMWindow = 0; //Configuration.ATIMWindow = pAdapter->PortCfg.AtimWin;
			if (pAdapter->PortCfg.Channel == 14)
				Configuration.DSConfig = 2484000;
			else
				Configuration.DSConfig = 2412000 + (pAdapter->PortCfg.Channel - 1) * 5000;

			DBGPRINT(RT_DEBUG_TRACE, "Query::OID_802_11_CONFIGURATION (BeaconPeriod=%d,AtimW=%d,Channel=%d) \n",
						Configuration.BeaconPeriod, Configuration.ATIMWindow, pAdapter->PortCfg.Channel);
			wrq->u.data.length = sizeof(NDIS_802_11_CONFIGURATION);
			copy_to_user(wrq->u.data.pointer, &Configuration, wrq->u.data.length);
			break;
		case RT_OID_802_11_PHY_MODE:					
			ulInfo = (ULONG)pAdapter->PortCfg.PhyMode;
			wrq->u.data.length = sizeof(ulInfo);
			copy_to_user(wrq->u.data.pointer, &ulInfo, wrq->u.data.length);
			DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_PHY_MODE (=%d)\n", ulInfo);
			break;
		case RT_OID_802_11_WPA_REKEY:            
            wrq->u.data.length = sizeof(RT_802_11_WPA_REKEY);
            copy_to_user(wrq->u.data.pointer, &pAdapter->PortCfg.WPAREKEY, wrq->u.data.length);
            DBGPRINT(RT_DEBUG_TRACE, "Query :: RT_OID_802_11_QUERY_WPA_REKEY\n");
            break;
        case RT_OID_802_11_EEPROM_VERSION:
            wrq->u.data.length = sizeof(pAdapter->PortCfg.EepromVersion);
            copy_to_user(wrq->u.data.pointer, &pAdapter->PortCfg.EepromVersion, wrq->u.data.length);
            DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_EEPROM_VERSION (=%d)\n", pAdapter->PortCfg.EepromVersion);
            break;
		#ifdef VINCENT_K
        case RT_OID_802_11_PCI_64B:
			for (i=0; i<16; i++)
			{
				BaseAds = (PCI_VENDOR_ID + 4*i);
				pci_read_config_dword(pAdapter->pPci_Dev, BaseAds, &PciAds);
				//printk("-------kernel msg: BaseAds: %x, PCI Address: %.8x\n", BaseAds, PciAds);
				Resu = (UCHAR *)(&PciAds);
				//printk("Kernel: %2x %2x %2x %2x\n", Resu[0], Resu[1], Resu[2], Resu[3]);
				sprintf(msg+strlen(msg), "%x.%x.%x.%x.", Resu[0], Resu[1], Resu[2], Resu[3]);
			}
			wrq->u.data.length = strlen(msg);
			copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
            DBGPRINT(RT_DEBUG_TRACE, "Query::RT_OID_802_11_PCI_64B\n");
            break;
		#endif
		default:
			DBGPRINT(RT_DEBUG_TRACE, "Query::unknown IOCTL's subcmd = 0x%08x\n", cmd);
			Status = -EOPNOTSUPP;
			break;
	}
	
	return Status;
}

INT RT2500_ioctl(
	IN	struct net_device	*net_dev, 
	IN	OUT	struct ifreq	*rq, 
	IN	INT					cmd)
{
    VIRTUAL_ADAPTER                     *pVirtualAd;
	RTMP_ADAPTER                        *pAdapter;
    struct iwreq                        *wrq = (struct iwreq *) rq;
    NDIS_802_11_SSID                    Ssid;
    NDIS_802_11_RTS_THRESHOLD           RtsThresh;
    NDIS_802_11_FRAGMENTATION_THRESHOLD FragThresh;
    BOOLEAN                             RestartAPIsRequired = FALSE;
    INT                                 Status = NDIS_STATUS_SUCCESS;
    USHORT                              subcmd, index;

    // determine this ioctl command is comming from which interface.
	// !!Vincent!! 
    if (!RTMPEqualMemory(net_dev->name, "wl0", 3))  // ioctl command from virtual interface
    {
        pVirtualAd = net_dev->priv;
        pAdapter = pVirtualAd->RtmpDev->priv;
        
        for (index = 1; index < pAdapter->PortCfg.SsidNum; index++)
    	{
    	    if (pAdapter->PortCfg.MSSID[index].MSSIDDev == net_dev)
    	    {
    	        pAdapter->PortCfg.IoctlIF = index;
    	        
    	        DBGPRINT(RT_DEBUG_INFO, "VirtualIF_ioctl I/F(ra%d): cmd = 0x%08x\n", index, cmd);
    	        break;
    	    }
    	}
        // Interface not found!
        if(index == pAdapter->PortCfg.SsidNum)
        {
            return -ENETDOWN;
        }
    }
    else    // ioctl command from I/F(ra0)
    {
        pAdapter= net_dev->priv;
        pAdapter->PortCfg.IoctlIF = MAIN_MSSID;
        DBGPRINT(RT_DEBUG_INFO, "VirtualIF_ioctl I/F(ra%d): cmd = 0x%08x\n", MAIN_MSSID, cmd);
    }

	if(!RTMP_TEST_FLAG(pAdapter, fRTMP_ADAPTER_INTERRUPT_IN_USE))
	{
		DBGPRINT(RT_DEBUG_TRACE, "INFO::Network is down!\n");
		return -ENETDOWN;
	}

	switch(cmd)
	{
        case SIOCGIFHWADDR:
			DBGPRINT(RT_DEBUG_ERROR, "IOCTLIOCTLIOCTL::SIOCGIFHWADDR\n");
			strcpy(wrq->u.name, pAdapter->CurrentAddress);   
			break;
		case SIOCGIWNAME:
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCGIWNAME\n");
			strcpy(wrq->u.name, "RT2500 SoftAP");   
			break;
		case SIOCSIWESSID:  //Set ESSID
			{
				struct iw_point *erq = &wrq->u.essid;

				memset(&Ssid, 0x00, sizeof(NDIS_802_11_SSID));

				if (erq->flags) 
				{
					if (erq->length > IW_ESSID_MAX_SIZE)
					{
						Status = -E2BIG;
						break;
					}

					if (copy_from_user(Ssid.Ssid, erq->pointer, erq->length-1))
					{
						Status = -EFAULT;
						break;
					}

					Ssid.SsidLength = erq->length-1;	 //minus null character.					
				}
				else
				{ //Any ssid
					NdisMoveMemory(Ssid.Ssid, "RT2500SoftAP-", 7);
					Ssid.Ssid[7] = (((pAdapter->CurrentAddress[5] >> 4) > 9) ? ((pAdapter->CurrentAddress[5] >> 4) - 10 + 'A') : ((pAdapter->CurrentAddress[5] >> 4) + '0'));
					Ssid.Ssid[8] = (((pAdapter->CurrentAddress[5] & 0x0F) > 9) ? ((pAdapter->CurrentAddress[5]& 0x0F) - 10 + 'A') : ((pAdapter->CurrentAddress[5]& 0x0F) + '0'));
					Ssid.SsidLength = 9;
				}

				NdisZeroMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, MAX_LEN_OF_SSID);
				NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, Ssid.Ssid, Ssid.SsidLength);
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen = (UCHAR) Ssid.SsidLength;
				RestartAPIsRequired = TRUE;
				DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCSIWESSID (Len=%d,Ssid=%s)\n", pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid);
				break;
			}
		case SIOCGIWESSID:  //Get ESSID
			{
				struct iw_point *erq = &wrq->u.essid;

				erq->flags=1;
				erq->length = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen;

				if(erq->pointer)
				{
					if(copy_to_user(erq->pointer, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, erq->length))
					{
						Status = -EFAULT;
						break;
					}
				}
				DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCGIWESSID (Len=%d, ssid=%s...)\n", erq->length, erq->pointer);
			}
			break;
		case SIOCGIWNWID: // get network id 
			Status = -EOPNOTSUPP;
			break;
		case SIOCSIWNWID: // set network id (the cell)
			Status = -EOPNOTSUPP;
			break;
		case SIOCGIWFREQ: // get channel/frequency (Hz)
			wrq->u.freq.m = pAdapter->PortCfg.Channel;
			wrq->u.freq.e = 0;
			wrq->u.freq.i = 0;
			break; 
		case SIOCSIWFREQ: //set channel/frequency (Hz)
			{
				struct iw_freq *frq = &wrq->u.freq;
				int chan = -1;

				if((frq->e == 0) && (frq->m <= 1000))
				{
					// Setting by channel number
					chan = frq->m;
				}
				else
				{
					// Setting by frequency - search the table
					int mult = 1;
					int i;

					for(i = 0; i < (6 - frq->e); i++)
						mult *= 10;

					for(i = 0; i < NUM_CHANNELS; i++)
						if(frq->m == (channel_frequency[i] * mult))
							chan = i+1;
				}

				pAdapter->PortCfg.Channel = chan;
				RestartAPIsRequired = TRUE;
				DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCSIWFREQ (Channel=%d)\n", pAdapter->PortCfg.Channel);				
			}
			break;
		case SIOCGIWNICKN:
			Status = -EOPNOTSUPP;
			break;
		case SIOCSIWNICKN: //set node name/nickname
			Status = -EOPNOTSUPP;
			break;
		case SIOCGIWRATE:  //get default bit rate (bps)
			wrq->u.bitrate.value = RateIdTo500Kbps[pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].TxRate] * 500000;
			wrq->u.bitrate.disabled = 0;
			break;
		case SIOCSIWRATE:  //set default bit rate (bps)
			{
				int i;
				BOOLEAN  IsSupportRate = FALSE;
				
				for(i=0; i<NUM_RateIdTo500Kbps; i++)
				{
					if( wrq->u.bitrate.value == RateIdTo500Kbps[i] * 500000)
					{
						pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].TxRate = i;
						IsSupportRate = TRUE;
						break;
					}
				}
	
				if( !IsSupportRate )
				{
					Status = -EINVAL;
					break;
				}
				else
					DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCSIWRATE (rates=%d)\n", pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].TxRate);
			}
			break;
		case SIOCGIWRTS:  // get RTS/CTS threshold (bytes)
			wrq->u.rts.value = (INT) pAdapter->PortCfg.RtsThreshold;
			wrq->u.rts.disabled = (wrq->u.rts.value == MAX_RTS_THRESHOLD);
			wrq->u.rts.fixed = 1;
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCGIWRTS (=%d)\n",wrq->u.rts.value);
			break;
		case SIOCSIWRTS:  //set RTS/CTS threshold (bytes)
			RtsThresh = wrq->u.rts.value;
			if(wrq->u.rts.disabled)
				RtsThresh = MAX_RTS_THRESHOLD;
			if (RtsThresh > MAX_RTS_THRESHOLD || RtsThresh <= 0 )
			{
				Status = -EINVAL;
				break;
			}
			else
				pAdapter->PortCfg.RtsThreshold = (USHORT)RtsThresh;
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCSIWRTS (=%d)\n",RtsThresh);
			break;
		case SIOCGIWFRAG:  //get fragmentation thr (bytes)
			wrq->u.frag.value = pAdapter->PortCfg.FragmentThreshold;
			wrq->u.frag.disabled = (wrq->u.frag.value >= MAX_FRAG_THRESHOLD);
			wrq->u.frag.fixed = 1;
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCGIWFRAG (=%d)\n", wrq->u.frag.value);
			break;
		case SIOCSIWFRAG:  //set fragmentation thr (bytes)
			FragThresh = wrq->u.frag.value;
			if(wrq->u.frag.disabled)
				FragThresh = MAX_FRAG_THRESHOLD;

			if((FragThresh < MIN_FRAG_THRESHOLD) || (FragThresh > MAX_FRAG_THRESHOLD))
			{
				Status = -EINVAL;
				break;
			}
			else
			{
				pAdapter->PortCfg.FragmentThreshold = (USHORT)FragThresh;
				pAdapter->PortCfg.AllowFragSize = (pAdapter->PortCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;
				DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCSIWFRAG (=%d)\n", FragThresh);
			}
			break;
		case SIOCGIWENCODE:  //get encoding token & mode
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				if ((index < 0) || (index >= NR_WEP_KEYS))
					index = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId;  	// Default key for tx (shared key)
				
				wrq->u.encoding.flags = (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus ==  Ndis802_11WEPEnabled) ? IW_ENCODE_RESTRICTED : IW_ENCODE_OPEN;
				if(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus ==  Ndis802_11WEPDisabled)
					wrq->u.encoding.flags |= IW_ENCODE_DISABLED;
				if(wrq->u.encoding.pointer)
				{
					wrq->u.encoding.length = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].KeyLen;
					if (copy_to_user(wrq->u.encoding.pointer, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].Key, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].KeyLen))
					{
						Status = -EFAULT;
						break;
					}
					wrq->u.encoding.flags |= (index + 1);
				}
			}
			break;
		case SIOCSIWENCODE:  //set encoding token & mode
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				/* take the old default key if index is invalid */
				if((index < 0) || (index >= NR_WEP_KEYS))
					index = pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId;  	// Default key for tx (shared key)
				if(wrq->u.encoding.pointer)
				{
					int len = wrq->u.encoding.length;

					if(len > WEP_LARGE_KEY_LEN)
						len = WEP_LARGE_KEY_LEN;

					memset(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].Key, 0, MAX_LEN_OF_KEY);
					if(copy_from_user(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].Key, wrq->u.encoding.pointer, len)) 
					{
						pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].KeyLen = 0;
						Status = -EFAULT;
						break;
					}
					else
						pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[index].KeyLen = (len <= WEP_SMALL_KEY_LEN ) ? WEP_SMALL_KEY_LEN : WEP_LARGE_KEY_LEN;
				}
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId = (UCHAR) index;
                pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = (wrq->u.encoding.flags & IW_ENCODE_DISABLED) ? 1 : 0;

				if (wrq->u.encoding.flags & IW_ENCODE_RESTRICTED)
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = AUTH_MODE_KEY;
				if ((wrq->u.encoding.flags & IW_ENCODE_OPEN) || wrq->u.encoding.flags & IW_ENCODE_NOKEY)
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = AUTH_MODE_OPEN;

				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].CapabilityInfo = CAP_GENERATE((pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus != Ndis802_11EncryptionDisabled), 1, pAdapter->PortCfg.UseShortSlotTime);
				// re-built BEACON frame format
				AsicDisableSync(pAdapter);
				MakeBssBeacon(pAdapter);
				UpdateBeaconFrame(pAdapter);
				AsicEnableBssSync(pAdapter);
			}
			break;
		case SIOCGIWAP:  //get access point MAC addresses
			wrq->u.ap_addr.sa_family = ARPHRD_ETHER;
			memcpy(wrq->u.ap_addr.sa_data, &pAdapter->PortCfg.Bssid, ETH_ALEN);
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::SIOCGIWAP(=%02x:%02x:%02x:%02x:%02x:%02x)\n",
				pAdapter->PortCfg.Bssid.Octet[0],pAdapter->PortCfg.Bssid.Octet[1],pAdapter->PortCfg.Bssid.Octet[2],
				pAdapter->PortCfg.Bssid.Octet[3],pAdapter->PortCfg.Bssid.Octet[4],pAdapter->PortCfg.Bssid.Octet[5]);
			break;
		case SIOCSIWAP:  //set access point MAC addresses
			Status = -EOPNOTSUPP;   //SoftAP didn't support.
			break;
		case SIOCGIWMODE:  //get operation mode
			wrq->u.mode = IW_MODE_INFRA;   //SoftAP always on INFRA mode.
			break;
		case SIOCSIWMODE:  //set operation mode
			Status = -EOPNOTSUPP;   //SoftAP didn't support.
			break;
		case SIOCGIWSENS:   //get sensitivity (dBm)
		case SIOCSIWSENS:	//set sensitivity (dBm)
		case SIOCGIWPOWER:  //get Power Management settings
		case SIOCSIWPOWER:  //set Power Management settings
		case SIOCGIWTXPOW:  //get transmit power (dBm)
		case SIOCSIWTXPOW:  //set transmit power (dBm)
#if WIRELESS_EXT >= 12
		case SIOCGIWRETRY:  //get retry limits and lifetime
		case SIOCSIWRETRY:  //set retry limits and lifetime
#endif	//WIRELESS_EXT >= 12
			Status = -EOPNOTSUPP;
			break;
		case RT_PRIV_IOCTL:
			subcmd = wrq->u.data.flags;
			if( subcmd & OID_GET_SET_TOGGLE)
				Status = RTMPSetInformation(pAdapter, rq, subcmd);
			else
				Status = RTMPQueryInformation(pAdapter, rq, subcmd);
			break;
		case SIOCGIWPRIV:
			if (wrq->u.data.pointer) {
				if ( verify_area(VERIFY_WRITE, wrq->u.data.pointer, sizeof(privtab)) )
					break;
				wrq->u.data.length = sizeof(privtab) / sizeof(privtab[0]);
				if (copy_to_user(wrq->u.data.pointer, privtab, sizeof(privtab)))
					Status = -EFAULT;
			}
			break;
		case RTPRIV_IOCTL_SET:
			{				
				char *this_char;
				char *value;

				if( verify_area(VERIFY_READ, wrq->u.data.pointer, wrq->u.data.length) )
					break;

				while ((this_char = strsep(&wrq->u.data.pointer, ",")) != NULL) 
				{
					if (!*this_char)
						 continue;

					if ((value = strchr(this_char, '=')) != NULL)
						*value++ = 0;

					if (!value || !*value)
						continue;  							

					for (PRTMP_PRIVATE_SET_PROC = RTMP_PRIVATE_SUPPORT_PROC; PRTMP_PRIVATE_SET_PROC->name; PRTMP_PRIVATE_SET_PROC++)
					{
						if (!strcmp(this_char, PRTMP_PRIVATE_SET_PROC->name)) 
						{						
							if(!PRTMP_PRIVATE_SET_PROC->set_proc(pAdapter, value))
							{   //FALSE:Set private failed then return Invalid argument 								
								Status = -EINVAL;							
							}
							break;  //Exit for loop.
						}
					}

					if(PRTMP_PRIVATE_SET_PROC->name == NULL)
					{  //Not found argument
						Status = -EINVAL;
						DBGPRINT(RT_DEBUG_TRACE, "IOCTL::(iwpriv) Not Support Set Command [%s=%s]\n", this_char, value);
						break;
					}	
				}
			}
			break;
		case RTPRIV_IOCTL_SITESURVEY:
			RTMPIoctlSiteSurvey(pAdapter,wrq);
		    break;
		    
		case RTPRIV_IOCTL_GET_MAC_TABLE:
			RTMPIoctlGetMacTable(pAdapter,wrq);
		    break;
		    
		case RTPRIV_IOCTL_GSITESURVEY:
			RTMPIoctlGetSiteSurvey(pAdapter,wrq);
		    break;
		
		case RTPRIV_IOCTL_STATISTICS:
		    RTMPIoctlStatistics(pAdapter, wrq);
		    break;

#ifdef RALINK_ATE
		case RTPRIV_IOCTL_BBP:
			RTMPIoctlBBP(pAdapter, wrq);
			break;
			
		case RTPRIV_IOCTL_MAC:
			RTMPIoctlMAC(pAdapter, wrq);
			break;

		case RTPRIV_IOCTL_E2P:
			RTMPIoctlE2PROM(pAdapter, wrq);
			break;
#endif
		default:
			DBGPRINT(RT_DEBUG_TRACE, "IOCTL::unknown IOCTL's cmd = 0x%08x\n", cmd);
			Status = -EOPNOTSUPP;
			break;
	}

	if( RestartAPIsRequired )
	{
		ApStop(pAdapter);
		ApStartUp(pAdapter);
	}

	return Status;
}

/*
	========================================================================
	
	Routine Description:
		Change NIC PHY mode. Re-association may be necessary.

	Arguments:
		pAdapter						Pointer to our adapter
        phmode
		
	========================================================================
*/
VOID	RTMPSetPhyMode(
	IN	PRTMP_ADAPTER	pAdapter,
	IN  ULONG phymode)
{
    // if no change, do nothing
    if (pAdapter->PortCfg.PhyMode == phymode)
        return;
    
    pAdapter->PortCfg.PhyMode = (UCHAR)phymode;
    switch (phymode) {
        case PHY_11B:
            pAdapter->PortCfg.SupportedRates[0]  = 0x82;    // 1 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[1]  = 0x84;    // 2 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[2]  = 0x0B;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[3]  = 0x16;    // 11 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRatesLen  = 4;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
            break;
        case PHY_11BG_MIXED:
            // 2004-1-12 to keep maximum compatibility with 802.11b-only clients, we should not turn on
            // BASIC RATE bit of any OFDM rates in outgoing BEACON and ProbrResponse frames
            pAdapter->PortCfg.SupportedRates[0]  = 0x82;    // 1 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[1]  = 0x84;    // 2 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[2]  = 0x8B;    // 5.5 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[3]  = 0x96;    // 11 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[4]  = 0x0C;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[5]  = 0x12;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[6]  = 0x18;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[7]  = 0x24;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[8]  = 0x30;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[9]  = 0x48;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[10] = 0x60;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[11] = 0x6C;    // 54 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRatesLen  = 12;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 72;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 96;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 108;   // 54 mbps, in units of 0.5 Mbps
            break;
        case PHY_11G:
            // 2004-1-12 to keep maximum compatibility with 802.11b-only clients, we should not turn on
            // BASIC RATE bit of any OFDM rates in outgoing BEACON and ProbrResponse frames
            pAdapter->PortCfg.SupportedRates[0]  = 0x82;    // 1 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[1]  = 0x84;    // 2 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[2]  = 0x8B;    // 5.5 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[3]  = 0x96;    // 11 mbps, in units of 0.5 Mbps, basic rate
            pAdapter->PortCfg.SupportedRates[4]  = 0x8C;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[5]  = 0x12;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[6]  = 0x98;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[7]  = 0x24;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[8]  = 0xB0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[9]  = 0x48;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[10] = 0x60;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRates[11] = 0x6C;    // 54 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.SupportedRatesLen  = 12;
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 72;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 96;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 108;   // 54 mbps, in units of 0.5 Mbps
            break;
        default:
            break;
    }
    
    MlmeUpdateTxRates(pAdapter);
}

/* 
    ==========================================================================
    Description:
        Set Country Region
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT Set_CountryRegion_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG region;
	int   success = FALSE;
		
	region = simple_strtol(arg, 0, 10);
	if((region >= REGION_MIN) && (region <= REGION_MAX))
	{
		pAdapter->PortCfg.CountryRegion = (UCHAR) region;
		DBGPRINT(RT_DEBUG_TRACE, "Set_CountryRegion_Proc::(CountryRegion=%d)\n", pAdapter->PortCfg.CountryRegion);

		success = TRUE;
	}
	else
		success = FALSE;

	return success;
}

/* 
    ==========================================================================
    Description:
        Set SSID
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_SSID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	int   success = FALSE;

	if(strlen(arg) <= MAX_LEN_OF_SSID)
	{
		NdisZeroMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, MAX_LEN_OF_SSID);
		NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, arg, strlen(arg));
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen = (UCHAR)strlen(arg);
		success = TRUE;

		// AP need to restart only if set to Main SSID.
		if(pAdapter->PortCfg.IoctlIF == MAIN_MSSID)
		{
    		ApStop(pAdapter);
    		ApStartUp(pAdapter);
		}
		DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_SSID_Proc::(Len=%d,Ssid=%s)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid);
	}
	else
		success = FALSE;

	return success;
}

/* 
    ==========================================================================
    Description:
        Set Wireless Mode
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_WirelessMode_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG	WirelessMode;
	int   success = TRUE;

	WirelessMode = simple_strtol(arg, 0, 10);

	if(WirelessMode == PHY_11B)
		pAdapter->PortCfg.PhyMode = PHY_11B;
	else if(WirelessMode == PHY_11BG_MIXED)
		pAdapter->PortCfg.PhyMode = PHY_11BG_MIXED;
	else if(WirelessMode == PHY_11G)
		pAdapter->PortCfg.PhyMode = PHY_11G;
	else
		success = FALSE;

	if(success)
	{
		ApStop(pAdapter);
		ApStartUp(pAdapter);
		DBGPRINT(RT_DEBUG_TRACE, "Set_WirelessMode_Proc::(=%d)\n", pAdapter->PortCfg.PhyMode);
	}
	
	return success;
}

/* 
    ==========================================================================
    Description:
        Set TxRate
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT Set_TxRate_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	switch(simple_strtol(arg, 0, 10))
	{
		case 0: // Auto 
			switch(pAdapter->PortCfg.PhyMode)
			{
				case PHY_11BG_MIXED: // B/G Mixed
				case PHY_11G: // G Only
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x6c; // 54Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x60; // 48Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x48; // 36Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x30; // 24Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x16; // 11Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x0b; // 5.5Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x04; // 2Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x02; // 1Mbps
					break;
				case PHY_11B: // B Only
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x16; // 11Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x0b; // 5.5Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x04; // 2Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x02; // 1Mbps
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
					pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
					break;
			}
			break;
		case 1:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x02;	// 1M	Mandatory
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 2:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x04;	// 2M	Mandatory
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 3:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x0b;	// 5.5M	Mandatory
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 4:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x16;	// 11M Mandatory
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;	
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 5:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x0c; // 6M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 6:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x12; // 9M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 7:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x18; // 12M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 8:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x24; // 18M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 9:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x30; // 24M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 10:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x48; // 36M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 11:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x60; // 48M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 12:
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0] = 0x6c; // 54M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6] = 0x00;
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7] = 0x00;
			break;
		case 13://1M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 0;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 0;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 0;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 0;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 14://2M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 0;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 0;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 0;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 15://5.5M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 0;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 0;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 16://11M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 0;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 17://6M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 0;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 18://9M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 0;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 19://12M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 0;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 20://18M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 0;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 21://24M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 0;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 22://36M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 72;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 0;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 23://48M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 72;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 96;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 0;
			break;
		case 24://54M
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0]  = 2;     // 1 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1]  = 4;     // 2 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2]  = 11;    // 5.5 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3]  = 22;    // 11 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4]  = 12;    // 6 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5]  = 18;    // 9 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6]  = 24;    // 12 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7]  = 36;    // 18 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8]  = 48;    // 24 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9]  = 72;    // 36 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10] = 96;    // 48 mbps, in units of 0.5 Mbps
            pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11] = 108;
			break;
		default:
			return FALSE;  //Invalid argument 
	}
	
	MlmeUpdateTxRates(pAdapter);

	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_TxRate_Proc::(%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x,%02x)\n", pAdapter->PortCfg.IoctlIF,
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[0],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[1],
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[2],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[3],
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[4],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[5],
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[6],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[7],
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[8],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[9],
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[10],pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DesiredRates[11]);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set BasicRate
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT Set_BasicRate_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
    pAdapter->PortCfg.BasicRateBitMap = (ULONG) simple_strtol(arg, 0, 10);

    if (pAdapter->PortCfg.BasicRateBitMap > 4095) // (2 ^ MAX_LEN_OF_SUPPORTED_RATES) -1
        return FALSE;

    MlmeUpdateTxRates(pAdapter);
    
    return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Channel
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Channel_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	UCHAR channel;	

	channel = (UCHAR) simple_strtol(arg, 0, 10);

	pAdapter->PortCfg.Channel = channel;
#ifdef VINCENT_K
    AsicSwitchChannel(pAdapter, pAdapter->PortCfg.Channel);
    AsicLockChannel(pAdapter, pAdapter->PortCfg.Channel);
#else
	ApStop(pAdapter);
	ApStartUp(pAdapter);
#endif
	DBGPRINT(RT_DEBUG_TRACE, "Set_Channel_Proc::(Channel=%d)\n", pAdapter->PortCfg.Channel);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Beacon Period
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_BeaconPeriod_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	USHORT BeaconPeriod;
	int   success = FALSE;

	BeaconPeriod = (USHORT) simple_strtol(arg, 0, 10);
	if((BeaconPeriod >= 20) && (BeaconPeriod < 1024))
	{
		pAdapter->PortCfg.BeaconPeriod = BeaconPeriod;
		success = TRUE;		
		AsicDisableSync(pAdapter);
		AsicEnableBssSync(pAdapter);
		DBGPRINT(RT_DEBUG_TRACE, "Set_BeaconPeriod_Proc::(BeaconPeriod=%d)\n", pAdapter->PortCfg.BeaconPeriod);
	}
	else
		success = FALSE;

	return success;
}

/* 
    ==========================================================================
    Description:
        Set Dtim Period
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_DtimPeriod_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	USHORT DtimPeriod;
	int   success = FALSE;

	DtimPeriod = (USHORT) simple_strtol(arg, 0, 10);
	if((DtimPeriod >= 1) && (DtimPeriod <= 255))
	{
		pAdapter->Mlme.DtimPeriod = DtimPeriod;
		success = TRUE;		
		DBGPRINT(RT_DEBUG_TRACE, "Set_DtimPeriod_Proc::(DtimPeriod=%d)\n", pAdapter->Mlme.DtimPeriod);
	}
	else
		success = FALSE;

	return success;
}

/* 
    ==========================================================================
    Description:
        Set Tx power
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_TxPower_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG TxPower;
	int   success = FALSE;

	TxPower = (ULONG) simple_strtol(arg, 0, 10);
	if (TxPower <= 100)
    {
		pAdapter->PortCfg.TxPowerPercentage = TxPower;
		success = TRUE;		
		DBGPRINT(RT_DEBUG_TRACE, "Set_TxPower_Proc::(TxPowerPercentage=%d)\n", pAdapter->PortCfg.TxPowerPercentage);
	}
	else
		success = FALSE;

	return success;
}

#ifdef RT2500_DBG
/* 
    ==========================================================================
    Description:
        For Debug information
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Debug_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)

{
	DBGPRINT(RT_DEBUG_TRACE, "**************************************************************\n");
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_Debug_Proc\n");

    if(simple_strtol(arg, 0, 10) <= RT_DEBUG_LOUD)
        RTDebugLevel = simple_strtol(arg, 0, 10);

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_Debug_Proc(RTDebugLevel = %d)\n", RTDebugLevel);
	DBGPRINT(RT_DEBUG_TRACE, "**************************************************************\n");
	return TRUE;
}
#endif

/* 
    ==========================================================================
    Description:
        Set 11B/11G Protection
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_BGProtection_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)

{
	switch (simple_strtol(arg, 0, 10))
	{
		case 0: //AUTO
			pAdapter->PortCfg.UseBGProtection = 0;
			break;
		case 1: //Always On
			pAdapter->PortCfg.UseBGProtection = 1;
			break;
		case 2: //Always OFF
			pAdapter->PortCfg.UseBGProtection = 2;
			break;		
		default:  //Invalid argument 
			return FALSE;
	}
	ApUpdateCapabilityAndErpIe(pAdapter);
	DBGPRINT(RT_DEBUG_TRACE, "Set_BGProtection_Proc::(BGProtection=%d)\n", pAdapter->PortCfg.UseBGProtection);	

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set TxAntenna
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_TxAntenna_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG					BbpCsr1;
	UCHAR					TxValue;

	RTMP_IO_READ32(pAdapter, BBPCSR1, &BbpCsr1);

	switch (simple_strtol(arg, 0, 10))
	{
		case 0: //Diversity
			pAdapter->PortCfg.CurrentTxAntenna = 0xFF;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Tx_Configure, &TxValue);
			TxValue = (TxValue & 0xFC) | 0x01;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Tx_Configure, TxValue);
			BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00010001;
			RTMP_IO_WRITE32(pAdapter, BBPCSR1, BbpCsr1);
			break;
		case 1: //Antenna A
			pAdapter->PortCfg.CurrentTxAntenna = 0x00;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Tx_Configure, &TxValue);
			TxValue = (TxValue & 0xFC) | 0x00;
			BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00000000;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Tx_Configure, TxValue);
			RTMP_IO_WRITE32(pAdapter, BBPCSR1, BbpCsr1);
			break;
		case 2: //Antenna B
			pAdapter->PortCfg.CurrentTxAntenna = 0x01;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Tx_Configure, &TxValue);
			TxValue = (TxValue & 0xFC) | 0x02;
			BbpCsr1 = (BbpCsr1 & 0xFFFCFFFC) | 0x00020002;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Tx_Configure, TxValue);
			RTMP_IO_WRITE32(pAdapter, BBPCSR1, BbpCsr1);
			break;
		default:  //Invalid argument 
			return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "Set_TxAntenna_Proc::(TxAntenna=%d)\n", pAdapter->PortCfg.CurrentTxAntenna);	

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set RxAntenna
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_RxAntenna_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG					BbpCsr1;
	UCHAR					RxValue;
	
	RTMP_IO_READ32(pAdapter, BBPCSR1, &BbpCsr1);

	switch (simple_strtol(arg, 0, 10))
	{
		case 0: //Diversity
			pAdapter->PortCfg.CurrentRxAntenna = 0xFF;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Rx_Configure, &RxValue);
			RxValue = (RxValue & 0xFC) | 0x01;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Rx_Configure, RxValue);	
			break;
		case 1: //Antenna A
			pAdapter->PortCfg.CurrentRxAntenna = 0x00;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Rx_Configure, &RxValue);
			RxValue = (RxValue & 0xFC) | 0x00;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Rx_Configure, RxValue);
			break;
		case 2: //Antenna B
			pAdapter->PortCfg.CurrentRxAntenna = 0x01;
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, BBP_Rx_Configure, &RxValue);
			RxValue = (RxValue & 0xFC) | 0x02;
			RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, BBP_Rx_Configure, RxValue);
			break;
		default: //Invalid argument 
			return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "Set_RxAntenna_Proc::(RxAntenna=%d)\n", pAdapter->PortCfg.CurrentRxAntenna);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set TxPreamble
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_TxPreamble_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	switch (simple_strtol(arg, 0, 10))
	{
		case Rt802_11PreambleLong:
			pAdapter->PortCfg.TxPreamble = Rt802_11PreambleLong;
			// MlmeSetTxPreamble(pAdapter, Rt802_11PreambleLong);
			break;
		case Rt802_11PreambleShort:
			pAdapter->PortCfg.TxPreamble = Rt802_11PreambleShort;
			// MlmeSetTxPreamble(pAdapter, Rt802_11PreambleShort);
			break;
		case Rt802_11PreambleAuto:
			pAdapter->PortCfg.TxPreamble = Rt802_11PreambleAuto;
			// MlmeSetTxPreamble(pAdapter, Rt802_11PreambleAuto);
			break;
		default: //Invalid argument 
			return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "Set_TxPreamble_Proc::(TxPreamble=%d)\n", pAdapter->PortCfg.TxPreamble);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set RTS Threshold
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_RTSThreshold_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG	RtsThresh;

	RtsThresh = simple_strtol(arg, 0, 10);

	if((RtsThresh >= 1) && (RtsThresh <= MAX_RTS_THRESHOLD))
		pAdapter->PortCfg.RtsThreshold  = (USHORT)RtsThresh;
	else
		return FALSE; //Invalid argument 

	DBGPRINT(RT_DEBUG_TRACE, "Set_RTSThreshold_Proc::(RTSThreshold=%d)\n", pAdapter->PortCfg.RtsThreshold);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Fragment Threshold
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_FragThreshold_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG	FragThresh;

	FragThresh = simple_strtol(arg, 0, 10);

	if ((FragThresh >= MIN_FRAG_THRESHOLD) || (FragThresh <= MAX_FRAG_THRESHOLD))
	{
		pAdapter->PortCfg.FragmentThreshold = FragThresh;
		pAdapter->PortCfg.AllowFragSize = (pAdapter->PortCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;
		pAdapter->PortCfg.bFragmentZeroDisable = TRUE;
	}
	else
		return FALSE; //Invalid argument 

	DBGPRINT(RT_DEBUG_TRACE, "Set_FragThreshold_Proc::(FragThreshold=%d)\n", pAdapter->PortCfg.FragmentThreshold);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set TxBurst
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_TxBurst_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG TxBurst;

	TxBurst = simple_strtol(arg, 0, 10);

	if (TxBurst == 1)
		pAdapter->PortCfg.EnableTxBurst = TRUE;
	else if (TxBurst == 0)
		pAdapter->PortCfg.EnableTxBurst = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_TxBurst_Proc::(TxBurst=%d)\n", pAdapter->PortCfg.EnableTxBurst);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set TurboRate Enable or Disable
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_TurboRate_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG TurboRate;

	TurboRate = simple_strtol(arg, 0, 10);

	if (TurboRate == 1)
		pAdapter->PortCfg.EnableTurboRate = TRUE;
	else if (TurboRate == 0)
		pAdapter->PortCfg.EnableTurboRate = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_TurboRate_Proc::(TurboRate=%d)\n", pAdapter->PortCfg.EnableTurboRate);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set No Forwarding Enable or Disable
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_NoForwarding_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG NoForwarding;

	NoForwarding = simple_strtol(arg, 0, 10);

	if (NoForwarding == 1)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IsolateInterStaTraffic = TRUE;
	else if (NoForwarding == 0)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IsolateInterStaTraffic = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_NoForwarding_Proc::(NoForwarding=%d)\n", pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IsolateInterStaTraffic);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set No Forwarding between each SSID
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_NoForwardingBTNSSID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG NoForwarding;

	NoForwarding = simple_strtol(arg, 0, 10);

	if (NoForwarding == 1)
		pAdapter->PortCfg.IsolateInterStaTrafficBTNSSID = TRUE;
	else if (NoForwarding == 0)
		pAdapter->PortCfg.IsolateInterStaTrafficBTNSSID = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_NoForwardingBTNSSID_Proc::(NoForwarding=%d)\n", pAdapter->PortCfg.IsolateInterStaTrafficBTNSSID);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Hide SSID Enable or Disable
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_HideSSID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG HideSsid;

	HideSsid = simple_strtol(arg, 0, 10);

	if (HideSsid == 1)
		HideSsid = TRUE;
	else if (HideSsid == 0)
		HideSsid = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	if (pAdapter->PortCfg.HideSsid != HideSsid)
	{
		pAdapter->PortCfg.HideSsid = HideSsid;
		// re-built BEACON frame format
		AsicDisableSync(pAdapter);
		MakeBssBeacon(pAdapter);
		UpdateBeaconFrame(pAdapter);
		AsicEnableBssSync(pAdapter);
	}

	DBGPRINT(RT_DEBUG_TRACE, "Set_HideSSID_Proc::(HideSSID=%d)\n", pAdapter->PortCfg.HideSsid);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Short Slot Time Enable or Disable
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ShortSlot_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG ShortSlot;

	ShortSlot = simple_strtol(arg, 0, 10);

	if (ShortSlot == 1)
		pAdapter->PortCfg.UseShortSlotTime = TRUE;
	else if (ShortSlot == 0)
		pAdapter->PortCfg.UseShortSlotTime = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_ShortSlot_Proc::(ShortSlot=%d)\n", pAdapter->PortCfg.UseShortSlotTime);

	return TRUE;
}

INT	Set_IEEE8021X_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
    ULONG ieee8021x;

	DBGPRINT(RT_DEBUG_TRACE, "==> Set_IEEE8021X_Proc\n");
	DBGPRINT(RT_DEBUG_TRACE, "arg=%s\n", arg);

	ieee8021x = simple_strtol(arg, 0, 10);

	if (ieee8021x == 1)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IEEE8021X = TRUE;
	else if (ieee8021x == 0)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IEEE8021X = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_IEEE8021X_Proc::(IEEE8021X=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].IEEE8021X);

	return TRUE;
}

INT	Set_PreAuth_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
    ULONG PreAuth;

	DBGPRINT(RT_DEBUG_TRACE, "==> Set_PreAuth_Proc\n");
	DBGPRINT(RT_DEBUG_TRACE, "arg=%s\n", arg);

	PreAuth = simple_strtol(arg, 0, 10);

	if (PreAuth == 1)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PreAuth = TRUE;
	else if (PreAuth == 0)
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PreAuth = FALSE;
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_PreAuth_Proc::(PreAuth=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PreAuth);

	return TRUE;
}

INT	Set_VLANID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_VLANID_Proc\n");
	DBGPRINT(RT_DEBUG_TRACE, "arg=%s\n", arg);

	pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].VLAN_VID = simple_strtol(arg, 0, 10);
	
	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_VLANID_Proc::(VLAN_VID=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].VLAN_VID);

	return TRUE;
}

INT	Set_VLANPriority_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_VLANPriority_Proc\n");
	DBGPRINT(RT_DEBUG_TRACE, "arg=%s\n", arg);

	pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].VLAN_Priority = simple_strtol(arg, 0, 10);

	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_VLANPriority_Proc::(VLAN_Priority=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].VLAN_Priority);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Authentication mode
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_AuthMode_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	if ((strcmp(arg, "OPEN") == 0) || (strcmp(arg, "open") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeOpen;
	else if ((strcmp(arg, "SHARED") == 0) || (strcmp(arg, "shared") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeShared;
	else if ((strcmp(arg, "WPAPSK") == 0) || (strcmp(arg, "wpapsk") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPAPSK;
	else if ((strcmp(arg, "WPA") == 0) || (strcmp(arg, "wpa") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPA;
	else if ((strcmp(arg, "WPA2PSK") == 0) || (strcmp(arg, "wpa2psk") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPA2PSK;
	else if ((strcmp(arg, "WPA2") == 0) || (strcmp(arg, "wpa2") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPA2;
	else if ((strcmp(arg, "WPA1WPA2") == 0) || (strcmp(arg, "wpa1wpa2") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPA1WPA2;
	else if ((strcmp(arg, "WPAPSKWPA2PSK") == 0) || (strcmp(arg, "wpapskwpa2psk") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode = Ndis802_11AuthModeWPAPSKWPA2PSK;
	else
		return FALSE;  

	pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PortSecured = WPA_802_1X_PORT_NOT_SECURED;
    RTMPMakeRSNIE(pAdapter, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus, pAdapter->PortCfg.IoctlIF);

	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_AuthMode_Proc::(AuthMode=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Encryption Type
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_EncrypType_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	if ((strcmp(arg, "NONE") == 0) || (strcmp(arg, "none") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = Ndis802_11WEPDisabled;
	else if ((strcmp(arg, "WEP") == 0) || (strcmp(arg, "wep") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = Ndis802_11WEPEnabled;
	else if ((strcmp(arg, "TKIP") == 0) || (strcmp(arg, "tkip") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = Ndis802_11Encryption2Enabled;
	else if ((strcmp(arg, "AES") == 0) || (strcmp(arg, "aes") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = Ndis802_11Encryption3Enabled;
	else if ((strcmp(arg, "TKIPAES") == 0) || (strcmp(arg, "tkipaes") == 0))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus = Ndis802_11Encryption4Enabled;
	else
		return FALSE;

    RTMPMakeRSNIE(pAdapter, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].AuthMode, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus, pAdapter->PortCfg.IoctlIF);
	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_EncrypType_Proc::(EncrypType=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].WepStatus);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Default Key ID
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_DefaultKeyID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG KeyIdx;

	KeyIdx = simple_strtol(arg, 0, 10);
	if((KeyIdx >= 1 ) && (KeyIdx <= 4))
		pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId = (UCHAR) (KeyIdx - 1 );
	else
		return FALSE;  //Invalid argument 
	
	DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_DefaultKeyID_Proc::(DefaultKeyID=%d)\n", pAdapter->PortCfg.IoctlIF, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].DefaultKeyId);

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set WEP KEY1
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Key1_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	int KeyLen;
	int i;

	KeyLen = strlen(arg);

	switch (KeyLen)
	{
		case 5: //wep 40 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key1_Proc::(Key1=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 10: //wep 40 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key1_Proc::(Key1=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		case 13: //wep 104 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key1_Proc::(Key1=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 26: //wep 104 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[0].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key1_Proc::(Key1=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key1_Proc::Invalid argument (=%s)\n", pAdapter->PortCfg.IoctlIF, arg);		
			return FALSE;
	}

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set WEP KEY2
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Key2_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	int KeyLen;
	int i;

	KeyLen = strlen(arg);

	switch (KeyLen)
	{
		case 5: //wep 40 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key2_Proc::(Key2=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 10: //wep 40 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key2_Proc::(Key2=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		case 13: //wep 104 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key2_Proc::(Key2=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 26: //wep 104 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[1].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key2_Proc::(Key2=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key2_Proc::Invalid argument (=%s)\n", pAdapter->PortCfg.IoctlIF, arg);		
			return FALSE;
	}

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set WEP KEY3
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Key3_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	int KeyLen;
	int i;

	KeyLen = strlen(arg);

	switch (KeyLen)
	{
		case 5: //wep 40 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key3_Proc::(Key3=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 10: //wep 40 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key3_Proc::(Key3=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		case 13: //wep 104 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key3_Proc::(Key3=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 26: //wep 104 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[2].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key3_Proc::(Key3=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_ERROR, "I/F(ra%d) Set_Key3_Proc::Invalid argument (=%s)\n", pAdapter->PortCfg.IoctlIF, arg);		
			return FALSE;
	}

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set WEP KEY4
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_Key4_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	int KeyLen;
	int i;

	KeyLen = strlen(arg);

	switch (KeyLen)
	{
		case 5: //wep 40 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key4_Proc::(Key4=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 10: //wep 40 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key4_Proc::(Key4=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		case 13: //wep 104 Ascii type
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].KeyLen = KeyLen;
			memcpy(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].Key, arg, KeyLen);	
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key4_Proc::(Key4=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Ascii");		
			break;
		case 26: //wep 104 Hex type
			for(i=0; i < KeyLen; i++)
			{
				if( !isxdigit(*(arg+i)) )
					return FALSE;  //Not Hex value;
			}
			pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].KeyLen = KeyLen / 2 ;
			AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SharedKey[3].Key, KeyLen / 2);
			DBGPRINT(RT_DEBUG_TRACE, "I/F(ra%d) Set_Key4_Proc::(Key4=%s and type=%s)\n", pAdapter->PortCfg.IoctlIF, arg, "Hex");		
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_ERROR, "I/F(ra%d) Set_Key4_Proc::Invalid argument (=%s)\n", pAdapter->PortCfg.IoctlIF, arg);		
			return FALSE;
	}

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set HotSpot counter policy
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_HSCounter_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	switch (simple_strtol(arg, 0, 10))
	{
		case 0: //Disable
			pAdapter->PortCfg.bEnableHSCounter = FALSE;
			break;
		case 1: //Enable
			pAdapter->PortCfg.bEnableHSCounter = TRUE;
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_ERROR, "Set_HSCounter_Proc::Invalid argument (=%s)\n", arg);		
			return FALSE;
	}
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_HSCounter_Proc::(bEnableHSCounter=%d)\n", pAdapter->PortCfg.bEnableHSCounter);

	return TRUE;	
}

/* 
    ==========================================================================
    Description:
        Set Access ctrol policy
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_AccessPolicy_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	switch (simple_strtol(arg, 0, 10))
	{
		case 0: //Disable
			pAdapter->PortCfg.AccessControlList.Policy = 0;
			break;
		case 1: //Allow All
			pAdapter->PortCfg.AccessControlList.Policy = 1;
			break;
		case 2: //Reject All
			pAdapter->PortCfg.AccessControlList.Policy = 2;
			break;
		default: //Invalid argument 
			DBGPRINT(RT_DEBUG_ERROR, "Set_AccessPolicy_Proc::Invalid argument (=%s)\n", arg);		
			return FALSE;
	}

	// check if ACL change affects any existent associtions
	ApUpdateAccessControlList(pAdapter);
	
	DBGPRINT(RT_DEBUG_TRACE, "Set_AccessPolicy_Proc::(AccessPolicy=%d)\n", pAdapter->PortCfg.AccessControlList.Policy);

	return TRUE;	
}

/* 
    ==========================================================================
    Description:
        Set Access ctrol mac table list
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_AccessControlList_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	MACADDR					macAddress;
	RT_802_11_ACL			acl;
	char					*this_char;
	char					*value;
	int						i,j;
	BOOLEAN					isDuplicat=FALSE;

	memset(&acl, 0x00, sizeof(RT_802_11_ACL));

	while ((this_char = strsep((char **)&arg, ";")) != NULL) 
	{
		if(strlen(this_char) != 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
			return FALSE;

         for (i=0, value = strtok(this_char,":"); value; value = strtok(NULL,":")) 
		 {
			if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
				return FALSE;  //Invalid

			AtoH(value, &macAddress.Octet[i++], 2);
		 }

		if(i != 6)
			return FALSE;  //Invalid

		//Check is duplicate
		isDuplicat = FALSE;
		for(j=0; j<acl.Num; j++)
			if(memcmp(acl.Entry[j].Addr, &macAddress, 6) == 0)
				isDuplicat = TRUE;

		if(!isDuplicat)
			memcpy(acl.Entry[acl.Num++].Addr, &macAddress, 6);

		if(acl.Num >= MAX_LEN_OF_MAC_TABLE)
			break;
	}

	acl.Policy = pAdapter->PortCfg.AccessControlList.Policy;
	memcpy(&pAdapter->PortCfg.AccessControlList, &acl, sizeof(RT_802_11_ACL));
	if (pAdapter->PortCfg.AccessControlList.Num > MAX_LEN_OF_MAC_TABLE)
		pAdapter->PortCfg.AccessControlList.Num = MAX_LEN_OF_MAC_TABLE;

	// check if ACL change affects any existent associtions
	ApUpdateAccessControlList(pAdapter);

	DBGPRINT(RT_DEBUG_TRACE, "Set::Set_AccessControlList_Proc(Policy=%d, Entry#=%d\n",
        pAdapter->PortCfg.AccessControlList.Policy, pAdapter->PortCfg.AccessControlList.Num);

#ifdef RT2500_DBG
	DBGPRINT(RT_DEBUG_TRACE, "=============== Entry ===============\n");
	for(i=0; i<pAdapter->PortCfg.AccessControlList.Num; i++)
	{
		DBGPRINT(RT_DEBUG_TRACE, "Entry #%02d: ", i+1);
		for(j=0; j<6; j++)
		   DBGPRINT(RT_DEBUG_TRACE, "%02X ", pAdapter->PortCfg.AccessControlList.Entry[i].Addr[j]);
		DBGPRINT(RT_DEBUG_TRACE, "\n");
	}
#endif
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set WPA PSK key

    Arguments:
        pAdapter            Pointer to our adapter
        arg                 WPA pre-shared key string

    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_WPAPSK_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	UCHAR       keyMaterial[40];

	DBGPRINT(RT_DEBUG_TRACE, "Set_WPAPSK_Proc::(WPAPSK=%s)\n", arg);

	if ((strlen(arg) < 8) || (strlen(arg) > 64))
	{
		DBGPRINT(RT_DEBUG_TRACE, "Set failed!!(WPAPSK=%s), WPAPSK key-string required 8 ~ 64 characters \n", arg);
		return FALSE;
	}

	if (strlen(arg) == 64)
	{
	    AtoH(arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK, 32);
	}
	else
	{
	    PasswordHash((char *)arg, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].Ssid, pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].SsidLen, keyMaterial);
	    NdisMoveMemory(pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].PMK, keyMaterial, 32);
	}

	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Reset statistics counter

    Arguments:
        pAdapter            Pointer to our adapter
        arg                 

    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ResetStatCounter_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
    UCHAR           apidx, i;
    MAC_TABLE_ENTRY *pEntry;
    
    DBGPRINT(RT_DEBUG_TRACE, "==>Set_ResetStatCounter_Proc\n");
    
    NdisZeroMemory(&pAdapter->WlanCounters, sizeof(COUNTER_802_11));
    NdisZeroMemory(&pAdapter->Counters, sizeof(COUNTER_802_3));
    NdisZeroMemory(&pAdapter->RalinkCounters, sizeof(COUNTER_RALINK));

    // Reset HotSpot counter
    for (apidx = 0; apidx < pAdapter->PortCfg.SsidNum; apidx++)
    {
        for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
        {
            pEntry = &pAdapter->PortCfg.MSSID[apidx].MacTab.Content[i];

            if ((pEntry->Valid == FALSE) || (pEntry->Sst != SST_ASSOC))
                continue;

            pEntry->HSCounter.LastDataPacketTime = 0;
            pEntry->HSCounter.TotalRxByteCount= 0;
            pEntry->HSCounter.TotalTxByteCount= 0;
        }
    }
    
    return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set Group Rekey interval type 

    Arguments:
        pAdapter            Pointer to our adapter
        arg                 

    Return:
          
    ==========================================================================
*/
INT	Set_RekeyMethod_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
    if ((strcmp(arg, "TIME") == 0) || (strcmp(arg, "time") == 0))
        pAd->PortCfg.WPAREKEY.ReKeyMethod = TIME_REKEY;
    else if ((strcmp(arg, "PKT") == 0) || (strcmp(arg, "pkt") == 0))
        pAd->PortCfg.WPAREKEY.ReKeyMethod = PKT_REKEY;
    else if ((strcmp(arg, "NONE") == 0) || (strcmp(arg, "none") == 0))
    {
        pAd->PortCfg.WPAREKEY.ReKeyMethod = DISABLE_REKEY;
        pAd->PortCfg.WPAREKEY.ReKeyInterval = 0;
    }
    else
        return FALSE;
    
    if ((pAd->PortCfg.REKEYTimerRunning == FALSE) && (pAd->PortCfg.WPAREKEY.ReKeyInterval != 0)
        && (pAd->PortCfg.WPAREKEY.ReKeyMethod != DISABLE_REKEY))
    {
        add_timer(&pAd->PortCfg.REKEYTimer);
        pAd->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
        pAd->PortCfg.REKEYTimerRunning = TRUE;
        pAd->PortCfg.REKEYCOUNTER = 0;
    }
    DBGPRINT(RT_DEBUG_TRACE, "Set_RekeyMethod_Proc::(RekeyMethod=%d)\n", pAd->PortCfg.WPAREKEY.ReKeyMethod);
    
    return TRUE;
}

INT	Set_RekeyInterval_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
    ULONG   period;
    
    period = simple_strtol(arg, 0, 10);
    if ((pAd->PortCfg.REKEYTimerRunning == TRUE) && (period == 0))
    {
        del_timer_sync(&pAd->PortCfg.REKEYTimer);
        pAd->PortCfg.WPAREKEY.ReKeyInterval = 0;
        pAd->PortCfg.REKEYTimerRunning = FALSE;
    }
    else if (period <= MAX_REKEY_INTER)
    {
        pAd->PortCfg.WPAREKEY.ReKeyInterval = period;
        if (pAd->PortCfg.REKEYTimerRunning == FALSE)
        {
            add_timer(&pAd->PortCfg.REKEYTimer);
            pAd->PortCfg.REKEYTimer.expires = jiffies + GUPDATE_EXEC_INTV;
            pAd->PortCfg.REKEYTimerRunning = TRUE;
            pAd->PortCfg.REKEYCOUNTER = 0;
        }
    }
    else
        return FALSE;

    DBGPRINT(RT_DEBUG_TRACE, "Set_RekeyInterval_Proc::(RekeyInterval=%d)\n", period);
    return TRUE;
}

INT	Set_PMKCachePeriod_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
    pAd->PortCfg.PMKCachePeriod = simple_strtol(arg, 0, 10) * 60 * HZ;
    
    DBGPRINT(RT_DEBUG_TRACE, "Set_PMKCachePeriod_Proc::(PMKCachePeriod=%d)\n", pAd->PortCfg.PMKCachePeriod);
    return TRUE;
}

#ifdef VINCENT_K
// RaLink support
INT	Set_Rx_Calibration_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
  	//step 1: RF Change to ch14_20Mhz=2504M

  	//step 2: R16=BDh, R17=14h
  	UCHAR   RSET1 = 0xbd;
  	UCHAR   RSET2 = 0x24;
  	UCHAR   tmp_R17;
  	UCHAR   value[5], avg = 0;
  	ULONG   sum = 0, i, A0, B8;
  	USHORT  R17, diff;
  	ULONG   Value;

  	//if(gWlanState != WLAN_STATE_INIT_OK)return VNIC_Not_Init(pOutRst, pOutData);

  	diff = 8;

  	pAd->ate.Mode = ATE_RXCALIBRATION;

  	RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 17, &tmp_R17);
  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 16, RSET1);
  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 17, RSET2);

  	//read FCS error and false CCA from MAC
  	RTMP_IO_READ32(pAd, CNT0, &A0);
  	RTMP_IO_READ32(pAd, CNT3, &B8);

 	RTMP_RF_IO_WRITE32(pAd, 0x94022020);
  	RTMP_RF_IO_WRITE32(pAd, 0x94000d2a);
  	RTMP_RF_IO_WRITE32(pAd, 0x94062d11);
  	RTMP_RF_IO_WRITE32(pAd, 0x94000a03);
  	//pAd->PortCfg.LatchRfRegs.Channel = Channel;
  	pAd->PortCfg.LatchRfRegs.R1 = 0x94022020;
  	pAd->PortCfg.LatchRfRegs.R2 = 0x94000d2a;
  	pAd->PortCfg.LatchRfRegs.R3 = 0x94062d11;
  	pAd->PortCfg.LatchRfRegs.R4 = 0x94000a03;

  	//step 3: Need receive error packets.
  	// read to clear counters
  	RTMP_IO_READ32(pAd, CNT0, &Value);
  	RTMP_IO_READ32(pAd, CNT1, &Value);
  	RTMP_IO_READ32(pAd, CNT2, &Value);
  	RTMP_IO_READ32(pAd, CNT4, &Value);

  	RTMP_IO_WRITE32(pAd, RXCSR0, 0x0);

  	//step 4: Read R51(RSSI)  5times/5=Nh
  	for (i = 0; i<5;i++)
  	{
    	RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 51, &value[i]);
    	sum += value[i];
  	}

  	RTMP_IO_READ32(pAd, CNT0, &A0);
  	RTMP_IO_READ32(pAd, CNT3, &B8);

  	avg = (UCHAR)sum/5;
  	if (avg > 0x24)
  		avg = 0x24;
  	else if (avg < 0xe)
    	avg = 0xe;

  	//step 5:-90dbm-(15h-Nh) = Xdbm
  	//step 6: Dec = -88dbm - Xdbm
  	//step 7: EEPROM offset 28h; R17 = 0x32-Dec(h)
  	R17 = 24 +avg + 0x1100;
  	RTMP_EEPROM_WRITE16(pAd, 0x28, R17);

  	//step 8: EEPROM offset 7Ch=0x79-Dec(h)
  	R17 =RTMP_EEPROM_READ16(pAd, 0x7C);
  	R17 &=0xff00;
  	R17 |= (95 + avg );

  	//back to initial.
  	RSET1 =0x2C;
  	RSET2 =24 + avg +diff ; //+8 for gary
  	if(RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_ATE_WRITE_BBP_R17))
    	RSET2=tmp_R17;

  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 16, RSET1);
  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 17, RSET2);
  	//channel 1
  	RTMP_RF_IO_WRITE32(pAd, 0x94022020);
  	RTMP_RF_IO_WRITE32(pAd, 0x94080c9e);
  	RTMP_RF_IO_WRITE32(pAd, 0x94060111);
  	RTMP_RF_IO_WRITE32(pAd, 0x94000a1b);

  	RTMP_IO_WRITE32(pAd, RXCSR0, 0xffffffff);
  	//*pOutRst = CMD_RTN_OK;

    DBGPRINT(RT_DEBUG_TRACE, "Set_Rx_Calibration_Proc::\n");
    return TRUE;
}

#endif

#ifdef RALINK_ATE
/* 
    ==========================================================================
    Description:
        Read / Write BBP
Arguments:
    pAdapter                    Pointer to our adapter
    wrq                         Pointer to the ioctl argument

    Return Value:
        None

    Note:
        Usage: 
               1.) iwpriv ra0 bbp               ==> read all BBP
               2.) iwpriv ra0 bbp 1,2,10,32     ==> raed BBP where ID=1,2,10,32
               3.) iwpriv ra0 bbp 1=10,17=3E    ==> write BBP R1=0x10, R17=0x3E
    ==========================================================================
*/
VOID RTMPIoctlBBP(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	struct iwreq	*wrq)
{
	char				*this_char;
	char				*value;
	int					i=0;
	int					count = 0;
	UCHAR				regBBP;
	char				msg[2048];
	char				arg[255];
	char				*ptr;
	LONG				bbpId;
	LONG				bbpValue;
	BOOLEAN				bIsPrintAllBBP = FALSE;

	DBGPRINT(RT_DEBUG_TRACE, "==>RTMPIoctlBBP\n");

	memset(msg, 0x00, 2048);
	if (wrq->u.data.length > 1) //No parameters.
	{
		memcpy(arg, wrq->u.data.pointer, (wrq->u.data.length > 255) ? 255 : wrq->u.data.length);
		ptr = arg;
		sprintf(msg, "\n");
		//Parsing Read or Write
		while ((this_char = strsep(&ptr, ",")) != NULL)
		{
			i++;
			DBGPRINT(RT_DEBUG_TRACE, "this_char=%s\n", this_char);
			if (!*this_char)
				continue;

			if ((value = strchr(this_char, '=')) != NULL)
				*value++ = 0;

			if (!value || !*value)
			{ //Read
				DBGPRINT(RT_DEBUG_TRACE, "this_char=%s, value=%s\n", this_char, value);
				if (sscanf(this_char, "%d", &(bbpId)) == 1)
				{
					if ((bbpId >=0) && (bbpId <= 74))
					{
						RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, bbpId, &regBBP);
						sprintf(msg+strlen(msg), "R%02d[0x%02X]:%02X  ", bbpId, bbpId*2, regBBP);
						count++;
						if (count%5 == 4)
							sprintf(msg+strlen(msg), "\n");
						DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
					}
					else
					{//Invalid parametes, so default printk all bbp
						bIsPrintAllBBP = TRUE;
						break;
					}
				}
				else
				{ //Invalid parametes, so default printk all bbp
					bIsPrintAllBBP = TRUE;
					break;
				}
			}
			else
			{ //Write
				DBGPRINT(RT_DEBUG_TRACE, "this_char=%s, value=%s\n", this_char, value);
				if ((sscanf(this_char, "%d", &(bbpId)) == 1) && (sscanf(value, "%x", &(bbpValue)) == 1))
				{
					DBGPRINT(RT_DEBUG_TRACE, "bbpID=%02d, value=0x%x\n", bbpId, bbpValue);
					if ((bbpId >=0) && (bbpId <= 74))
					{
						RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, (UCHAR) bbpId, (UCHAR) bbpValue);
						//Read it back for showing
						RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, bbpId, &regBBP);
						sprintf(msg+strlen(msg), "R%02d[0x%02X]:%02X  ", bbpId, bbpId*2, regBBP);
						count++;
						if (count%5 == 4)
							sprintf(msg+strlen(msg), "\n");
						DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
					}
					else
					{//Invalid parametes, so default printk all bbp
						bIsPrintAllBBP = TRUE;
						break;
					}
				}
				else
				{ //Invalid parametes, so default printk all bbp
					bIsPrintAllBBP = TRUE;
					break;
				}
			}
		}
	}
	else
		bIsPrintAllBBP = TRUE;

	if (bIsPrintAllBBP)
	{
		memset(msg, 0x00, 2048);
		sprintf(msg, "\n");
		for (i = 0; i <= 74; i++)
		{
			RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, i, &regBBP);
			sprintf(msg+strlen(msg), "R%02d[0x%02X]:%02X  ", i, i*2, regBBP);
			if (i%5 == 4)
				sprintf(msg+strlen(msg), "\n");
		}
		// Copy the information into the user buffer
		DBGPRINT(RT_DEBUG_TRACE, "strlen(msg) =%d\n", strlen(msg));
		wrq->u.data.length = strlen(msg);
		copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
	}
	else
	{
		DBGPRINT(RT_DEBUG_TRACE, "copy to user [msg=%s]\n", msg);
		// Copy the information into the user buffer
		DBGPRINT(RT_DEBUG_TRACE, "strlen(msg) =%d\n", strlen(msg));
		wrq->u.data.length = strlen(msg);
		copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
	}
	DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlBBP\n");
}

/* 
    ==========================================================================
    Description:
        Read / Write MAC
Arguments:
    pAdapter                    Pointer to our adapter
    wrq                         Pointer to the ioctl argument

    Return Value:
        None

    Note:
        Usage: 
               1.) iwpriv ra0 mac 0        ==> read MAC where Addr=0x0
               2.) iwpriv ra0 mac 0=12     ==> write MAC where Addr=0x0, value=12
    ==========================================================================
*/
VOID RTMPIoctlMAC(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
	char				*this_char;
	char				*value;
	int					j=0, k=0;
	int					count = 0;
	char				msg[1024];
	char				arg[255];
	char				*ptr;
	ULONG				macAddr = 0;
	UCHAR				temp[16], temp2[16];
	ULONG				macValue;

	DBGPRINT(RT_DEBUG_TRACE, "==>RTMPIoctlMAC\n");
	memset(msg, 0x00, 1024);
	if (wrq->u.data.length > 1) //No parameters.
	{
		memcpy(arg, wrq->u.data.pointer, (wrq->u.data.length > 255) ? 255 : wrq->u.data.length);
		ptr = arg;
		sprintf(msg, "\n");
		//Parsing Read or Write
		while ((this_char = strsep(&ptr, ",")) != NULL)
		{
			DBGPRINT(RT_DEBUG_TRACE, "this_char=%s\n", this_char);
			if (!*this_char)
				continue;

			if ((value = strchr(this_char, '=')) != NULL)
				*value++ = 0;

			if (!value || !*value)
			{ //Read
				DBGPRINT(RT_DEBUG_TRACE, "Read: this_char=%s, strlen=%d\n", this_char, strlen(this_char));

				// Sanity check
				if(strlen(this_char) > 4)
					break;

				j = strlen(this_char);
				while(j-- > 0)
				{
					if(this_char[j] > 'f' || this_char[j] < '0')
						return;
				}

				// Mac Addr
				k = j = strlen(this_char);
				while(j-- > 0)
				{
					this_char[4-k+j] = this_char[j];
				}
				
				while(k < 4)
					this_char[3-k++]='0';
				this_char[4]='\0';

				if(strlen(this_char) == 4)
				{
					AtoH(this_char, temp, 4);
					macAddr = *temp*256 + temp[1];					
					if (macAddr < 0xFFFF)
					{
						RTMP_IO_READ32(pAdapter, macAddr, &macValue);
						DBGPRINT(RT_DEBUG_TRACE, "macAddr=%x, regMAC=%x\n", macAddr, macValue);
						sprintf(msg+strlen(msg), "[0x%08X]:%08X  ", macAddr , macValue);
						count++;
						if (count%5 == 4)
							sprintf(msg+strlen(msg), "\n");
						DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
					}
					else
					{//Invalid parametes, so default printk all bbp
						break;
					}
				}
			}
			else
			{ //Write
				DBGPRINT(RT_DEBUG_TRACE, "Write: this_char=%s, strlen(value)=%d, value=%s\n", this_char, strlen(value), value);
				memcpy(&temp2, value, strlen(value));
				temp2[strlen(value)] = '\0';

				// Sanity check
				if((strlen(this_char) > 4) || strlen(temp2) > 8)
					break;

				j = strlen(this_char);
				while(j-- > 0)
				{
					if(this_char[j] > 'f' || this_char[j] < '0')
						return;
				}

				j = strlen(temp2);
				while(j-- > 0)
				{
					if(temp2[j] > 'f' || temp2[j] < '0')
						return;
				}

				//MAC Addr
				k = j = strlen(this_char);
				while(j-- > 0)
				{
					this_char[4-k+j] = this_char[j];
				}

				while(k < 4)
					this_char[3-k++]='0';
				this_char[4]='\0';

				//MAC value
				k = j = strlen(temp2);
				while(j-- > 0)
				{
					temp2[8-k+j] = temp2[j];
				}
				
				while(k < 8)
					temp2[7-k++]='0';
				temp2[8]='\0';

				{
					AtoH(this_char, temp, 4);
					macAddr = *temp*256 + temp[1];

					AtoH(temp2, temp, 8);
					macValue = *temp*256*256*256 + temp[1]*256*256 + temp[2]*256 + temp[3];

					DBGPRINT(RT_DEBUG_TRACE, "macAddr=%02x, macValue=0x%x\n", macAddr, macValue);
					
					RTMP_IO_WRITE32(pAdapter, macAddr, macValue);
					sprintf(msg+strlen(msg), "[0x%02X]:%02X  ", macAddr, macValue);
					count++;
					if (count%5 == 4)
						sprintf(msg+strlen(msg), "\n");
					DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
				}
			}
		}
	}

	if(strlen(msg) == 1)
		sprintf(msg+strlen(msg), "===>Error command format!");
	DBGPRINT(RT_DEBUG_TRACE, "copy to user [msg=%s]\n", msg);
	// Copy the information into the user buffer
	DBGPRINT(RT_DEBUG_TRACE, "strlen(msg) =%d\n", strlen(msg));
	wrq->u.data.length = strlen(msg);
	copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
	
	DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlMAC\n");
}

/* 
    ==========================================================================
    Description:
        Read / Write E2PROM
Arguments:
    pAdapter                    Pointer to our adapter
    wrq                         Pointer to the ioctl argument

    Return Value:
        None

    Note:
        Usage: 
               1.) iwpriv ra0 e2p 0     	==> read E2PROM where Addr=0x0
               2.) iwpriv ra0 e2p 0=1234    ==> write E2PROM where Addr=0x0, value=1234
    ==========================================================================
*/
VOID RTMPIoctlE2PROM(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
	char				*this_char;
	char				*value;
	int					j=0, k=0;
	int					count = 0;
	char				msg[1024];
	char				arg[255];
	char				*ptr;
	USHORT				eepAddr = 0;
	UCHAR				temp[16], temp2[16];
	USHORT				eepValue;

	DBGPRINT(RT_DEBUG_TRACE, "==>RTMPIoctlE2PROM\n");
	memset(msg, 0x00, 1024);
	if (wrq->u.data.length > 1) //No parameters.
	{
		memcpy(arg, wrq->u.data.pointer, (wrq->u.data.length > 255) ? 255 : wrq->u.data.length);
		ptr = arg;
		#ifndef VINCENT_K
		sprintf(msg, "\n");
		#endif
		//Parsing Read or Write
		while ((this_char = strsep(&ptr, ",")) != NULL)
		{
			DBGPRINT(RT_DEBUG_TRACE, "this_char=%s\n", this_char);
			if (!*this_char)
				continue;

			if ((value = strchr(this_char, '=')) != NULL)
				*value++ = 0;

			if (!value || !*value)
			{ //Read
				DBGPRINT(RT_DEBUG_TRACE, "Read: this_char=%s, strlen=%d\n", this_char, strlen(this_char));
				this_char[4]='\0'; //Vincent
				#ifdef VINCENT_K
				//printk("kernel: Read: this_char=%s, strlen=%d\n", this_char, strlen(this_char));
				#endif

				// Sanity check
				if(strlen(this_char) > 4)
					break;

				j = strlen(this_char);
				while(j-- > 0)
				{
					if(this_char[j] > 'f' || this_char[j] < '0')
						return;
				}

				// E2PROM addr
				k = j = strlen(this_char);
				while(j-- > 0)
				{
					this_char[4-k+j] = this_char[j];
				}
				
				while(k < 4)
					this_char[3-k++]='0';
				this_char[4]='\0';

				if(strlen(this_char) == 4)
				{
					AtoH(this_char, temp, 4);
					eepAddr = *temp*256 + temp[1];					
					if (eepAddr < 0xFFFF)
					{
						eepValue = RTMP_EEPROM_READ16(pAdapter, eepAddr);
						DBGPRINT(RT_DEBUG_TRACE, "eepAddr=%x, eepValue=%x\n", eepAddr, eepValue);
						#ifdef VINCENT_K
						//printk("eepAddr=%x, eepValue=%04x\n", eepAddr, eepValue);
						sprintf(msg+strlen(msg), "%04X ", eepValue);
						#else
						sprintf(msg+strlen(msg), "[0x%04X]:%04X  ", eepAddr , eepValue);
						#endif
						count++;
						if (count%5 == 4)
							sprintf(msg+strlen(msg), "\n");
						DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
					}
					else
					{//Invalid parametes, so default printk all bbp
						break;
					}
				}
			}
			else
			{ //Write
				DBGPRINT(RT_DEBUG_TRACE, "Write: this_char=%s, strlen(value)=%d, value=%s\n", this_char, strlen(value), value);
				#ifdef VINCENT_K
				//printk("kernel: Write: this_char=%s, strlen(value)=%d, value=%s\n", this_char, strlen(value), value);
				#endif
				memcpy(&temp2, value, strlen(value));
				temp2[strlen(value)] = '\0';

				// Sanity check
				if((strlen(this_char) > 4) || strlen(temp2) > 8)
					break;

				j = strlen(this_char);
				while(j-- > 0)
				{
					if(this_char[j] > 'f' || this_char[j] < '0')
						return;
				}
				j = strlen(temp2);
				while(j-- > 0)
				{
					if(temp2[j] > 'f' || temp2[j] < '0')
						return;
				}

				//MAC Addr
				k = j = strlen(this_char);
				while(j-- > 0)
				{
					this_char[4-k+j] = this_char[j];
				}

				while(k < 4)
					this_char[3-k++]='0';
				this_char[4]='\0';

				//MAC value
				k = j = strlen(temp2);
				while(j-- > 0)
				{
					temp2[4-k+j] = temp2[j];
				}
				
				while(k < 4)
					temp2[3-k++]='0';
				temp2[4]='\0';

				AtoH(this_char, temp, 4);
				eepAddr = *temp*256 + temp[1];

				AtoH(temp2, temp, 4);
				eepValue = *temp*256 + temp[1];

				RTMP_EEPROM_WRITE16(pAdapter, eepAddr, eepValue);
				DBGPRINT(RT_DEBUG_TRACE, "eepAddr=%02x, eepValue=0x%x\n", eepAddr, eepValue);
				#ifdef VINCENT_K
				//printk("kernel: eepAddr=%02x, eepValue=0x%x\n", eepAddr, eepValue);
				
				sprintf(msg+strlen(msg), "%02X ", eepValue);
				#else
				sprintf(msg+strlen(msg), "[0x%02X]:%02X  ", eepAddr, eepValue);
				#endif
				count++;
				if (count%5 == 4)
					sprintf(msg+strlen(msg), "\n");
				DBGPRINT(RT_DEBUG_TRACE, "msg=%s\n", msg);
			}
		}
	}

	if(strlen(msg) == 1)
		sprintf(msg+strlen(msg), "===>Error command format!");

    // Copy the information into the user buffer	
    DBGPRINT(RT_DEBUG_TRACE, "copy to user [msg=%s]\n", msg);	
	wrq->u.data.length = strlen(msg);
	copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
	
	DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlE2PROM\n");
}
#endif //#ifdef RALINK_ATE

/* 
    ==========================================================================
    Description:
        Read statistics counter
Arguments:
    pAdapter                    Pointer to our adapter
    wrq                         Pointer to the ioctl argument

    Return Value:
        None

    Note:
        Usage: 
               1.) iwpriv ra0 stat 0     	==> Read statistics counter
    ==========================================================================
*/
VOID RTMPIoctlStatistics(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
    char                msg[1024];
    ULONG               Value;
	u8					*ptr;
    
    DBGPRINT(RT_DEBUG_TRACE, "==>RTMPIoctlStatistics\n");

    RTMP_IO_READ32(pAdapter, CNT0, &Value);
    pAdapter->WlanCounters.FCSErrorCount += Value;

    RTMP_IO_READ32(pAdapter, CNT3, &Value);
	pAdapter->PrivateInfo.CCAErrCnt = (Value & 0x0000ffff);

    memset(msg, 0, 1024);
#ifdef VINCENT_K
	#ifdef VINCENT_K_DBG
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.TransmittedFragmentCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.MulticastTransmittedFrameCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.FailedCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.RetryCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.MultipleRetryCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.RTSSuccessCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.RTSFailureCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.ACKFailureCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.FrameDuplicateCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.ReceivedFragmentCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.MulticastReceivedFrameCount);
    sprintf(msg+strlen(msg), "%d.", (ULONG)pAdapter->WlanCounters.FCSErrorCount);
	#endif

	#ifdef LITTLE_ENDIAN
	ptr = (u8 *) &(pAdapter->WlanCounters.TransmittedFragmentCount);
    sprintf(msg+strlen(msg), "%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);

	ptr = (u8 *) &(pAdapter->WlanCounters.MulticastTransmittedFrameCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);

	ptr = (u8 *) &(pAdapter->WlanCounters.FailedCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);

	ptr = (u8 *) &(pAdapter->WlanCounters.RetryCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.MultipleRetryCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.RTSSuccessCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.RTSFailureCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.ACKFailureCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.FrameDuplicateCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.ReceivedFragmentCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.MulticastReceivedFrameCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.FCSErrorCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[0], ptr[1], ptr[2], ptr[3]);
	#endif
	#ifdef BIG_ENDIAN
	ptr = (u8 *) &(pAdapter->WlanCounters.TransmittedFragmentCount);
    sprintf(msg+strlen(msg), "%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);

	ptr = (u8 *) &(pAdapter->WlanCounters.MulticastTransmittedFrameCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);

	ptr = (u8 *) &(pAdapter->WlanCounters.FailedCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);

	ptr = (u8 *) &(pAdapter->WlanCounters.RetryCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.MultipleRetryCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.RTSSuccessCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.RTSFailureCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.ACKFailureCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.FrameDuplicateCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.ReceivedFragmentCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.MulticastReceivedFrameCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	
	ptr = (u8 *) &(pAdapter->WlanCounters.FCSErrorCount);
    sprintf(msg+strlen(msg), ".%x.%x.%x.%x", ptr[3], ptr[2], ptr[1], ptr[0]);
	#endif
#else
    sprintf(msg, "\n");

    sprintf(msg+strlen(msg), "Frame Trans success               = %d\n", (ULONG)pAdapter->WlanCounters.TransmittedFragmentCount);
    sprintf(msg+strlen(msg), "Frame Trans success without retry = %d\n", (ULONG)pAdapter->WlanCounters.TransmittedFragmentCount - (ULONG)pAdapter->WlanCounters.RetryCount);
    sprintf(msg+strlen(msg), "Frame Trans success after retry   = %d\n", (ULONG)pAdapter->WlanCounters.RetryCount);
    sprintf(msg+strlen(msg), "Fail to Rcv ACK after all retry   = %d\n", (ULONG)pAdapter->WlanCounters.FailedCount);
    sprintf(msg+strlen(msg), "RTS Success Rcv CTS               = %d\n", (ULONG)pAdapter->WlanCounters.RTSSuccessCount);
    sprintf(msg+strlen(msg), "RTS Fail Rcv CTS                  = %d\n", (ULONG)pAdapter->WlanCounters.RTSFailureCount);

    sprintf(msg+strlen(msg), "Frame Rcv success                 = %d\n", (ULONG)pAdapter->Counters.GoodReceives);
    sprintf(msg+strlen(msg), "Frame Rcv with CRC                = %d\n", (ULONG)pAdapter->WlanCounters.FCSErrorCount);
    sprintf(msg+strlen(msg), "Frame drop due to out of resource = %d\n", (ULONG)pAdapter->Counters.RxNoBuffer);
    sprintf(msg+strlen(msg), "Duplicate frame Rcv               = %d\n", (ULONG)pAdapter->WlanCounters.FrameDuplicateCount);

    sprintf(msg+strlen(msg), "False CCA (per query)             = %d\n\n", (ULONG)pAdapter->PrivateInfo.CCAErrCnt);
#endif
	
    // Copy the information into the user buffer
    DBGPRINT(RT_DEBUG_TRACE, "copy to user [msg=%s]\n", msg);
    wrq->u.data.length = strlen(msg);
    copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
    
    DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlStatistics\n");
}

/* 
    ==========================================================================
    Description:
        Read Site Survey results
Arguments:
    pAdapter                    Pointer to our adapter
    wrq                         Pointer to the ioctl argument

    Return Value:
        None

    Note:
        Usage: 
               1.) iwpriv ra0 site_survey   	==> Read site survey results
    ==========================================================================
*/

char *GetEncryptType(char enc)
{
    if(enc == Ndis802_11WEPDisabled)
        return "NONE";
    if(enc == Ndis802_11WEPEnabled)
    	return "WEP";
    if(enc == Ndis802_11Encryption2Enabled)
    	return "TKIP";
    if(enc == Ndis802_11Encryption3Enabled)
    	return "AES";
    else
    	return "UNKNOW";
}		
    	
VOID RTMPIoctlGetSiteSurvey(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
    #define 		LINE_LEN	(3+33+8+20)
    char                msg[MAX_LEN_OF_BSS_TABLE*LINE_LEN];
    int 		i=0;     
    char *GetEncryptType(char enc);
	int LOW=35, HIGH=75;
	int rssi_qual, rssi;
    extern void AP_SiteSurvey(IN	PRTMP_ADAPTER	pAdapter);
    
    memset(msg, 0 ,MAX_LEN_OF_BSS_TABLE*LINE_LEN );
    sprintf(msg,"%s","\n");
    for(i=0;i<MAX_LEN_OF_BSS_TABLE;i++)
	{
    	if(pAdapter->PortCfg.BssTab.BssEntry[i].Channel==0)
	    	break;
    	if(pAdapter->PortCfg.BssTab.BssEntry[i].SsidLen==0)
	    	continue;
		if((strlen(msg)+LINE_LEN ) >= (MAX_LEN_OF_BSS_TABLE*LINE_LEN) )
	    	break;	
	
		// RSSI to singal quality, 
		rssi = pAdapter->PortCfg.BssTab.BssEntry[i].Rssi;
		//rssi_qual = rssi - pAdapter->PortCfg.RssiToDbm;
		if (rssi > HIGH)
			rssi = HIGH;
		else if (rssi < LOW)
			rssi = LOW;
		rssi_qual = (rssi-LOW)*100/(HIGH-LOW);

		if(rssi_qual <= 0)
			rssi_qual = 1;

		sprintf(msg+strlen(msg),"%d", pAdapter->PortCfg.BssTab.BssEntry[i].Channel);
		sprintf(msg+strlen(msg),"\t%s", pAdapter->PortCfg.BssTab.BssEntry[i].Ssid);
		sprintf(msg+strlen(msg),"\t%02x%02x%02x%02x%02x%02x", 
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[0], 
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[1],
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[2], 
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[3], 
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[4], 
			pAdapter->PortCfg.BssTab.BssEntry[i].Bssid.Octet[5]);
		sprintf(msg+strlen(msg),"\t%-8s",GetEncryptType(pAdapter->PortCfg.BssTab.BssEntry[i].WepStatus));	
		sprintf(msg+strlen(msg),"\t%d\n", rssi_qual);
    }

    wrq->u.data.length = strlen(msg);
    copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);

#ifdef VINCENT_SITESURVEY
	// Re-Scan APs, Vincent 2005/05/13
    AP_SiteSurvey(pAdapter);
#endif
}

VOID RTMPIoctlSiteSurvey(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
    extern void AP_SiteSurvey(IN	PRTMP_ADAPTER	pAdapter);

    DBGPRINT(RT_DEBUG_TRACE, "==>RTMPIoctlSiteSurvey\n");
  
    AP_SiteSurvey(pAdapter);
    
    DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlSiteSurvey\n");
}

VOID RTMPIoctlGetMacTable(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	struct iwreq	*wrq)
{
	INT		i;
	CHAR	msg[MAX_LEN_OF_MAC_TABLE * MAC_ADDR_LEN];

	memset(msg, 0 , MAX_LEN_OF_MAC_TABLE * MAC_ADDR_LEN);
    sprintf(msg, "%s", "\n");

#ifdef VINCENT_WDSTKIP
	for (i=0; pAdapter->WdsTab.WdsEntry[i].Valid == TRUE; i++)
	{
		sprintf(msg+strlen(msg),"%02x:%02x:%02x:%02x:%02x:%02x\n", 
			pAdapter->WdsTab.WdsEntry[i].WdsAddr[0], pAdapter->WdsTab.WdsEntry[i].WdsAddr[1],
			pAdapter->WdsTab.WdsEntry[i].WdsAddr[2], pAdapter->WdsTab.WdsEntry[i].WdsAddr[3],
			pAdapter->WdsTab.WdsEntry[i].WdsAddr[4], pAdapter->WdsTab.WdsEntry[i].WdsAddr[5]);
	}
#else
	for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++)
	{
		if (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Valid && (pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Sst == SST_ASSOC))
		{
			sprintf(msg+strlen(msg),"%02x:%02x:%02x:%02x:%02x:%02x\n", 
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[0], 
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[1],
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[2], 
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[3], 
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[4], 
				pAdapter->PortCfg.MSSID[pAdapter->PortCfg.IoctlIF].MacTab.Content[i].Addr.Octet[5]);
		}
	}				
#endif
	wrq->u.data.length = strlen(msg);
	copy_to_user(wrq->u.data.pointer, msg, wrq->u.data.length);
    
    DBGPRINT(RT_DEBUG_TRACE, "<==RTMPIoctlGetMacTable (wrq->u.data.length=%d)\n", wrq->u.data.length);
}


#ifdef RALINK_ATE
UCHAR TempletFrame[24] = {0x08,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00,0xAA,0xBB,0x12,0x34,0x56,0x00,0x11,0x22,0xAA,0xBB,0xCC,0x00,0x00};	// 802.11 MAC Header, Type:Data, Length:24bytes 

#ifdef VINCENT_K
void send_test_data(IN PRTMP_ADAPTER pAd)
{
 	UINT i;
  	PTXD_STRUC pTxD;
  	unsigned char *pDest;
  	UINT j;
  	for (i = 0 ; (i < 1) && ((i < pAd->ate.TxCount) || (pAd->ate.TxCount == 0)) ; i++)
  	{
    	pTxD = (PTXD_STRUC)pAd->TxRing[pAd->CurEncryptIndex].va_addr;
    	pDest = (unsigned char *)pAd->TxRing[pAd->CurEncryptIndex].va_data_addr;

		// Prepare frame payload
		memcpy(pDest, &TempletFrame, LENGTH_802_11);
		for(j = LENGTH_802_11; j < PRIO_BUFFER_SIZE; j++)
			pDest[j] = 0xAA;
		memcpy(&pDest[10], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);
		memcpy(&pDest[16], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);

		RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
			SHORT_RETRY, IFS_BACKOFF, pAd->ate.TxRate, 4,
			pAd->ate.TxLength, pAd->PortCfg.TxPreambleInUsed, 0);

    	pAd->CurEncryptIndex++;
    	if (pAd->CurEncryptIndex >= TX_RING_SIZE)
    	{
      		pAd->CurEncryptIndex = 0;
    	}
		//printk("^^^^^^^^^kernel msg: send %d packets\n", i);
  	}
  	pAd->ate.TxDoneCount += i;

  	RTMP_IO_WRITE32(pAd, RXCSR0, 0xffffffff);
  	RTMP_IO_WRITE32(pAd, SECCSR1, 0x1);
}


void Tx_Loop_Proc(IN PRTMP_ADAPTER pAd)
{
	pAd->ate.Mode = ATE_TXFRAME;
	pAd->ate.TxDoneCount = 0;
	send_test_data(pAd);
}


void Tx_Cont_Proc(IN PRTMP_ADAPTER pAd)
{
   	PTXD_STRUC pTxD;
   	unsigned char *pDest;
   	UINT j;
	int i;
    
	pAd->PortCfg.HideSsid = 1;			// stop beacon
	pAd->WdsTab.Mode = WDS_BRIDGE_MODE; // Stop beacon and probe request
  	pAd->ate.Mode = ATE_TXCONT;
	pAd->ate.TxLength = 50;
	pAd->ate.TxRate = RATE_54;

//  for (i = 0 ; (i < TX_RING_SIZE) ; i++)
  	for (i = 0 ; (i < 1) ; i++)
  	{
    	pTxD = (PTXD_STRUC)pAd->TxRing[pAd->CurEncryptIndex].va_addr;
    	pDest = (PUCHAR)pAd->TxRing[pAd->CurEncryptIndex].va_data_addr;

    	// Prepare frame payload
    	memcpy(pDest, &TempletFrame, LENGTH_802_11);
    	for (j = LENGTH_802_11 ; j < PRIO_BUFFER_SIZE; j++)
      		pDest[j] = 0xAA;
		memcpy(&pDest[10], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);
		memcpy(&pDest[16], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);

    	RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
                                    SHORT_RETRY, IFS_BACKOFF, pAd->ate.TxRate, 4,
                                    pAd->ate.TxLength, pAd->PortCfg.TxPreambleInUsed, 0);

    	pAd->CurEncryptIndex++;
    	if (pAd->CurEncryptIndex >= TX_RING_SIZE)
    	{
      		pAd->CurEncryptIndex = 0;
    	}
  	}

  	RTMP_IO_WRITE32(pAd, RXCSR0, 0xffffffff);
  	RTMP_IO_WRITE32(pAd, SECCSR1, 0x1);
}

// Tx continue
INT	Set_Tx_Cont_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
  	USHORT  BbpData;
  	ULONG   MacData;
   	PTXD_STRUC pTxD;
   	unsigned char *pDest;
   	UINT j;
	int i;
    
	pAd->PortCfg.HideSsid = 1;			// stop beacon
	pAd->WdsTab.Mode = WDS_BRIDGE_MODE; // Stop beacon and probe request
  	pAd->ate.Mode = ATE_TXCONT;
	pAd->ate.TxLength = 50;
	pAd->ate.TxRate = RATE_54;

  	//if(gWlanState != WLAN_STATE_INIT_OK)return VNIC_Not_Init(pOutRst, pOutData);

  	AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
  	AsicLockChannel(pAd, pAd->PortCfg.Channel);

  	RTMP_IO_READ32(pAd, MACCSR1, &MacData);
  	RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 63, &BbpData);

  	BbpData = 0;
  	MacData &= 0xFBFFFFFF;
    
	// abort any previous action; Felix; 93.06.18
  	RTMP_IO_WRITE32(pAd, TXCSR0, 0x08);		// Abort Tx
	RTMP_IO_WRITE32(pAd, MACCSR1, MacData);
	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 63, BbpData);

  	BbpData |= 0x80;
  	MacData |= 0x04000000;

  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 63, BbpData);
  	RTMP_IO_WRITE32(pAd, MACCSR1, MacData);

  	//Tx_Cont_Proc(pAd);
	
//  for (i = 0 ; (i < TX_RING_SIZE) ; i++)
  	for (i = 0 ; (i < 1) ; i++)
  	{
    	pTxD = (PTXD_STRUC)pAd->TxRing[pAd->CurEncryptIndex].va_addr;
    	pDest = (PUCHAR)pAd->TxRing[pAd->CurEncryptIndex].va_data_addr;

    	// Prepare frame payload
    	memcpy(pDest, &TempletFrame, LENGTH_802_11);
    	for (j = LENGTH_802_11 ; j < PRIO_BUFFER_SIZE; j++)
      		pDest[j] = 0xAA;
		memcpy(&pDest[10], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);
		memcpy(&pDest[16], &pAd->PermanentAddress[0], ETH_LENGTH_OF_ADDRESS);

    	RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
                                    SHORT_RETRY, IFS_BACKOFF, pAd->ate.TxRate, 4,
                                    pAd->ate.TxLength, pAd->PortCfg.TxPreambleInUsed, 0);

    	pAd->CurEncryptIndex++;
    	if (pAd->CurEncryptIndex >= TX_RING_SIZE)
    	{
      		pAd->CurEncryptIndex = 0;
    	}
  	}

  	RTMP_IO_WRITE32(pAd, RXCSR0, 0xffffffff);
  	RTMP_IO_WRITE32(pAd, SECCSR1, 0x1);
  	//*pOutRst = CMD_RTN_OK;

    DBGPRINT(RT_DEBUG_TRACE, "Set_Tx_Cont_Proc::\n");
    return TRUE;
}

INT	Set_Tx_Start_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
  	USHORT  BbpData;
  	ULONG   MacData;

	pAd->ate.TxCount = simple_strtol(arg, 0, 10);

	if(0 == pAd->ate.TxCount)
		pAd->ate.TxCount = 0xFFFFFFFF;

  	//if(gWlanState != WLAN_STATE_INIT_OK)return VNIC_Not_Init(pOutRst, pOutData);

  	AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
  	AsicLockChannel(pAd, pAd->PortCfg.Channel);

  	RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 63, &BbpData);
  	RTMP_IO_READ32(pAd, MACCSR1, &MacData);

  	BbpData = 0;
  	MacData &= 0xFBFFFFFF;
    
	// abort any previous action; Felix; 93.06.18
  	RTMP_IO_WRITE32(pAd, TXCSR0, 0x08);		// Abort Tx
	RTMP_IO_WRITE32(pAd, MACCSR1, MacData);
	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 63, BbpData);

  	Tx_Loop_Proc(pAd);

  	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 63, BbpData);
  	RTMP_IO_WRITE32(pAd, MACCSR1, MacData);

  	//*pOutRst = CMD_RTN_OK;

    DBGPRINT(RT_DEBUG_TRACE, "Set_Tx_Cont_Proc::\n");
    return TRUE;
}

INT	Set_Rx_Start_Proc(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
	u16 flag;
  	USHORT  BbpData;
  	ULONG   MacData;

	flag = simple_strtol(arg, 0, 10);

  	AsicSwitchChannel(pAd, pAd->PortCfg.Channel);
  	AsicLockChannel(pAd, pAd->PortCfg.Channel);

  	RTMP_BBP_IO_READ32_BY_REG_ID(pAd, 63, &BbpData);
  	RTMP_IO_READ32(pAd, MACCSR1, &MacData);

  	BbpData = 0;
  	MacData &= 0xFBFFFFFF;
    
	// abort any previous action; Felix; 93.06.18
	RTMP_IO_WRITE32(pAd, MACCSR1, MacData);
	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAd, 63, BbpData);
  	RTMP_IO_WRITE32(pAd, TXCSR0, 0x08);		// Abort Tx

	if (0==flag)
	{
		RTMP_IO_WRITE32(pAd, RXCSR0, 0xFFFFFFFF); //stop rx
	} else
	{
		RTMP_IO_WRITE32(pAd, RXCSR0, 0x56); //start rx
	}

    DBGPRINT(RT_DEBUG_TRACE, "Set_Tx_Cont_Proc::\n");
    return TRUE;
}
INT Set_InitAsicFromEEPROM(
   	IN	PRTMP_ADAPTER	pAd, 
	IN	PUCHAR			arg)
{
	NICReadEEPROMParameters(pAd);
	NICInitAsicFromEEPROM(pAd);

	DBGPRINT(RT_DEBUG_TRACE, "Set_InitAsicFromEEPROM::\n");
	return TRUE;
}
#endif
/*
    ==========================================================================
    Description:
        Set ATE operation mode to
        0. APSTOP  = Stop AP Mode
        1. APSTART = Start AP Mode
        2. TXCONT  = Continuous Transmit
        3. TXCARR  = Transmit Carrier
        4. TXFRAME = Transmit Frames
        5. RXFRAME = Receive Frames
    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	USHORT			BbpData;
	ULONG			MacData;
	PTXD_STRUC		pTxD;
	PUCHAR			pDest;
	UINT			i, j;
	
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_Proc (arg = %s)\n", arg);

	mdelay(5);

    ATEReset(pAdapter);
	mdelay(10);

	AsicSwitchChannel(pAdapter, pAdapter->ate.Channel);
	AsicLockChannel(pAdapter, pAdapter->ate.Channel);
	mdelay(5);

	RTMP_BBP_IO_READ32_BY_REG_ID(pAdapter, 63, &BbpData);
	RTMP_IO_READ32(pAdapter, MACCSR1, &MacData);

	BbpData = 0;
	MacData &= 0xFBFFFFFF;

	if (!strcmp(arg, "APSTOP")) 
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: APSTOP\n");
		
		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);
		
		pAdapter->ate.Mode = ATE_APSTOP;

		netif_stop_queue(pAdapter->net_dev);
        del_timer_sync(&pAdapter->Mlme.PeriodicTimer);
		ApStop(pAdapter);
		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0xffffffff);	// Stop Rx
	}
	else if (!strcmp(arg, "APSTART")) 
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: APSTART\n");
		
		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);
		
		pAdapter->ate.Mode = ATE_APSTART;

		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0x56);	// Start Rx
		netif_start_queue(pAdapter->net_dev);

		pAdapter->Mlme.PeriodicTimer.expires = jiffies + MLME_TASK_EXEC_INTV;
		add_timer(&pAdapter->Mlme.PeriodicTimer);

		ApStartUp(pAdapter);
	}
	else if (!strcmp(arg, "TXCONT")) 		// Continuous Tx
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: TXCONT\n");
		
		pAdapter->ate.Mode = ATE_TXCONT;

		BbpData |= 0x80;
		MacData |= 0x04000000;

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
		RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);

		for (i = 0; (i < TX_RING_SIZE) && (i < pAdapter->ate.TxCount); i++)
		{
			pTxD = (PTXD_STRUC)pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
			pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;

			// Prepare frame payload
			memcpy(pDest, &TempletFrame, LENGTH_802_11);
			for(j = LENGTH_802_11; j < pAdapter->ate.TxLength; j++)
				pDest[j] = 0xAA;
			memcpy(&pDest[4], pAdapter->ate.Addr1, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[10], pAdapter->ate.Addr2, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[16], pAdapter->ate.Addr3, ETH_LENGTH_OF_ADDRESS);

			RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
				SHORT_RETRY, IFS_SIFS, pAdapter->ate.TxRate, 4,
				pAdapter->ate.TxLength, pAdapter->PortCfg.TxPreambleInUsed, 0);

			pAdapter->CurEncryptIndex++;
	        if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
	        {
	            pAdapter->CurEncryptIndex = 0;
	        }
		}

		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0xffffffff);
		RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
	}
	else if (!strcmp(arg, "TXCARR"))			// Tx Carrier -------------------------------------
	{
		DBGPRINT(RT_DEBUG_TRACE, "ATE: TXCARR\n");
		pAdapter->ate.Mode = ATE_TXCARR;

		BbpData |= 0x40;
		MacData |= 0x04000000;

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
		RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);

		for (i = 0; (i < TX_RING_SIZE) && (i < pAdapter->ate.TxCount); i++)
		{
			pTxD = (PTXD_STRUC)pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
			pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;

			// Prepare frame payload
			memcpy(pDest, &TempletFrame, LENGTH_802_11);
			for(j = LENGTH_802_11; j < pAdapter->ate.TxLength; j++)
				pDest[j] = 0xAA;
			memcpy(&pDest[4], pAdapter->ate.Addr1, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[10], pAdapter->ate.Addr2, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[16], pAdapter->ate.Addr3, ETH_LENGTH_OF_ADDRESS);

			RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
				SHORT_RETRY, IFS_BACKOFF, pAdapter->ate.TxRate, 4,
				pAdapter->ate.TxLength, pAdapter->PortCfg.TxPreambleInUsed, 0);

			pAdapter->CurEncryptIndex++;
	        if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
	        {
	            pAdapter->CurEncryptIndex = 0;
	        }
		}

		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0xffffffff);
		RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
	}
	else if (!strcmp(arg, "TXFRAME"))			// Tx Frames --------------------------------------
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: TXFRAME(Count=%d)\n", pAdapter->ate.TxCount);
		pAdapter->ate.Mode = ATE_TXFRAME;

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);

		pAdapter->ate.TxDoneCount = 0;
		
		for (i = 0; (i < TX_RING_SIZE) && (i < pAdapter->ate.TxCount); i++)
		{
			pTxD = (PTXD_STRUC)pAdapter->TxRing[pAdapter->CurEncryptIndex].va_addr;
			pDest = (PUCHAR) pAdapter->TxRing[pAdapter->CurEncryptIndex].va_data_addr;

			// Prepare frame payload
			memcpy(pDest, &TempletFrame, LENGTH_802_11);
			for(j = LENGTH_802_11; j < pAdapter->ate.TxLength; j++)
				pDest[j] = 0xAA;
			memcpy(&pDest[4], pAdapter->ate.Addr1, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[10], pAdapter->ate.Addr2, ETH_LENGTH_OF_ADDRESS);
			memcpy(&pDest[16], pAdapter->ate.Addr3, ETH_LENGTH_OF_ADDRESS);

			RTMPWriteTxDescriptor(pTxD, TRUE, CIPHER_NONE, FALSE, FALSE, FALSE,
				SHORT_RETRY, IFS_BACKOFF, pAdapter->ate.TxRate, 4,
				pAdapter->ate.TxLength, pAdapter->PortCfg.TxPreambleInUsed, 0);

			pAdapter->CurEncryptIndex++;
	        if (pAdapter->CurEncryptIndex >= TX_RING_SIZE)
	        {
	            pAdapter->CurEncryptIndex = 0;
	        }
		}
		pAdapter->ate.TxDoneCount += i;

		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0xffffffff); //stop rx
		RTMP_IO_WRITE32(pAdapter, SECCSR1, 0x1);
	}
	else if (!strcmp(arg, "RXFRAME")) 			// Rx Frames --------------------------------------
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: RXFRAME\n");

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);
    	
		pAdapter->ate.Mode = ATE_RXFRAME;
		pAdapter->ate.TxDoneCount = pAdapter->ate.TxCount;
		
		RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x08);		// Abort Tx
		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0x56);		// Start Rx
	}
	#ifdef VINCENT_K
	else if (!strcmp(arg, "RXSTART")) 			// Rx Frames --------------------------------------
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: RXSTART\n");

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);
    	
		pAdapter->ate.Mode = ATE_RXFRAME;
		pAdapter->ate.TxDoneCount = pAdapter->ate.TxCount;
		
		RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x08);		// Abort Tx
		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0x56);		// Start Rx
	}
	else if (!strcmp(arg, "RXSTOP")) 			// Rx Frames --------------------------------------
	{						
		DBGPRINT(RT_DEBUG_TRACE, "ATE: RXSTOP\n");

		RTMP_IO_WRITE32(pAdapter, MACCSR1, MacData);
    	RTMP_BBP_IO_WRITE32_BY_REG_ID(pAdapter, 63, BbpData);
    	
		pAdapter->ate.Mode = ATE_RXSTOP;
		pAdapter->ate.TxDoneCount = pAdapter->ate.TxCount;
		
		RTMP_IO_WRITE32(pAdapter, TXCSR0, 0x08);		// Abort Tx
		RTMP_IO_WRITE32(pAdapter, RXCSR0, 0xffffffff);		// Stop Rx
	}
	#endif
	else
	{	
		DBGPRINT(RT_DEBUG_TRACE, "ATE:	Invalid arg!\n");
		return FALSE;
	}
	mdelay(5);

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_Proc\n");
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE ADDR1=DA for TxFrames    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_DA_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	char				*value;
	int					i;
	
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_DA_Proc (arg = %s)\n", arg);
	
	if(strlen(arg) != 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
		return FALSE;

    for (i=0, value = strtok(arg,":"); value; value = strtok(NULL,":")) 
	{
		if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
			return FALSE;  //Invalid

		AtoH(value, &pAdapter->ate.Addr1[i++], 2);
	}

	if(i != 6)
		return FALSE;  //Invalid
		
	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_DA_Proc (DA = %2X:%2X:%2X:%2X:%2X:%2X)\n", pAdapter->ate.Addr1[0], pAdapter->ate.Addr1[1], pAdapter->ate.Addr1[2], pAdapter->ate.Addr1[3], pAdapter->ate.Addr1[4], pAdapter->ate.Addr1[5]);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE ADDR2=SA for TxFrames    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_SA_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	char				*value;
	int					i;
	
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_SA_Proc (arg = %s)\n", arg);
	
	if(strlen(arg) != 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
		return FALSE;

    for (i=0, value = strtok(arg,":"); value; value = strtok(NULL,":")) 
	{
		if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
			return FALSE;  //Invalid

		AtoH(value, &pAdapter->ate.Addr2[i++], 2);
	}

	if(i != 6)
		return FALSE;  //Invalid

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_SA_Proc (DA = %2X:%2X:%2X:%2X:%2X:%2X)\n", pAdapter->ate.Addr2[0], pAdapter->ate.Addr2[1], pAdapter->ate.Addr2[2], pAdapter->ate.Addr2[3], pAdapter->ate.Addr2[4], pAdapter->ate.Addr2[5]);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE ADDR3=BSSID for TxFrames    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_BSSID_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	char				*value;
	int					i;
	
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_BSSID_Proc (arg = %s)\n", arg);
	
	if(strlen(arg) != 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
		return FALSE;

    for (i=0, value = strtok(arg,":"); value; value = strtok(NULL,":")) 
	{
		if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
			return FALSE;  //Invalid

		AtoH(value, &pAdapter->ate.Addr3[i++], 2);
	}

	if(i != 6)
		return FALSE;  //Invalid

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_BSSID_Proc (DA = %2X:%2X:%2X:%2X:%2X:%2X)\n", pAdapter->ate.Addr3[0], pAdapter->ate.Addr3[1], pAdapter->ate.Addr3[2], pAdapter->ate.Addr3[3], pAdapter->ate.Addr3[4], pAdapter->ate.Addr3[5]);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE Tx Length    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_CHANNEL_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_CHANNEL_Proc (arg = %s)\n", arg);
	
	pAdapter->ate.Channel = simple_strtol(arg, 0, 10);

	if((pAdapter->ate.Channel < 1) || (pAdapter->ate.Channel > 14))
	{
		pAdapter->ate.Channel = 1;
		return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_CHANNEL_Proc (ATE Channel = %d)\n", pAdapter->ate.Channel);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE Tx Power    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_TX_POWER_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	ULONG R3;
	
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_TX_POWER_Proc\n (arg = %s)\n", arg);
	
	pAdapter->ate.TxPower = simple_strtol(arg, 0, 10);

	if(pAdapter->ate.TxPower >= 32)
	{
		pAdapter->ate.TxPower = pAdapter->PortCfg.ChannelTxPower[pAdapter->ate.Channel - 1];;
		return FALSE;
	}

	R3 = pAdapter->ate.TxPower;
    R3 = R3 << 9; // shift TX power control to correct RF register bit position

	R3 |= (pAdapter->PortCfg.LatchRfRegs.R3 & 0xffffc1ff);
	RTMP_RF_IO_WRITE32(pAdapter, R3);

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_TX_POWER_Proc (TxPower = %d)\n", pAdapter->ate.TxPower);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE Tx Length    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_TX_LENGTH_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_TX_LENGTH_Proc (arg = %s)\n", arg);
	
	pAdapter->ate.TxLength = simple_strtol(arg, 0, 10);

	if((pAdapter->ate.TxLength < 24) || (pAdapter->ate.TxLength > 1500))
	{
		pAdapter->ate.TxLength = 1500;
		return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_TX_LENGTH_Proc (TxLength = %d)\n", pAdapter->ate.TxLength);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE Tx Count    Return:
        TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_TX_COUNT_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_TX_COUNT_Proc (arg=%s)\n", arg);
	
	pAdapter->ate.TxCount = simple_strtol(arg, 0, 10);

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_TX_COUNT_Proc (TxCount = %d)\n", pAdapter->ate.TxCount);
	
	return TRUE;
}

/* 
    ==========================================================================
    Description:
        Set ATE Tx Rate
        Return:
        	TRUE if all parameters are OK, FALSE otherwise
    ==========================================================================
*/
INT	Set_ATE_TX_RATE_Proc(
	IN	PRTMP_ADAPTER	pAdapter, 
	IN	PUCHAR			arg)
{
	DBGPRINT(RT_DEBUG_TRACE, "==> Set_ATE_TX_RATE_Proc (arg = %s)\n", arg);
	
	pAdapter->ate.TxRate = simple_strtol(arg, 0, 10);

	if(pAdapter->ate.TxRate > RATE_54)
	{
		pAdapter->ate.TxRate = RATE_11;
		return FALSE;
	}

	DBGPRINT(RT_DEBUG_TRACE, "<== Set_ATE_TX_RATE_Proc (TxRate = %d)\n", pAdapter->ate.TxRate);
	
	return TRUE;
}
#endif	//#ifdef RALINK_ATE

